; /////////////////////////////////////////////////////////////////////////////
; //
; // Fidonet Mail Access: DOS System dependant routines module
; //
; // DosMisc.asm
; //
; // Copyright (c) 1999 by Pete Kvitek
; //
; // This program is free software; you can redistribute it and/or modify
; // it under the terms of the GNU General Public License as published by
; // the Free Software Foundation; either version 2 of the License, or
; // (at your option) any later version. See GNU_GPL.TXT for details.
; //
; // Created: 04/Oct/1999
; // Updated: 04/Oct/1999
; //
; // History: 18/Sep/1991 kvitekp - Initially created for DW-TEXT project.
; //
; /////////////////////////////////////////////////////////////////////////////

INCLUDE DosMisc.inc

; *//////////////////////////////////////////////////////////////////////////*
; * M o d u l e   s p e c i f i c   d e c l a r a t i o n s                  *
; *//////////////////////////////////////////////////////////////////////////*

		; Operating system types (see FmaSys.h)

OSTYPE_DOS              = 0001h         ; Good old DOS
OSTYPE_WINDOWS          = 0002h         ; Microsoft Windows DosBox
OSTYPE_WINNT            = 0003h         ; Microsoft Windows NT DosBox
OSTYPE_DESQVIEW         = 0004h         ; QuarterDesk DesqView
OSTYPE_OS2DOS           = 0005h         ; IBM OS/2 DosBox
OSTYPE_OS2              = 0006h         ; Native IBM OS/2
OSTYPE_W32              = 0007h         ; Native Win32
OSTYPE_MASK             = 000Fh         ; OS Type mask
OSTYPE_PROTMODE         = 8000h         ; Protected mode

		; DPMI translation services data structure

DPMI_TRN STRUC
dtrn_EDI        dd      ?
dtrn_ESI        dd      ?
dtrn_EBP        dd      ?
dtrn_reserved   dd      ?
dtrn_EBX        dd      ?
dtrn_EDX        dd      ?
dtrn_ECX        dd      ?
dtrn_EAX        dd      ?
dtrn_flags      dw      ?
dtrn_ES         dw      ?
dtrn_DS         dw      ?
dtrn_FS         dw      ?
dtrn_GS         dw      ?
dtrn_IP         dw      ?
dtrn_CS         dw      ?
dtrn_SP         dw      ?
dtrn_SS         dw      ?
DPMI_TRN ENDS

; *//////////////////////////////////////////////////////////////////////////*
; * D a t a   s e g m e n t                                                  *
; *//////////////////////////////////////////////////////////////////////////*

		.DATA

; *//////////////////////////////////////////////////////////////////////////*
; * C o d e   s e g m e n t                                                  *
; *//////////////////////////////////////////////////////////////////////////*

		.CODE

; ****************************************************************************
; * This routine returns operating system type
; *
; * Prototype:
; *
; *   USHORT APIENTRY DosGetOSType(VOID);
; *
; * Expects: none
; * Returns: ax = OSTYPE_ flag
; * Updates: none
; *

DosGetOSType    PROC FAR USES si di

		; ---
		; Check for MS Windows

		mov     ax, 1600h       ; MS Windows installation check
		int     2Fh
		or      al, al
		jz      @F
		mov     ax, OSTYPE_WINDOWS
		jmp     @@More
@@:
		; ---
		; Check for MS Windows NT

		mov     ax, 3306h       ; Get true DOS version number
		mov     bx, 0000h
		int     21h
		cmp     bx, 3205h       ; Windows NT returns v5.50
		jne     @F
		mov     ax, OSTYPE_WINNT
		jmp     @@More
@@:
		; ---
		; Check for IBM OS/2 DOS Box

		mov     ax, 4010h       ; OS/2 v2.0+ installation check
		mov     bx, 0000h
		int     2Fh
		cmp     ax, 4010h
		je      @F
		mov     ax, OSTYPE_OS2DOS
		jmp     @@More
@@:
		; ---
		; Check for QuarterDesk DesqView

		mov     ax, 1022h       ; Get DesqView version
		mov     bx, 0000h
		int     15h
		or      bx, bx
		jz      @F
		mov     ax, OSTYPE_DESQVIEW
		jmp     @@More
@@:
		; ---
		; Assume general DOS

		mov     ax, OSTYPE_DOS

		; ---
		; Check if running in real or protected mode

@@More: push    ax              ; preserve os type
		mov     ax, 1686h       ; DPMI protected mode ?
		int     2Fh
		or      ax, ax          ; check if in protected mode
		pop     ax              ; restore os type
		jnz     @F              ; jump if not in protected mode
		or      ax, OSTYPE_PROTMODE
@@:

@@Exit: ret

DosGetOSType    ENDP

; ****************************************************************************
; * This routine gives out time slice given the operating system type
; *
; * Prototype:
; *
; *   VOID APIENTRY DosGiveUpTimeSlice(USHORT OSType);
; *
; * Expects: OSTYPE_ flag
; * Returns: none
; * Updates: none
; *

DosGiveUpTimeSlice      PROC FAR USES si di,    \
						OSType:word

		; Local variables on stack

		LOCAL   dtrn:DPMI_TRN   ; dpmi translation services data

		; Get operating system type

		mov     ax, OSType
		and     ax, OSTYPE_MASK

		; ---
		; Check for MS Windows

		cmp     ax, OSTYPE_WINDOWS
		jne     @F
		mov     ax, 1680h       ; Give out time slice
		int     2Fh
		jmp     @@Exit
@@:
		; ---
		; Check for MS Windows NT

		cmp     ax, OSTYPE_WINNT
		jne     @F

		; Check if we're running in real or protected mode.
		; This is necessary since Win/NT does not reflect
		; int 2Fh when in protected mode

		test    OSType, OSTYPE_PROTMODE
		jnz     @@Prot
		mov     ax, 1680h       ; Give out time slice
		int     2Fh
		jmp     @@Exit

		; Clean up dpmi translation services structure

@@Prot: cld
		mov     cx, size DPMI_TRN
		lea     di, dtrn
		mov     ax, ss
		mov     es, ax
		xor     ax, ax
		rep     stosb

		mov     word ptr dtrn.dtrn_EAX, 1680h   ; Give out time slice
		lea     di, dtrn        ; es:di = dpmi translation services data
		mov     ax, 0300h       ; dpmi simulate real mode interrupt
		mov     bx, 002Fh       ; flags and real mode interrupt number
		mov     cx, 0           ; number of words to copy from stack
		int     31h             ; simulate real mode interrupt

		jmp     @@Exit
@@:
		; ---
		; Check for IBM OS/2 DOS Box

		cmp     ax, OSTYPE_OS2DOS
		jne     @F
		mov     ax, 1680h       ; Give out time slice
		int     2Fh
		jmp     @@Exit
@@:
		; ---
		; Check for QuarterDesk DesqView

		cmp     ax, OSTYPE_DESQVIEW
		jne     @F
		mov     ax, 101Ah       ; Switch to task stack
		int     15h
		mov     ax, 1000h       ; Give out time slice
		int     15h
		mov     ax, 1025h       ; Switch back to user stack
		int     15h
		jmp     @@Exit
@@:
		; ---
		; Check for general DOS

		cmp     ax, OSTYPE_DOS
		jne     @F
		mov     ah, 01h         ; Poll keyboard
		int     16h
		int     28h             ; DOS Idle Interrupt
@@:

@@Exit: ret

DosGiveUpTimeSlice      ENDP

		END

; *//////////////////////////////////////////////////////////////////////////*
; * End of DosMisc.asm                                                       *
; *//////////////////////////////////////////////////////////////////////////*
