/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Main header file
//
// FmaApi.h
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 10/Feb/2000
//
// History: 17/Jan/1999 kvitekp - Initially created for IMA project.
//
//          29/Jan/2000 kvitekp - Added FmaDupeMsg(pfmm)
//
//          10/Feb/2000 kvitekp - Added FmaAddFmpFormatV()
//
//          11/Feb/2000 kvitekp - Added Pkt base stuff
//
//          22/Feb/2000 kvitekp - Added namespace compatibility #ifdefs
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _FMA_API_INCLUDED
#define _FMA_API_INCLUDED

  // Declare C/C++ frame

#ifdef __cplusplus
extern "C" {
#endif

  // Include types declaratons

#include "FmaDef.h"

/////////////////////////////////////////////////////////////////////////////
// Fidonet mail message property declarations
/////////////////////////////////////////////////////////////////////////////

  // Fidonet mail message property structure

  typedef struct _FMP {         /* fmp */
	struct  _FMP * pfmpPrev;            // Pointer to prev property
	struct  _FMP * pfmpNext;            // Pointer to next property
	ULONG   id;                         // Property id (FMP_)
	ULONG   cb;                         // Property data size
  //BYTE    ab[cb];                     // Property data
  } FMP, FAR * PFMP;

  // Macros to manage fidonet mail message property data

#define PFMPDATA(pfmp) (PVOID)((PBYTE)(pfmp) + sizeof(FMP)) // Poiner to property data
#define PFMPLONG(pfmp) (PLONG)(PFMPDATA(pfmp))              // Poiner to property long
#define PFMPPPSZ(pfmp) (PSZ *)(PFMPDATA(pfmp))              // Poiner to property psz
#define PFMPDTTM(pfmp) (PDTTM)(PFMPDATA(pfmp))              // Poiner to property dttm

  // Fidonet mail message property id masks

#define FMP_FLGMASK     0xff000000L     // Flags mask
#define FMP_POSMASK     0x00ff0000L     // Position mask
#define FMP_TAGMASK     0x0000ffffL     // Tag mask

  // Fidonet mail message property id macros

#define FMP_FLG(id)     (((id) & FMP_FLGMASK)      )
#define FMP_POS(id)     (((id) & FMP_POSMASK) >> 16)
#define FMP_TAG(id)     (((id) & FMP_TAGMASK)      )

  // Fidonet mail message property id flags

#define FMP_ASZ          0x01000000L    // Asciiz string
#define FMP_PSZ          0x02000000L    // Pointer(s) to asciiz string(s)
#define FMP_U32          0x04000000L    // Unsigned 32-bit value(s)
#define FMP_BIN          0x08000000L    // Counted binary of cb size
#define FMP_DTM          0x10000000L    // Date/time info (DTTM format)
#define FMP_EOM          0x20000000L    // Goes past the message text
#define FMP_ONE          0x40000000L    // Only one instance is allowed
#define FMP_KLU (FMP_ASZ|0x80000000L)   // FNT kludge property

  // Macro to define property ids

#define _FMP(tag, pos, flg)  (ULONG)(         \
	(( ((ULONG)(tag))      ) & FMP_TAGMASK) | \
	(( ((ULONG)(pos)) << 16) & FMP_POSMASK) | \
	(( ((ULONG)(flg))      ) & FMP_FLGMASK) )

  // Fidonet mail message basic property ids

#define FMP_MESSAGETEXT   _FMP( 0, 0, FMP_PSZ | FMP_ONE)        // Message text
#define FMP_SENDERNAME    _FMP( 1, 1, FMP_ASZ | FMP_ONE)        // Author's name
#define FMP_RECEVRNAME    _FMP( 2, 2, FMP_ASZ | FMP_ONE)        // Recipient's name
#define FMP_ORIGADDR      _FMP( 3, 3, FMP_ASZ | FMP_ONE)        // Originating address
#define FMP_DESTADDR      _FMP( 4, 4, FMP_ASZ | FMP_ONE)        // Destination address
#define FMP_GATEADDR      _FMP( 5, 5, FMP_ASZ | FMP_ONE)        // Gate address
#define FMP_SUBJECT       _FMP( 6, 6, FMP_ASZ | FMP_ONE)        // Message subject
#define FMP_FILEATTACH    _FMP( 7, 7, FMP_ASZ          )        // File attachment
#define FMP_FILEREQUEST   _FMP( 8, 8, FMP_ASZ          )        // File request
#define FMP_FILEUPDATEREQ _FMP( 9, 9, FMP_ASZ          )        // File update request

  // Fidonet mail message additional property ids

#define FMP_DTTMWRITTEN   _FMP(11,11, FMP_DTM | FMP_ONE)        // When msg was written
#define FMP_DTTMRECEIVED  _FMP(12,12, FMP_DTM | FMP_ONE)        // When msg was received
#define FMP_DTTMPROCESSED _FMP(13,13, FMP_DTM | FMP_ONE)        // When msg was processed
#define FMP_MSGREPLYTO    _FMP(14,14, FMP_U32 | FMP_ONE)        // Msg which this msg replies to
#define FMP_MSGREPLIES    _FMP(15,15, FMP_U32 | FMP_ONE)        // Msgs which reply to this msg
#define FMP_MSGREPLYNEXT  _FMP(16,16, FMP_U32 | FMP_ONE)        // Next msg in reply chain if any
#define FMP_TIMESREAD     _FMP(17,17, FMP_U32 | FMP_ONE)        // Number of times message was read
#define FMP_PASSWORDCRC   _FMP(18,18, FMP_U32 | FMP_ONE)        // CRC-32 of password to access msg
#define FMP_MESSAGECOST   _FMP(19,19, FMP_U32 | FMP_ONE)        // Cost of message

  // Fidonet mail message kludge property ids. 
  // Note: ^aFMPT,^aTOPT,^aINTL and ^aDOMAIN kludge info should be processed and
  // placed in appropriate properties. Also, all ^aFLAGS flags which have matching 
  // FMMA_ attributes should not be stored in FMP_FLAGS to prevent ambiguities

#define FMP_AREA          _FMP(29,29, FMP_KLU | FMP_ONE)        // FTN ^aAREA kludge info
#define FMP_MSGID         _FMP(30,30, FMP_KLU | FMP_ONE)        // FTN ^aMSGID kludge info
#define FMP_REPLYID       _FMP(31,30, FMP_KLU | FMP_ONE)        // FTN ^aREPLYID kludge info
#define FMP_PID           _FMP(32,30, FMP_KLU | FMP_ONE)        // FTN ^aPID kludge info
#define FMP_TID           _FMP(33,30, FMP_KLU | FMP_ONE)        // FTN ^aTID kludge info
#define FMP_TZUTC         _FMP(34,30, FMP_KLU | FMP_ONE)        // FTN ^aTZUTC kludge info ([+|-]NNNN)
#define FMP_FLAGS         _FMP(35,30, FMP_KLU | FMP_ONE)        // FTN ^aFLAGS kludge info
#define FMP_CHARSET       _FMP(36,30, FMP_KLU | FMP_ONE)        // FTN ^aCHRS kludge info
#define FMP_REALNAME      _FMP(37,30, FMP_KLU | FMP_ONE)        // FTN ^aREALNAME kludge info
#define FMP_REPLYADDR     _FMP(38,30, FMP_KLU | FMP_ONE)        // FTN ^aREPLYADDR kludge info
#define FMP_REPLYTO       _FMP(39,30, FMP_KLU | FMP_ONE)        // FTN ^aREPLYTO kludge info
#define FMP_FTSKLUDGE     _FMP(40,30, FMP_KLU          )        // Unknown kludge line

#define FMP_SEENBY2D      _FMP(70,70, FMP_KLU | FMP_EOM)        // 2D (net/node) SEEN-BY information
#define FMP_PATH2D        _FMP(71,71, FMP_KLU | FMP_EOM)        // 2D (net/node) PATH information
#define FMP_EOMKLUDGE     _FMP(80,80, FMP_KLU | FMP_EOM)        // Unknown kludge line at eom

/////////////////////////////////////////////////////////////////////////////
// Fidonet mail message declarations
/////////////////////////////////////////////////////////////////////////////

  // Fidonet mail message object structure

  typedef struct _FMM { /* fmm */ 
	UINT    cb;                         // Size of this structure
	UINT    type;                       // Msg native type (FMMT_)
	ULONG   umsg;                       // Unique message id or number
	ULONG   attr;                       // Msg attributes (MA_)
	PFMP    pfmp;                       // Msg property list
#ifdef FMM_EXTENSION
	FMM_EXTENSION                       // Extension declarations
#endif
  } FMM, FAR * PFMM;

  // Fidonet mail message types

#define FMMT_NATIVE      0              // Msg is native fmm message
#define FMMT_FTS         1              // Msg is from FTS-001 message base
#define FMMT_SQUISH      2              // Msg is from Squish message base
#define FMMT_JAM         3              // Msg is from JAM message base
#define FMMT_PKT         4              // Msg is from Pkt message base

  // Fidonet mail message attributes in fmm.attr

#define MA_LOCAL         0x00000001L    // Msg created locally
#define MA_INTRANSIT     0x00000002L    // Msg is in-transit
#define MA_PRIVATE       0x00000004L    // Private message
#define MA_READ          0x00000008L    // Read by addressee
#define MA_SENT          0x00000010L    // Sent to remote
#define MA_KILLSENT      0x00000020L    // Kill when sent
#define MA_ARCHIVESENT   0x00000040L    // Archive when sent
#define MA_HOLD          0x00000080L    // Hold for pick-up
#define MA_CRASH         0x00000100L    // Crash
#define MA_IMMEDIATE     0x00000200L    // Send Msg now, ignore restrictions
#define MA_DIRECT        0x00000400L    // Send directly to destination
#define MA_GATE          0x00000800L    // Send via gateway
#define MA_FILEREQUEST   0x00001000L    // File request
#define MA_FILEATTACH    0x00002000L    // File(s) attached to Msg
#define MA_TRUNCFILE     0x00004000L    // Truncate file(s) when sent
#define MA_KILLFILE      0x00008000L    // Delete file(s) when sent
#define MA_RECEIPTREQ    0x00010000L    // Return receipt requested
#define MA_CONFIRMREQ    0x00020000L    // Confirmation receipt requested
#define MA_ORPHAN        0x00040000L    // Unknown destination
#define MA_ENCRYPT       0x00080000L    // Msg text is encrypted
#define MA_COMPRESS      0x00100000L    // Msg text is compressed
#define MA_ESCAPED       0x00200000L    // Msg text is seven bit ASCII
#define MA_FORCEPICKUP   0x00400000L    // Force pickup
#define MA_TYPELOCAL     0x00800000L    // Msg is for local use only (not for export)
#define MA_TYPEECHO      0x01000000L    // Msg is for conference distribution
#define MA_TYPENET       0x02000000L    // Msg is direct network mail
#define MA_SCANNED       0x04000000L    // Msg is scanned (squish)
#define MA_FILEUPDATEREQ 0x08000000L    // File update request
#define MA_AUDITREQUEST  0x10000000L    // Audit request
#define MA_NODISPLAY     0x20000000L    // Msg may not be displayed to user
#define MA_LOCKED        0x40000000L    // Msg is locked, no editing possible
#define MA_DELETED       0x80000000L    // Msg is deleted

  // Function prototypes

  FMAAPI_(PFMM) FmaCreateMsg(UINT cbExtra);
  FMAAPI_(BOOL) FmaDestroyMsg(PFMM pfmm);
  FMAAPI_(PFMM) FmaDupeMsg(PFMM pfmm);
  FMAAPI_(BOOL) FmaDumpMsg(PFMM pfmm);

  FMAAPI_(PFMP) FmaGetFmp(PFMM pfmm, ULONG id, UINT ifmp);
  FMAAPI_(BOOL) FmaDelFmp(PFMM pfmm, ULONG id, UINT ifmp);
  FMAAPI_(BOOL) FmaDumpFmp(PFMP pfmp);
  FMAAPI_(PFMP) FmaAddFmp(PFMM pfmm, ULONG id, PVOID pbData, ULONG cbData);
  FMAAPIV_(PFMP) FmaAddFmpFormat(PFMM pfmm, ULONG id, PSZ pszFormat, ...);
  FMAAPI_(PFMP)  FmaAddFmpFormatV(PFMM pfmm, ULONG id, PSZ pszFormat, va_list parg);
  FMAAPI_(ULONG) FmaGetFmpTextLen(PFMP pfmp);

/////////////////////////////////////////////////////////////////////////////
// Abstract message base api declarations
/////////////////////////////////////////////////////////////////////////////

  // FmaOpenArea() open modes

#define OPENAREA_EXISTING           0   // Open existing, fail if does not exist
#define OPENAREA_ALWAYS             1   // Open existing or create if does not exist
#define OPENAREA_CREATE_NEW         2   // Create new, fail if already exists
#define OPENAREA_CREATE_ALWAYS      3   // Create new, truncate if exists

  // FmaOpenArea() flags

#define FOA_LOCKAREA      0x00000001L   // Lock area upon opening
#define FOA_READONLY      0x00000002L   // Open area in readonly mode
#define FOA_DENYALL       0x00000004L   // Open area in exclusive mode
#define FOA_NETMAIL       0x00000100L   // Area contains netmail
#define FOA_ECHOMAIL      0x00000200L   // Area contains echomail
#define FOA_HIGHREAD      0x00000800L   // Maintain highest read message (Fts/Squish)
#define FOA_KEEPUMSG      0x00010000L   // Keep unique message id in smh.umsg (Squish)
#define FOA_DELAYTRIMMING 0x00020000L   // Delay maxmsg trimming until close area (Squish)
#define FOA_CREATEATEOF   0x00040000L   // Create new msgs at end of base (Squish)

  // FmaEnumMsgs() callback info structure

  typedef struct _ENUMMSGINFO { /* emi */ 
	MSGN    msgn;                       // Message number
	UMSG    umsg;                       // Message uid
	ULONG   imsg;                       // Message index
	ULONG   offs;                       // Message offset (sqi.lfm or jdx.lmh)
	ULONG   hash;                       // Message hash (sqi.hash or jdx.crc)
	HAREA   harea;                      // Message area handle
  } ENUMMSGINFO, FAR * PENUMMSGINFO;

  // FmaEnumMsgs() callback function prototype

  typedef BOOL (FMACALLBACK * PFNENUMMSGS)(PENUMMSGINFO pemi, PVOID pvCallbackData);

  // FmaEnumMsgs() flags

#define FENUMMSGS_REVERSE   0x00000001  // Enumerate in reverse order
#define FENUMMSGS_SAFEENUM  0x00010000  // Allows message map modification

  // FmaOpenMsg() open modes

#define OPENMSG_CREATE      0           // Create new msg for reading and writing, truncate if exists
#define OPENMSG_READ        1           // Open existing msg for reading, fail if none
#define OPENMSG_WRITE       2           // Open existing msg for writing, fail if none
#define OPENMSG_READWRITE   3           // Open existing msg for reading and writing, fail if none

  // FmaOpenMsg() flags

#define FOM_KEEPUMSG        0x00000001  // Keep unique message id in smh.umsg

  // FmaOpenMsg() special message number values

#define MSGN_CUR            (MSGN)-1    // Open current message, fail if does not exist
#define MSGN_PREV           (MSGN)-2    // Open previous message, fail if does not exist
#define MSGN_NEXT           (MSGN)-3    // Open next message, fail if does not exist
#define MSGN_CURORPREV      (MSGN)-4    // Open current or previous if current does not exist
#define MSGN_CURORNEXT      (MSGN)-5    // Open current or next if current does not exist

  // FmaUidToMsgn() flags

#define UMSG_EXACT          0x0001      // Return exact message number
#define UMSG_NEXT           0x0002      // Return next message number
#define UMSG_PREV           0x0004      // Return previous message number

  // FmaLoadMsg() flags

#define FLM_NOTEXT          0x0001      // Don't load message text

  // FmaSaveMsg() flags

#define FSM_NOTEXT          0x0001      // Don't save message text
#define FSM_FORCEINTL       0x0002      // Always add ^aINTL kludge
#define FSM_AUTODOMAIN      0x0010      // Add ^aDOMAIN if both domains are known
#define FSM_AUTOMSGID       0x0020      // Add ^aMSGID automatically
#define FSM_AUTOPID         0x0040      // Add ^aPID automatically

  // FmaReadMsg() special values

#define FRM_GETCTRLCOPY  0              // *pcchCtrl: Allocate a copy of msg control info
#define FRM_GETCTRLPTR  ((ULONG)-1)     // *pcchCtrl: Return a readonly pointer to msg control info
#define FRM_GETALLTEXT  ((ULONG)-1)     // *pcchText: read entire message text

  // FmaGet/SetLRInfo structure

  typedef struct _LRINFO { /* lri */
	UMSG    umsgLast;                   // Last read message uid
	UMSG    umsgHigh;                   // Highest read message uid
  } LRINFO, FAR * PLRINFO;

  // Function prototypes

  FMAAPI_(HAREA) FmaOpenArea(PSZ pszPath, UINT type, UINT fOpenMode, ULONG fl,
							 NETADDR * pnetAddr);

  FMAAPI_(BOOL) FmaCloseArea(HAREA harea);
  FMAAPI_(BOOL) FmaLockArea(HAREA harea, BOOL bLock);

  FMAAPI_(BOOL) FmaValidateMsg(HAREA harea, MSGN msgn, PULONG pimsg);
  FMAAPI_(BOOL) FmaCreateMsgMap(HAREA harea, UMSG * paumsg[], PULONG pcumsg);
  FMAAPI_(BOOL) FmaEnumMsgs(HAREA harea, PFNENUMMSGS pfnCallback, 
							PVOID pvCallbackData, ULONG fl);

  FMAAPI_(BOOL) FmaDeleteMsg(HAREA harea, MSGN msgn);
  FMAAPI_(BOOL) FmaSetCurMsg(HAREA harea, MSGN msgn);
  FMAAPI_(MSGN) FmaGetCurMsg(HAREA harea);
  FMAAPI_(MSGN) FmaGetFirstMsg(HAREA harea);
  FMAAPI_(MSGN) FmaGetHighMsg(HAREA harea);
  FMAAPI_(LONG) FmaGetMsgCount(HAREA harea);

  FMAAPI_(UMSG) FmaMsgnToUid(HAREA harea, MSGN msgn);
  FMAAPI_(MSGN) FmaUidToMsgn(HAREA harea, UMSG umsg, UINT fMode);

  FMAAPI_(HMSG) FmaOpenMsg(HAREA harea, UINT mode, MSGN msgn, ULONG fl);
  FMAAPI_(BOOL) FmaCloseMsg(HMSG hmsg);

  FMAAPI_(BOOL) FmaLoadMsg(HMSG hmsg, PFMM * ppfmm, ULONG fl);
  FMAAPI_(BOOL) FmaSaveMsg(HMSG hmsg, PFMM pfmm, ULONG fl, ULONG cchTextTotal);

  FMAAPI_(BOOL) FmaReadMsg(HMSG hmsg, PVOID pmsgh, PVOID * ppchCtrl, ULONG * pcchCtrl,
						   PCH * ppchText, ULONG * pcchText, ULONG * poffText);

  FMAAPI_(BOOL) FmaWriteMsg(HMSG hmsg, PVOID pmsgh, PVOID pchCtrl, ULONG cchCtrl,
							PCH pchText, ULONG cchText, ULONG * poffText, 
							ULONG cchTextTotal);

  FMAAPI_(BOOL) FmaGetMsgCtrlLen(HMSG hmsg, ULONG * pcchCtrl);
  FMAAPI_(BOOL) FmaGetMsgTextLen(HMSG hmsg, ULONG * pcchText);

  FMAAPI_(BOOL) FmaGetLRInfo(HAREA harea, ULONG uid, PLRINFO plri);
  FMAAPI_(BOOL) FmaSetLRInfo(HAREA harea, ULONG uid, PLRINFO plri, 
							 PSZ pszUserName);

/////////////////////////////////////////////////////////////////////////////
// Api management declarations
/////////////////////////////////////////////////////////////////////////////

  // Message base types

#define MBTYPE_NONE       0             // No message base
#define MBTYPE_FTS        1             // FTS-001 message base
#define MBTYPE_SQU        2             // Squish message base
#define MBTYPE_JAM        3             // JAM message base
#define MBTYPE_PKT        4             // Pkt message base
#define MBTYPE_USER       5             // User message base id
#define MBTYPE_MAXNO     10             // Max number of message bases

  // Message base api structure

  typedef struct _API { /* api */
	UINT   cb;                          // Size of this structure
	HAREA (FMAAPI * pfnOpenArea)(PSZ, UINT, ULONG, NETADDR *);
	BOOL  (FMAAPI * pfnCloseArea)(HAREA);
	BOOL  (FMAAPI * pfnLockArea)(HAREA, BOOL);
	BOOL  (FMAAPI * pfnValidateMsg)(HAREA, MSGN, PULONG);
	BOOL  (FMAAPI * pfnCreateMsgMap)(HAREA, UMSG **, PULONG);
	BOOL  (FMAAPI * pfnEnumMsgs)(HAREA, PFNENUMMSGS, PVOID, ULONG);
	BOOL  (FMAAPI * pfnDeleteMsg)(HAREA, MSGN);
	BOOL  (FMAAPI * pfnSetCurMsg)(HAREA, MSGN);
	MSGN  (FMAAPI * pfnGetCurMsg)(HAREA);
	MSGN  (FMAAPI * pfnGetFirstMsg)(HAREA);
	MSGN  (FMAAPI * pfnGetHighMsg)(HAREA);
	LONG  (FMAAPI * pfnGetMsgCount)(HAREA);
	UMSG  (FMAAPI * pfnMsgnToUid)(HAREA, MSGN);
	MSGN  (FMAAPI * pfnUidToMsgn)(HAREA, UMSG, UINT);
	HMSG  (FMAAPI * pfnOpenMsg)(HAREA, UINT, MSGN, ULONG);
	BOOL  (FMAAPI * pfnCloseMsg)(HMSG);
	BOOL  (FMAAPI * pfnLoadMsg)(HMSG, PFMM *, ULONG);
	BOOL  (FMAAPI * pfnSaveMsg)(HMSG, PFMM, ULONG, ULONG);
	BOOL  (FMAAPI * pfnReadMsg)(HMSG, PVOID, PVOID, ULONG *, PCH *, ULONG *, ULONG *);
	BOOL  (FMAAPI * pfnWriteMsg)(HMSG, PVOID, PVOID, ULONG, PCH, ULONG, ULONG *, ULONG);
	BOOL  (FMAAPI * pfnGetMsgCtrlLen)(HMSG, ULONG *);
	BOOL  (FMAAPI * pfnGetMsgTextLen)(HMSG, ULONG *);
	BOOL  (FMAAPI * pfnGetLRInfo)(HAREA, ULONG, PLRINFO);
	BOOL  (FMAAPI * pfnSetLRInfo)(HAREA, ULONG, PLRINFO, PSZ);
  } API, FAR * PAPI;

  // Function prototypes

  FMAAPI_(BOOL) FmaRegisterApi(UINT type, PAPI (FMAAPI * pfnGetApi)(VOID));

/////////////////////////////////////////////////////////////////////////////
// Miscellaneous declarations
/////////////////////////////////////////////////////////////////////////////

  // Define newline string

#ifndef CRLF
#define CRLF  "\r\n"
#endif

  // Check if pointer points to newline string

#ifndef ISCRLF
#define ISCRLF(pch) ((pch)[0] == '\r' && (pch)[1] == '\n')
#endif

  // Check for linear white space character

#ifndef ISLWSP
#define ISLWSP(ch) ((ch) == ' ' || (ch) == '\t')
#endif

  // Check for end-of-line character

#ifndef ISEOL
#define ISEOL(ch)  ((ch) == '\r' || (ch) == '\n') 
#endif

  // Version info structure

  typedef struct _FMAVER { /* fmaVer */
	UINT    verMajor;                   // Major version number
	UINT    verMinor;                   // Minor version number
	UINT    verBuild;                   // Build number
	CHAR    achName[64];                // Build name
  } FMAVER, FAR * PFMAVER;

  // Initialization structure

  typedef struct _FMAINIT { /* fmaInit */
	PVOID  (FMAAPI * pfnMemAlloc)(size_t cb, BOOL bZero);
	PVOID  (FMAAPI * pfnMemRealloc)(PVOID p, size_t cb);
	VOID   (FMAAPI * pfnMemFree)(PVOID p);
  } FMAINIT, FAR * PFMAINIT;

  // Function prototypes

  FMAAPI_(BOOL) FmaInitialize(PFMAINIT pfmaInit);
  FMAAPI_(BOOL) FmaTerminate(VOID);
  FMAAPI_(BOOL) FmaGetVersion(PFMAVER pfmaVer);

/////////////////////////////////////////////////////////////////////////////
// Error codes returned by FmaGetLastError() 
/////////////////////////////////////////////////////////////////////////////

  // Common error codes

#define FMAERROR_NONE                   0
#define FMAERROR_INTERNAL_ERROR         1
#define FMAERROR_NOT_SUPPORTED          2
#define FMAERROR_INVALID_PARAM          3
#define FMAERROR_INVALID_DATA           4
#define FMAERROR_INVALID_SIZE           5
#define FMAERROR_NOT_ENOUGH_MEMORY      6
#define FMAERROR_NOT_FOUND              7
#define FMAERROR_ALREADY_EXISTS         8
#define FMAERROR_STILL_OPEN             9
#define FMAERROR_CANNOT_CREATE          10
#define FMAERROR_CANNOT_OPEN            11
#define FMAERROR_CANNOT_SEEK            12
#define FMAERROR_CANNOT_READ            13
#define FMAERROR_CANNOT_WRITE           14
#define FMAERROR_CANNOT_DELETE          15
#define FMAERROR_CANNOT_LOCK            16
#define FMAERROR_CANNOT_UNLOCK          17
#define FMAERROR_NO_DISK_SPACE          18
#define FMAERROR_NOT_LOCKED             19
#define FMAERROR_NOT_EXCLUSIVE          20
#define FMAERROR_NOT_CACHING            21
#define FMAERROR_FRAME_IN_USE           22
#define FMAERROR_INVALID_FRAME          23
#define FMAERROR_INVALID_INDEX          24
#define FMAERROR_ACCESS_DENIED          25
#define FMAERROR_MISSING_FILES          26
#define FMAERROR_INVALID_ADDRESS        27
#define FMAERROR_INVALID_KLUDGE         28
#define FMAERROR_INVALID_USERID         29
#define FMAERROR_NOT_REGISTERED         30
#define FMAERROR_DATA_SIZE_LIMIT        31

#define FMAERROR_CUSTOM_ERRORS_BASE     100

  // Function prototypes

  FMAAPI_(UINT) FmaSetLastError(UINT code);
  FMAAPI_(UINT) FmaGetLastError(VOID);

/////////////////////////////////////////////////////////////////////////////
// Global data
/////////////////////////////////////////////////////////////////////////////

  // Global initialization structure

#if defined(_FMA_DLL_BUILD) && !defined(__BORLANDC__)
  FMAGLOBAL_(FMAINIT) fmaInit;
#else
  extern FMAINIT fmaInit;
#endif

/////////////////////////////////////////////////////////////////////////////
// Header cleanup
/////////////////////////////////////////////////////////////////////////////

  // Terminate C/C++ frame

#ifdef __cplusplus
}
#endif

#endif  // #ifndef _FMA_API_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// End of FmaApi.h
/////////////////////////////////////////////////////////////////////////////
