/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Private header file
//
// FmaApiP.h
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 10/Feb/2000
//
// History: 17/Jan/1999 kvitekp - Initially created for IMA project.
//
//          10/Feb/2000 kvitekp - Added FmaAssertV()
//
//          22/Feb/2000 kvitekp - Added namespace compatibility #ifdefs
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _FMA_APIP_INCLUDED
#define _FMA_APIP_INCLUDED

  // Declare C/C++ frame

#ifdef __cplusplus
extern "C" {
#endif

  // Include private headers

#include "FmaDef.h"
#include "FmaSys.h"
#include "FmaUti.h"

/////////////////////////////////////////////////////////////////////////////
// Private data declarations
/////////////////////////////////////////////////////////////////////////////

  // Runtime info structure

  typedef struct _FMARTM { /* fmaRtm */
	LONG    nOpenAreas;                         // Number of open areas
#if _MT
#ifdef __W32__
	CRITICAL_SECTION csMsgId;                   // ^aMSGID stamp generation
#endif
#endif
  } FMARTM, FAR * PFMARTM;

  extern FMARTM fmaRtm;                         // Private runtime info

/////////////////////////////////////////////////////////////////////////////
// Define some useful macros 
/////////////////////////////////////////////////////////////////////////////

  // Endless loop until break

#ifndef loop
#define loop  for (;;)
#endif

  // Number of elements in array

#ifndef numbof
#define numbof(a)  (sizeof(a)/sizeof(a[0]))
#endif

  // Length of the const string

#ifndef lengof
#define lengof(s)  (numbof(s) - 1)
#endif

  // Structure field offset

#ifndef fieldoffs
#define fieldoffs(type, field)  ((int)&(((type *)0)->field))
#endif
  
  // Structure field size

#ifndef fieldsize
#define fieldsize(type, field)  ((int)sizeof(*(&(((type *)1)->field))))
#endif

  // Define bit manipulation macros

#ifndef SETFLAG
#define SETFLAG(f,b)  (f)|= (b)
#endif

#ifndef RESETFLAG
#define RESETFLAG(f,b)  (f)&=~(b)
#endif

#ifndef SETFLAGTO
#define SETFLAGTO(f,b,c)  if (c) SETFLAG(f,b); else RESETFLAG(f,b);
#endif

/////////////////////////////////////////////////////////////////////////////
// Debugging support
/////////////////////////////////////////////////////////////////////////////

  // Set error code macro

#define FMAERROR(err) FmaSetLastError(FMAERROR_##err)

  // Define the assertion macro

#ifdef ASSERT
#undef ASSERT
#endif

#ifdef NDEBUG
#  define ASSERT(p) ((void)0)
#else
#  define ASSERT(p) ((p) ? (void)0 : (void) FmaAssert(          \
					"Assertion failed: %s, file %s, line %d\n", \
					#p, __FILE__, __LINE__ ) )
#endif

#ifdef __W32__
#define FMA_BREAK_ON_FAILED_ASSERT
#endif

  // Define logging controls

#if !defined NDEBUG && !defined FMA_LOG
#define FMA_LOG
#endif

  // Function prototypes

  FMAAPIV_(VOID) FmaLog(PSZ pszFormat, ...);
  FMAAPI_(VOID)  FmaLogV(PSZ pszFormat, va_list parg);
  FMAAPIV_(VOID) FmaAssert(PSZ pszFormat, ...);
  FMAAPI_(VOID)  FmaAssertV(PSZ pszFormat, va_list parg);

  // Map functions to common names

#ifndef FMA_NO_FUNC_DEFS
#define Log        FmaLog
#define LogV       FmaLogV
#define LogAssert  FmaLogAssert
#define LogAssertV FmaLogAssertV
#endif

/////////////////////////////////////////////////////////////////////////////
// More compiler dependant stuff
/////////////////////////////////////////////////////////////////////////////

  // Macro to check for data size limit

#if UINT_MAX != ULONG_MAX
#define CHECK_DATA_SIZE_LIMIT(cb) ((ULONG)(cb) < (ULONG)(UINT)UINT_MAX)
#endif

  // Borland compilers bark if non pvoid parameter function pointer gets 
  // assigned to pvoid one so provide a pvoid casting to shut them down, 
  // while having proper type checking for other compilers

#ifdef __BORLANDC__
#define XPV (PVOID)
#else
#define XPV
#endif

/////////////////////////////////////////////////////////////////////////////
// Header cleanup
/////////////////////////////////////////////////////////////////////////////

  // Terminate C/C++ frame

#ifdef __cplusplus
}
#endif

#endif  // #ifndef _FMA_APIP_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// End of FmaApiP.h
/////////////////////////////////////////////////////////////////////////////
