/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: C++ header file
//
// FmaCpp.h
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 23/Mar/2000
//
// History: 27/Aug/1999 kvitekp - Initially created
//
//          23/Mar/2000 kvitekp - Added CFts, CSqu, CJam and CPkt
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _FMA_CPP_INCLUDED
#define _FMA_CPP_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// CFma class contains wrappers for Fma functions 
/////////////////////////////////////////////////////////////////////////////

class CFma 
{
public:

#ifdef _FMA_API_INCLUDED

	// Fidonet mail message functions

	static 
	PFMM CreateMsg(UINT cbExtra = 0) { 
		return ::FmaCreateMsg(cbExtra); 
	}

	static 
	BOOL DestroyMsg(PFMM pfmm) { 
		return ::FmaDestroyMsg(pfmm); 
	} 

	static 
	PFMM DupeMsg(PFMM pfmm) { 
		return ::FmaDupeMsg(pfmm); 
	}

	static 
	BOOL DumpMsg(PFMM pfmm) { 
		return ::FmaDumpMsg(pfmm); 
	}

	static 
	PFMP GetFmp(PFMM pfmm, ULONG id, UINT ifmp = 0) { 
		return ::FmaGetFmp(pfmm, id, ifmp); 
	}

	static 
	BOOL DelFmp(PFMM pfmm, ULONG id, UINT ifmp = 0) { 
		return ::FmaDelFmp(pfmm, id, ifmp); 
	}

	static 
	BOOL DumpFmp(PFMP pfmp) { 
		return ::FmaDumpFmp(pfmp); 
	}

	static 
	PFMP AddFmp(PFMM pfmm, ULONG id, PVOID pbData, ULONG cbData) {
		return ::FmaAddFmp(pfmm, id, pbData, cbData);
	} 

	static 
	PFMP AddFmpFormat(PFMM pfmm, ULONG id, LPCSTR pszFormat, ...) {
		va_list parg; va_start(parg, pszFormat);
		return ::FmaAddFmpFormatV(pfmm, id, const_cast<PSZ>(pszFormat), parg);
	}

	static 
	PFMP AddFmpFormatV(PFMM pfmm, ULONG id, LPCSTR pszFormat, va_list parg) {
		return ::FmaAddFmpFormatV(pfmm, id, const_cast<PSZ>(pszFormat), parg);
	}

	static 
	ULONG GetFmpTextLen(PFMP pfmp) {
		return ::FmaGetFmpTextLen(pfmp);
	}

	// Abstract message base api functions

	static 
	HAREA OpenArea(LPCSTR pszPath, UINT type, UINT fOpenMode, 
				   ULONG fl = 0, NETADDR * pnetAddr = NULL) {
		return ::FmaOpenArea(const_cast<PSZ>(pszPath), type, fOpenMode, fl, pnetAddr);
	}

	static 
	BOOL CloseArea(HAREA harea) {
		return ::FmaCloseArea(harea);
	}

	static 
	BOOL LockArea(HAREA harea, BOOL bLock) {
		return ::FmaLockArea(harea, bLock);
	}

	static 
	BOOL ValidateMsg(HAREA harea, MSGN msgn, PULONG pimsg = NULL) {
		return ::FmaValidateMsg(harea, msgn, pimsg);
	}

	static 
	BOOL CreateMsgMap(HAREA harea, UMSG * paumsg[], PULONG pcumsg) {
		return ::FmaCreateMsgMap(harea, paumsg, pcumsg);
	}
	
	static 
	BOOL EnumMsgs(HAREA harea, PFNENUMMSGS pfnCallback, 
				  PVOID pvCallbackData = NULL, ULONG fl = 0) {
		return ::FmaEnumMsgs(harea, pfnCallback, pvCallbackData, fl);
	}

	static 
	BOOL DeleteMsg(HAREA harea, MSGN msgn) {
		return ::FmaDeleteMsg(harea, msgn);
	}

	static 
	BOOL SetCurMsg(HAREA harea, MSGN msgn) {
		return ::FmaSetCurMsg(harea, msgn);
	}
	
	static 
	MSGN GetCurMsg(HAREA harea) {
		return ::FmaGetCurMsg(harea);
	}
	
	static 
	MSGN GetFirstMsg(HAREA harea) {
		return ::FmaGetFirstMsg(harea);
	}

	static 
	MSGN GetHighMsg(HAREA harea) {
		return ::FmaGetHighMsg(harea);
	}

	static 
	LONG GetMsgCount(HAREA harea) {
		return ::FmaGetMsgCount(harea);
	}

	static 
	UMSG MsgnToUid(HAREA harea, MSGN msgn) {
		return ::FmaMsgnToUid(harea, msgn);
	}
	
	static 
	MSGN UidToMsgn(HAREA harea, UMSG umsg, UINT fMode = UMSG_EXACT) {
		return ::FmaUidToMsgn(harea, umsg, fMode);
	}

	static 
	HMSG OpenMsg(HAREA harea, UINT mode, MSGN msgn, ULONG fl) {
		return ::FmaOpenMsg(harea, mode, msgn, fl);
	}

	static 
	BOOL CloseMsg(HMSG hmsg) {
		return ::FmaCloseMsg(hmsg);
	}

	static 
	BOOL LoadMsg(HMSG hmsg, PFMM * ppfmm, ULONG fl = 0) {
		return ::FmaLoadMsg(hmsg, ppfmm, fl);
	}

	static 
	BOOL SaveMsg(HMSG hmsg, PFMM pfmm, ULONG fl = 0, 
				 ULONG cchTextTotal = (ULONG)-1) {
		return ::FmaSaveMsg(hmsg, pfmm, fl, cchTextTotal);
	}

	static 
	BOOL ReadMsg(HMSG hmsg, PVOID pmsgh, 
				 PVOID * ppchCtrl = NULL, ULONG * pcchCtrl = NULL, 
				 PCH * ppchText = NULL, ULONG * pcchText = NULL, 
				 ULONG * poffText = NULL) {
		return ::FmaReadMsg(hmsg, pmsgh, ppchCtrl, pcchCtrl, ppchText, pcchText, poffText);
	}

	static 
	BOOL WriteMsg(HMSG hmsg, PVOID pmsgh, 
				  PVOID pchCtrl = NULL, ULONG cchCtrl = 0, 
				  PCH pchText = NULL, ULONG cchText = 0, ULONG * poffText = NULL, 
				  ULONG cchTextTotal = (ULONG)-1) {
		return ::FmaWriteMsg(hmsg, pmsgh, pchCtrl, cchCtrl, pchText, cchText, poffText, cchTextTotal);
	}

	static 
	BOOL GetMsgCtrlLen(HMSG hmsg, ULONG * pcchCtrl) {
		return ::FmaGetMsgCtrlLen(hmsg, pcchCtrl);
	}

	static 
	BOOL GetMsgTextLen(HMSG hmsg, ULONG * pcchText) {
		return ::FmaGetMsgTextLen(hmsg, pcchText);
	}

	static 
	BOOL GetLRInfo(HAREA harea, ULONG uid, PLRINFO plri) {
		return ::FmaGetLRInfo(harea, uid, plri);
	}

	static 
	BOOL SetLRInfo(HAREA harea, ULONG uid, PLRINFO plri, LPCSTR pszUserName = NULL) {
		return ::FmaSetLRInfo(harea, uid, plri, const_cast<PSZ>(pszUserName));
	}

	static 
	BOOL RegisterApi(UINT type, PAPI (FMAAPI * pfnGetApi)(VOID)) {
		return ::FmaRegisterApi(type, pfnGetApi);
	}

	static 
	BOOL Initialize(PFMAINIT pfmaInit = NULL) {
		return ::FmaInitialize(pfmaInit);
	}

	static 
	BOOL Terminate(VOID) {
		return ::FmaTerminate();
	}

	static 
	BOOL GetVersion(PFMAVER pfmaVer) {
		return ::FmaGetVersion(pfmaVer);
	}

	static 
	UINT SetLastError(UINT code = FMAERROR_NONE) {
		return ::FmaSetLastError(code);
	}

	static 
	UINT GetLastError(VOID) {
		return ::FmaGetLastError();
	}

#endif // #ifdef _FMA_API_INCLUDED

#ifdef _FMA_APIP_INCLUDED

	// Debugging functions

	static 
	void Log(LPCSTR pszFormat, ...) {
		va_list parg; va_start(parg, pszFormat);
		::FmaLogV(const_cast<PSZ>(pszFormat), parg);
	}

	static 
	void LogV(LPCSTR pszFormat, va_list parg) {
		::FmaLogV(const_cast<PSZ>(pszFormat), parg);
	}

	static 
	void Assert(LPCSTR pszFormat, ...) {
		va_list parg; va_start(parg, pszFormat);
		::FmaAssertV(const_cast<PSZ>(pszFormat), parg);
	}

	static 
	void AssertV(LPCSTR pszFormat, va_list parg) {
		::FmaAssertV(const_cast<PSZ>(pszFormat), parg);
	}
#endif // #ifdef _FMA_APIP_INCLUDED

#ifdef _FMA_SYS_INCLUDED

	// Memory management functions
  
	static 
	PVOID FmaAlloc(size_t cb, BOOL bZero = TRUE) {
		return ::FmaMemAlloc(cb, bZero);
	}

	static 
	PVOID FmaRealloc(PVOID p, size_t cb) {
		return ::FmaMemRealloc(p, cb);
	}
	
	static 
	void FmaFree(PVOID p) {
		if (p) ::FmaMemFree(p);
	}

	// File management functions

	static 
	BOOL OpenFile(LPCSTR pszFile, UINT fOpen, HFILE * phfile) {
		return ::FmaOpenFile(const_cast<PSZ>(pszFile), fOpen, phfile);
	}

	static 
	BOOL CloseFile(HFILE * phfile) {
		return ::FmaCloseFile(phfile);
	}

	static 
	BOOL ReadFile(HFILE hfile, LONG offs, PVOID pb, ULONG cb, 
				  ULONG * pcbRead = NULL) {
		return ::FmaReadFile(hfile, offs, pb, cb, pcbRead);
	}

	static 
	BOOL WriteFile(HFILE hfile, LONG offs, const PVOID pb, ULONG cb, 
				   ULONG * pcbWritten = NULL) {
		return ::FmaWriteFile(hfile, offs, pb, cb, pcbWritten);
	}
	
	static 
	LONG GetFilePos(HFILE hfile) {
		return ::FmaGetFilePos(hfile);
	}

	static 
	LONG SetFilePos(HFILE hfile, LONG offs) {
		return ::FmaSetFilePos(hfile, offs);
	}
	
	static 
	LONG GetFileSize(HFILE hfile) {
		return ::FmaGetFileSize(hfile);
	}

	static 
	BOOL SetFileSize(HFILE hfile, ULONG cbFile) {
		return ::FmaSetFileSize(hfile, cbFile);
	}
	
	static 
	BOOL Lock(HFILE hfile, LONG offs, LONG cb, BOOL bLock) {
		return ::FmaLock(hfile, offs, cb, bLock);
	}

	static 
	BOOL DeleteFile(LPCSTR pszFile) {
		return ::FmaDeleteFile(const_cast<PSZ>(pszFile));
	}

	static 
	BOOL EnumFiles(LPCSTR pszPath, LPCSTR pszMask, PFNENUMFILES pfnCallback, 
				   PVOID pvCallbackData = NULL, ULONG fl = 0) {
		return ::FmaEnumFiles(const_cast<PSZ>(pszPath), const_cast<PSZ>(pszMask), 
							  pfnCallback, pvCallbackData, fl);
	}

	static 
	BOOL FileExists(LPCSTR pszPath) {
		return ::FmaFileExists(const_cast<PSZ>(pszPath));
	}
	
	static 
	BOOL FmaDirExists(LPCSTR pszPath) {
		return ::FmaDirExists(const_cast<PSZ>(pszPath));
	}

	static 
	BOOL FmaMakeDir(LPCSTR pszPath) {
		return ::FmaMakeDir(const_cast<PSZ>(pszPath));
	}

	// Miscellaneous declarations

	static 
	void InterlockedIncrement(LONG * plv) {
		::FmaInterlockedIncrement(plv);
	}
	
	static 
	void InterlockedDecrement(LONG * plv) {
		::FmaInterlockedDecrement(plv);
	}

	static 
	BOOL GetLocalDttm(PDTTM pdttm) {
		return ::FmaGetLocalDttm(pdttm);
	}

	static 
	ULONG GetLocalTime(VOID) {
		return ::FmaGetLocalTime();
	}

	static 
	void Sleep(ULONG milliseconds = 0) {
		::FmaSleep(milliseconds);
	}
	
	static 
	ULONG GetMsgIdStamp(VOID) {
		return ::FmaGetMsgIdStamp();
	}

	static 
	UINT QueryOSType(VOID) {
		return ::FmaQueryOSType();
	}

	static 
	void GiveOutTimeSlice(VOID) {
		::FmaGiveOutTimeSlice();
	}

#endif // #ifdef _FMA_SYS_INCLUDED

#ifdef _FMA_UTI_INCLUDED

	static 
	PCH ScanNetAddr(NETADDR * pnetAddr, LPCSTR psz) {
		return ::FmaScanNetAddr(pnetAddr, const_cast<PSZ>(psz));
	}

	static 
	UINT FormatNetAddr(LPCSTR pch, UINT cchMax, const NETADDR * pnetAddr, LPCSTR pszDomain) {
		return ::FmaFormatNetAddr(const_cast<PCH>(pch), cchMax, 
								  const_cast<NETADDR*>(pnetAddr), 
								  const_cast<PSZ>(pszDomain));
	}

	static 
	BOOL CheckDttm(const PDTTM pdttm) {
		return ::FmaCheckDttm(pdttm);
	}
	
	static 
	BOOL AsciiToDttm(LPCSTR psz, PDTTM pdttm) {
		return ::FmaAsciiToDttm(const_cast<PSZ>(psz), pdttm);
	}

	static 
	ULONG GetTimeFromDttm(const PDTTM pdttm) {
		return ::FmaGetTimeFromDttm(pdttm);
	}

	static 
	BOOL GetDttmFromTime(PDTTM pdttm, ULONG time) {
		return ::FmaGetDttmFromTime(pdttm, time);
	}

	static 
	UINT FormatDttm(PCH pchDttm, UINT cchMax, const PDTTM pdttm, ULONG fl = 0, 
					CHAR chDateSep = 0, CHAR chTimeSep = 0, LPCSTR pszDateTimeSep = NULL) {
		return ::FmaFormatDttm(pchDttm, cchMax, const_cast<PDTTM>(pdttm), fl, 
							   chDateSep, chTimeSep, const_cast<PSZ>(pszDateTimeSep));
	}

	static 
	ULONG GetHash(LPCSTR lpsz) {
		return ::FmaGetHash(const_cast<PSZ>(lpsz));
	}

	static
	ULONG GetCrc32(const PBYTE pb, ULONG cb, ULONG crc = (ULONG)-1, BOOL bLowerCase = FALSE) {
		return ::FmaGetCrc32(const_cast<PBYTE>(pb), cb, crc, bLowerCase);
	}

	static
	ULONG GetCrc32(const PBYTE pb, const PBYTE pbEnd, ULONG crc = (ULONG)-1, BOOL bLowerCase = FALSE) {
		return ::FmaGetCrc32(const_cast<PBYTE>(pb), (ULONG)(pbEnd - pb), crc, bLowerCase);
	}

	static
	ULONG GetCrc32(LPCSTR lpsz, ULONG cb = (ULONG)-1, ULONG crc = (ULONG)-1, BOOL bLowerCase = FALSE) {
		return ::FmaGetCrc32((PBYTE)const_cast<PSZ>(lpsz), cb, crc, bLowerCase);
	}

#endif // #ifndef _FMA_UTI_INCLUDED
};

/////////////////////////////////////////////////////////////////////////////
// CFts class contains wrappers for Fts functions 
/////////////////////////////////////////////////////////////////////////////

class CFts
{
public:

#ifdef _FTS_API_INCLUDED

	// Fts api functions

	static 
	PAPI GetApi() {
		return ::FtsGetApi();
	}

	static 
	HAREA OpenArea(LPCSTR pszPath, UINT fOpenMode, 
				   ULONG fl = 0, NETADDR * pnetAddr = NULL) {
		return ::FtsOpenArea(const_cast<PSZ>(pszPath), fOpenMode, fl, pnetAddr);
	}

	static 
	BOOL CloseArea(HAREA harea) {
		return ::FtsCloseArea(harea);
	}

	static 
	BOOL LockArea(HAREA harea, BOOL bLock) {
		return ::FtsLockArea(harea, bLock);
	}

	static 
	BOOL ValidateMsg(HAREA harea, MSGN msgn, PULONG pimsg = NULL) {
		return ::FtsValidateMsg(harea, msgn, pimsg);
	}

	static 
	BOOL CreateMsgMap(HAREA harea, UMSG * paumsg[], PULONG pcumsg) {
		return ::FtsCreateMsgMap(harea, paumsg, pcumsg);
	}
	
	static 
	BOOL EnumMsgs(HAREA harea, PFNENUMMSGS pfnCallback, 
				  PVOID pvCallbackData = NULL, ULONG fl = 0) {
		return ::FtsEnumMsgs(harea, pfnCallback, pvCallbackData, fl);
	}

	static 
	BOOL DeleteMsg(HAREA harea, MSGN msgn) {
		return ::FtsDeleteMsg(harea, msgn);
	}

	static 
	BOOL SetCurMsg(HAREA harea, MSGN msgn) {
		return ::FtsSetCurMsg(harea, msgn);
	}
	
	static 
	MSGN GetCurMsg(HAREA harea) {
		return ::FtsGetCurMsg(harea);
	}
	
	static 
	MSGN GetFirstMsg(HAREA harea) {
		return ::FtsGetFirstMsg(harea);
	}

	static 
	MSGN GetHighMsg(HAREA harea) {
		return ::FtsGetHighMsg(harea);
	}

	static 
	LONG GetMsgCount(HAREA harea) {
		return ::FtsGetMsgCount(harea);
	}

	static 
	UMSG MsgnToUid(HAREA harea, MSGN msgn) {
		return ::FtsMsgnToUid(harea, msgn);
	}
	
	static 
	MSGN UidToMsgn(HAREA harea, UMSG umsg, UINT fMode = UMSG_EXACT) {
		return ::FtsUidToMsgn(harea, umsg, fMode);
	}

	static 
	HMSG OpenMsg(HAREA harea, UINT mode, MSGN msgn, ULONG fl) {
		return ::FtsOpenMsg(harea, mode, msgn, fl);
	}

	static 
	BOOL CloseMsg(HMSG hmsg) {
		return ::FtsCloseMsg(hmsg);
	}

	static 
	BOOL LoadMsg(HMSG hmsg, PFMM * ppfmm, ULONG fl = 0) {
		return ::FtsLoadMsg(hmsg, ppfmm, fl);
	}

	static 
	BOOL SaveMsg(HMSG hmsg, PFMM pfmm, ULONG fl = 0, 
				 ULONG cchTextTotal = (ULONG)-1) {
		return ::FtsSaveMsg(hmsg, pfmm, fl, cchTextTotal);
	}

	static 
	BOOL ReadMsg(HMSG hmsg, PFMH pfmh, 
				 PCH * ppchCtrl = NULL, ULONG * pcchCtrl = NULL, 
				 PCH * ppchText = NULL, ULONG * pcchText = NULL, 
				 ULONG * poffText = NULL) {
		return ::FtsReadMsg(hmsg, pfmh, ppchCtrl, pcchCtrl, ppchText, pcchText, poffText);
	}

	static 
	BOOL WriteMsg(HMSG hmsg, PFMH pfmh, 
				  PCH pchCtrl = NULL, ULONG cchCtrl = 0, 
				  PCH pchText = NULL, ULONG cchText = 0, ULONG * poffText = NULL, 
				  ULONG cchTextTotal = (ULONG)-1) {
		return ::FtsWriteMsg(hmsg, pfmh, pchCtrl, cchCtrl, pchText, cchText, poffText, cchTextTotal);
	}

	static 
	BOOL GetMsgCtrlLen(HMSG hmsg, ULONG * pcchCtrl) {
		return ::FtsGetMsgCtrlLen(hmsg, pcchCtrl);
	}

	static 
	BOOL GetMsgTextLen(HMSG hmsg, ULONG * pcchText) {
		return ::FtsGetMsgTextLen(hmsg, pcchText);
	}

	static 
	BOOL GetLRInfo(HAREA harea, ULONG uid, PLRINFO plri) {
		return ::FtsGetLRInfo(harea, uid, plri);
	}

	static 
	BOOL SetLRInfo(HAREA harea, ULONG uid, PLRINFO plri, LPCSTR pszUserName = NULL) {
		return ::FtsSetLRInfo(harea, uid, plri, const_cast<PSZ>(pszUserName));
	}

#endif // #ifdef _FTS_API_INCLUDED
};

/////////////////////////////////////////////////////////////////////////////
// CSqu class contains wrappers for Squish functions 
/////////////////////////////////////////////////////////////////////////////

class CSqu
{
public:

#ifdef _SQU_API_INCLUDED

	// Squish api functions

	static 
	PAPI GetApi() {
		return ::SquGetApi();
	}

	static 
	HAREA OpenArea(LPCSTR pszPath, UINT fOpenMode, 
				   ULONG fl = 0, NETADDR * pnetAddr = NULL) {
		return ::SquOpenArea(const_cast<PSZ>(pszPath), fOpenMode, fl, pnetAddr);
	}

	static 
	BOOL CloseArea(HAREA harea) {
		return ::SquCloseArea(harea);
	}

	static 
	BOOL LockArea(HAREA harea, BOOL bLock) {
		return ::SquLockArea(harea, bLock);
	}

	static 
	BOOL ValidateMsg(HAREA harea, MSGN msgn, PULONG pimsg = NULL) {
		return ::SquValidateMsg(harea, msgn, pimsg);
	}

	static 
	BOOL CreateMsgMap(HAREA harea, UMSG * paumsg[], PULONG pcumsg) {
		return ::SquCreateMsgMap(harea, paumsg, pcumsg);
	}
	
	static 
	BOOL EnumMsgs(HAREA harea, PFNENUMMSGS pfnCallback, 
				  PVOID pvCallbackData = NULL, ULONG fl = 0) {
		return ::SquEnumMsgs(harea, pfnCallback, pvCallbackData, fl);
	}

	static 
	BOOL DeleteMsg(HAREA harea, MSGN msgn) {
		return ::SquDeleteMsg(harea, msgn);
	}

	static 
	BOOL SetCurMsg(HAREA harea, MSGN msgn) {
		return ::SquSetCurMsg(harea, msgn);
	}
	
	static 
	MSGN GetCurMsg(HAREA harea) {
		return ::SquGetCurMsg(harea);
	}
	
	static 
	MSGN GetFirstMsg(HAREA harea) {
		return ::SquGetFirstMsg(harea);
	}

	static 
	MSGN GetHighMsg(HAREA harea) {
		return ::SquGetHighMsg(harea);
	}

	static 
	LONG GetMsgCount(HAREA harea) {
		return ::SquGetMsgCount(harea);
	}

	static 
	UMSG MsgnToUid(HAREA harea, MSGN msgn) {
		return ::SquMsgnToUid(harea, msgn);
	}
	
	static 
	MSGN UidToMsgn(HAREA harea, UMSG umsg, UINT fMode = UMSG_EXACT) {
		return ::SquUidToMsgn(harea, umsg, fMode);
	}

	static 
	HMSG OpenMsg(HAREA harea, UINT mode, MSGN msgn, ULONG fl) {
		return ::SquOpenMsg(harea, mode, msgn, fl);
	}

	static 
	BOOL CloseMsg(HMSG hmsg) {
		return ::SquCloseMsg(hmsg);
	}

	static 
	BOOL LoadMsg(HMSG hmsg, PFMM * ppfmm, ULONG fl = 0) {
		return ::SquLoadMsg(hmsg, ppfmm, fl);
	}

	static 
	BOOL SaveMsg(HMSG hmsg, PFMM pfmm, ULONG fl = 0, 
				 ULONG cchTextTotal = (ULONG)-1) {
		return ::SquSaveMsg(hmsg, pfmm, fl, cchTextTotal);
	}

	static 
	BOOL ReadMsg(HMSG hmsg, PSMH psmh, 
				 PCH * ppchCtrl = NULL, ULONG * pcchCtrl = NULL, 
				 PCH * ppchText = NULL, ULONG * pcchText = NULL, 
				 ULONG * poffText = NULL) {
		return ::SquReadMsg(hmsg, psmh, ppchCtrl, pcchCtrl, ppchText, pcchText, poffText);
	}

	static 
	BOOL WriteMsg(HMSG hmsg, PSMH psmh, 
				  PCH pchCtrl = NULL, ULONG cchCtrl = 0, 
				  PCH pchText = NULL, ULONG cchText = 0, ULONG * poffText = NULL, 
				  ULONG cchTextTotal = (ULONG)-1) {
		return ::SquWriteMsg(hmsg, psmh, pchCtrl, cchCtrl, pchText, cchText, poffText, cchTextTotal);
	}

	static 
	BOOL GetMsgCtrlLen(HMSG hmsg, ULONG * pcchCtrl) {
		return ::SquGetMsgCtrlLen(hmsg, pcchCtrl);
	}

	static 
	BOOL GetMsgTextLen(HMSG hmsg, ULONG * pcchText) {
		return ::SquGetMsgTextLen(hmsg, pcchText);
	}

	static 
	BOOL GetLRInfo(HAREA harea, ULONG uid, PLRINFO plri) {
		return ::SquGetLRInfo(harea, uid, plri);
	}

	static 
	BOOL SetLRInfo(HAREA harea, ULONG uid, PLRINFO plri, LPCSTR pszUserName = NULL) {
		return ::SquSetLRInfo(harea, uid, plri, const_cast<PSZ>(pszUserName));
	}

#endif // #ifdef _SQU_API_INCLUDED
};

/////////////////////////////////////////////////////////////////////////////
// CJam class contains wrappers for JAM functions 
/////////////////////////////////////////////////////////////////////////////

class CJam
{
public:

#ifdef _JAM_API_INCLUDED

	// JAM api functions

	static 
	PAPI GetApi() {
		return ::JamGetApi();
	}

	static 
	HAREA OpenArea(LPCSTR pszPath, UINT fOpenMode, 
				   ULONG fl = 0, NETADDR * pnetAddr = NULL) {
		return ::JamOpenArea(const_cast<PSZ>(pszPath), fOpenMode, fl, pnetAddr);
	}

	static 
	BOOL CloseArea(HAREA harea) {
		return ::JamCloseArea(harea);
	}

	static 
	BOOL LockArea(HAREA harea, BOOL bLock) {
		return ::JamLockArea(harea, bLock);
	}

	static 
	BOOL ValidateMsg(HAREA harea, MSGN msgn, PULONG pimsg = NULL) {
		return ::JamValidateMsg(harea, msgn, pimsg);
	}

	static 
	BOOL CreateMsgMap(HAREA harea, UMSG * paumsg[], PULONG pcumsg) {
		return ::JamCreateMsgMap(harea, paumsg, pcumsg);
	}
	
	static 
	BOOL EnumMsgs(HAREA harea, PFNENUMMSGS pfnCallback, 
				  PVOID pvCallbackData = NULL, ULONG fl = 0) {
		return ::JamEnumMsgs(harea, pfnCallback, pvCallbackData, fl);
	}

	static 
	BOOL DeleteMsg(HAREA harea, MSGN msgn) {
		return ::JamDeleteMsg(harea, msgn);
	}

	static 
	BOOL SetCurMsg(HAREA harea, MSGN msgn) {
		return ::JamSetCurMsg(harea, msgn);
	}
	
	static 
	MSGN GetCurMsg(HAREA harea) {
		return ::JamGetCurMsg(harea);
	}
	
	static 
	MSGN GetFirstMsg(HAREA harea) {
		return ::JamGetFirstMsg(harea);
	}

	static 
	MSGN GetHighMsg(HAREA harea) {
		return ::JamGetHighMsg(harea);
	}

	static 
	LONG GetMsgCount(HAREA harea) {
		return ::JamGetMsgCount(harea);
	}

	static 
	UMSG MsgnToUid(HAREA harea, MSGN msgn) {
		return ::JamMsgnToUid(harea, msgn);
	}
	
	static 
	MSGN UidToMsgn(HAREA harea, UMSG umsg, UINT fMode = UMSG_EXACT) {
		return ::JamUidToMsgn(harea, umsg, fMode);
	}

	static 
	HMSG OpenMsg(HAREA harea, UINT mode, MSGN msgn, ULONG fl) {
		return ::JamOpenMsg(harea, mode, msgn, fl);
	}

	static 
	BOOL CloseMsg(HMSG hmsg) {
		return ::JamCloseMsg(hmsg);
	}

	static 
	BOOL LoadMsg(HMSG hmsg, PFMM * ppfmm, ULONG fl = 0) {
		return ::JamLoadMsg(hmsg, ppfmm, fl);
	}

	static 
	BOOL SaveMsg(HMSG hmsg, PFMM pfmm, ULONG fl = 0, 
				 ULONG cchTextTotal = (ULONG)-1) {
		return ::JamSaveMsg(hmsg, pfmm, fl, cchTextTotal);
	}

	static 
	BOOL ReadMsg(HMSG hmsg, PJMH pjmh, 
				 PJSF * ppjsf = NULL, ULONG * pcbjsf = NULL,
				 PCH * ppchText = NULL, ULONG * pcchText = NULL, 
				 ULONG * poffText = NULL) {
		return ::JamReadMsg(hmsg, pjmh, ppjsf, pcbjsf, ppchText, pcchText, poffText);
	}

	static 
	BOOL WriteMsg(HMSG hmsg, PJMH pjmh, 
				  PJSF pjsf = NULL, ULONG cbjsf = 0,
				  PCH pchText = NULL, ULONG cchText = 0, ULONG * poffText = NULL, 
				  ULONG cchTextTotal = (ULONG)-1) {
		return ::JamWriteMsg(hmsg, pjmh, pjsf, cbjsf, pchText, cchText, poffText, cchTextTotal);
	}

	static 
	BOOL GetMsgCtrlLen(HMSG hmsg, ULONG * pcchCtrl) {
		return ::JamGetMsgCtrlLen(hmsg, pcchCtrl);
	}

	static 
	BOOL GetMsgTextLen(HMSG hmsg, ULONG * pcchText) {
		return ::JamGetMsgTextLen(hmsg, pcchText);
	}

	static 
	BOOL GetLRInfo(HAREA harea, ULONG uid, PLRINFO plri) {
		return ::JamGetLRInfo(harea, uid, plri);
	}

	static 
	BOOL SetLRInfo(HAREA harea, ULONG uid, PLRINFO plri, LPCSTR pszUserName = NULL) {
		return ::JamSetLRInfo(harea, uid, plri, const_cast<PSZ>(pszUserName));
	}

#endif // #ifdef _JAM_API_INCLUDED
};

/////////////////////////////////////////////////////////////////////////////
// CPkt class contains wrappers for Pkt functions 
/////////////////////////////////////////////////////////////////////////////

class CPkt
{
public:

#ifdef _PKT_API_INCLUDED

	// Pkt api functions

	static 
	PAPI GetApi() {
		return ::PktGetApi();
	}

	static 
	HAREA OpenArea(LPCSTR pszPath, UINT fOpenMode, 
				   ULONG fl = 0, NETADDR * pnetAddr = NULL) {
		return ::PktOpenArea(const_cast<PSZ>(pszPath), fOpenMode, fl, pnetAddr);
	}

	static 
	BOOL CloseArea(HAREA harea) {
		return ::PktCloseArea(harea);
	}

	static 
	BOOL LockArea(HAREA harea, BOOL bLock) {
		return ::PktLockArea(harea, bLock);
	}

	static 
	BOOL ValidateMsg(HAREA harea, MSGN msgn, PULONG pimsg = NULL) {
		return ::PktValidateMsg(harea, msgn, pimsg);
	}

	static 
	BOOL CreateMsgMap(HAREA harea, UMSG * paumsg[], PULONG pcumsg) {
		return ::PktCreateMsgMap(harea, paumsg, pcumsg);
	}
	
	static 
	BOOL EnumMsgs(HAREA harea, PFNENUMMSGS pfnCallback, 
				  PVOID pvCallbackData = NULL, ULONG fl = 0) {
		return ::PktEnumMsgs(harea, pfnCallback, pvCallbackData, fl);
	}

	static 
	BOOL DeleteMsg(HAREA harea, MSGN msgn) {
		return ::PktDeleteMsg(harea, msgn);
	}

	static 
	BOOL SetCurMsg(HAREA harea, MSGN msgn) {
		return ::PktSetCurMsg(harea, msgn);
	}
	
	static 
	MSGN GetCurMsg(HAREA harea) {
		return ::PktGetCurMsg(harea);
	}
	
	static 
	MSGN GetFirstMsg(HAREA harea) {
		return ::PktGetFirstMsg(harea);
	}

	static 
	MSGN GetHighMsg(HAREA harea) {
		return ::PktGetHighMsg(harea);
	}

	static 
	LONG GetMsgCount(HAREA harea) {
		return ::PktGetMsgCount(harea);
	}

	static 
	UMSG MsgnToUid(HAREA harea, MSGN msgn) {
		return ::PktMsgnToUid(harea, msgn);
	}
	
	static 
	MSGN UidToMsgn(HAREA harea, UMSG umsg, UINT fMode = UMSG_EXACT) {
		return ::PktUidToMsgn(harea, umsg, fMode);
	}

	static 
	HMSG OpenMsg(HAREA harea, UINT mode, MSGN msgn, ULONG fl) {
		return ::PktOpenMsg(harea, mode, msgn, fl);
	}

	static 
	BOOL CloseMsg(HMSG hmsg) {
		return ::PktCloseMsg(hmsg);
	}

	static 
	BOOL LoadMsg(HMSG hmsg, PFMM * ppfmm, ULONG fl = 0) {
		return ::PktLoadMsg(hmsg, ppfmm, fl);
	}

	static 
	BOOL SaveMsg(HMSG hmsg, PFMM pfmm, ULONG fl = 0, 
				 ULONG cchTextTotal = (ULONG)-1) {
		return ::PktSaveMsg(hmsg, pfmm, fl, cchTextTotal);
	}

	static 
	BOOL ReadMsg(HMSG hmsg, PFMH pfmh, 
				 PCH * ppchCtrl = NULL, ULONG * pcchCtrl = NULL, 
				 PCH * ppchText = NULL, ULONG * pcchText = NULL, 
				 ULONG * poffText = NULL) {
		return ::PktReadMsg(hmsg, pfmh, ppchCtrl, pcchCtrl, ppchText, pcchText, poffText);
	}

	static 
	BOOL WriteMsg(HMSG hmsg, PFMH pfmh, 
				  PCH pchCtrl = NULL, ULONG cchCtrl = 0, 
				  PCH pchText = NULL, ULONG cchText = 0, ULONG * poffText = NULL, 
				  ULONG cchTextTotal = (ULONG)-1) {
		return ::PktWriteMsg(hmsg, pfmh, pchCtrl, cchCtrl, pchText, cchText, poffText, cchTextTotal);
	}

	static 
	BOOL GetMsgCtrlLen(HMSG hmsg, ULONG * pcchCtrl) {
		return ::PktGetMsgCtrlLen(hmsg, pcchCtrl);
	}

	static 
	BOOL GetMsgTextLen(HMSG hmsg, ULONG * pcchText) {
		return ::PktGetMsgTextLen(hmsg, pcchText);
	}

	static 
	BOOL GetLRInfo(HAREA harea, ULONG uid, PLRINFO plri) {
		return ::PktGetLRInfo(harea, uid, plri);
	}

	static 
	BOOL SetLRInfo(HAREA harea, ULONG uid, PLRINFO plri, LPCSTR pszUserName = NULL) {
		return ::PktSetLRInfo(harea, uid, plri, const_cast<PSZ>(pszUserName));
	}

#endif // #ifdef _PKT_API_INCLUDED
};

#endif  // #ifndef _FMA_CPP_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// End of FmaApi.h
/////////////////////////////////////////////////////////////////////////////
