/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Control info routines module
//
// FmaCtrl.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 26/Sep/1999
// Updated: 08/Sep/2000
//
// History: 01/Aug/1992 kvitekp - Initially created for MadMED
//
//          29/Jan/2000 kvitekp - Fixed space between ^aAREA: and area tag
//
//          30/Jan/2000 kvitekp - Fixed bug in FmaLoadAddrKludges which
//                                sometimes caused zero origination zone
//
//          11/Feb/2000 kvitekp - ^aAREA: is now always the first kludge
//
//          08/Sep/2000 kvitekp - Only 1st of FMP_ONE kludges will be loaded
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"
#include "FmaCtrl.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // aFLAGS kludge flags mapping table entry macro

#define KFLG_(attr, psz) { attr, psz },

  // aFLAGS kludge flags mapping table entry structure

  typedef struct _KFLGMAP {
	ULONG   attr;               // Fidonet message attribute (MA_)
	PSZ     psz;                // ^aFLAGS kludge flag
  } KFLGMAP, FAR * PKFLGMAP;

  // ^aFLAGS kludge flags mapping table. Note: these are all the flags 
  //  we recognize when loading a message from any type of message base.
  //  Any other flag will be placed into FMP_FLAGS property.

  static KFLGMAP aFlag[] = {
//  KFLG_(MA_LOCAL,              NULL)
//  KFLG_(MA_INTRANSIT,          NULL)
	KFLG_(MA_PRIVATE,           "PVT")
	KFLG_(MA_READ,              "RCV")
	KFLG_(MA_SENT,              "SNT")
	KFLG_(MA_KILLSENT,          "K/S")
	KFLG_(MA_ARCHIVESENT,       "A/S")
	KFLG_(MA_HOLD,              "HLD")
	KFLG_(MA_CRASH,             "CRA")
	KFLG_(MA_IMMEDIATE,         "IMM")
	KFLG_(MA_DIRECT,            "DIR")
	KFLG_(MA_GATE,              "ZON")
	KFLG_(MA_FILEREQUEST,       "FRQ")
	KFLG_(MA_FILEATTACH,        "FIL")
	KFLG_(MA_TRUNCFILE,         "TFS")
	KFLG_(MA_KILLFILE,          "KFS")
	KFLG_(MA_RECEIPTREQ,        "RRQ")
	KFLG_(MA_CONFIRMREQ,        "CFM")
//  KFLG_(MA_ORPHAN,             NULL)
	KFLG_(MA_ENCRYPT,           "ENC")
	KFLG_(MA_COMPRESS,          "CMP")
	KFLG_(MA_ESCAPED,           "ESC")
	KFLG_(MA_FORCEPICKUP,       "FPU")
//  KFLG_(MA_TYPELOCAL,          NULL)
//  KFLG_(MA_TYPEECHO,           NULL)
//  KFLG_(MA_TYPENET,            NULL)
//  KFLG_(MA_SCANNED,            NULL)
//  KFLG_(MA_FILEUPDATEREQ,      NULL)
	KFLG_(MA_AUDITREQUEST,      "ARQ")
	KFLG_(MA_NODISPLAY,         "NOD")
	KFLG_(MA_LOCKED,            "LOK")
//  KFLG_(MA_DELETED,            NULL)
  };

  // Kludge mapping table table used both when 
  // loading and saving messages

  static KLMAP aCtrl[] = {
	KLMAP_(0,              "FMPT ",      KLI_FMPT)
	KLMAP_(0,              "TOPT ",      KLI_TOPT)
	KLMAP_(0,              "INTL ",      KLI_INTL)
	KLMAP_(0,              "DOMAIN ",    KLI_DOMAIN)
	KLMAP_(0,              "FLAGS ",     KLI_FLAGS)
	KLMAP_(FMP_AREA,       "AREA:",      0)
	KLMAP_(FMP_MSGID,      "MSGID:",     KLI_MSGID)
	KLMAP_(FMP_REPLYID,    "REPLY:",     KLI_REPLY)
	KLMAP_(FMP_PID,        "PID:",       0)
	KLMAP_(FMP_TID,        "TID:",       0)
	KLMAP_(FMP_TZUTC,      "TZUTC:",     0)
	KLMAP_(FMP_CHARSET,    "CHRS:",      0)
	KLMAP_(FMP_REALNAME,   "REALNAME:",  0)
	KLMAP_(FMP_REPLYADDR,  "REPLYADDR:", 0)
	KLMAP_(FMP_REPLYTO,    "REPLYTO:",   0)
  };

  // End of message kludge mapping table table 
  // used both when loading and saving messages

  static KLMAP aEom[] = {
	KLMAP_(FMP_SEENBY2D,   "SEEN-BY:",   KLI_NOPREFIX)
	KLMAP_(FMP_PATH2D,     "PATH:",      0)
	KLMAP_(FMP_EOMKLUDGE,  "Via ",       0)
	KLMAP_(FMP_EOMKLUDGE,  "Recd ",      0)
	KLMAP_(FMP_EOMKLUDGE,  "Forwarded ", 0)
  };

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine scans until the next space
//

  static UINT FMASUB DoScanToSpace(PCH * ppch, UINT * pcch)
  {
	UINT cchNonSpace = 0;
	UINT cch = *pcch;
	PCH  pch = *ppch;

	// Scan non spaces and check if ok

	while (cch && !ISLWSP(*pch)) { pch++; cch--; cchNonSpace++; }

	// Update pointer and counter for caller

	*ppch = pch; *pcch = cch;

	return cchNonSpace;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine compares net/node of two network addresses
//

  static BOOL FMASUB DoEqNN(NETADDR * pnetAddr1, NETADDR * pnetAddr2)
  {
	return pnetAddr1->net  == pnetAddr2->net  &&
		   pnetAddr1->node == pnetAddr2->node;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine compares net/node/point of two network addresses
//

  static BOOL FMASUB DoEqNNP(NETADDR * pnetAddr1, NETADDR * pnetAddr2)
  {
	return pnetAddr1->net == pnetAddr2->net &&
		   pnetAddr1->node == pnetAddr2->node &&
		   pnetAddr1->point == pnetAddr2->point;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine copies zone/net/node from one network address to another
//

  static VOID FMASUB DoCopyZNN(NETADDR * pnetAddr1, NETADDR * pnetAddr2)
  {
	pnetAddr1->zone = pnetAddr2->zone;
	pnetAddr1->net  = pnetAddr2->net ;
	pnetAddr1->node = pnetAddr2->node;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds end of message kludge
//

  static BOOL FMASUB DoAddEomKludge(PFMM pfmm, PCH pch, UINT cch, BOOL bHasPrefix)
  {
	PKLMAP pklme;
	ULONG id;

	// Check if this is a known eom kludge and and get associated propery id,
	// otherwise assume generic eom kludge property id

	if ((pklme = FmaGetEomKludgeFromString(pch, cch, bHasPrefix)) == NULL)
	  id = FMP_EOMKLUDGE;
	else
	  id = pklme->id;

	// Check if the kludge line has prefix (SEEN-BYs don't!) and skip over it

	if (pch[0] == '\01') {
	  pch++; cch--;
	}

	// Check if this eom kludge has a specific id and 
	// skip over it's name

	if (id != FMP_EOMKLUDGE) {
	  pch+= pklme->cch;
	  cch-= pklme->cch;
	}

	// Strip leading spaces if any

	while (cch && ISLWSP(pch[0])) {
	  pch++; cch--;
	}

	// Strip trailing spaces if any

	while (cch && ISLWSP(pch[cch - 1])) {
	  cch--;
	}

	// Add this eom kludge and check if ok

	if (!FmaAddFmp(pfmm, id, pch, cch))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads ^aFLAGS kludge flag
//

  static BOOL FMASUB DoLoadFlag(PFMM pfmm, PCH pch, UINT cch)
  {
	UINT iFlag;

	// Scan the flags mapping table looking for a matching flag and 
	// if found, set the appropriate fidonet mail message attribute

	for (iFlag = 0; iFlag < numbof(aFlag); iFlag++) {
	  if (aFlag[iFlag].psz && 
		  FmaEqPch(aFlag[iFlag].psz, pch, cch)) {
		SETFLAG(pfmm->attr, aFlag[iFlag].attr);
		return TRUE;
	  }
	}

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine scans ^aDOMAIN <dDomain> <dAddr> <oDomain> <oAddr>
//

  static BOOL FMASUB DoScanDOMAIN(PFMM pfmm, PKLI pkli)
  {
	struct {
	  PCH pch; UINT cch;
	} a[4];

	UINT cch, i;
	PCH pch;

	// Get kludge data on stack

	cch = pkli->cch[KLI_DOMAIN]; 
	pch = pkli->pch[KLI_DOMAIN];

	// Check if we have the domain kludge info

	if (!pch || !cch) return FALSE;

	// Scan in ^aDOMAIN info in four steps

	for (i = 0; i < 4; i++) {

	  // Skip spaces and check for more

	  while (cch && ISLWSP(*pch)) { pch++; cch--; }
	  if (!cch) return FALSE;

	  // Scan in this step info and check if ok

	  a[i].pch = pch;
	  a[i].cch = DoScanToSpace(&pch, &cch);
	  if (!a[i].cch) return FALSE;
	}

	// Add destination address property and check if ok

	if (!FmaAddFmpFormat(pfmm, FMP_DESTADDR, "%.*s@%.*s", 
		 a[1].cch, a[1].pch, a[0].cch, a[0].pch))
	  return FALSE;

	// Add origination address property and check if ok

	if (!FmaAddFmpFormat(pfmm, FMP_ORIGADDR, "%.*s@%.*s", 
		 a[3].cch, a[3].pch, a[2].cch, a[2].pch))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine copies zone/net/node from one network address to another
//

  static PSZ FMASUB DoGetKludgeFromFmp(PFMP pfmp)
  {
	UINT iCtrl;

	ASSERT((FMP_FLG(pfmp->id) & FMP_KLU) == FMP_KLU);

	// Scan the known kludges table looking for an id match

	for (iCtrl = 0; iCtrl < numbof(aCtrl); iCtrl++) {
	  if (aCtrl[iCtrl].id == pfmp->id)
		return aCtrl[iCtrl].psz;
	}

	return NULL;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns property for the specified kludge
//

  static PFMP FMASUB DoGetFmpFromKludge(PFMM pfmm, PSZ pszKludge)
  {
	UINT ifmp, cchKludge = strlen(pszKludge);
	PFMP pfmp;

	// Loop through all the fts kludge properties

	for (ifmp = 0; (pfmp = FmaGetFmp(pfmm, FMP_FTSKLUDGE, ifmp)) != NULL; ifmp++) {
	  if (pfmp->cb >= cchKludge && FmaEqPch(PFMPDATA(pfmp), pszKludge, cchKludge)) 
		return pfmp;
	}

	return NULL;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine gets the binary network address from teh given property
//

  FMAAPI_(BOOL) FmaGetNetAddr(PFMM pfmm, ULONG id, NETADDR * pnetAddr,
							  PCH * ppchDomain)
  {
	PFMP pfmp;
	PCH pch;

	// Get the specified property and check if ok

	if ((pfmp = FmaGetFmp(pfmm, id, 0)) == NULL) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Scan the network address and check if ok

	if ((pch = FmaScanNetAddr(pnetAddr, PFMPDATA(pfmp))) == NULL) {
	  FMAERROR(INVALID_ADDRESS);
	  return FALSE;
	}

	// Check if caller needs domain and pass it 

	if (ppchDomain) *ppchDomain = (*pch == '@') ? pch : NULL; 

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads message header
//

  FMAAPI_(BOOL) FmaLoadMhdr(PFMM pfmm, PVOID pmhdr, MHDRMAP aMhdr[], UINT nMhdr)
  {
	CHAR ach[64];
	UINT iMhdr;
	DTTM dttm;
	ULONG lv;
	PVOID pb;
	UINT cb;

	// Scan the message header mapping table adding properties

	for (iMhdr = 0; iMhdr < nMhdr; iMhdr++) {

	  // Calculate field address in the message header

	  pb = ((PBYTE)pmhdr) + aMhdr[iMhdr].offs;

	  // Process field according to its type

	  switch (aMhdr[iMhdr].type) {

		case MHDR_UL32: 
		  cb = sizeof(ULONG);
		  break;

		case MHDR_US16: 
		  lv = (ULONG)(*((PUSHORT)pb));
		  pb = &lv;
		  cb = sizeof(ULONG);
		  break;

		case MHDR_DTTM: 
		  cb = sizeof(DTTM);
		  break;

		case MHDR_TIME: 
		  FmaGetDttmFromTime(&dttm, *((PULONG)pb));
		  pb = &dttm;
		  cb = sizeof(DTTM);
		  break;

		case MHDR_ASCZ: 
		  cb = strlen(pb);
		  break;

		case MHDR_ADDR: 
		  cb = FmaFormatNetAddr(ach, lengof(ach), (NETADDR *)pb, NULL);
		  pb = ach;
		  break;

		default:        
		  FMAERROR(NOT_SUPPORTED);
		  return FALSE;
	  }

	  // Add the property and check if ok

	  if (!FmaAddFmp(pfmm, aMhdr[iMhdr].id, pb, cb)) 
		return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message header
//

  FMAAPI_(BOOL) FmaSaveMhdr(PFMM pfmm, PVOID pmhdr, MHDRMAP aMhdr[], UINT nMhdr)
  {
	NETADDR netAddr;
	UINT iMhdr;
	PFMP pfmp;
	PVOID pb;
	UINT cb;

	// Scan the message header mapping table

	for (iMhdr = 0; iMhdr < nMhdr; iMhdr++) {

	  // Get the property and check if ok

	  if ((pfmp = FmaGetFmp(pfmm, aMhdr[iMhdr].id, 0)) == NULL) 
		continue;

	  // Calculate field address in the message header

	  pb = ((PBYTE)pmhdr) + aMhdr[iMhdr].offs;

	  // Process field according to its type

	  switch (aMhdr[iMhdr].type) {

		case MHDR_UL32: 
		  *((PULONG)pb) = *PFMPLONG(pfmp);
		  break;

		case MHDR_US16: 
		  *((PUSHORT)pb) = (USHORT)(*PFMPLONG(pfmp));
		  break;

		case MHDR_DTTM: 
		  *((PDTTM)pb) = *PFMPDTTM(pfmp);
		  break;

		case MHDR_TIME: 
		  *((PULONG)pb) = FmaGetTimeFromDttm(PFMPDTTM(pfmp));
		  break;

		case MHDR_ASCZ: 
		  cb = strlen(PFMPDATA(pfmp));
		  strncpy(pb, PFMPDATA(pfmp), aMhdr[iMhdr].cb);
		  ((PCH)pb)[aMhdr[iMhdr].cb - 1] = '\0';
		  break;

		case MHDR_ADDR: 
		  if (FmaScanNetAddr(&netAddr, PFMPDATA(pfmp))) {
			*((NETADDR *)pb) = netAddr;
		  } else {
			FMAERROR(INVALID_ADDRESS);
			return FALSE;
		  }
		  break;

		default:        
		  FMAERROR(NOT_SUPPORTED);
		  return FALSE;
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads ^aFLAGS kludge flags
//

  FMAAPI_(BOOL) FmaLoadFLAGS(PFMM pfmm, PCH pchData, UINT cchData)
  {
	DSZ dsz = DSZ_INIT;
	UINT cch, cchEnd;
	PCH pch, pchEnd;

	// Get the ^aFLAGS kludge data on stack

	pch = pchData;
	cch = cchData;

	// Check if we have the ^aFLAGS kludge data

	if (!pch || !cch) return TRUE;

	// Scan the ^aFLAGS kludge data

	loop {

	  // Skip leading spaces if any

	  while (cch && ISLWSP(pch[0])) {
		pch++; cch--;
	  }

	  // Check if we have any data 

	  if (!cch) break;

	  // Scan to the end of this flag catching possible end of kludge

	  for (pchEnd = pch; !EOKL(*pchEnd) && !ISLWSP(*pchEnd); pchEnd++);

	  // Calculate flag length and check it against flags mapping table and
	  // no match found, add the flag to unrecognized flags collection.
	  // Note: if dsz allocation fails, ignore as this is minor failure

	  cchEnd = (UINT)(pchEnd - pch);
	  if (!DoLoadFlag(pfmm, pch, cchEnd)) {
		AddDszFormat(&dsz, "%s%.*s", dsz.ich ? " " : "", cchEnd, pch);
	  }

	  // Advance to the next flag

	  pch = pchEnd;
	  cch-= cchEnd;
	}

	// Check if we have any unrecognized flags and 
	// add them to the message as FLAGS property
	// Note: don't fail on this as this it's minor

	if (dsz.psz) {
	  FmaAddFmp(pfmm, FMP_FLAGS, dsz.psz, dsz.ich);
	  DelDsz(&dsz);
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads message control info kludge
//

  FMAAPI_(BOOL) FmaLoadKludge(PFMM pfmm, PCH pch, UINT cch, PKLI pkli)
  {
	UINT iCtrl, cchData;
	PCH pchData;

	// Scan the known kludges table 

	for (iCtrl = 0; iCtrl < numbof(aCtrl); iCtrl++) {
	  if (FmaEqPch(pch, aCtrl[iCtrl].psz, aCtrl[iCtrl].cch)) {

		// Make up the kludge data pointer and length

		pchData = pch + aCtrl[iCtrl].cch;
		cchData = cch - aCtrl[iCtrl].cch;

		// Strip leading spaces if any

		while (cchData && ISLWSP(pchData[0])) {
		  pchData++; cchData--;
		}

		// Strip trailing spaces if any

		while (cchData && ISLWSP(pchData[cchData - 1])) {
		  cchData--;
		}

		// Check if this kludge has an associated property and add it
		// unless it's already there, thus we prefer first occurence
		// in case of duplicate kludges

		if (aCtrl[iCtrl].id) {
		  if ((aCtrl[iCtrl].id & FMP_ONE) && FmaGetFmp(pfmm, aCtrl[iCtrl].id, 0))
		    return TRUE;
		  if (!FmaAddFmp(pfmm, aCtrl[iCtrl].id, pchData, cchData))
		    return FALSE;
		} 

		// Check if this kludge info has to be saved for later
		// processing (i.e. address kludges, etc)

		if (aCtrl[iCtrl].type) {
		  pkli->pch[aCtrl[iCtrl].type] = pchData;
		  pkli->cch[aCtrl[iCtrl].type] = cchData;
		}

		return TRUE;
	  }
	}

	// If we got here the kludge is not known, so check if we have seen
	// any of known eom kludges or this one is a known eom kludges and 
	// add accordingly, otherwise add as a generic kludge.

	if (pkli->bSeenEomKludge || FmaGetEomKludgeFromString(pch, cch, TRUE)) {
	  DoAddEomKludge(pfmm, pch, cch, TRUE);
	  pkli->bSeenEomKludge = TRUE;
	} else {
	  if (!FmaAddFmp(pfmm, FMP_FTSKLUDGE, pch, cch))
		return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves kludge
//

  FMAAPI_(BOOL) FmaSaveKludge(PFMM pfmm, PFMP pfmp, PDSZ pdszCtrl, BOOL bAddCR)
  {
	PSZ pszCR = bAddCR ? "\r" : "";
	PSZ pszKludge;
	PSZ pszCtrl;

	ASSERT((FMP_FLG(pfmp->id) & FMP_KLU) == FMP_KLU);

	// Check if this is unknown kludge and add it as is

	if (pfmp->id == FMP_FTSKLUDGE) {
	  if (!AddDszFormat(pdszCtrl, "\x01""%s%s", PFMPDATA(pfmp), pszCR)) {
Fail:     FMAERROR(NOT_ENOUGH_MEMORY);
		return FALSE;
	  }
	} else {

	  // Get the name for the known kludge and if ok, add it

	  if ((pszKludge = DoGetKludgeFromFmp(pfmp)) != NULL) {

		// Check if this is an ^aAREA kludge and add it before any other 
		// kludges without space between kludge name and area tag, otherwise
		// add the kludge as usual and check if ok

		if (pfmp->id == FMP_AREA) {
		  if ((pszCtrl = pdszCtrl->psz) == NULL) {
			if (!AddDszFormat(pdszCtrl, "\x01""%s%s%s", pszKludge, PFMPDATA(pfmp), pszCR))
			  goto Fail;
		  } else {
			pdszCtrl->psz = NULL; pdszCtrl->ich = pdszCtrl->cch = 0;
			if (!AddDszFormat(pdszCtrl, "\x01""%s%s%s%s", pszKludge, PFMPDATA(pfmp), pszCR, pszCtrl)) {
			  MemFree(pszCtrl);
			  goto Fail;
			} else
			  MemFree(pszCtrl);
		  }
		} else {
		  if (!AddDszFormat(pdszCtrl, "\x01""%s %s%s", pszKludge, PFMPDATA(pfmp), pszCR))
			goto Fail;
		}
	  } else {

		// Unknown kludge with existing property id, 
		// time to update kludge mapping tables

		FMAERROR(INVALID_KLUDGE);
		return FALSE;
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads address kludges info ^aFMPT,^aTOPT,^aINTL and ^aDOMAIN
//

  FMAAPI_(BOOL) FmaLoadAddrKludges(PFMM pfmm, PKLI pkli, NETADDR * pnetOrig, 
								   NETADDR * pnetDest, NETADDR * pnetDefault,
								   BOOL bEchoMail)
  {
	UINT cch;
	PCH pch;

	// Stuff these in a structure so we can clear it in one shot

	struct {
	  NETADDR oAddr, dAddr;
	  NETADDR oIntl, dIntl;
	  NETADDR gate, msgid, reply;
	  PCH pchODomain; UINT cchODomain;
	  PCH pchDDomain; UINT cchDDomain;
	} net;

	// Scan in the ^aDOMAIN kludge and leave if ok since ^aDOMAIN kludge
	// information supercedes everything and no further messing is needed.

	if (DoScanDOMAIN(pfmm, pkli)) 
	  return TRUE;

	// Clear working structure in one shot

	memset(&net, 0, sizeof(net));

	// Get the current origination and destination addresses

	net.oAddr = *pnetOrig;
	net.dAddr = *pnetDest;

	// Scan in the ^aFMPT kludge if any

	if (pkli->cch[KLI_FMPT] && (pch = pkli->pch[KLI_FMPT]) != NULL) {
	  net.oAddr.point = (USHORT)atoi(pch);
	}

	// Scan in the ^aTOPT kludge if any

	if (pkli->cch[KLI_TOPT] && (pch = pkli->pch[KLI_TOPT]) != NULL) {
	  net.dAddr.point = (USHORT)atoi(pch);
	}

	// Scan in the ^aINTL kludge if any

	if (pkli->cch[KLI_INTL] && (pch = pkli->pch[KLI_INTL]) != NULL) {
	  if ((pch = FmaScanNetAddr(&net.dIntl, pch)) == NULL ||
		  (pch = FmaScanNetAddr(&net.oIntl, pch)) == NULL) {
		net.dIntl.zone = net.oIntl.zone = 0;
	  } else {

		// Update origination address zone, net and node

		DoCopyZNN(&net.oAddr, &net.oIntl);

		// Check if not zonegating, that is net/node are the same
		// in binary header and in ^aINTL kludge

		if (DoEqNN(&net.dAddr, &net.dIntl)) {
		  net.dAddr.zone = net.dIntl.zone;
		} else {

		  // When zonegating, the binary header destination is actially
		  // a gate address, so preserve it for later processing

		  net.gate = net.dAddr;
		  net.gate.zone = net.oAddr.zone;
		  net.gate.point = 0;

		  // Update destination address zone, net and node

		  DoCopyZNN(&net.dAddr, &net.dIntl);
		}
	  }
	}

	// Scan in the ^aMSGID kludge if any

	if (pkli->cch[KLI_MSGID] && (pch = pkli->pch[KLI_MSGID]) != NULL) {
	  if ((pch = FmaScanNetAddr(&net.msgid, pch)) != NULL) {

		// Scan in the origination domain

		net.pchODomain = pch; 
		cch = pkli->cch[KLI_MSGID] - (UINT)(pch - pkli->pch[KLI_MSGID]);
		net.cchODomain = DoScanToSpace(&pch, &cch);

		// Check if this is an echo mail area and if so, replace origination
		// address with the one in ^aMSGID since tosser placed an uplink address
		// in place of origination address

		if (bEchoMail) {
		  net.oAddr = net.msgid;
		} else

		// Check if we have no origination zone and ^aMSGID net, node and
		// point address match the ones in our origination address and if so,
		// assume the origination zone

		if (!net.oAddr.zone && DoEqNNP(&net.oAddr, &net.msgid)) {
		  net.oAddr.zone = net.msgid.zone;
		} 
	  }
	}

	// Scan in the ^aREPLY kludge if any

	if (pkli->cch[KLI_REPLY] && (pch = pkli->pch[KLI_REPLY]) != NULL) {
	  if ((pch = FmaScanNetAddr(&net.reply, pch)) != NULL) {

		// Scan in the destination domain

		net.pchDDomain = pch; 
		cch = pkli->cch[KLI_REPLY] - (UINT)(pch - pkli->pch[KLI_REPLY]);
		net.cchDDomain = DoScanToSpace(&pch, &cch);

		// Check if this is an echo mail area and if so, replace destination
		// address with the one in ^aREPLY since tosser placed our system address
		// in place of destination address

		if (bEchoMail) {
		  net.dAddr = net.reply;
		} else

		// Check if we have no destination zone and ^aREPLY net, node and
		// point address match the ones in our destination address and if so,
		// assume the destination zone

		if (!net.dAddr.zone && DoEqNNP(&net.dAddr, &net.reply)) {
		  net.dAddr.zone = net.msgid.zone;
		} 
	  }
	}

	// Check if we still have no zone for origination or destination address
	// and assume the default zone for this area

	if (pnetDefault) {
	  if (!net.oAddr.zone) net.oAddr.zone = pnetDefault->zone;
	  if (!net.dAddr.zone) net.dAddr.zone = pnetDefault->zone;
	}

	// Add the origination address property and check if ok

	if (!FmaAddFmpFormat(pfmm, FMP_ORIGADDR, "%u:%u/%u.%u%.*s",
						 net.oAddr.zone, net.oAddr.net, net.oAddr.node, net.oAddr.point,
						 net.cchODomain, net.pchODomain))
	  return FALSE;

	// Add the destination address property and check if ok

	if (!FmaAddFmpFormat(pfmm, FMP_DESTADDR, "%u:%u/%u.%u%.*s",
						 net.dAddr.zone, net.dAddr.net, net.dAddr.node, net.dAddr.point,
						 net.cchDDomain, net.pchDDomain))
	  return FALSE;

	// Check if we have gate address and add appropriate property

	if (net.gate.zone) {
	  if (!FmaAddFmpFormat(pfmm, FMP_GATEADDR, "%u:%u/%u.%u",
						   net.gate.zone, net.gate.net, net.gate.node, net.gate.point))
		return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves address kludges info ^aFMPT,^aTOPT,^aINTL and ^aDOMAIN
//

  FMAAPI_(BOOL) FmaSaveAddrKludges(PFMM pfmm, PDSZ pdszCtrl, NETADDR * pnetOrig, 
								   NETADDR * pnetDest, NETADDR * pnetDefault,
								   BOOL bNoMhdrZoneInfo, BOOL bNoMhdrPointInfo, 
								   ULONG fl, BOOL bAddCR)
  {
	PSZ pszCR = bAddCR ? "\r" : "";

	// Stuff these in a structure so we can clear it in one shot

	struct {
	  NETADDR oAddr, dAddr;
	  PCH oDomain, dDomain;
	} net;

	// Clear working structure in one shot

	memset(&net, 0, sizeof(net));

	// Get the origination/destination addresses/domains and check if ok

	if (!FmaGetNetAddr(pfmm, FMP_ORIGADDR, &net.oAddr, &net.oDomain) ||
		!FmaGetNetAddr(pfmm, FMP_DESTADDR, &net.dAddr, &net.dDomain))
	  return FALSE;

	// Check if we need to add point info

	if (bNoMhdrPointInfo) {

	  // Check if origination point is not zero and add ^aFMPT kludge
	  // unless it's already there

	  if (net.oAddr.point && !DoGetFmpFromKludge(pfmm, "FMPT ")) {
		if (!AddDszFormat(pdszCtrl, "\x01""FMPT %u%s", net.oAddr.point, pszCR))
		  goto Fail;
	  }

	  // Check if destination point is not zero and add ^aTOPT kludge
	  // unless it's already there

	  if (net.dAddr.point && !DoGetFmpFromKludge(pfmm, "TOPT ")) {
		if (!AddDszFormat(pdszCtrl, "\x01""TOPT %u%s", net.dAddr.point, pszCR))
		  goto Fail;
	  }
	}

	// Check if we need to add ^aINTL kludge and add it 
	// unless it's already there

	if (((fl & FSM_FORCEINTL) || (bNoMhdrZoneInfo &&
		(net.oAddr.zone != net.dAddr.zone ||
		 net.oAddr.zone != pnetDefault->zone ||
		 net.dAddr.zone != pnetDefault->zone))) && 
		!DoGetFmpFromKludge(pfmm, "INTL ")) {
	  if (!AddDszFormat(pdszCtrl, "\x01""INTL %u:%u/%u %u:%u/%u%s", 
						net.dAddr.zone, net.dAddr.net, net.dAddr.node, 
						net.oAddr.zone, net.oAddr.net, net.oAddr.node, 
						pszCR)) {
Fail:   FMAERROR(NOT_ENOUGH_MEMORY);
		return FALSE;
	  }
	}

	// Check if we have both domains and add ^aDOMAIN kludge
	// unless it's already there

	if ((fl & FSM_AUTODOMAIN) && (net.oDomain && net.dDomain) && 
		!DoGetFmpFromKludge(pfmm, "DOMAIN ")) {

	  // Skip over leading @ characters

	  if (net.oDomain[0] == '@') net.oDomain++;
	  if (net.dDomain[0] == '@') net.dDomain++;

	  // Add the domain kludge and check if ok

	  if (!AddDszFormat(pdszCtrl, "\x01""DOMAIN %s %u:%u/%u.%u %s %u:%u/%u.%u%s", 
						net.dDomain, net.dAddr.zone, net.dAddr.net, net.dAddr.node, net.dAddr.point,
						net.oDomain, net.oAddr.zone, net.oAddr.net, net.oAddr.node, net.oAddr.point,
						pszCR))
		goto Fail;
	}

	// Pass the addresses to caller

	if (pnetOrig) *pnetOrig = net.oAddr;
	if (pnetDest) *pnetDest = net.dAddr;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves end of message kludges
//

  FMAAPI_(BOOL) FmaLoadEomKludges(PFMM pfmm, ULONG fl)
  {
	PCH pch, pchText, pchLine, pch1stEomKludge = NULL;
	UINT cchScan, cchText, cchLine;
	PFMP pfmp;

	// Get message text property and check if ok

	if ((pfmp = FmaGetFmp(pfmm, FMP_MESSAGETEXT, 0)) == NULL) {
	  FMAERROR(NOT_SUPPORTED);
	  return FALSE;
	}

	// Get length of the message text and check if ok

	pchText = *PFMPPPSZ(pfmp);
	if ((cchText = strlen(pchText)) == 0) 
	  return TRUE;

	// Prepare for reverse scan

	pch = pchText + cchText - 1;
	cchScan = cchText - 1;
	cchLine = 0;

	// Skip all empty lines at eom

	while (cchScan && *pch == '\r') {
	  cchScan--; pch--;
	}

	// Ckeck if we still have anything to scan

	if (!cchScan) return TRUE;

	// Scan back checking line by line accepting any line starting with ^a and
	// those which start with known eom kludge without ^a (e.g. SEEN-BY)

	loop {

	  // Check if hit end of previous line 

	  if (*pch == '\r') {
		pchLine = pch + 1;
Check:  if (pchLine[0] != '\x01' && !FmaGetEomKludgeFromString(pchLine, cchLine, -1)) break;
		pch1stEomKludge = pchLine;
		cchLine = 0;
	  } else

	  // Check if hit beginning of the message text

	  if (!cchScan) {
		pchLine = pch; cchLine++;
		goto Check;
	  } else
		cchLine++;

	  // Get back one char and check if ok

	  if (cchScan-- > 0)
		pch--;
	  else
		break;
	}

	// Check if we've found any eom kludges and leave if not

	if (!pch1stEomKludge) return TRUE;

	// Add all the eom kludges. Note that we want to make sure that the very
	// last line is properly processed even if it does not have trailing cr

	for (pchLine = pch1stEomKludge; pchLine != NULL; ) {

	  // Get length of the eom kludge line

	  if ((pch = strchr(pchLine, '\r')) != NULL) {
		cchLine = (UINT)(pch - pchLine);
	  } else
		cchLine = strlen(pchLine);

	  // Check if line is not empty and add eom kludge line

	  if (cchLine) {
		if (!DoAddEomKludge(pfmm, pchLine, cchLine, -1))
		  return FALSE;
	  }

	  // Advance to the next eom kludge line

	  pchLine+= cchLine;
	  if (pchLine[0] == '\r')
		pchLine++;
	  else
		break;
	}

	// Now chop the message text at the first eom kludge

	cchText = (UINT)(pch1stEomKludge - pchText) + 1;
	if ((pchText = MemRealloc(pchText, cchText)) != NULL) {
	  *PFMPPPSZ(pfmp) = pchText;
	  pch1stEomKludge[0] = '\0';
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves end of message kludges
//

  FMAAPI_(BOOL) FmaSaveEomKludges(PFMM pfmm, PDSZ pdszEom)
  {
	PSZ pszPrefix;
	PKLMAP pklme;
	PFMP pfmp;

	// Loop through all the properties

	for (pfmp = pfmm->pfmp; pfmp != NULL; pfmp = pfmp->pfmpNext) {

	  // Skip non eom kludge properties 

	  if ((FMP_FLG(pfmp->id) & FMP_KLU) != FMP_KLU || !(FMP_FLG(pfmp->id) & FMP_EOM))
		continue;

	  // Check if this eom kludge does not have the specific property id

	  if (pfmp->id == FMP_EOMKLUDGE) {

		// Add generic eom kludge and check if ok

		if (!AddDszFormat(pdszEom, "\x01""%s\r", PFMPDATA(pfmp))) {
		  FMAERROR(NOT_ENOUGH_MEMORY);
		  return FALSE;
		}

	  } else {

		// For the eom kludge which has specific propery is get associated 
		// kludge mapping table entry and check if ok

		if ((pklme = FmaGetEomKludgeFromId(pfmp->id)) == NULL) {
		  FMAERROR(INVALID_KLUDGE);
		  return FALSE;
		}

		// Check if this kludge has prefix and add its length

		pszPrefix = (pklme->type & KLI_NOPREFIX) ? "" : "\x01";

		// Add eom kludge with specific property id and check if ok

		if (!AddDszFormat(pdszEom, "%s%s %s\r", pszPrefix, pklme->psz, PFMPDATA(pfmp))) {
		  FMAERROR(NOT_ENOUGH_MEMORY);
		  return FALSE;
		}
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns known ^aFLAGS flag string given attribute
//

  FMAAPI_(PSZ) FmaGetFlagFromAttr(ULONG attr)
  {
	UINT iFlag;

	// Scan the flags mapping table looking for an id match

	for (iFlag = 0; iFlag < numbof(aFlag); iFlag++) {
	  if (aFlag[iFlag].attr & attr)
		return aFlag[iFlag].psz;
	}

	return NULL;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns known ^aFLAGS flag attribute given flag string
//

  FMAAPI_(ULONG) FmaGetAttrFromFlag(PCH pch, UINT cch)
  {
	UINT iFlag;

	// Check if empty flag and fail

	if (!pch || !cch) return FALSE;

	// Scan the flags mapping table looking for a string match

	for (iFlag = 0; iFlag < numbof(aFlag); iFlag++) {
	  if (FmaEqPch(pch, aFlag[iFlag].psz, cch))
		return aFlag[iFlag].attr;
	}

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns known kludge given property id
//

  FMAAPI_(PKLMAP) FmaGetKludgeFromId(ULONG id)
  {
	UINT iCtrl;

	ASSERT(id != 0);

	// Scan the kludges mapping table looking for an id match

	for (iCtrl = 0; iCtrl < numbof(aCtrl); iCtrl++) {
	  if (aCtrl[iCtrl].id == id)
		return &aCtrl[iCtrl];
	}

	return NULL;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns known end of message kludge given string
//

  FMAAPI_(PKLMAP) FmaGetKludgeFromString(PCH pch, UINT cch)
  {
	UINT iCtrl;

	// Check if empty line and fail

	if (!pch || !cch) return NULL;

	// Scan the kludges mapping table looking for a string match

	for (iCtrl = 0; iCtrl < numbof(aCtrl); iCtrl++) {
	  if (FmaEqPch(pch, aCtrl[iCtrl].psz, aCtrl[iCtrl].cch))
		return &aCtrl[iCtrl];
	}

	return NULL;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns known end of message kludge given property id
//

  FMAAPI_(PKLMAP) FmaGetEomKludgeFromId(ULONG id)
  {
	UINT iEom;

	ASSERT(id != FMP_EOMKLUDGE);

	// Scan the eom kludges table looking for an id match

	for (iEom = 0; iEom < numbof(aEom); iEom++) {
	  if (aEom[iEom].id == id)
		return &aEom[iEom];
	}

	return NULL;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns known end of message kludge given string
//

  FMAAPI_(PKLMAP) FmaGetEomKludgeFromString(PCH pch, UINT cch, BOOL bHasPrefix)
  {
	UINT iEom;

	// Check if empty line and fail

	if (!pch || !cch) return NULL;

	// Check if the line has ^a prefix and skip it saving this fact,
	// unless this info is provided by the caller

	if (bHasPrefix == -1) {
	  if (pch[0] == '\01') {
		bHasPrefix = TRUE;
		pch++; cch--;
	  } else
		bHasPrefix = FALSE;
	}

	// Scan the eom kludges table looking for a match

	for (iEom = 0; iEom < numbof(aEom); iEom++) {
	  if (cch >= aEom[iEom].cch) {

		// Check the prefix match and skip if not

		if (aEom[iEom].type & KLI_NOPREFIX) {
		  if (bHasPrefix) continue;
		} else {
		  if (!bHasPrefix) continue;
		}

		// Check the kludge match and return is so

		if (FmaEqPch(pch, aEom[iEom].psz, aEom[iEom].cch))
		  return &aEom[iEom];
	  }
	}

	return NULL;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds automatic FMP_MSGID property
//

  FMAAPI_(PFMP) FmaAddAutoMSGID(PFMM pfmm)
  {
	ULONG lStamp;
	PSZ pszAddr;
	PFMP pfmp;

	// Check if we already have FMP_MSGID property and leave

	if ((pfmp = FmaGetFmp(pfmm, FMP_MSGID, 0)) != NULL)
	  return pfmp;

	// Get the origination address and check if ok

	if ((pfmp = FmaGetFmp(pfmm, FMP_ORIGADDR, 0)) == NULL) {
	  FMAERROR(INVALID_ADDRESS);
	  return NULL;
	} else
	  pszAddr = PFMPDATA(pfmp);

	// Generate unique message id stamp

	lStamp = FmaGetMsgIdStamp();

	// Add FMP_MSGID property 

	pfmp = FmaAddFmpFormat(pfmm, FMP_MSGID, "%s %08lx", pszAddr, lStamp);

	return pfmp;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds automatic FMP_PID property
//

  FMAAPI_(PFMP) FmaAddAutoPID(PFMM pfmm)
  {
	FMAVER fmaVer;
	PFMP pfmp;

	static CHAR achFmt[] = 
	  "FMA v%02u.%02u.%u"
#ifdef _DEBUG
	  " (debug)"
#endif
	  ;

	// Check if we already have FMP_PID property and leave

	if ((pfmp = FmaGetFmp(pfmm, FMP_PID, 0)) != NULL)
	  return pfmp;

	// Get Fidonet Mail Access version info

	FmaGetVersion(&fmaVer);

	// Add FMP_PID property 

	pfmp = FmaAddFmpFormat(pfmm, FMP_PID, achFmt, 
						   fmaVer.verMajor, fmaVer.verMinor, 
						   fmaVer.verBuild);

	return pfmp;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds automatic properties
//

  FMAAPI_(BOOL) FmaAddAutoProp(PFMM pfmm, ULONG fl)
  {
	// Check if we need to add automatic MSGID

	if (fl & FSM_AUTOMSGID) {
	  if (!FmaAddAutoMSGID(pfmm))
		return FALSE;
	}

	// Check if we need to add automatic PID

	if (fl & FSM_AUTOPID) {
	  if (!FmaAddAutoPID(pfmm))
		return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of FmaCtrl.c
/////////////////////////////////////////////////////////////////////////////
