/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Control info routines header file
//
// FmaCtrl.h
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 26/Sep/1999
// Updated: 12/Oct/1999
//
// History: 26/Sep/1999 kvitekp - Initially created
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _FMA_CTRL_INCLUDED
#define _FMA_CTRL_INCLUDED

  // Declare C/C++ frame

#ifdef __cplusplus
extern "C" {
#endif

  // Include types declaratons

#include "FmaDef.h"

/////////////////////////////////////////////////////////////////////////////
// Control info parser declarations
/////////////////////////////////////////////////////////////////////////////

  // End-of-kludge macro

#define EOKL(ch) ((UCHAR)(ch) <= 0x01 || (ch) == '\r')

  // Attribute mapping table entry macro

#define ATTR_(attr, attrNative, flag) { attr, attrNative, flag },

  // Attribute mapping table structure

  typedef struct _ATTRMAP {
	ULONG   attr;               // Fidonet mail attribute (MA_)
	ULONG   attrNative;         // Native message attribute
	PSZ     pszFlag;            // Matching ^aFLAGS kludge flag
  } ATTRMAP, FAR * PATTRMAP;

  // Message header mapping table entry types

#define MHDR_UL32       1       // Unsigned long 32 bit value
#define MHDR_US16       2       // Unsigned short 16 bit value
#define MHDR_DTTM       3       // Date/time info 32-bit, DTTM format
#define MHDR_TIME       4       // Date/time info 32-bit, time format
#define MHDR_ASCZ       5       // Asciiz string, bSave == cchMax
#define MHDR_ADDR       6       // Binary netword address (NETADDR)
#define MHDR_LAST       6       // Use this for defining private types

  // Message header mapping table entry macro

#define MHDR_(mhdr, field, type, id) { fieldoffs(mhdr, field), fieldsize(mhdr, field), type, id },

  // Message header mapping table entry structure

  typedef struct _MHDRMAP {
	UINT    offs;               // Offset into the message header
	UINT    cb;                 // Size of the message header field
	UINT    type;               // Message header entry type (MHDR_)
	ULONG   id;                 // Fidonet mail message property id (FMP_)
  } MHDRMAP, FAR * PMHDRMAP;

  // Kludge info structure types. Note that since these
  // are also used as flags in tables, its one based

#define KLI_FMPT          1
#define KLI_TOPT          2
#define KLI_INTL          3
#define KLI_DOMAIN        4
#define KLI_FLAGS         5
#define KLI_MSGID         6
#define KLI_REPLY         7
#define KLI_NOPREFIX 0x8000

  // Kludge info structure

  typedef struct _KLI { /* kli */
	PCH     pch[8];             // Kludge data pointer array indexed by KLI_
	UINT    cch[8];             // Kludge data length array indexed by KLI_
	BOOL    bSeenEomKludge;     // TRUE is have seen any of known eom kludges
  } KLI, FAR * PKLI;

  // Kludge mapping table entry macro

#define KLMAP_(id, name, type) { id, name, lengof(name), type } ,

  // Kludge mapping table entry structure

  typedef struct _KLMAP { /* klme */
	ULONG   id;                 // Fidonet mail property id (FMP_)
	PSZ     psz;                // Kludge name pointer
	UINT    cch;                // Kludge name length
	UINT    type;               // Kludge type (KLI_)
  } KLMAP, FAR * PKLMAP;

  // Function prototypes

  FMAAPI_(BOOL) FmaGetNetAddr(PFMM pfmm, ULONG id, NETADDR * pnetAddr,
							  PCH * ppchDomain);

  FMAAPI_(BOOL) FmaLoadMhdr(PFMM pfmm, PVOID pmhdr, MHDRMAP aMhdr[], UINT nMhdr);
  FMAAPI_(BOOL) FmaSaveMhdr(PFMM pfmm, PVOID pmhdr, MHDRMAP aMhdr[], UINT nMhdr);

  FMAAPI_(BOOL) FmaLoadFLAGS(PFMM pfmm, PCH pchData, UINT cchData);

  FMAAPI_(BOOL) FmaLoadKludge(PFMM pfmm, PCH pch, UINT cch, PKLI pkli);
  FMAAPI_(BOOL) FmaSaveKludge(PFMM pfmm, PFMP pfmp, PDSZ pdszCtrl, BOOL bAddCR);

  FMAAPI_(BOOL) FmaLoadAddrKludges(PFMM pfmm, PKLI pkli, NETADDR * pnetOrig, 
								   NETADDR * pnetDest, NETADDR * pnetDefault,
								   BOOL bEchoMail);

  FMAAPI_(BOOL) FmaSaveAddrKludges(PFMM pfmm, PDSZ pdszCtrl, NETADDR * pnetOrig, 
								   NETADDR * pnetDest, NETADDR * pnetDefault,
								   BOOL bNoMhdrZoneInfo, BOOL bNoMhdrPointInfo, 
								   ULONG fl, BOOL bAddCR);

  FMAAPI_(BOOL) FmaLoadEomKludges(PFMM pfmm, ULONG fl);
  FMAAPI_(BOOL) FmaSaveEomKludges(PFMM pfmm, PDSZ pdszEom);

  FMAAPI_(PSZ) FmaGetFlagFromAttr(ULONG attr);
  FMAAPI_(ULONG) FmaGetAttrFromFlag(PCH pch, UINT cch);

  FMAAPI_(PKLMAP) FmaGetKludgeFromId(ULONG id);
  FMAAPI_(PKLMAP) FmaGetKludgeFromString(PCH pch, UINT cch);

  FMAAPI_(PKLMAP) FmaGetEomKludgeFromId(ULONG id);
  FMAAPI_(PKLMAP) FmaGetEomKludgeFromString(PCH pch, UINT cch, BOOL bHasPrefix);

  FMAAPI_(PFMP) FmaAddAutoMSGID(PFMM pfmm);
  FMAAPI_(PFMP) FmaAddAutoPID(PFMM pfmm);
  FMAAPI_(BOOL) FmaAddAutoProp(PFMM pfmm, ULONG fl);

/////////////////////////////////////////////////////////////////////////////
// Header cleanup
/////////////////////////////////////////////////////////////////////////////

  // Terminate C/C++ frame

#ifdef __cplusplus
}
#endif

#endif  // #ifndef _FMA_CTRL_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// End of FmaCtrl.h
/////////////////////////////////////////////////////////////////////////////
