/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Common constants and types declaration 
//
// FmaDef.h
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 10/Nov/1999
//
// History: 17/Jan/1999 kvitekp - Initially created for IMA project.
//
//          22/Feb/2000 kvitekp - Added namespace compatibility #ifdefs
//
/////////////////////////////////////////////////////////////////////////////
// 
// Global platform specification defines:
//
// __DOS__    - generic MS-DOS
// __OS2__    - native OS/2
// __W32__    - native Win32
// __MAC__    - generic Mac
// __UNIX__   - generic Unix
//
// Global platform modification defines:
//
// __DOSOVL__ - overlayed version (DOS only)
// __DPMI__   - DPMI protercted mode version (DOS only)
// __FLAT__   - flat 32-bit address model
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _FMA_DEF_INCLUDED
#define _FMA_DEF_INCLUDED

  // Declare C/C++ frame

#ifdef __cplusplus
extern "C" {
#endif

/////////////////////////////////////////////////////////////////////////////
// Catch common platform declarations early
/////////////////////////////////////////////////////////////////////////////

#if !defined(__W32__) && defined(WIN32)
#define __W32__
#endif

#if !defined(__OS2__) && defined(OS2)
#define __OS2__
#endif

#if !defined(__UNIX__) && defined(UNIX)
#define __UNIX__
#endif

/////////////////////////////////////////////////////////////////////////////
// Include standard headers
/////////////////////////////////////////////////////////////////////////////

  // Include standard headers

#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>

  // Include DOS declarations if compiling for DOS

#ifdef __DOS__
#include <dos.h>
#endif

  // Include Windows declarations if compiling for Win32

#ifdef __W32__
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

  // Include OS/2 declarations if compiling for OS/2

#ifdef __OS2__
#define INCL_NOCOMMON
#define INCL_NOPM
#define INCL_DOSFILEMGR
#define INCL_DOSINFOSEG
#define INCL_DOSPROCESS
#define INCL_DOSDATETIME
#define INCL_DOSERRORS
#define INCL_DOSMISC
#include <os2.h>
#endif

/////////////////////////////////////////////////////////////////////////////
// Define compiler dependant keywords
/////////////////////////////////////////////////////////////////////////////

#ifndef __COMMON_COMPILER_DEFS__
#define __COMMON_COMPILER_DEFS__

#if   defined __BORLANDC__
  #define __FAR     far
  #define __NEAR    near
  #define __EXPORT _export
  #define __PASCAL  pascal
  #define __CDECL  cdecl
#if (__BORLANDC__ >= 0x500)
  #define __STDCALL __stdcall
  #define __INLINE  __inline
#else
  #define __STDCALL
  #define __INLINE
#endif

#elif defined _MSC_VER
  #define __FAR    _far
  #define __NEAR   _near
  #define __EXPORT _export
  #define __PASCAL _pascal
  #define __CDECL  _cdecl
#if (_MSC_VER >= 800)
  #define __STDCALL __stdcall
  #define __INLINE  __inline
#else
  #define __STDCALL
  #define __INLINE
#endif

#elif defined __WATCOMC__
  #define __FAR    _far
  #define __NEAR   _near
  #define __EXPORT _export
  #define __PASCAL _pascal
  #define __CDECL  _cdecl
  #define __STDCALL __stdcall
  #define __INLINE  __inline

#else
  #define __FAR    
  #define __NEAR   
  #define __EXPORT 
  #define __PASCAL 
  #define __STDCALL
  #define __CDECL
  #define __INLINE
#endif

#endif // #ifndef __COMMON_COMPILER_DEFS__

/////////////////////////////////////////////////////////////////////////////
// Define platform dependant declarations
/////////////////////////////////////////////////////////////////////////////

#if defined __W32__ && !defined __FLAT__
  #define __FLAT__
#endif

#ifndef FAR
#ifndef __FLAT__
  #define FAR   __FAR
#else
  #define FAR
#endif
#endif

#ifndef NEAR
#ifndef __FLAT__
  #define NEAR  __NEAR
#else
  #define NEAR
#endif
#endif

/////////////////////////////////////////////////////////////////////////////
// Define basic types
/////////////////////////////////////////////////////////////////////////////

  // Catch environments which declare common types we use and bypass
  // private declaration

#if !defined(BASETYPES) && !defined(OS2DEF_INCLUDED) && !defined(DW_SAA_DEFS)

  #define VOID    void
  #define CHAR    char              /* ch  */
  #define SHORT   short             /* s   */
  #define LONG    long              /* l   */
  #define INT     int               /* i   */

  typedef unsigned char  UCHAR;     /* uch */
  typedef unsigned short USHORT;    /* us  */
  typedef unsigned long  ULONG;     /* ul  */
  typedef unsigned int   UINT;      /* ui  */

  typedef unsigned char  BYTE;      /* b   */
  typedef unsigned short WORD;      /* w   */
  typedef unsigned long  DWORD;     /* dw  */

  typedef unsigned char FAR * PSZ;  
  typedef unsigned char FAR * PCH;  

  typedef BYTE   FAR * PBYTE;
  typedef CHAR   FAR * PCHAR;
  typedef SHORT  FAR * PSHORT;
  typedef LONG   FAR * PLONG;
  typedef INT    FAR * PINT;

  typedef UCHAR  FAR * PUCHAR;
  typedef USHORT FAR * PUSHORT;
  typedef ULONG  FAR * PULONG;
  typedef UINT   FAR * PUINT;

  typedef VOID   FAR * PVOID;

  typedef int BOOL;                 /* b   */
  typedef BOOL FAR * PBOOL;

  typedef int HFILE;                /* hfile */
  typedef HFILE FAR * PHFILE;

#endif

/////////////////////////////////////////////////////////////////////////////
// Define linkage
/////////////////////////////////////////////////////////////////////////////

  // Define dynamic link library import/export declarations

#if _MSC_VER >= 1000
#define __DECL_DLL_EXPORT __declspec(dllexport)
#else
#define __DECL_DLL_EXPORT __EXPORT
#endif

  // Define API entry point declaration for static and dll linkage

#ifdef _FMA_DLL_BUILD
#define FMADECL   __DECL_DLL_EXPORT
#else
#define FMADECL
#endif

  // Define global data declaration for static and dll linkage

#ifdef _FMA_DLL_BUILD
#define FMAGLOBAL __DECL_DLL_EXPORT
#else
#define FMAGLOBAL
#endif

  // Define calling convertion declaratinos

#if defined(_STDCALL_SUPPORTED) || (_MSC_VER >= 800) || (__BORLANDC__ >= 0x500) || (__WATCOMC__ >= 110)
#define __FMAAPI    __STDCALL
#define __FMAAPIV   __CDECL
#else
#define __FMAAPI    __PASCAL
#define __FMAAPIV   __CDECL
#endif

  // Define calling conventions

#define FMAAPI      __FMAAPI            // API entry point
#define FMAAPIV     __FMAAPIV           // API entry point with var param
#define FMACALLBACK __FMAAPI            // Callback entry point
#define FMAINLINE   __INLINE            // Inline routine

  // Define local calling conventions. Note: most overlay managers 
  // require far entry points even for statically defined routines

#ifndef __DOSOVL__
#define FMASUB      __FMAAPI  NEAR
#define FMASUBV     __FMAAPIV NEAR
#else
#define FMASUB      __FMAAPI
#define FMASUBV     __FMAAPIV 
#endif

  // Define api function declaration macros

#if _MSC_VER >= 1000
#define FMAAPI_(type)  FMADECL type FMAAPI
#define FMAAPIV_(type) FMADECL type FMAAPIV
#else
#define FMAAPI_(type)  type FMADECL FMAAPI
#define FMAAPIV_(type) type FMADECL FMAAPIV
#endif

  // Define global data declaration macros

#if _MSC_VER >= 1000
#define FMAGLOBAL_(type)  FMAGLOBAL type
#else
#define FMAGLOBAL_(type)  type FMAGLOBAL
#endif

/////////////////////////////////////////////////////////////////////////////
// More compiler dependant stuff
/////////////////////////////////////////////////////////////////////////////

  // Define use of private string formatter. Despite the fact that many
  // compilers support _snprintf and _vsnprintf, some implemetations expose
  // subtle differences in trailing null handling, so we always use private 
  // formatter for consistent behaveor. Note that if FMA_PRIVATE_FORMAT is 
  // not defined, it is necessary to define Format macro in FmaUti.h to map 
  // them appropriate function calls.

//#if !(_MSC_VER >= 800) && !(__WATCOMC__ >= 110)
#define FMA_PRIVATE_FORMAT
//#endif

  // Define used of #pragma pack(push, n) and pack(pop). Note these are only
  // used in message base definition headers, while in all other places default
  // structure packing is in effect.

#if (_MSC_VER >= 800) || (__BORLANDC__ >= 0x500) || (__WATCOMC__ >= 110)
#define FMA_PUSHPOP_PACKING
#endif

  // Microsoft Visual C++ warning control

#if _MSC_VER >= 800
//#pragma warning(disable:4068) // unknown pragma
//#pragma warning(disable:4761) // integral type mismatch
//#pragma warning(disable:4146) // unary minus to unsigned type
//#pragma warning(disable:4229) // modifiers on data are ignored
//#pragma warning(disable:4102) // unreferenced label
//#pragma warning(disable:4127) // conditional expression is constant
#pragma warning(disable:4100)   // unreferenced formal parameter
#endif

  // Borland C++ warning control

#ifdef __BORLANDC__
//#pragma warn -ucp             // mixing pointers to signed/unsigned char
//#pragma warn -pin             // initialization is only partially braketed
//#pragma warn -bbf             // bitfields must be signed or unsigned ints
//#pragma warn -ccc             // condition is always true/false
//#pragma warn -pck             // structure packing size has changed
#pragma warn -aus               // 'ident' assigned a value which is never used	
#pragma warn -par               // parameter 'ident' is never used	
#pragma warn -use               // 'ident' declared but never used	
#endif

/////////////////////////////////////////////////////////////////////////////
// Miscellaneous declarations
/////////////////////////////////////////////////////////////////////////////

  // Define common constants and macros

#ifndef FALSE
#define FALSE   0
#endif

#ifndef TRUE
#define TRUE    1
#endif

#ifndef max
#define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a,b) (((a) < (b)) ? (a) : (b))
#endif

  // Define maximum path length

#ifndef MAX_PATH
#if   defined __DOS__
#define MAX_PATH  80
#elif defined __OS2__
#define MAX_PATH  260
#elif defined __W32__
#define MAX_PATH  260
#elif defined PATH_MAX
#define MAX_PATH  PATH_MAX
#else
#error Unknown MAX_PATH
#endif
#endif

  // Define path component delimiter

#ifndef PATH_SEP
#if   defined __DOS__
#define PATH_SEP  '\\'
#elif defined __OS2__
#define PATH_SEP  '\\'
#elif defined __W32__
#define PATH_SEP  '\\'
#elif defined __MAC__
#define PATH_SEP  ':'
#elif defined __UNIX__
#define PATH_SEP  '/'
#else
#error Unknown PATH_SEP
#endif
#endif

/////////////////////////////////////////////////////////////////////////////
// Additional types declaration
/////////////////////////////////////////////////////////////////////////////

  // Common types

#define HAREA  PVOID                    // Message area handle
#define HMSG   PVOID                    // Message handle
#define MSGN   ULONG                    // Physical message number
#define UMSG   ULONG                    // Unique message id

  // Set structure packing for common byte packed structures

#ifdef FMA_PUSHPOP_PACKING
#pragma pack(push, 1)
#endif

  // Network address structure

#ifndef _NETADDR_DEFINED
#define _NETADDR_DEFINED

  typedef struct _NETADDR { /* net */
	USHORT  zone;                       // Zone number
	USHORT  net;                        // Network number
	USHORT  node;                       // Node number
	USHORT  point;                      // Point number
  } NETADDR, FAR * PNETADDR;

#endif 

  // Date and time structure

#ifndef _DTTM_DEFINED
#define _DTTM_DEFINED

  typedef struct _DTTM { /* dttm */
	UINT    da:5;                       // Day of the month (1-31)
	UINT    mo:4;                       // Month (1 = January, 2 = February, and so on)
	UINT    yr:7;                       // Year offset from 1980 (add 1980 to get actual year) 
	UINT    ss:5;                       // Second divided by 2
	UINT    mm:6;                       // Minute (0-59)
	UINT    hh:5;                       // Hour (0-23 on a 24-hour clock)
  } DTTM, FAR * PDTTM;

#endif 

  // Restore structure packing

#ifdef FMA_PUSHPOP_PACKING
#pragma pack(pop)
#endif

/////////////////////////////////////////////////////////////////////////////
// Header cleanup
/////////////////////////////////////////////////////////////////////////////

  // Terminate C/C++ frame

#ifdef __cplusplus
}
#endif

#endif  // #ifndef FMA_DEF_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// End of FmaDef.h
/////////////////////////////////////////////////////////////////////////////
