/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Dynamic string routines module
//
// FmaDsz.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 27/Aug/1999
//
// History: 01/Jun/1992 kvitekp - Initially created for MadMED project.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine adds specified string to the dynamic string
//

  FMAAPI_(BOOL) FmaAddDsz(PDSZ pdsz, PCH pch, UINT cch)
  {
	UINT cchAlloc, cchGrow = DSZ_GROW;
	PSZ pszAlloc;

	ASSERT(pdsz != NULL);

	// Check for null string

	if (!pch) return TRUE;

	// Check if adding an asciiz string and get its lengs

	if (cch == (UINT)-1) cch = strlen(pch);

	// Check if string is not yet allocated or currently allocated string
	// can't accomodate characters caller wants to add

	if (pdsz->psz == NULL || pdsz->ich + cch > pdsz->cch) {

	  // Tweak allocation grow factor to prevent excessive heap fragmentation

	  if (pdsz->cch > DSZ_GROW * DSZ_GROWTWEAK) {
		cchGrow+= pdsz->cch / DSZ_GROWTWEAK;
#ifdef _DEBUG
//Log("FmaAddDsz: dsz.cch=%u cchGrow=%u", pdsz->cch, cchGrow);
#endif
	  }

	  // Reallocate string buffer and check if ok

	  cchAlloc = pdsz->ich + max(cch, cchGrow);
	  if ((pszAlloc = MemRealloc(pdsz->psz, cchAlloc + 1)) != NULL) {
		pdsz->cch = cchAlloc;
		pdsz->psz = pszAlloc;
	  } else
		return FALSE;
	}

	// Add characters and fix string end

	memcpy(pdsz->psz + pdsz->ich, pch, cch);
	pdsz->ich+= cch;
	pdsz->psz[pdsz->ich] = '\0';

	ASSERT(pdsz->ich <= pdsz->cch);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds formatted string to the dynamic string given var arg
//

  FMAAPIV_(BOOL) FmaAddDszFormat(PDSZ pdsz, PSZ pszFormat, ...)
  {
	va_list parg;
	va_start(parg, pszFormat);

	return FmaAddDszFormatV(pdsz, pszFormat, parg);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds formatted string to the dynamic string given arg pointer
//

  FMAAPI_(BOOL) FmaAddDszFormatV(PDSZ pdsz, PSZ pszFormat, va_list parg)
  {
	UINT cch, cchAlloc, cchGrow = DSZ_GROW;
	PSZ pszAlloc;

	ASSERT(pdsz != NULL);
	ASSERT(pszFormat != NULL);

	// Check if there is any buffer allocation and try to format
	// string into into it, otherwise assume we can't fit

	cch = pdsz->cch - pdsz->ich; 
	if (pdsz->psz && cch > strlen(pszFormat)) {
	  cch = FormatV(pdsz->psz + pdsz->ich, cch, pszFormat, parg);
	} else
	  cch = (ULONG)-1;

	// Loop while we can't fit the formatted string into the buffer

	while (cch == (UINT)-1) {

	  // Tweak allocation grow factor to prevent excessive heap fragmentation

	  if (pdsz->cch > DSZ_GROW * DSZ_GROWTWEAK) {
		cchGrow+= pdsz->cch / DSZ_GROWTWEAK;
#ifdef _DEBUG
//Log("FmaAddDszFormatV: dsz.cch=%u cchGrow=%u", pdsz->cch, cchGrow);
#endif
	  }

	  // Reallocate dynamic string buffer and check if ok

	  cchAlloc = pdsz->cch + cchGrow;
	  if ((pszAlloc = MemRealloc(pdsz->psz, cchAlloc + 1)) != NULL) {
		pdsz->psz = pszAlloc;
		pdsz->cch = cchAlloc;
	  } else 
		return FALSE;

	  // Format string and check if it fits into the buffer

	  cch = pdsz->cch - pdsz->ich;
	  cch = FormatV(pdsz->psz + pdsz->ich, cch, pszFormat, parg);
	}

	// Adjust dynamic string length

	pdsz->ich+= cch;

	ASSERT(pdsz->ich <= pdsz->cch);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets specified string to the dynamic string
//

  FMAAPI_(BOOL) FmaSetDsz(PDSZ pdsz, PCH pch, UINT cch)
  {
	ASSERT(pdsz != NULL);

	pdsz->ich = 0;

	return FmaAddDsz(pdsz, pch, cch);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets formatted string to the dynamic string given var arg
//

  FMAAPIV_(BOOL) FmaSetDszFormat(PDSZ pdsz, PSZ pszFormat, ...)
  {
	va_list parg;
	va_start(parg, pszFormat);

	ASSERT(pdsz != NULL);

	pdsz->ich = 0;

	return FmaAddDszFormatV(pdsz, pszFormat, parg);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets formatted string to the dynamic string given arg pointer
//

  FMAAPI_(BOOL) FmaSetDszFormatV(PDSZ pdsz, PSZ pszFormat, va_list parg)
  {
	ASSERT(pdsz != NULL);

	pdsz->ich = 0;

	return FmaAddDszFormatV(pdsz, pszFormat, parg);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine shrinks dynamic string allocation
//

  FMAAPI_(BOOL) FmaFixDsz(PDSZ pdsz)
  {
	PSZ psz;

	ASSERT(pdsz != NULL);

	// Check if dynamic string length is less than its allocation and
	// leave if not

	if (!(pdsz->ich < pdsz->cch))
	  return FALSE;

	// Check if dynamic string is empty and free its allocation

	if (!pdsz->ich) {
	  FmaDelDsz(pdsz);
	  return TRUE;
	}

	// Reallocate dynamic string to match its length

	if ((psz = MemRealloc(pdsz->psz, pdsz->ich + 1)) != NULL) {
	  pdsz->cch = pdsz->ich;
	  pdsz->psz = psz;
	}

	ASSERT(pdsz->psz != NULL);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes dynamic string
//

  FMAAPI_(VOID) FmaDelDsz(PDSZ pdsz)
  {
	ASSERT(pdsz != NULL);

	// Free dynamic string buffer

	if (pdsz->psz) MemFree(pdsz->psz);

	// Reset dynamic list controls

	pdsz->cch = 0;
	pdsz->ich = 0;
	pdsz->psz = NULL;
  }

/////////////////////////////////////////////////////////////////////////////
// End of FmaDsz.c
/////////////////////////////////////////////////////////////////////////////
