/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Date/time routines module
//
// FmaDttm.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 26/Sep/1999
// Updated: 10/Feb/2000
//
// History: 27/Sep/1998 kvitekp - Initially created for MBS project
//
//          10/Feb/2000 kvitekp - Fixed Y2K problem in FmaAsciiToDttm()
//
/////////////////////////////////////////////////////////////////////////////
//
// Note: Some of this code is based on JAM time routines published by 
//       JAM message base authors in JAMAPI.ARJ
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Convenient defines

#define SECS_IN_DAY  (24L * 60L * 60L)

  // Short month names

  static PSZ aMonth[] = {"Jan","Feb","Mar","Apr","May","Jun",
						 "Jul","Aug","Sep","Oct","Nov","Dec"};

  //  Number of days since January 1, the 1st of each month

  static int aDaysFromJan1st[13] = {
	0,                                  // Jan
	31,                                 // Feb
	31+28,                              // Mar
	31+28+31,                           // Apr
	31+28+31+30,                        // May
	31+28+31+30+31,                     // Jun
	31+28+31+30+31+30,                  // Jul
	31+28+31+30+31+30+31,               // Aug
	31+28+31+30+31+30+31+31,            // Sep
	31+28+31+30+31+30+31+31+30,         // Oct
	31+28+31+30+31+30+31+31+30+31,      // Nov
	31+28+31+30+31+30+31+31+30+31+30,   // Dec
	31+28+31+30+31+30+31+31+30+31+30+31 // Jan
  };

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine performs the simple string pattern matching
//

  static BOOL FMASUB DoCheckPattern(PSZ psz, PSZ pszPattern)
  {
	// Scan string to match the pattern

	for ( ; *pszPattern; pszPattern++, psz++) {
	  switch (*pszPattern) {
		case 'D': if (!ISDIGIT(*psz))                   return FALSE; break;
		case 'd': if (!ISDIGIT(*psz) && (*psz != ' '))  return FALSE; break;
		case 'A': if (!ISALPHA(*psz))                   return FALSE; break;
		case 'a': if (!ISALPHA(*psz) && (*psz != ' '))  return FALSE; break;
		case '/': if (!strchr("-/ ",  *psz))            return FALSE; break;
		default : if (*psz != *pszPattern)              return FALSE; break;
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns month number for a given month string
//

  static BOOL FMASUB DoGetMonth(PSZ psz)
  {
	UINT iMonth;

	// Scan the months string array looking for a match

	for (iMonth = 0; iMonth < numbof(aMonth); iMonth++) {
	  if (FmaEqPch(psz, aMonth[iMonth], 3))
		return 1 + iMonth;
	}

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine checks the date/time info 
//

  FMAAPI_(BOOL) FmaCheckDttm(PDTTM pdttm)
  {
	DTTM dttm = *pdttm;

	// Validate date time specification

	if (!(                dttm.hh <= 23)) return FALSE;
	if (!(                dttm.mm <= 59)) return FALSE;
	if (!(                dttm.ss <= 29)) return FALSE;

	if (!(dttm.da >= 1 && dttm.da <= 31)) return FALSE;
	if (!(dttm.mo >= 1 && dttm.mo <= 12)) return FALSE;
	if (!(                dttm.yr <= 99)) return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine converts ascii date into dttm
//

  FMAAPI_(BOOL) FmaAsciiToDttm(PSZ psz, PDTTM pdttm)
  {
	static CHAR achDateSep[] = " /-";
	DTTM dttm;

	// Clear target structure

	*((PULONG)&dttm) = 0;

	// Scan in various formats of day and month

	if (DoCheckPattern(psz, "dD/AAA/")) {
	  dttm.da = atoi(psz);
	  dttm.mo = DoGetMonth(psz + lengof("dD/"));
	  psz+= lengof("dD/AAA/");
	} else

	if (DoCheckPattern(psz, "dD/DD/")) {
	  dttm.da = atoi(psz);
	  dttm.mo = atoi(psz + lengof("dD/"));
	  psz+= lengof("dD/DD/");
	} else
	  return FALSE;

	// Scan in various formats of year

	if (DoCheckPattern(psz, "DD ")) {
	  int year = atoi(psz);
	  dttm.yr = (year < 80) ? year + 20 : year - 80;
	  psz+= lengof("DD ");
	} else

	if (DoCheckPattern(psz, "DDDD ")) {
	  dttm.yr = atoi(psz) - 1980;
	  psz+= lengof("DDDD ");
	} else
	  return FALSE;

	// Skip spaces

	while (*psz == ' ') psz++;

	// Scan in various formats of time

	if (DoCheckPattern(psz, "D:DD:DD")) {
	  dttm.hh = atoi(psz);
	  dttm.mm = atoi(psz + lengof("D:"));
	  dttm.ss = atoi(psz + lengof("D:DD:")) / 2;
	} else

	if (DoCheckPattern(psz, "DD:DD:DD")) {
	  dttm.hh = atoi(psz);
	  dttm.mm = atoi(psz + lengof("DD:"));
	  dttm.ss = atoi(psz + lengof("DD:DD:")) / 2;
	} else
	  return FALSE;

	// Verify resulting dttm

	if (!FmaCheckDttm(&dttm)) return FALSE;

	// Pass dttm to caller

	if (pdttm) *pdttm = dttm;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine formats date/time string
//

  FMAAPI_(UINT) FmaFormatDttm(PCH pchDttm, UINT cchMax, PDTTM pdttm, ULONG fl,
							  CHAR chDateSep, CHAR chTimeSep, 
							  PSZ pszDateTimeSep)
  {
	CHAR achTime[16], achDate[16];
	DTTM dttm = *pdttm;
	UINT cch, year;

	// Fix up invalid dttm

	if (!FmaCheckDttm(&dttm)) {
	  *((PULONG)&dttm) = 0;
	  dttm.da = 1;
	  dttm.mo = 1;
	}

	// Use default separators if not specified

	if (!chDateSep) chDateSep = '/';
	if (!chTimeSep) chTimeSep = ':';
	if (!pszDateTimeSep) pszDateTimeSep = " ";

	// Check if we need to format time

	if (!(fl & FDTTM_DATEONLY)) {

	  // Format time as specified

	  if (fl & FDTTM_TIMESECONDS) {
		Format(achTime, lengof(achTime), "%02u%c%02u%c%02u", 
			   dttm.hh, chTimeSep,
			   dttm.mm, chTimeSep,
			   dttm.ss * 2);
	  } else {
		Format(achTime, lengof(achTime), "%02u%c%02u%", 
			   dttm.hh, chTimeSep,
			   dttm.mm);
	  }
	}

	// Check if we need to format date

	if (!(fl & FDTTM_TIMEONLY)) {

	  // Format day and month as specified

	  if (fl & FDTTM_MONTHNUMBER) {
		cch = Format(achDate, lengof(achDate), "%02u%c%02u%c", 
					 dttm.da, chDateSep, 
					 dttm.mo, chDateSep);
	  } else {
		cch = Format(achDate, lengof(achDate), "%02u%c%s%c", 
					 dttm.da, chDateSep,
					 aMonth[dttm.mo - 1], chDateSep);
	  }

	  // Format year as specified

	  if (fl & FDTTM_4DIGITYEAR) {
		year = 1980 + dttm.yr;
		cch+= Format(achDate + cch, lengof(achDate) - cch, "%04u", year);
	  } else {
		year = 1980 + dttm.yr;
		year = year < 2000 ? year - 1900 : year - 2000;
		cch+= Format(achDate + cch, lengof(achDate) - cch, "%02u", year);
	  }
	}

	// Format the final string for caller

	if (fl & FDTTM_TIMEONLY) {
	  cch = Format(pchDttm, cchMax, "%s", achTime);
	} else 
	if (fl & FDTTM_DATEONLY) {
	  cch = Format(pchDttm, cchMax, "%s", achDate);
	} else 
	if (fl & FDTTM_TIMEFIRST) {
	  cch = Format(pchDttm, cchMax, "%s%s%s", achTime, pszDateTimeSep, achDate);
	} else {
	  cch = Format(pchDttm, cchMax, "%s%s%s", achDate, pszDateTimeSep, achTime);
	}

	return cch;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns time given dttm
//

  FMAAPI_(ULONG) FmaGetTimeFromDttm(PDTTM pdttm)
  {
	UINT hour, min, sec;
	UINT day, mon, year;
	ULONG nDaysSince1970;
	UINT nYearsSince1970;
	DTTM dttm = *pdttm;

	// Fix up invalid dttm

	if (!FmaCheckDttm(&dttm)) {
	  *((PULONG)&dttm) = 0;
	  dttm.da = 1;
	  dttm.mo = 1;
	}

	// Pick up date and time on stack and massage 
	// them a little bit for convenience

	hour = dttm.hh;
	min  = dttm.mm;
	sec  = dttm.ss * 2;

	day  = dttm.da;
	mon  = dttm.mo;
	year = dttm.yr + 80;        // Years since 1900

	// Get number of years since 1970

	nYearsSince1970 = year - 70;

	// Calculate number of days during these years,
	// including extra days for leap years

	nDaysSince1970 = nYearsSince1970 * 365 + ((nYearsSince1970 + 1) / 4);

	// Add the number of days during this year,
	// including extra day if in leap year

	nDaysSince1970+= aDaysFromJan1st[mon - 1] + (day - 1);
	if ((year & 3) == 0 && mon > 2) nDaysSince1970++;

	// Convert to seconds, and add the number of seconds this day

	return (((ULONG) nDaysSince1970 * (24L * 60L * 60L)) 
		  + ((ULONG) hour * (60L * 60L))
		  + ((ULONG) min * 60L) 
		  +  (ULONG) sec);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns dttm given time
//

  FMAAPI_(BOOL) FmaGetDttmFromTime(PDTTM pdttm, ULONG time)
  {
	int sec, min, hour, day, mon, year;
	int nDayOfYear;
	int nLeapDay;
	LONG n;

	// Calculate hour, minute and second

	n = time;
	sec  = (UINT)(n % 60); n/= 60;
	min  = (UINT)(n % 60); n/= 60;
	hour = (UINT)(n % 24); n/= 24;

	// Now n is number of days since Jan 1st, 1970,
	// find out year since 1970

	year = (UINT)(n / 365 + 1);
	do {
	  year--;
	  nDayOfYear = (int)(n - year * 365 - ((year + 1) / 4));
	} while (nDayOfYear < 0);

	// Convert to year since 1900 and calculate leap day

	year+= 70;
	nLeapDay = ((year & 3) == 0 && nDayOfYear >= aDaysFromJan1st[2]);

	// Find out month and day by scanning months table

	for (mon = day = 0; day == 0; mon++) {
	  if (nDayOfYear < aDaysFromJan1st[mon + 1] + nLeapDay) {
		day = nDayOfYear + 1 - (aDaysFromJan1st[mon] + (mon != 1 ? nLeapDay : 0));
	  }
	}

	// Check if we've got a year which can be converted to dttm year

	if (year < 80) {
	  if (pdttm) *((PLONG)pdttm) = 0;
	  FMAERROR(INVALID_DATA);
	  return FALSE;
	}

	// Convert what we've got to dttm 

	if (pdttm) {

	  pdttm->hh = hour;
	  pdttm->mm = min;
	  pdttm->ss = sec / 2;    // Dttm seconds are half of real seconds

	  pdttm->da = day;
	  pdttm->mo = mon ;
	  pdttm->yr = year - 80;  // Convert to year since 1980

	  ASSERT(FmaCheckDttm(pdttm));
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of FmaDttm.c
/////////////////////////////////////////////////////////////////////////////
