/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Fidonet Mail Message routines module
//
// FmaFmm.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 25/Sep/1999
// Updated: 10/Feb/2000
//
// History: 25/Sep/1999 kvitekp - Initially created.
//
//          29/Jan/2000 kvitekp - Added FmaDupeMsg(pfmm)
//
//          10/Feb/2000 kvitekp - Added FmaAddFmpFormatV()
//
//          25/Feb/2000 kvitekp - Fixes ss*2 in FmaDumpFmp()
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine creates fidonet mail message property object
//

  static PFMP FMASUB DoCreateFmp(UINT cbData)
  {
	PFMP pfmp;
	UINT cb;

	// Allocate memory to hold fmp object and check if ok

	cb = sizeof(FMP) + cbData;
	if ((pfmp = MemAlloc(cb, MA_CLEAR)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return NULL;
	}

	// Initialize fmm object

	pfmp->cb = cbData;

	return pfmp;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine destroys fidonet mail message property object
//

  static BOOL FMASUB DoDestroyFmp(PFMP pfmp)
  {
	PSZ * ppsz;
	ULONG cpsz;

	// Check if no object specified

	if (!pfmp) return TRUE;

	// Check if this is a string pointer property and
	// free all associated strings

	if (FMP_FLG(pfmp->id) & FMP_PSZ) {
	  ppsz = PFMPDATA(pfmp);
	  cpsz = pfmp->cb / sizeof(PSZ);
	  while (cpsz--) {
		if (ppsz) MemFree(*ppsz);
		ppsz++;
	  }
	}

	// Free object memory

	MemFree(pfmp);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Fidonet mail message routines
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine creates fidonet mail message object
//

  FMAAPI_(PFMM) FmaCreateMsg(UINT cbExtra)
  {
	PFMM pfmm;
	UINT cb;

	// Allocate memory to hold fmm object and check if ok

	cb = sizeof(FMM) + cbExtra;
	if ((pfmm = MemAlloc(cb, MA_CLEAR)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return NULL;
	}

	// Initialize fmm object

	pfmm->cb = cb;

	return pfmm;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine destroys fidonet mail message object
//

  FMAAPI_(BOOL) FmaDestroyMsg(PFMM pfmm)
  {
	// Check if no object specified

	if (!pfmm) return TRUE;

	// Free associated properties

	while (pfmm->pfmp) {
	  FmaDelFmp(pfmm, pfmm->pfmp->id, 0);
	}

	// Free object memory

	MemFree(pfmm);    

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine duplicates fidonet mail message object
//

  FMAAPI_(PFMM) FmaDupeMsg(PFMM pfmm)
  {
	PFMP pfmpDupe, pfmp;
	PFMM pfmmDupe;

	// Check if no object specified

	if (!pfmm) return NULL;

	// Create new fmm and check if ok

	if ((pfmmDupe = FmaCreateMsg(pfmm->cb)) == NULL)
	  return NULL;

	// Copy the fmm structure info and reset copied 
	// properties list pointer since it belongs to 
	// the original message

	memcpy(pfmmDupe, pfmm, pfmm->cb);
	pfmmDupe->pfmp = NULL;

	// Loop through all the properties

	for (pfmp = pfmm->pfmp; pfmp != NULL; pfmp = pfmp->pfmpNext) {

	  // Add a copy of the original property

	  pfmpDupe = FmaAddFmp(pfmmDupe, pfmp->id, PFMPDATA(pfmp), pfmp->cb);
	  if (!pfmpDupe) {
Fail:   FmaDestroyMsg(pfmmDupe);
		return NULL;
	  }

	  // Check if this is a string pointer array and reallocate it

	  if (pfmp->id & FMP_PSZ) {
		PSZ * ppsz = PFMPDATA(pfmp); int n; 
		PSZ * ppszDupe = PFMPDATA(pfmpDupe); 
		memset(PFMPDATA(pfmpDupe), 0, pfmpDupe->cb);
		for (n = 0; n < (int)(pfmp->cb / sizeof(PSZ)); n++) {
		  UINT cb = strlen(*ppsz) + 1;
		  if ((*ppszDupe = MemAlloc(cb, 0)) == NULL) {
			FMAERROR(NOT_ENOUGH_MEMORY);
			goto Fail;
		  }
		  memcpy(*ppszDupe, *ppsz, cb);
		  ppsz++; ppszDupe++;
		}
	  }
	}

	return pfmmDupe;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine dumps fidonet mail message object
//

  FMAAPI_(BOOL) FmaDumpMsg(PFMM pfmm)
  {
#ifndef FMA_LOG	

	FMAERROR(NOT_SUPPORTED);
	return FALSE;

#else
	PFMP pfmp;

	// Check if no object specified

	if (!pfmm) return TRUE;

	// Dump the message header

	Log("\nFmaDumpMsg ------>: [%09Fp] umsg=%ld type=%d attr=%08lx", pfmm, pfmm->umsg, pfmm->type, pfmm->attr);

	// Dump the message properties

	for (pfmp = pfmm->pfmp; pfmp != NULL; pfmp = pfmp->pfmpNext) {
	  FmaDumpFmp(pfmp);
	}

	// Finish dump

	Log("FmaDumpMsg ------<: [%09Fp] ", pfmm);

	return TRUE;
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// Fidonet mail message property routines
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine gets fidonet mail message property
//

  FMAAPI_(PFMP) FmaGetFmp(PFMM pfmm, ULONG id, UINT ifmp)
  {
	PFMP pfmp;

	ASSERT(pfmm != NULL);

	// Loop through the property list

	for (pfmp = pfmm->pfmp; pfmp != NULL; pfmp = pfmp->pfmpNext) {

	  // Check for a property id match and skip specified amount 
	  // of such property objects before returning one

	  if (pfmp->id == id && !ifmp--)
		return pfmp;
	}

	// Set error code

	FMAERROR(NOT_FOUND);

	return NULL;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds fidonet mail message property
//

  FMAAPI_(PFMP) FmaAddFmp(PFMM pfmm, ULONG id, PVOID pbData, ULONG cbData)
  {
	ULONG cbAlloc = cbData;
	PFMP pfmp, pfmpThis;
	UINT ifmp, ilst;

	ASSERT(pfmm != NULL);

	// Check if property is an asciiz string and get its length,
	// then add a byte to the allocation for the null terminator

	if (FMP_FLG(id) & FMP_ASZ) {
	  if (cbData == (ULONG)-1) cbData = pbData ? strlen(pbData) : 0;
	  cbAlloc = cbData + 1;
	}

	// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
	if (!CHECK_DATA_SIZE_LIMIT(cbAlloc + sizeof(FMP))) {
	  FMAERROR(DATA_SIZE_LIMIT);
	  return NULL;
	}
#endif

	// Create property object and check if ok

	if ((pfmpThis = DoCreateFmp((int)cbAlloc)) == NULL) 
	  return NULL;

	// Initialize new property header

	pfmpThis->id = id;
	pfmpThis->cb = cbData;

	// Initialize new property data

	if (pbData) memcpy(PFMPDATA(pfmpThis), pbData, (int)cbData);

	// Check if this is single instance property and remove 
	// any existing instances of the same property

	if (FMP_FLG(id) & FMP_ONE) {
	  while (FmaDelFmp(pfmm, id, 0));
	}

	// Loop through the property list looking for the property
	// with the higher position than this one or list end

	for (pfmp = pfmm->pfmp, ifmp = 0; pfmp != NULL; 
		 pfmp = pfmp->pfmpNext, ifmp++) {
	  if (FMP_POS(pfmp->id) > FMP_POS(id))
		break;
	}

	// Link new property to the property list

	ilst = ListAdd(&pfmm->pfmp, pfmpThis, ifmp);

	ASSERT(ilst == ifmp);

	return pfmpThis;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds formatted fidonet mail message property given var arg
//

  FMAAPIV_(PFMP) FmaAddFmpFormat(PFMM pfmm, ULONG id, PSZ pszFormat, ...)
  {
	va_list parg;
	va_start(parg, pszFormat);

	// Fall to arg ptr backend routine to do the job

	return FmaAddFmpFormatV(pfmm, id, pszFormat, parg);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds formatted fidonet mail message property given arg ptr
//

  FMAAPI_(PFMP) FmaAddFmpFormatV(PFMM pfmm, ULONG id, PSZ pszFormat, va_list parg)
  {
	DSZ dsz = DSZ_INIT;
	PFMP pfmp;

	ASSERT(FMP_FLG(id) & FMP_ASZ);

	// Format the property into dynamic string and check if ok

	if (!AddDszFormatV(&dsz, pszFormat, parg)) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return NULL;
	}

	// Add fidonet mail message property

	pfmp = FmaAddFmp(pfmm, id, dsz.psz, dsz.ich);

	// Free dynamic string

	DelDsz(&dsz);

	return pfmp;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes fidonet mail message property
//

  FMAAPI_(BOOL) FmaDelFmp(PFMM pfmm, ULONG id, UINT ifmp)
  {
	PFMP pfmp, pfmpDel;
	UINT ilst;

	ASSERT(pfmm != NULL);

	// Loop through the property list

	for (pfmp = pfmm->pfmp, ilst = 0; pfmp != NULL; 
		 pfmp = pfmp->pfmpNext, ilst++) {

	  // Check for a property id match and skip specified
	  // amout of such ids before deleting one

	  if (pfmp->id == id && !ifmp--) {
		pfmpDel = ListDel(&pfmm->pfmp, ilst);
		ASSERT(pfmpDel == pfmp);
		DoDestroyFmp(pfmp);
		return TRUE;
	  }
	}

	// Set error code

	FMAERROR(NOT_FOUND);

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine dumps fidonet mail message property
//

  FMAAPI_(BOOL) FmaDumpFmp(PFMP pfmp)
  {
#ifndef FMA_LOG	

	FMAERROR(NOT_SUPPORTED);
	return FALSE;

#else
	ULONG fl = FMP_FLG(pfmp->id);
	DSZ dsz = DSZ_INIT;
	PSZ * ppsz;
	DTTM dttm;
	PSZ psz; 
	int n;

	// Check if no object specified

	if (!pfmp) return TRUE;

	// Make up the fidonet message property name string

	switch (pfmp->id) {
	  case FMP_MESSAGETEXT:     psz = "FMP_MESSAGETEXT";        break;
	  case FMP_SENDERNAME:      psz = "FMP_SENDERNAME";         break;
	  case FMP_RECEVRNAME:      psz = "FMP_RECEVRNAME";         break;
	  case FMP_ORIGADDR:        psz = "FMP_ORIGADDR";           break;
	  case FMP_DESTADDR:        psz = "FMP_DESTADDR";           break;
	  case FMP_GATEADDR:        psz = "FMP_GATEADDR";           break;
	  case FMP_SUBJECT:         psz = "FMP_SUBJECT";            break;
	  case FMP_FILEATTACH:      psz = "FMP_FILEATTACH";         break;
	  case FMP_FILEREQUEST:     psz = "FMP_FILEREQUEST";        break;
	  case FMP_FILEUPDATEREQ:   psz = "FMP_FILEUPDATEREQ";      break;
	  case FMP_DTTMWRITTEN:     psz = "FMP_DTTMWRITTEN";        break;
	  case FMP_DTTMRECEIVED:    psz = "FMP_DTTMRECEIVED";       break;
	  case FMP_DTTMPROCESSED:   psz = "FMP_DTTMPROCESSED";      break;
	  case FMP_MSGREPLYTO:      psz = "FMP_MSGREPLYTO";         break;
	  case FMP_MSGREPLIES:      psz = "FMP_MSGREPLIES";         break;
	  case FMP_MSGREPLYNEXT:    psz = "FMP_MSGREPLYNEXT";       break;
	  case FMP_TIMESREAD:       psz = "FMP_TIMESREAD";          break;
	  case FMP_PASSWORDCRC:     psz = "FMP_PASSWORDCRC";        break;
	  case FMP_MESSAGECOST:     psz = "FMP_MESSAGECOST";        break;
	  case FMP_AREA:            psz = "FMP_AREA";               break;
	  case FMP_MSGID:           psz = "FMP_MSGID";              break;
	  case FMP_REPLYID:         psz = "FMP_REPLYID";            break;
	  case FMP_PID:             psz = "FMP_PID";                break;
	  case FMP_TID:             psz = "FMP_TID";                break;
	  case FMP_TZUTC:           psz = "FMP_TZUTC";              break;
	  case FMP_FLAGS:           psz = "FMP_FLAGS";              break;
	  case FMP_CHARSET:         psz = "FMP_CHARSET";            break;
	  case FMP_REALNAME:        psz = "FMP_REALNAME";           break;
	  case FMP_REPLYADDR:       psz = "FMP_REPLYADDR";          break;
	  case FMP_REPLYTO:         psz = "FMP_REPLYTO";            break;
	  case FMP_FTSKLUDGE:       psz = "FMP_FTSKLUDGE";          break;
	  case FMP_EOMKLUDGE:       psz = "FMP_EOMKLUDGE";          break;
	  case FMP_SEENBY2D:        psz = "FMP_SEENBY2D";           break;
	  case FMP_PATH2D:          psz = "FMP_PATH2D";             break;
	  default:                  psz = "<unknown>";              break;
	}

	// Make up the asciiz string property data string

	if (fl & FMP_ASZ) {
		AddDszFormat(&dsz, "ASZ: \"%s\"", PFMPDATA(pfmp));
	} else

	// Make up the asciiz string pointer property data string

	if (fl & FMP_PSZ) {
	  AddDszFormat(&dsz, "PSZ: ");
	  ppsz = PFMPDATA(pfmp);
	  for (n = 0; n < (int)(pfmp->cb / sizeof(PSZ)); n++) {
		AddDszFormat(&dsz, "\"%.128s\"", *ppsz++);
	  }
	} else

	// Make up the 32-bit value property data string

	if (fl & FMP_U32) {
	  AddDszFormat(&dsz, "U32: ");
	  for (n = 0; n < (int)(pfmp->cb / sizeof(ULONG)); n++) {
		AddDszFormat(&dsz, "%08lx ", ((PULONG)PFMPDATA(pfmp))[n]);
	  }
	} else

	// Make up the counted binary array property data string

	if (fl & FMP_BIN) {
	  AddDszFormat(&dsz, "BIN: ");
	} else

	// Make up the date/time property data string

	if (fl & FMP_DTM) {
	  dttm = *((PDTTM)PFMPDATA(pfmp));
	  AddDszFormat(&dsz, "DTM: %02u:%02u:%02u %02u/%02u/%04u",
				   dttm.hh, dttm.mm, dttm.ss * 2,
				   dttm.da, dttm.mo, dttm.yr + 1980);
	} else

	// Unknown property

	{
	  AddDszFormat(&dsz, "???: ");
	}

	// Dump the message header

	Log("FmaDumpFmp       >: [%09Fp] id=%08lx cb=%-3ld %-20s %s", pfmp, pfmp->id, pfmp->cb, psz, dsz.psz);

	// Free dynamic string

	DelDsz(&dsz);

	return TRUE;
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns specified string property text length
//

  FMAAPI_(ULONG) FmaGetFmpTextLen(PFMP pfmp)
  {
	ULONG cch = 0;
	PSZ * ppsz;
	ULONG cpsz;

	// Check if no object specified

	if (!pfmp) return FALSE;

	// Check if this is an asciiz property

	if (FMP_FLG(pfmp->id) & FMP_ASZ) {

	  // Get length of the asciiz string. Note: don't count
	  // on property size since string may be shorter 

	  cch = strlen(PFMPDATA(pfmp));

	} else 

	// Check if this is an psz array property

	if (FMP_FLG(pfmp->id) & FMP_PSZ) {

	  // Loop through all the string pointers

	  ppsz = PFMPDATA(pfmp);
	  cpsz = pfmp->cb / sizeof(PSZ);
	  while (cpsz--) {
		if (ppsz) cch+= strlen(*ppsz);
		ppsz++;
	  }

	} else  {

	  // This is not a string property

	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	return cch;
  }

/////////////////////////////////////////////////////////////////////////////
// End of FmaFmm.c
/////////////////////////////////////////////////////////////////////////////
