/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Formatting routines module
//
// FmaFmt.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 27/Oct/1999
//
// History: 05/Mar/1995 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

  // Include standard headers

#include <limits.h>

  // Don't compile this module unless using private formatting

#ifdef FMA_PRIVATE_FORMAT

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Borland compilers can't expand inline functions with while statement

#if defined(__BORLANDC__)
#undef  FMAINLINE
#define FMAINLINE
#endif

  // Microsoft Visual C++ warning control

#if _MSC_VER >= 800
#pragma warning(disable:4701)   // local variable may be used without having been initialized
#endif

  // Sizes of int/long/short/pointer

#ifdef __FLAT__
#define LONG_IS_INT      1      // long is same size as int
#define SHORT_IS_INT     0      // short is same size as int
#define PTR_IS_INT       1      // ptr is same size as int
#define PTR_IS_LONG      1      // ptr is same size as long
#else
#define LONG_IS_INT      0      // long is same size as int
#define SHORT_IS_INT     1      // short is same size as int
#define PTR_IS_INT       0      // ptr is same size as int
#define PTR_IS_LONG      1      // ptr is same size as long
#endif

  // Define argument retrieval macros

#define GET_ARG_INT(parg)    (int)va_arg(parg, int)

#if LONG_IS_INT
#define GET_ARG_LONG(parg)   (long)GET_ARG_INT(parg)
#else
#define GET_ARG_LONG(parg)   (long)va_arg(parg, long)
#endif

#if SHORT_IS_INT
#define GET_ARG_SHORT(parg)  (short)GET_ARG_INT(parg)
#else
#define GET_ARG_SHORT(parg)  va_arg(parg, short)
#endif

#if PTR_IS_INT
#define GET_ARG_PTR(parg)    (void *)GET_ARG_INT(parg)
#elif PTR_IS_LONG
#define GET_ARG_PTR(parg)    (void *)GET_ARG_LONG(parg)
#else  
#error Pointer size must be size of int or long
#endif  

  // Formatting parameter block

  typedef struct _FPB { /* fpb */
	PCH     pch;              // Pointer into pchTarged where to place next character
	int     cchMax;           // Count of characters allowed to be written into pch
	int     cchOut;           // Count of characters written or -1 if error occurres
  } FPB, * PFPB;

  // Define macros to call output routines

#define WRITE_CHAR(ch)            DoWriteChar(ch, &fpb)
#define WRITE_MULTI_CHAR(ch, cch) DoWriteMultiChar(ch, cch, &fpb)
#define WRITE_STRING(pch, cch)    DoWriteString(pch, cch, &fpb)

  // Size of conversion buffer (ANSI-specified minimum is 509)

#define BUFFERSIZE    512

  // Flag definitions

#define FL_SIGN       0x0001  // put plus or minus in front
#define FL_SIGNSP     0x0002  // put space or minus in front
#define FL_LEFT       0x0004  // left justify
#define FL_LEADZERO   0x0008  // pad with leading zeros
#define FL_LONG       0x0010  // long value given
#define FL_SHORT      0x0020  // short value given
#define FL_SIGNED     0x0040  // signed data given
#define FL_ALTERNATE  0x0080  // alternate form requested
#define FL_NEGATIVE   0x0100  // value is negative
#define FL_FORCEOCTAL 0x0200  // force leading '0' for octals

  // Character type values

  enum CHARTYPE {
	CH_OTHER,                 // 0: character with no special meaning
	CH_PERCENT,               // 1: '%'
	CH_DOT,                   // 2: '.'
	CH_STAR,                  // 3: '*'
	CH_ZERO,                  // 4: '0'
	CH_DIGIT,                 // 5: '1'..'9'
	CH_FLAG,                  // 6: ' ', '+', '-', '#'
	CH_SIZE,                  // 7: 'l', 'L', 'N', 'F'
	CH_TYPE                   // 8: type specifying character
  };

  // Macro to retrieve character type

#define GET_CHAR_TYPE(c)  ((c) < ' ' || (c) > 'x'               \
							? CH_OTHER                          \
							: achTab[(c) - ' '] & 0x0F)
  // State values

  enum STATE {
	ST_NORMAL,                // 0: normal state; outputting literal chars
	ST_PERCENT,               // 1: just read '%'
	ST_FLAG,                  // 2: just read flag character
	ST_WIDTH,                 // 3: just read width specifier
	ST_DOT,                   // 4: just read '.'
	ST_PRECIS,                // 5: just read precision specifier
	ST_SIZE,                  // 6: just read size specifier
	ST_TYPE                   // 7: just read type specifier
  };

#define NUMBER_OF_STATES (ST_TYPE + 1)

  // Macro to retrieve next state

#define GET_NEXT_STATE(type, state) (achTab[(type) * NUMBER_OF_STATES + (state)] >> 4)

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   d a t a                                       //
/////////////////////////////////////////////////////////////////////////////

  // The state table. This table is actually two tables combined into one.
  // The lower nibble of each byte gives the character type:
  //
  //    chartype = achTab[char - ' '] & 0x0F;
  //
  // The upper nibble of the byte gives the next state to enter:
  //
  //    newstate = achTab[chartype][oldstate] >> 4;
  //

  //norm  perc  flag  width dot   prec  size  type  // cur state

  static char achTab[89] = {                        // cur char
  // ' '   '!'   '"'   '#'   '$'   '%'   '&'   '''  
	0x06, 0x00, 0x00, 0x06, 0x00, 0x01, 0x00, 0x00, // CH_OTHER

  // '('   ')'   '*'   '+'   ','   '-'   '.'   '/'
	0x10, 0x00, 0x03, 0x06, 0x00, 0x06, 0x02, 0x10, // CH_PERCENT

  // '0'   '1'   '2'   '3'   '4'   '5'   '6'   '7'
	0x04, 0x45, 0x45, 0x45, 0x05, 0x05, 0x05, 0x05, // CH_DOT

  // '8'   '9'   ':'   ';'   '<'   '='   '>'   '?'
	0x05, 0x35, 0x30, 0x00, 0x50, 0x00, 0x00, 0x00, // CH_STAR

  // '@'   'A'   'B'   'C'   'D'   'E'   'F'   'G'
	0x00, 0x20, 0x20, 0x30, 0x50, 0x50, 0x07, 0x00, // CH_ZERO

  // 'H'   'I'   'J'   'K'   'L'   'M'   'N'   'O'
	0x00, 0x30, 0x30, 0x30, 0x57, 0x50, 0x07, 0x00, // CH_DIGIT

  // 'P'   'Q'   'R'   'S'   'T'   'U'   'V'   'W'
	0x00, 0x20, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, // CH_FLAG

  // 'X'   'Y'   'Z'   '['   '\'   ']'   '^'   '_'
	0x08, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x00, // CH_SIZE

  // '`'   'a'   'b'   'c'   'd'   'e'   'f'   'g'
	0x00, 0x70, 0x70, 0x78, 0x78, 0x70, 0x70, 0x00, // CH_TYPE

  // 'h'   'i'   'j'   'k'   'l'   'm'   'n'   'o'
	0x00, 0x08, 0x00, 0x00, 0x07, 0x00, 0x08, 0x08,

  // 'p'   'q'   'r'   's'   't'   'u'   'v'   'w'
	0x08, 0x00, 0x00, 0x08, 0x00, 0x08, 0x00, 0x00,

  // 'x'   'y'   'z'   '{'   '|'   '}'   '~'   ''
	0x08
  };

  // String to print on null pointer

  static char achNull[] = "(null)";

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   s u b r o u t i n e s                         //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine writes out single character
//

  static FMAINLINE VOID FMASUB DoWriteChar(int ch, PFPB pfpb)
  {
	if (pfpb->cchMax-- > 0) {
	  *pfpb->pch++ = (CHAR)ch; pfpb->cchOut++;
	} else
	  pfpb->cchOut = -1;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes out multiple characters
//

  static FMAINLINE VOID FMASUB DoWriteMultiChar(int ch, int cch, PFPB pfpb)
  {
	while (cch-- > 0) DoWriteChar(ch, pfpb);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes out string
//

  static FMAINLINE VOID FMASUB DoWriteString(PCH pch, int cch, PFPB pfpb)
  {
	while (cch-- > 0) DoWriteChar(*pch++, pfpb);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine is a formatting engine
//

  static int FMASUB DoFormatV(PCH pchTarget, UINT cchTarget, 
								 PSZ pszFormat, va_list parg)
  {
	enum STATE state = ST_NORMAL; // current state
	enum CHARTYPE chartype;       // current character type

	FPB fpb;

	CHAR ch;            // character just read
	CHAR achPrefix[2];  // numeric prefix -- up to two characters
	int cchPrefix;      // length of prefix -- 0 means no prefix

	int fl;             // Current flags (FL)

	int hexAdd;         // offset to add to number to get 'a'..'f'
	int radix;          // current conversion radix
	int cchField;       // selected field width -- 0 means default
	int cchPrec;        // selected precision  -- -1 means default

	int fNoOutput;      // non-zero = produce no output for this specifier

	PCH pchText;        // pointer text to be printed, not zero terminated
	int cchText = 0;    // length of the text in bytes to be printed.

	CHAR achText[BUFFERSIZE];

	ASSERT(pchTarget != NULL);
	ASSERT(pszFormat != NULL);
	ASSERT(parg != NULL);

	// Initialize formatting parameter block

	fpb.pch = pchTarget;
	fpb.cchMax = cchTarget;
	fpb.cchOut = 0;

	// Loop while format character exist and no errors

	while ((ch = *pszFormat++) != '\0' && fpb.cchOut >= 0) {

	  // Find character class

	  chartype = GET_CHAR_TYPE(ch);

	  // Find next state

	  state = GET_NEXT_STATE(chartype, state);

	  // Execute code for each state

	  switch (state) {

		case ST_NORMAL:

		  // Just write character

		  WRITE_CHAR(ch);

		  break;

		case ST_PERCENT:

		  // Set default values of conversion parameters

		  cchPrefix = cchField = 0; cchPrec = -1;
		  fNoOutput = FALSE;
		  fl = 0; 

		  break;

		case ST_FLAG:

		  // Set flag based on flag character

		  switch (ch) {
			case '-': fl|= FL_LEFT;       break;  // left justify
			case '+': fl|= FL_SIGN;       break;  // force sign indicator
			case ' ': fl|= FL_SIGNSP;     break;  // force sign or space
			case '#': fl|= FL_ALTERNATE;  break;  // alternate form
			case '0': fl|= FL_LEADZERO;   break;  // pad with leading zeros
		  }

		  break;

		case ST_WIDTH:

		  // Update field width value

		  if (ch == '*') {

			// Get width from arg list

			cchField = GET_ARG_INT(parg);

			// ANSI says neg fld width means '-' flag and pos width

			if (cchField < 0) {
			  fl|= FL_LEFT; cchField = -cchField;
			}
		  } else {

			// Add digit to current field width

			cchField = cchField * 10 + (ch - '0');
		  }

		  break;

		case ST_DOT:

		  // Zero the precision, since dot with no number means 0
		  // not default, according to ANSI

		  cchPrec = 0;

		  break;

		case ST_PRECIS:

		  // Update precison value

		  if (ch == '*') {

			// Get precision from arg list

			cchPrec = GET_ARG_INT(parg);

			// ANSI says neg precision means default

			if (cchPrec < 0) cchPrec = -1; 

		  } else {

			// Add digit to current precision

			cchPrec = cchPrec * 10 + (ch - '0');
		  }

		  break;

		case ST_SIZE:

		  // Just read a size specifier, set the flags based on it

		  switch (ch) {
#if !LONG_IS_INT
			case 'l': fl|= FL_LONG;   break;  // long int
#endif

#if !SHORT_IS_INT
			case 'h': fl|= FL_SHORT;  break;  // short int
#endif
		  }

		  break;

		case ST_TYPE:

		  // Act according to the type character

		  switch (ch) {

			case 'c': {

			  // Print a single character specified by an int argument

			  achText[0] = (char) GET_ARG_INT(parg);
			  pchText = achText;
			  cchText = 1;

			  } break;

			case 's': {

			  // Output a string

			  int cch;
			  PCH pch;

			  // Get string pointer and check if ok

			  if ((pchText = GET_ARG_PTR(parg)) == NULL)
				pchText = achNull;

			  // At this point it is tempting to use strlen(), but
			  // if a precision is specified, we're not allowed to
			  // scan past there, because there might be no null  
			  // at all. Thus, we must do our own scan.          

			  cch = (cchPrec == -1) ? INT_MAX : cchPrec;
			  pch = pchText;

			  // Scan for null up to cch characters

			  while (cch-- && *pch) {
				++pch;
			  }

			  cchText = (UINT)(pch - pchText);

			  } break;

			case 'n': {

			  // Write count of characters seen so far into
			  // short/int/long through pointer in args

			  void * p;

			  p = GET_ARG_PTR(parg);

			  // Store chars out into short/long/int depending on flags

#if !LONG_IS_INT
			  if (fl & FL_LONG)
				*(long *)p = fpb.cchOut;
			  else
#endif

#if !SHORT_IS_INT
			  if (fl & FL_SHORT)
				*(short *)p = (short) fpb.cchOut;
			  else
#endif
				*(int *)p = fpb.cchOut;

			  // Force no output

			  fNoOutput = TRUE;

			  } break;


			case 'd':
			case 'i':

			  // Signed decimal number

			  fl |= FL_SIGNED;

			  // fall through

			case 'u':

			  // Unsigned decimal number

			  radix = 10;
			  goto goInt;

			case 'p':

			  // Pointer

			  cchPrec = 2 * sizeof(void *);     // number of hex digits needed

#if !PTR_IS_INT
			  fl|= FL_LONG;                     // assume we're converting a long
#endif

			  // fall through

			case 'X':

			  // Unsigned upper hex output

			  hexAdd = 'A' - '9' - 1;
			  goto goHex;

			case 'x':

			  // Unsigned lower hex output

			  hexAdd = 'a' - '9' - 1;

			  // fall through

goHex:        radix = 16;

			  // Alternate form means '0x' prefix

			  if (fl & FL_ALTERNATE) {
				achPrefix[0] = '0';
				achPrefix[1] = (CHAR)('x' - 'a' + '9' + 1 + hexAdd);  // 'x' or 'X'
				cchPrefix = 2;
			  }

			  goto goInt;

			case 'o':

			  // Unsigned octal output

			  radix = 8;

			  // Alternate form means force a leading 0

			  if (fl & FL_ALTERNATE) {
				fl|= FL_FORCEOCTAL;
			  }

			  // fall through

goInt:        {

			  // Common number formatting routine

			  unsigned long number;   // number to convert
			  int digit;              // ascii value of digit
			  long l;                 // temp long value

			  // Phase 1: read argument into l, sign extend as needed

#if !LONG_IS_INT
			  if (fl & FL_LONG)
				l = GET_ARG_LONG(parg);
			  else
#endif

#if !SHORT_IS_INT
			  if (fl & FL_SHORT) {
				if (fl & FL_SIGNED)
				  l = (short) GET_ARG_INT(parg);               // sign extend
				else
				  l = (unsigned short) GET_ARG_INT(parg);      // zero-extend
			  } else
#endif
			  {
				if (fl & FL_SIGNED)
				  l = GET_ARG_INT(parg);                       // sign extend
				else
				  l = (unsigned int) GET_ARG_INT(parg);        // zero-extend
			  }

			  // Phase 2: check for negative; copy into number

			  if ((fl & FL_SIGNED) && l < 0) {
				number = -l;
				fl|= FL_NEGATIVE;   // remember negative sign
			  } else {
				number = l;
			  }

			  // Phase 3: check precision value for default; non-default
			  //          turns off 0 flag, according to ANSI.

			  if (cchPrec < 0)
				cchPrec = 1;  // default precision
			  else
				fl&=~FL_LEADZERO;

			  // 4. Check if data is 0; if so, turn off hex prefix

			  if (number == 0)
				cchPrefix = 0;

			  // 5. Convert data to ASCII -- note if precision is zero
			  //    and number is zero, we get no digits at all.      

			  pchText = &achText[BUFFERSIZE - 1];                   // last digit at end of buffer

			  while (cchPrec-- > 0 || number != 0) {
				digit = (int)(number % radix) + '0';
				number/= radix;                                     // reduce number
				if (digit > '9') digit+= hexAdd;                    // a hex digit, make it a letter
				*pchText-- = (CHAR)digit;                           // store the digit
			  }

			  cchText = (CHAR *)&achText[BUFFERSIZE - 1] - (CHAR *)pchText; // compute length of number
			  ++pchText;                                            // text points to first digit now


			  // 6. Force a leading zero if FL_FORCEOCTAL flag set

			  if ((fl & FL_FORCEOCTAL) && (pchText[0] != '0' || cchText == 0)) {
				*--pchText = '0';
				++cchText;      // add a zero
			  }

			  } break;

		  }

		  // At this point, we have done the specific conversion, and 'pchText' 
		  // points to text to print; 'cchText' is length.  Now we justify it, 
		  // put on prefixes, leading zeros, and then print it.

		  if (!fNoOutput) {
			int cchPadding;

			// Make up the prefix

			if (fl & FL_SIGNED) {
			  if (fl & FL_NEGATIVE) {
				achPrefix[0] = '-'; cchPrefix = 1;
			  } else 
			  if (fl & FL_SIGN) {
				achPrefix[0] = '+'; cchPrefix = 1; 
			  } else 
			  if (fl & FL_SIGNSP) {
				achPrefix[0] = ' '; cchPrefix = 1;
			  }
			}

			// Calculate amount of padding -- might be negative,
			// but this will just mean zero

			cchPadding = cchField - cchText - cchPrefix;

			// Write out padding

			if (!(fl & (FL_LEFT | FL_LEADZERO))) {
			  WRITE_MULTI_CHAR(' ', cchPadding);
			}

			// Write out prefix

			WRITE_STRING(achPrefix, cchPrefix);

			// Write out leading zeros

			if ((fl & FL_LEADZERO) && !(fl & FL_LEFT)) {
			  WRITE_MULTI_CHAR('0', cchPadding);
			}

			// Write out text

			WRITE_STRING(pchText, cchText);

			// Pad on right with blanks

			if (fl & FL_LEFT) {
			  WRITE_MULTI_CHAR(' ', cchPadding);
			}
		  }

		  break;
	  }
	}

	// Fix trailing null if there is space

	if (fpb.cchMax-- > 0) {
	  fpb.pch[0] = '\0';
	} else
	  return -1;

	// Return number of characters written

	return fpb.cchOut;
  }

#endif // FMA_PRIVATE_FORMAT

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine formats string given variable argument list
//

  FMAAPIV_(INT) FmaFormat(PCH pch, UINT cch, PSZ pszFormat, ...)
  {
	va_list parg;
	va_start(parg, pszFormat);

#ifdef FMA_PRIVATE_FORMAT
	return DoFormatV(pch, cch, pszFormat, parg);
#else
	return FormatV(pch, cch, pszFormat, parg);
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine formats string given argument list pointer
//

  FMAAPI_(INT) FmaFormatV(PCH pch, UINT cch, PSZ pszFormat, va_list parg)
  {
#ifdef FMA_PRIVATE_FORMAT
	return DoFormatV(pch, cch, pszFormat, parg);
#else
	return FormatV(pch, cch, pszFormat, parg);
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// End of FmaFmt.c
/////////////////////////////////////////////////////////////////////////////
