/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Linked string list management routines
//
// FmaList.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 27/Aug/1999
//
// History: 20/Aug/1990 kvitekp - Initially created for DW-TEXT project.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   s u b r o u t i n e s                         //
/////////////////////////////////////////////////////////////////////////////

  // None so far ...

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine adds entry to the linked list
//

  FMAAPI_(UINT) FmaListAdd(PLST_ENTRY * pple, PLST_ENTRY ple, UINT ile)
  {
	PLST_ENTRY pleAfter;
	UINT ileReturn = 0;

	ASSERT(pple != NULL);
	ASSERT(ple != NULL);
	ASSERT(ple->plePrev == NULL);
	ASSERT(ple->pleNext == NULL);

	////////////////////////////////////
	// Check if no entries in list    //
	////////////////////////////////////

	if (*pple == NULL) {

	  // Insert very first entry

	  ASSERT (ile == 0 || ile == LIST_END);

	  *pple = ple;                      // Point to first

	} else

	////////////////////////////////////
	// Check if inserting at the end  //
	////////////////////////////////////

	if (ile == LIST_END) {

	  // Scan to the end of the list

	  pleAfter = *pple;
	  while (ileReturn++, pleAfter->pleNext != NULL) {
		pleAfter = pleAfter->pleNext;
	  }

	  // Update links

	  pleAfter->pleNext = ple;          // Prev to this
	  ple->plePrev = pleAfter;          // This to prev

	} else

	////////////////////////////////////
	// Check if inserting at the head //
	////////////////////////////////////

	if (ile == 0) {

	  // Insert at the head of list. Note 
	  // that there is at least one entry
	  // in list due to the previous logic

	  ple->pleNext = *pple;             // This to next
	  ple->pleNext->plePrev = ple;      // Next to this
	  *pple = ple;                      // Point to first

	} else {

	////////////////////////////////////
	// Otherwise insert in the middle //
	////////////////////////////////////

	  // Insert at the middle of list.
	  // Note that this may be the last 
	  // entry and there is at list one 
	  // entry in list

	  pleAfter = *pple;
	  while (++ileReturn != ile) {
		if (pleAfter->pleNext == NULL) {
		  ASSERT(pleAfter->pleNext);
		  break;
		} else
		  pleAfter = pleAfter->pleNext;
	  }

	  // Update links

	  ple->pleNext = pleAfter->pleNext; // This to next
	  pleAfter->pleNext = ple;          // Prev to this
	  ple->plePrev = pleAfter;          // This to prev
	  if (ple->pleNext != NULL)         // If not the last
		ple->pleNext->plePrev = ple;    // Next to this
	}

	return ileReturn;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes entry in the linked list
//

  FMAAPI_(PLST_ENTRY) FmaListDel(PLST_ENTRY * pple, UINT ile)
  {
	PLST_ENTRY ple;

	ASSERT(pple != NULL);

	// Check if empty list

	if ((ple = *pple) == NULL)
	  return NULL;

	// Check if we have to unlink last entry
	// Note that this can be the first entry

	if (ile == LIST_END) {

	  while (ple->pleNext != NULL) {    // Get to the last
		ple = ple->pleNext;
	  }

	  // Update links

	  if (ple->plePrev != NULL)         // If not the first
		ple->plePrev->pleNext = NULL;   // fix prev's next ptr;
	  else                              // otherwise reset ptr to
		*pple = NULL;                   // the first entry

	} else {

	  // Unlink the middle entry. Note that it may be
	  // also the last or the first entry in list

	  while (ile--) {                   // Get to the entry
		if ((ple = ple->pleNext) == NULL)
		  return NULL;
	  }

	  // Update links

	  if (ple->plePrev == NULL)                 // If first el in list,
		*pple = ple->pleNext;                   // fix ptr to first
	  else                                      // otherwise;
		ple->plePrev->pleNext = ple->pleNext;   // point prev to next link;

	  if (ple->pleNext != NULL)                 // If not the last,
		ple->pleNext->plePrev = ple->plePrev;   // point next to prev link
	}

	// Reset linke so we can add the same elemnet later

	ple->plePrev = NULL;
	ple->pleNext = NULL;

	return ple;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns number of list entries
//

  FMAAPI_(UINT) FmaListCount(PLST_ENTRY ple)
  {
	UINT ile;

	for (ile = 0; ple != NULL; ple = ple->pleNext, ile++);

	return ile;
  }

/////////////////////////////////////////////////////////////////////////////
// End of FmaList.c
/////////////////////////////////////////////////////////////////////////////
