/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Linked string list routines module
//
// FmaLsz.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 27/Aug/1999
//
// History: 01/Jun/1992 kvitekp - Initially created for MadMED project.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine adds a string to a linked list of strings
//

  FMAAPI_(PLSZ) FmaAddLsz(PLSZ * pplsz, PCH pch, UINT cch, UINT ilsz)
  {
	PLSZ plsz;

	ASSERT(pch != NULL);
	ASSERT(pplsz != NULL);

	// Check if asciiz string and get its length

	if (cch == (UINT)-1) cch = strlen(pch);

	// Create new string list element and check if ok

	if ((plsz = MemAlloc(sizeof(LSZ) + cch, MA_CLEAR)) == NULL)
	  return NULL;

	// Initialize the string list element

	memcpy(plsz->ach, pch, cch);
	plsz->hash = GetHash(plsz->ach);

	// Add the new string list entry

	ListAdd(pplsz, plsz, ilsz);

#ifdef FMA_LOG
//Log("AddLsz: pplsz=[%09Fp] '%s'", pplsz, plsz->ach);
#endif

	return plsz;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets a string from a linked list of strings
//

  FMAAPI_(PLSZ) FmaGetLsz(PLSZ plsz, PSZ psz)
  {
	ULONG hash;

	ASSERT(psz != NULL);

	// Get the hash value for the search string

	hash = GetHash(psz);

	// Search the string list

	while (plsz != NULL) {
	  if (plsz->hash == hash && EqPsz(plsz->ach, psz))
		return plsz;
	  else
		plsz = plsz->plszNext;
	}

	return NULL;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes a string from a linked list of strings
//

  FMAAPI_(BOOL) FmaDelLsz(PLSZ * pplsz, UINT ilsz)
  {
	PLSZ plsz;

	ASSERT(pplsz != NULL);

	// Delete entry from list and check if ok

	if ((plsz = ListDel(pplsz, ilsz)) == NULL)
	  return FALSE;

#ifdef FMA_LOG
//Log("DelLsz: pplsz=[%09Fp] '%s'", pplsz, plsz->ach);
#endif

	MemFree(plsz);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes the linked list of strings
//

  FMAAPI_(VOID) FmaDelLszList(PLSZ * pplsz)
  {
	PLSZ plsz;

	ASSERT(pplsz != NULL);

	// Delete the linked list of strings

	while (*pplsz != NULL) {
	  plsz = ListDel(pplsz, 0);

	  ASSERT(plsz != NULL);

#ifdef FMA_LOG
//Log("DelLszList: pplsz=[%09Fp] '%s'", pplsz, plsz->ach);
#endif

	  MemFree(plsz);
	}
  }

/////////////////////////////////////////////////////////////////////////////
// End of FmaLsz.c
/////////////////////////////////////////////////////////////////////////////
