/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: System dependant routines module
//
// FmaSys.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 09/Feb/2000
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
//          09/Feb/2000 kvitekp - Added FmaGet/SetFilePos
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

  // Include standard headers

#if defined(__DOS__) || defined(__W32__) || defined(__OS2__)
#include <io.h>
#include <time.h>
#include <fcntl.h>
#include <share.h>
#include <errno.h>
#include <direct.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <sys\locking.h>

#elif defined(__UNIX__)

#include <unistd.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>

#define O_BINARY      0
#define O_NOINHERIT   0
#define SH_DENYRW     O_EXLOCK
#define SH_DENYNO     0
#define locking       lockf
#define LK_NBLCK      F_TLOCK
#define LK_UNLCK      F_ULOCK
#define chsize        ftruncate

#elif defined(__MAC__)

#include "fss_misc.h"

#define O_NOINHERIT   0
#define S_IREAD       0
#define S_IWRITE      0

#endif

  // DOS specific 16-bit assembler routines from DosMisc.asm

#ifdef __DOS__
  USHORT __PASCAL DosGetOSType(VOID);
  VOID __PASCAL DosGiveUpTimeSlice(USHORT OSType);
#endif

  // Heap allocation macros

#define _MEM_ALLOC    malloc
#define _MEM_REALLOC  realloc
#define _MEM_FREE     free

  // Global memory allocation counter

#ifdef _DEBUG
  ULONG cMemAlloc;
#endif

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Memory management routines
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine allocates and clears specified memory block
//

  FMAAPI_(PVOID) FmaMemAlloc(size_t cb, BOOL bZero)
  {
	PBYTE p;

	// Allocate memory block of the specified length and check if ok

	if ((p = _MEM_ALLOC(cb)) == NULL)
	  return NULL;

	// Clear the alloctated memory block if requested

	if (bZero)
	  memset(p, 0, cb);

#ifdef _DEBUG
++cMemAlloc;
#endif

	return p;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine frees specified memory block
//

  FMAAPI_(VOID) FmaMemFree(PVOID p)
  {
	_MEM_FREE(p);

#ifdef _DEBUG
--cMemAlloc;
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine reallocates specified memory block
//

  FMAAPI_(PVOID) FmaMemRealloc(PVOID p, size_t cb)
  {
#ifdef _DEBUG
if (!p) ++cMemAlloc;
if (!cb) --cMemAlloc;
#endif

	// Reallocate specified memory block and check if ok

	if ((p = _MEM_REALLOC(p, cb)) == NULL) {
	  return NULL;
	}

	return p;
  }

/////////////////////////////////////////////////////////////////////////////
// File management routines
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine opens the file in a specified mode
// 

  FMAAPI_(BOOL) FmaOpenFile(PSZ pszFile, UINT fOpen, HFILE * phfile)
  {
	int oflag  = O_BINARY | O_NOINHERIT;
	int shflag = 0;
	HFILE hfile;
	UINT nErr;

	ASSERT(pszFile != NULL);
	ASSERT(phfile != NULL);

	// Check if file is already open and leave

	if (*phfile != -1) {
	  FMAERROR(ALREADY_EXISTS);
	  return TRUE;
	}

	// Make up the open flags

	if (fOpen & FOPENFILE_READONLY) {
	  oflag|= O_RDONLY;
	} else 
	if (fOpen & FOPENFILE_WRITEONLY) {
	  oflag|= O_WRONLY;
	} else 
	if (fOpen & FOPENFILE_READWRITE) {
	  oflag|= O_RDWR;
	} else {
BadP: FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Make up the creation flags

	if (fOpen & FOPENFILE_CREATE) {
	  oflag|= O_CREAT;
	} else 
	if (fOpen & FOPENFILE_TRUNCATE) {
	  oflag|= O_TRUNC;
	} 

	// Make up the sharing flags

#ifndef __MAC__
	if (fOpen & FOPENFILE_DENYNONE) {
	  shflag|= SH_DENYNO;
	} else 
	if (fOpen & FOPENFILE_DENYALL) {
	  shflag|= SH_DENYRW;
	} else
	  goto BadP;
#endif 

	// Open the file and check if ok

#if !defined __UNIX__ && !defined __MAC__
	if ((hfile = sopen(pszFile, oflag, shflag, S_IREAD | S_IWRITE)) != -1) {
#else
	if ((hfile = open(pszFile, oflag | shflag | S_IREAD | S_IWRITE)) != -1) {
#endif 
	  *phfile = hfile;
	  return TRUE;
	}

	// Make up an error code

	if (errno == EACCES) {
	  nErr = FMAERROR_ACCESS_DENIED;
	} else 
	if (fOpen & FOPENFILE_CREATE) {
	  nErr = FMAERROR_CANNOT_CREATE;
	} else
	  nErr = FMAERROR_CANNOT_OPEN;

	// Set error code and leave

	FmaSetLastError(nErr);

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine closes the file
// 

  FMAAPI_(BOOL) FmaCloseFile(HFILE * phfile)
  {
	ASSERT(phfile != NULL);

	// Check if file is open and close it,
	// then reset it's handle

	if (*phfile != -1) {
	  close(*phfile);
	  *phfile = -1;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets the specified file position
// 

  FMAAPI_(LONG) FmaGetFilePos(HFILE hfile)
  {
	LONG offs;
	
	ASSERT(hfile != -1);

	// Call the standard routine and check if ok

	if ((offs = tell(hfile)) == -1) {
	  FMAERROR(INVALID_PARAM);
	  return -1;
	}

	return offs;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets the specified file position
// 

  FMAAPI_(LONG) FmaSetFilePos(HFILE hfile, LONG offs)
  {
	ASSERT(hfile != -1);

	// Call the standard routine and check if ok

	if ((offs = lseek(hfile, offs, SEEK_SET)) == -1) {
	  FMAERROR(INVALID_PARAM);
	  return -1;
	}

	return offs;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets the specified file size
// 

  FMAAPI_(LONG) FmaGetFileSize(HFILE hfile)
  {
#if defined __UNIX__

	struct stat sb;

	ASSERT(hfile != -1);

	// Call the standard routine and check if ok

	if (fstat(hfile, &sb)){
	  FMAERROR(INVALID_PARAM);
	  return -1;
	}

	return sb.st_size;    

#else

	ULONG cbFile;

	ASSERT(hfile != -1);

	// Call the standard routine and check if ok

	if ((cbFile = filelength(hfile)) == (ULONG)-1) {
	  FMAERROR(INVALID_PARAM);
	  return -1;
	}

	return cbFile;

#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets the specified file size
// 

  FMAAPI_(BOOL) FmaSetFileSize(HFILE hfile, ULONG cbFile)
  {
	ASSERT(hfile != -1);

	// Call the standard routine and check if ok

	if (chsize(hfile, cbFile) == -1) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine reads specified number of bytes from the given file offset
// 

  FMAAPI_(BOOL) FmaReadFile(HFILE hfile, LONG offs, PVOID pb, ULONG cb, 
							ULONG * pcbRead)
  {
	int cbRead;
	UINT nErr;

	ASSERT(hfile != -1);

	// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
	if (!CHECK_DATA_SIZE_LIMIT(cb)) {
	  FMAERROR(DATA_SIZE_LIMIT);
	  return FALSE;
	}
#endif

	// Seek to specified offset and check if ok

	if (lseek(hfile, offs, SEEK_SET) != offs) {
	  nErr = FMAERROR_CANNOT_SEEK;
Fail: if (errno == EACCES) nErr = FMAERROR_ACCESS_DENIED;
	  FmaSetLastError(nErr);
	  return FALSE;
	}

	// Read specified amount of bytes and check if ok

	if ((cbRead = read(hfile, pb, (int)cb)) == -1) {
	  nErr = FMAERROR_CANNOT_READ;
	  goto Fail;
	}

	// Set number of bytes read for caller

	if (pcbRead) *pcbRead = (ULONG)((UINT)cbRead);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes specified number of bytes at the given file offset
// 

  FMAAPI_(BOOL) FmaWriteFile(HFILE hfile, LONG offs, PVOID pb, ULONG cb, 
							 ULONG * pcbWritten)
  {
	int cbWritten;
	UINT nErr;

	ASSERT(hfile != -1);

	// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
	if (!CHECK_DATA_SIZE_LIMIT(cb)) {
	  FMAERROR(DATA_SIZE_LIMIT);
	  return FALSE;
	}
#endif

	// Seek to specified offset and check if ok

	if (lseek(hfile, offs, SEEK_SET) != offs) {
	  nErr = FMAERROR_CANNOT_SEEK;
Fail: if (errno == EACCES) nErr = FMAERROR_ACCESS_DENIED;
	  FmaSetLastError(nErr);
	  return FALSE;
	}

	// Write specified amount of bytes and check if ok

	if ((cbWritten = write(hfile, pb, (int)cb)) == -1) {

	  if (errno == ENOSPC)
		nErr = FMAERROR_NO_DISK_SPACE;
	  else
		nErr = FMAERROR_CANNOT_WRITE;

	  goto Fail;
	}

	// Set number of bytes written for caller

	if (pcbWritten) *pcbWritten = (ULONG)((UINT)cbWritten);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine locks the specified file region 
// 

  FMAAPI_(BOOL) FmaLock(HFILE hfile, LONG offs, LONG cb, BOOL bLock)
  {
	ASSERT(hfile != -1);

	// Seek to the specified location and check if ok

	if (lseek(hfile, offs, SEEK_SET) == -1) {
Fail: if (bLock) FMAERROR(CANNOT_LOCK);
	  else FMAERROR(CANNOT_UNLOCK);
	  return FALSE;
	}

	// Lock or unlock the file region and check if ok

#ifndef __MAC__
	if (locking(hfile, bLock ? LK_NBLCK : LK_UNLCK, cb) == -1) 
	  goto Fail;
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes the file
// 

  FMAAPI_(BOOL) FmaDeleteFile(PSZ pszFile)
  {
	BOOL bOk = TRUE;
	UINT nErr;

	ASSERT(pszFile != NULL);

	// Reset error code since some callers may be interested if
	// the file was actually deleted or did not exist before

	nErr = FMAERROR_NONE;

	// Delete the file and check if ok

	if (unlink(pszFile) != -1) {
	  ASSERT(!FmaFileExists(pszFile));
	} else {

	  // Check if file does not exist

	  if (errno == ENOENT) {
		nErr = FMAERROR_NOT_FOUND;
	  } else {

		bOk = FALSE;

		// Check for sharing violation

		if (errno == EACCES) {
		  nErr = FMAERROR_ACCESS_DENIED;
		} else {
		  nErr = FMAERROR_CANNOT_DELETE;
		}
	  }
	}

	// Set error code and leave

	FmaSetLastError(nErr);

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine enumerates files matching specified mask
// 

  FMAAPI_(BOOL) FmaEnumFiles(PSZ pszPath, PSZ pszMask, PFNENUMFILES pfnCallback, 
							 PVOID pvCallbackData, ULONG fl)
  {
#if defined __DOS__

	struct find_t ff;
	CHAR achPath[MAX_PATH];
	BOOL bOk = FALSE;
	UINT cchPath;
	PSZ pszFile;

	// Validate parameters

	if (!pfnCallback) {
BadP: FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check for null path or mask and substitute dummies

	if (!pszPath) pszPath = "";
	if (!pszMask) pszMask = "*.*";

	// Copy path to the local storage and append mask

	if ((cchPath = strlen(pszPath)) + strlen(pszMask) >= MAX_PATH) goto BadP;
	memcpy(achPath, pszPath, cchPath); strcpy(achPath + cchPath, pszMask);

	// Find first matching file and check if ok

	if (_dos_findfirst(achPath, 0, &ff) != 0) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Loop enumerating matching files

	loop {

	  // Compose the file name

	  if (fl & FENUMFILES_FULLPATH) {
		strcpy(achPath + cchPath, ff.name);
		pszFile = achPath;
	  } else 
		pszFile = ff.name;

	  // Call the callback routine and cancel  if requested

	  if (!pfnCallback(pszFile, (ULONG)ff.size, (ULONG)ff.attrib, 
					   pvCallbackData)) 
		return FALSE;

	  // Find next matching file and check if ok

	  if (_dos_findnext(&ff) != 0) {
		if (errno == ENOENT) bOk = TRUE;
		break;
	  }
	} 

	return bOk;

#elif defined __OS2__

	FILEFINDBUF ffb;
	HDIR hFind = HDIR_CREATE;
	CHAR achPath[MAX_PATH];
	UINT code, nFind = 1;
	BOOL bOk = FALSE;
	UINT cchPath;
	PSZ pszFile;

	// Validate parameters

	if (!pfnCallback) {
BadP: FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check for null path or mask and substitute dummies

	if (!pszPath) pszPath = "";
	if (!pszMask) pszMask = "*.*";

	// Copy path to the local storage and append mask

	if ((cchPath = strlen(pszPath)) + strlen(pszMask) >= MAX_PATH) goto BadP;
	memcpy(achPath, pszPath, cchPath); strcpy(achPath + cchPath, pszMask);

	// Find first matching file and check if ok

	if (DosFindFirst(achPath, &hFind, FILE_NORMAL, &ffb, sizeof(ffb), &nFind, 0)) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Loop enumerating matching files

	loop {

	  // Compose the file name

	  if (fl & FENUMFILES_FULLPATH) {
		strcpy(achPath + cchPath, ffb.achName);
		pszFile = achPath;
	  } else 
		pszFile = ffb.achName;

	  // Call the callback routine and cancel  if requested

	  if (!pfnCallback(pszFile, ffb.cbFile, ffb.attrFile, 
					   pvCallbackData)) 
		return FALSE;

	  // Find next matching file and check if ok

	  nFind = 1;
	  if ((code = DosFindNext(hFind, &ffb, sizeof(ffb), &nFind)) != 0) {
		if (code == ERROR_NO_MORE_FILES) bOk = TRUE;
		break;
	  }
	} 

	// Close the search handle

	DosFindClose(hFind);

	return bOk;

#elif defined __W32__

	WIN32_FIND_DATA wfd;
	CHAR achPath[MAX_PATH];
	BOOL bOk = FALSE;
	HANDLE hFind;
	UINT cchPath;
	PSZ pszFile;

	// Validate parameters

	if (!pfnCallback) {
BadP: FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check for null path or mask and substitute dummies

	if (!pszPath) pszPath = "";
	if (!pszMask) pszMask = "*.*";

	// Copy path to the local storage and append mask

	if ((cchPath = strlen(pszPath)) + strlen(pszMask) >= MAX_PATH) goto BadP;
	memcpy(achPath, pszPath, cchPath); strcpy(achPath + cchPath, pszMask);

	// Find first matching file and check if ok

	if ((hFind = FindFirstFile(achPath, &wfd)) == INVALID_HANDLE_VALUE) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Loop enumerating matching files

	loop {

	  // Compose the file name

	  if (fl & FENUMFILES_FULLPATH) {
		strcpy(achPath + cchPath, wfd.cFileName);
		pszFile = achPath;
	  } else 
		pszFile = wfd.cFileName;

	  // Call the callback routine and cancel  if requested

	  if (!pfnCallback(pszFile, wfd.nFileSizeLow, wfd.dwFileAttributes, 
					   pvCallbackData)) 
		return FALSE;

	  // Find next matching file and check if ok

	  if (!FindNextFile(hFind, &wfd)) {
		if (GetLastError() == ERROR_NO_MORE_FILES) bOk = TRUE;
		break;
	  }
	} 

	// Close the search handle

	FindClose(hFind);

	return bOk;

#elif defined __UNIX__
#error FmaEnumFiles

#elif defined __MAC__

	_finddata_t ff;
	CHAR achPath[MAX_PATH];
	BOOL bOk = FALSE;
	UINT cchPath;
	PSZ pszFile;

	// Validate parameters

	if (!pfnCallback) {
BadP: FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check for null path or mask and substitute dummies

	if (!pszPath) pszPath = (PSZ)"";
	if (!pszMask) pszMask = (PSZ)"*.*";

	// Find first matching file and check if ok

	if (_findfirst((PCHAR)pszPath, (PCHAR)pszMask, &ff) < 0) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Loop enumerating matching files

	loop {

	  // Compose the file name

	  if (fl & FENUMFILES_FULLPATH) {
		strcpy(achPath, (PCHAR)pszPath);
		strcat(achPath, ff.name);
		pszFile = (PSZ)achPath;
	  } else 
		pszFile = (PSZ)ff.name;

	  // Call the callback routine and cancel  if requested

	  if (!pfnCallback(pszFile, (ULONG)ff.size, (ULONG)ff.attrib, 
					   pvCallbackData)) 
		return FALSE;

	  // Find next matching file and check if ok

	  if (_findnext(0, &ff) < 0) {
		bOk = TRUE;
		break;
	  }
	} 

	return bOk;

#else
#error FmaEnumFiles
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks if the specified file exists 
// 

  FMAAPI_(BOOL) FmaFileExists(PSZ pszFile)
  {
#if defined __DOS__

	struct find_t ff;

	ASSERT(pszFile != NULL);

	// Find first matching file and check if ok

	if (_dos_findfirst(pszFile, 0, &ff) != 0) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Return positive if the path is not a directory or label

	return !(ff.attrib & (_A_SUBDIR | _A_VOLID));

#elif defined __OS2__

	FILESTATUS fs;

	// Get path info and check if ok

	if (DosQPathInfo(pszFile, FIL_STANDARD, (PBYTE)&fs, sizeof(fs), 0)) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Return positive if the path is not a directory

	return !(fs.attrFile & FILE_DIRECTORY);

#elif defined __W32__

	DWORD dwAttr;

	ASSERT(pszFile != NULL);

	// Get the specified dfile attributes and check if ok

	if ((dwAttr = GetFileAttributes(pszFile)) == (DWORD)-1)
	  return FALSE;

	// Return positive if the path is not a directory or offline

	return !(dwAttr & (FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_OFFLINE));

#elif defined __UNIX__

	struct stat sb;

	// Get the file status info

	if (stat(pszFile, &sb)){
	  return FALSE;
	}

	return TRUE;

#elif defined __MAC__

	FSSpec  spec;
	FInfo   fndrInfo;
	FXInfo  itsFXInfo;
	Boolean isDirectory;
	OSErr   err;


	ASSERT(pszFile != NULL);

	if ((err = _c_path2fss((PCHAR)pszFile, &spec)) == noErr)
		err = get_finder_info (&spec, &fndrInfo, &itsFXInfo, &isDirectory);

	return ((err == noErr) && !isDirectory) ? TRUE : FALSE;

#else
#error FmaFileExists
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks if the specified directory exists 
// 

  FMAAPI_(BOOL) FmaDirExists(PSZ pszPath)
  {
#if defined __DOS__

	struct find_t ff;
	CHAR achPath[MAX_PATH];
	UINT cch;

	ASSERT(pszPath != NULL);

	// Copy path to the local storage and remove trailing backslash if any

	strncpy(achPath, pszPath, lengof(achPath)); 
	achPath[lengof(achPath)] = '\0';
	if ((cch = strlen(achPath)) > 0 && achPath[cch -1] == PATH_SEP)
	  achPath[cch - 1] = '\0';

	// Find first matching file and check if ok

	if (_dos_findfirst(achPath, _A_SUBDIR, &ff) != 0) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Return positive if the path is a directory

	return (ff.attrib & _A_SUBDIR) ? TRUE : FALSE;

#elif defined __OS2__

	FILESTATUS fs;
	CHAR achPath[MAX_PATH];
	UINT cch;

	// Copy path to the local storage and remove trailing backslash if any

	strncpy(achPath, pszPath, lengof(achPath)); 
	achPath[lengof(achPath)] = '\0';
	if ((cch = strlen(achPath)) > 0 && achPath[cch -1] == PATH_SEP)
	  achPath[cch - 1] = '\0';

	// Get path info and check if ok

	if (DosQPathInfo(achPath, FIL_STANDARD, (PBYTE)&fs, sizeof(fs), 0)) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Return positive if the path is a directory

	return (fs.attrFile & FILE_DIRECTORY) ? TRUE : FALSE;

#elif defined __W32__

	DWORD dwAttr;

	ASSERT(pszPath != NULL);

	// Get the specified file attributes and check if ok

	if ((dwAttr = GetFileAttributes(pszPath)) == (DWORD)-1)
	  return FALSE;

	// Return positive if the path is a directory

	return (dwAttr & FILE_ATTRIBUTE_DIRECTORY) ? TRUE : FALSE;

#elif defined __UNIX__

	struct stat sb;

	// Get the path status info

	if (stat(pszPath, &sb)){
	  return FALSE;
	}

	return TRUE;

#elif defined __MAC__

	FSSpec  spec;
	FInfo   fndrInfo;
	FXInfo  itsFXInfo;
	Boolean isDirectory;
	OSErr   err;
	char    pszPathCopy[256];


	ASSERT(pszPath != NULL);

	strcpy (pszPathCopy, (PCHAR)pszPath);
	if (pszPathCopy[strlen(pszPathCopy)-1] == PATH_SEP)
		pszPathCopy[strlen(pszPathCopy)-1] = 0;

	if ((err = _c_path2fss((PCHAR)pszPathCopy, &spec)) == noErr || err == notAFileErr)
		err = get_finder_info (&spec, &fndrInfo, &itsFXInfo, &isDirectory);

	return ((err == noErr) && isDirectory) ? TRUE : FALSE;

#else
#error FmaDirExists
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates the specified directory tree
// 

  FMAAPI_(BOOL) FmaMakeDir(PSZ pszPath)
  {
	CHAR achPath[MAX_PATH];
	BOOL bDone = FALSE;
	UINT cch;
	PCH pch;

	// Validate parameters

	if (!pszPath || !pszPath[0] || (cch = strlen(pszPath)) > MAX_PATH) {
BadP: FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Copy path to the local storage and check if we have drive or 
	// share specification and skip it if so

	strcpy(achPath, pszPath);
	if (achPath[1] == ':') {
	  if (cch <= lengof("x:")) goto BadP;
	  pch = achPath + lengof("x:");
	} else
	if (achPath[0] == PATH_SEP && achPath[1] == PATH_SEP) {
	  if ((pch = strchr(achPath + lengof("//"), PATH_SEP)) == NULL) goto BadP;
	  if ((pch = strchr(pch + 1, PATH_SEP)) == NULL) goto BadP;
	} else
	  pch = achPath;

	// Check if we're looking at leading backslash and skip it

	if (*pch == PATH_SEP) pch++;

	// Loop creating the directory tree

	while (!bDone) {

	  // Locate the next backslash and temporarily replace it with null

	  if ((pch = strchr(pch, PATH_SEP)) != NULL) {
		*pch = '\0';
	  } else
		bDone = TRUE;

	  // Check if this directory exists and create it if not

	  if (!FmaDirExists(achPath)) {
#if defined __UNIX__ || defined __MAC__
		if (mkdir(achPath, 0x1ff) != 0) {
#else
		if (mkdir(achPath) != 0) {
#endif
		  FMAERROR(CANNOT_CREATE);
		  return FALSE;
		}
	  }

	  // Restore the zeroed backslash and continue

	  if (!bDone) *pch++ = PATH_SEP;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// Miscellaneous routines
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine returns local date and time in dttm format known as DOS time
// 

  FMAAPI_(BOOL) FmaGetLocalDttm(PDTTM pdttm)
  {
#if defined __DOS__

	time_t tmSecGMT;
	struct tm * ptmLocal;

	// Get GMT time since Jan 1st, 1970 and convert it to local time

	time(&tmSecGMT);
	ptmLocal = localtime(&tmSecGMT);

	// Make up the caller's dttm structure

	pdttm->hh = ptmLocal->tm_hour;
	pdttm->mm = ptmLocal->tm_min;
	pdttm->ss = ptmLocal->tm_sec / 2;
	pdttm->da = ptmLocal->tm_mday;
	pdttm->mo = ptmLocal->tm_mon + 1;
	pdttm->yr = ptmLocal->tm_year - 80;

	return TRUE;

#elif defined __OS2__

	DATETIME datetime;
	DTTM dttm;
	LONG tm;

	// Get the OS/2 date and time and check if ok

	DosGetDateTime(&datetime);

	// Make up the dttm structure

	dttm.hh = datetime.hours;
	dttm.mm = datetime.minutes;
	dttm.ss = datetime.seconds / 2;
	dttm.da = datetime.day;
	dttm.mo = datetime.month;
	dttm.yr = datetime.year - 1980;

	// Check if timezone is specified and if so, get number of seconds 
	// since Jan 1st, 1970 GMT, adjust it to local time and convert it 
	// back to dttm

	if (datetime.timezone != -1) {
	  tm = FmaGetTimeFromDttm(&dttm);
	  tm-= (LONG)datetime.timezone * 60L;
	  FmaGetDttmFromTime(&dttm, tm);
	}

	// Make up the caller's dttm structure

	*pdttm = dttm;

	return TRUE;

#elif defined __W32__

	SYSTEMTIME st;

	ASSERT(pdttm != NULL);

	// Get local time

	GetLocalTime(&st);

	// Make up the caller's dttm structure

	pdttm->hh = st.wHour;
	pdttm->mm = st.wMinute;
	pdttm->ss = st.wSecond / 2;
	pdttm->da = st.wDay;
	pdttm->mo = st.wMonth;
	pdttm->yr = st.wYear - 1980;

	return TRUE;

#elif defined __UNIX__ || defined __MAC__

	time_t tmSecGMT;
	struct tm * ptmLocal;

	// Get GMT time since Jan 1st, 1970 and convert it to local time

	time(&tmSecGMT);
	ptmLocal = localtime(&tmSecGMT);

	// Make up the caller's dttm structure

	pdttm->hh = ptmLocal->tm_hour;
	pdttm->mm = ptmLocal->tm_min;
	pdttm->ss = ptmLocal->tm_sec / 2;
	pdttm->da = ptmLocal->tm_mday;
	pdttm->mo = ptmLocal->tm_mon + 1;
	pdttm->yr = ptmLocal->tm_year - 80;

	return TRUE;

#else
#error FmaGetLocalDttm
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns local date and time in time format known as Unix time
// 

  FMAAPI_(ULONG) FmaGetLocalTime(VOID)
  {
	DTTM dttm;

	FmaGetLocalDttm(&dttm);

	return FmaGetTimeFromDttm(&dttm);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sleeps for specified amount of seconds
// 

  FMAAPI_(VOID) FmaSleep(ULONG milliseconds)
  {
#if defined __DOS__

	time_t tm;

	if (!milliseconds) {
	  FmaGiveOutTimeSlice();
	} else {
	  tm = time(NULL) + (milliseconds / 1000L);
	  while (time(NULL) < tm + 1) {
		FmaGiveOutTimeSlice();
	  }
	}

#elif defined __OS2__

	DosSleep(milliseconds);

#elif defined __W32__

	Sleep(milliseconds);

#elif defined __UNIX__

	sleep(milliseconds / 1000L);

#elif defined __MAC__

	//  Nothing to do...

#else
#error FmaSleep
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns unique message id stamp value
//

  FMAAPI_(ULONG) FmaGetMsgIdStamp(VOID)
  {
	time_t tmStart = time(NULL);
	ULONG lStamp = 0, lSecs, lHund;
	static ULONG lStampPrev;

#if defined __DOS__
	union REGS regs;

#elif defined __OS2__
	SEL selgis, sellis;
	PGINFOSEG pgis;
	PLINFOSEG plis;

#elif defined __W32__
	SYSTEMTIME st;

#ifdef _MT
	EnterCriticalSection(&fmaRtm.csMsgId);
#endif

#elif defined __UNIX__
	struct timeval tv;
	struct timezone tz;

#elif defined __MAC__

#else
#error FmaGetMsgIdStamp
#endif

	// Under OS2 get pointers to the global and local info segments

#ifdef __OS2__
	  DosGetInfoSeg(&selgis, &sellis);
	  pgis = MAKEPGINFOSEG(selgis);
	  plis = MAKEPLINFOSEG(sellis);
#endif

#ifdef FMA_LOG
//Log("FmaGetMsgIdStamp >: lStampPrev=%08lx", lStampPrev);
#endif

	// Loop for no more than 5 seconds pretending to be multitasking friendly

	for (tmStart = time(NULL); time(NULL) < tmStart + 5; FmaGiveOutTimeSlice()) {

	  // Get the number of seconds since Jan 1, 1970 and 
	  // current system time hundreds of seconds. Note that
	  // we can't just use system clock since it needs to be 
	  // incremental and in sync with time.

#if   defined __OS2__
	  lSecs = (ULONG) pgis->time;
	  lHund = (ULONG) pgis->hundredths;
#elif defined __W32__
	  GetSystemTime(&st);
	  lSecs = (ULONG) time(NULL);
	  lHund = (ULONG) st.wMilliseconds / 10;
#elif defined __DOS__
	  lSecs = (ULONG) time(NULL);
	  regs.h.ah = 0x2c; intdos(&regs, &regs);
	  lHund = (ULONG) regs.h.dl;
#elif defined __UNIX__
	  gettimeofday(&tv, &tz);
	  lSecs = (ULONG) tv.tv_sec;
	  lHund = (ULONG) tv.tv_usec; // TODO: check this -- could be zero !!!
#elif defined __MAC__
	  lSecs = (ULONG) LMGetTime();
	  lHund = (ULONG) LMGetTicks() & 0x0000003FL;
#endif

	  ASSERT(lHund <= 0x07f);

	  // Make up time stamp out of number of seconds since Jan 1, 1970
	  // shifted 7 bits to the left OR'ed with current system clock.
	  // Note: this gives us 388 days cycle and reasonable resolution.

	  lStamp = (lSecs << 7) | (lHund & 0x07f);

#ifdef FMA_LOG
//Log("FmaGetMsgIdStamp  : lStamp=%08lx/%08lx lSecs=%08lx lHund=%08lx", lStamp, lStampPrev, lSecs, lHund);
#endif

	  // Check if we got a new ascending stamp value and break the loop

	  if (lStamp > lStampPrev)
		break;
	}

	// Check if we finally have unique ascending stamp and
	// if not, just incremented largest stamp value

	if (!(lStamp > lStampPrev)) lStamp = lStampPrev + 1;

#ifdef FMA_LOG
//Log("FmaGetMsgIdStamp <: lStamp=%08lx/%08lx", lStamp, lStampPrev);
#endif

	// Save the returned stamp value for the next turn

	lStampPrev = lStamp;

	// Leave critical section if in multitasking environment

#ifdef _MT

#if defined __W32__
	LeaveCriticalSection(&fmaRtm.csMsgId);

#else 
#error FmaGetMsgIdStamp CRITICAL SECTION SUPPORT
#endif

#endif

	// Return generated stamp value

	return lStamp;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine is increments a long value in a thread safe manner
//

  FMAAPI_(VOID) FmaInterlockedIncrement(LONG * plv)
  {
#ifndef _MT
	++(*plv);
#else
#if defined __W32__
	InterlockedIncrement(plv);
#else
#error FmaInterlockedIncrement
#endif
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine is decrements a long value in a thread safe manner
//

  FMAAPI_(VOID) FmaInterlockedDecrement(LONG * plv)
  {
#ifndef _MT
	--(*plv);
#else
#if defined __W32__
	InterlockedDecrement(plv);
#else
#error FmaInterlockedDecrement
#endif
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns host operating system type
//

  FMAAPI_(UINT) FmaQueryOSType(VOID)
  {
#if   defined __DOS__

	return DosGetOSType();

#elif defined __OS2__

	return (OSTYPE_OS2 | OSTYPE_PROTMODE);

#elif defined __W32__

	return (OSTYPE_W32 | OSTYPE_PROTMODE);

#elif defined __UNIX__

	return OSTYPE_UNIX;

#elif defined __MAC__

	return OSTYPE_MACOS;

#else
#error FmaQueryOSType
#endif
  }


/////////////////////////////////////////////////////////////////////////////
// This routine gives out time slice to the host operating system type
//

  FMAAPI_(VOID) FmaGiveOutTimeSlice(VOID)
  {
#if   defined __DOS__

	static UINT OSType;
	if (!OSType) OSType = DosGetOSType();
	DosGiveUpTimeSlice(OSType);

#elif defined __OS2__

	DosSleep(0);

#elif defined __W32__

	Sleep(0);

#elif defined __UNIX__ || defined __MAC__

	//  Nothing to do...

#else
#error FmaGiveOutTimeSlice
#endif
  }

/////////////////////////////////////////////////////////////////////////////
// End of FmaSys.c
/////////////////////////////////////////////////////////////////////////////
