/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: System dependant routines header file
//
// FmaSys.h
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 09/Feb/2000
//
// History: 17/Jan/1999 kvitekp - Initially created for IMA project.
//
//          09/Feb/2000 kvitekp - Added FmaGet/SetFilePos
//
//          22/Feb/2000 kvitekp - Added namespace compatibility #ifdefs
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _FMA_SYS_INCLUDED
#define _FMA_SYS_INCLUDED

  // Declare C/C++ frame

#ifdef __cplusplus
extern "C" {
#endif

  // Include types declaratons

#include "FmaDef.h"

/////////////////////////////////////////////////////////////////////////////
// Memory management declarations
/////////////////////////////////////////////////////////////////////////////

  // Function prototypes

  FMAAPI_(PVOID) FmaMemAlloc(size_t cb, BOOL bZero);
  FMAAPI_(PVOID) FmaMemRealloc(PVOID p, size_t cb);
  FMAAPI_(VOID)  FmaMemFree(PVOID p);

  // Flags used by MemAlloc

#ifndef MA_CLEAR
#define MA_CLEAR    TRUE
#endif

  // Map functions to common names redirected through global thunk

#ifndef FMA_NO_FUNC_DEFS
#define MemAlloc    fmaInit.pfnMemAlloc
#define MemRealloc  fmaInit.pfnMemRealloc
#define MemFree     fmaInit.pfnMemFree
#endif

/////////////////////////////////////////////////////////////////////////////
// File management declarations
/////////////////////////////////////////////////////////////////////////////

  // FmaOpenFile() open flags

#define FOPENFILE_READONLY    0x0001        // Open for read only access
#define FOPENFILE_WRITEONLY   0x0002        // Open for write only access
#define FOPENFILE_READWRITE   0x0004        // Open for read/write access
#define FOPENFILE_CREATE      0x0010        // Create if does not exist
#define FOPENFILE_TRUNCATE    0x0020        // Truncate if does exists
#define FOPENFILE_DENYNONE    0x0100        // Deny none other opens
#define FOPENFILE_DENYALL     0x0200        // Deny all other opens

  // Function prototypes

  FMAAPI_(BOOL) FmaOpenFile(PSZ pszFile, UINT fOpen, HFILE * phfile);
  FMAAPI_(BOOL) FmaCloseFile(HFILE * phfile);

  FMAAPI_(BOOL) FmaReadFile(HFILE hfile, LONG offs, PVOID pb, ULONG cb, 
							ULONG * pcbRead);

  FMAAPI_(BOOL) FmaWriteFile(HFILE hfile, LONG offs, PVOID pb, ULONG cb, 
							 ULONG * pcbWritten);

  FMAAPI_(LONG) FmaGetFilePos(HFILE hfile);
  FMAAPI_(LONG) FmaSetFilePos(HFILE hfile, LONG offs);
  FMAAPI_(LONG) FmaGetFileSize(HFILE hfile);
  FMAAPI_(BOOL) FmaSetFileSize(HFILE hfile, ULONG cbFile);
  FMAAPI_(BOOL) FmaLock(HFILE hfile, LONG offs, LONG cb, BOOL bLock);
  FMAAPI_(BOOL) FmaDeleteFile(PSZ pszFile);

  // FmaEnumFiles() callback function prototype

  typedef BOOL (FMACALLBACK * PFNENUMFILES)(PSZ pszFile, ULONG cbFile, ULONG attr, 
											PVOID pvCallbackData);

  // FmaEnumFiles() flags

#define FENUMFILES_FULLPATH 0x00000001

  // Function prototypes

  FMAAPI_(BOOL) FmaEnumFiles(PSZ pszPath, PSZ pszMask, PFNENUMFILES pfnCallback, 
							 PVOID pvCallbackData, ULONG fl);

  FMAAPI_(BOOL) FmaFileExists(PSZ pszPath);
  FMAAPI_(BOOL) FmaDirExists(PSZ pszPath);
  FMAAPI_(BOOL) FmaMakeDir(PSZ pszPath);

/////////////////////////////////////////////////////////////////////////////
// Miscellaneous declarations
/////////////////////////////////////////////////////////////////////////////

  // Flags used by FmaGetOSType() and FmaGiveUpTimeSlice()

#define OSTYPE_DOS              0x0001          // Plain DOS
#define OSTYPE_WINDOWS          0x0002          // Microsoft Windows DosBox
#define OSTYPE_WINNT            0x0003          // Microsoft Windows NT DosBox
#define OSTYPE_DESQVIEW         0x0004          // QuarterDesk DesqView
#define OSTYPE_OS2DOS           0x0005          // IBM OS/2 DosBox
#define OSTYPE_OS2              0x0006          // Native IBM OS/2
#define OSTYPE_W32              0x0007          // Native Win32
#define OSTYPE_MACOS            0x0008          // Native Mac OS
#define OSTYPE_UNIX             0x0009          // Generic UNIX
#define OSTYPE_MASK             0x000F          // OS type mask
#define OSTYPE_PROTMODE         0x8000          // Protected mode

  // Function prototypes

  FMAAPI_(VOID)  FmaInterlockedIncrement(LONG * plv);
  FMAAPI_(VOID)  FmaInterlockedDecrement(LONG * plv);
  FMAAPI_(BOOL)  FmaGetLocalDttm(PDTTM pdttm);
  FMAAPI_(ULONG) FmaGetLocalTime(VOID);
  FMAAPI_(VOID)  FmaSleep(ULONG milliseconds);
  FMAAPI_(ULONG) FmaGetMsgIdStamp(VOID);
  FMAAPI_(UINT)  FmaQueryOSType(VOID);
  FMAAPI_(VOID)  FmaGiveOutTimeSlice(VOID);

/////////////////////////////////////////////////////////////////////////////
// Header cleanup
/////////////////////////////////////////////////////////////////////////////

  // Terminate C/C++ frame

#ifdef __cplusplus
}
#endif

#endif  // #ifndef _FMA_SYS_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// End of FmaSys.h
/////////////////////////////////////////////////////////////////////////////
