/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Testing app
//
// Test.c
//
// Copyright (c) 1999 by Pete Kvitek
//
// Created: 27/Aug/1999
// Updated: 27/Aug/1999
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include standard headers

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

  // Include common and private headers

#include "FmaApi.h"
#include "FtsApi.h"
#include "FtsApiP.h"
#include "SquApi.h"
#include "SquApiP.h"
#include "JamApi.h"
#include "JamApiP.h"
#include "PktApi.h"
#include "PktApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

static NETADDR netAddr = {2, 5020, 6, 0};

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine initializes the app
//

  static BOOL FMASUB DoInitApp(VOID)
  {
	UINT type;

	// Initialize FMA and check if ok

	if (!FmaInitialize(NULL)) {
	  printf("FmaInitialize() failed, code=%u\n", FmaGetLastError());
	  return FALSE;
	}

	// Register all core message base apis

	if (!FmaRegisterApi(type = MBTYPE_FTS, FtsGetApi) ||
		!FmaRegisterApi(type = MBTYPE_SQU, SquGetApi) ||
		!FmaRegisterApi(type = MBTYPE_JAM, JamGetApi) ||
		!FmaRegisterApi(type = MBTYPE_PKT, PktGetApi)) {
	  printf("FmaRegisterApi() failed, type=%d, code=%u\n", type, FmaGetLastError());
	  FmaTerminate();
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests date time routines
//

  static BOOL FMASUB DoTestDttm(VOID)
  {
	ULONG fl = FDTTM_4DIGITYEAR | FDTTM_TIMESECONDS;
	ULONG time1, time2;
	CHAR ach[128];
	DTTM dttm;

	// Check formatting routine

	dttm.hh = 22; dttm.mm = 21; dttm.ss = 20 / 2;
	dttm.da = 01; dttm.mo = 02; dttm.yr = 1999 - 1980;
	FmaFormatDttm(ach, lengof(ach), &dttm, fl, 0, 0, NULL);

	// Check invalid time to dttm conversion and vise versa

	time1 = 0;
	FmaGetDttmFromTime(&dttm, time1);
	FmaFormatDttm(ach, lengof(ach), &dttm, fl, 0, 0, NULL);
	time2 = FmaGetTimeFromDttm(&dttm);
	FmaGetDttmFromTime(&dttm, time2);
	FmaFormatDttm(ach, lengof(ach), &dttm, fl, 0, 0, NULL);

	// Check local time to dttm conversion and vise versa

	FmaGetLocalDttm(&dttm);
	FmaFormatDttm(ach, lengof(ach), &dttm, fl, 0, 0, NULL);
	time1 = FmaGetTimeFromDttm(&dttm);
	FmaGetDttmFromTime(&dttm, time1);
	FmaFormatDttm(ach, lengof(ach), &dttm, fl, 0, 0, NULL);
	time2 = FmaGetTimeFromDttm(&dttm);

	ASSERT(time1 == time2);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests some basic fmm functinality
//

  static BOOL FMASUB DoTestFmm(VOID)
  {
	PFMM pfmm;
	PFMP pfmp;
	ULONG crc = -1;
	PSZ psz = NULL;

	// Create fidonet message and check if ok

	if ((pfmm = FmaCreateMsg(0)) == NULL) {
	  printf("FmaCreateMsg() failed, code=%u\n", FmaGetLastError());
	  return FALSE;
	} 

	// Add some string properties

	pfmp = FmaAddFmp(pfmm, FMP_SENDERNAME, NULL, -1);
	pfmp = FmaAddFmp(pfmm, FMP_SENDERNAME, "Pete Kvitek", -1);

	// Add some binary properties

	crc = -1;
	pfmp = FmaAddFmp(pfmm, FMP_PASSWORDCRC, &crc, sizeof(crc));

	FmaAllocString(&psz, "Message text", -1);
	pfmp = FmaAddFmp(pfmm, FMP_MESSAGETEXT, &psz, sizeof(psz));

	// Add some formatted properties

	pfmp = FmaAddFmpFormat(pfmm, FMP_FTSKLUDGE, "COOL: %s", "kludge");

	// Get some properties

	pfmp = FmaGetFmp(pfmm, FMP_RECEVRNAME, 0);
	pfmp = FmaGetFmp(pfmm, FMP_SENDERNAME, 0);
	pfmp = FmaGetFmp(pfmm, FMP_PASSWORDCRC, 0);
	pfmp = FmaGetFmp(pfmm, FMP_MESSAGETEXT, 0);

	// Dump fidonet mail message

	FmaDumpMsg(pfmm);

	// Destroy fidonet message and check if ok

	if (!FmaDestroyMsg(pfmm)) {
	  printf("FmaDestroyMsg() failed, code=%u\n", FmaGetLastError());
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests msgid timestamp generation
//

  static BOOL FMASUB DoTestMsgIdStamp(VOID)
  {
	UINT n;

	// Loop generating and printing msgid time stamps

	for (n = 0; n < 10; n++) {
	  printf("Msgid stamp: %08lx\n", FmaGetMsgIdStamp());
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests open area apis
//

  static BOOL FMASUB DoOpenArea(PSZ pszPath, UINT type)
  {
	ULONG fl = 0;
	HAREA harea;
	HMSG hmsg;
	MSGN msgn;

	// Inform caller

	printf("\nTouching area: %s, type=%d\n", pszPath, type);

	// Open and close an area and check if ok

	if ((harea = FmaOpenArea(pszPath, type, OPENAREA_ALWAYS, fl, &netAddr)) == NULL) {
	  printf("FmaOpenArea(\"%s\") failed, code=%u\n", pszPath, FmaGetLastError());
	  return FALSE;
	}

	// Display what these return

	printf("FmaGetFirstMsg()=%d\n", FmaGetFirstMsg(harea));
	printf("FmaGetHighMsg()=%d\n", FmaGetHighMsg(harea));
	printf("FmaGetMsgCount()=%d\n", FmaGetMsgCount(harea));
	printf("FmaGetCurMsg()=%d\n", FmaGetCurMsg(harea));

	// Open the current message (this should fail after open)

	msgn = MSGN_CUR;
	if ((hmsg = FmaOpenMsg(harea, OPENMSG_READ, msgn, 0)) == NULL) {
	  printf("FmaOpenMsg() failed, code=%d\n", FmaGetLastError());
	} else
	  FmaCloseMsg(hmsg);

	// Loop reading next msgs

	msgn = MSGN_NEXT;
	while ((hmsg = FmaOpenMsg(harea, OPENMSG_READ, msgn, 0)) != NULL) {
	  printf("Opened msgn=%d\n", FmaGetCurMsg(harea));
	  FmaCloseMsg(hmsg);
	}

	// Loop reading prev msgs

	msgn = MSGN_PREV;
	while ((hmsg = FmaOpenMsg(harea, OPENMSG_READ, msgn, 0)) != NULL) {
	  printf("Opened msgn=%d\n", FmaGetCurMsg(harea));
	  FmaCloseMsg(hmsg);
	}

	// Close an area and check if ok

	if (!FmaCloseArea(harea)) {
	  printf("FmaCloseArea(\"%s\") failed, code=%u\n", pszPath, FmaGetLastError());
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message loading
//

  static BOOL FMASUB DoLoadMsg(PSZ pszPath, UINT type, MSGN msgn)
  {
	ULONG fl = 0;
	HAREA harea;
	HMSG hmsg;
	PFMM pfmm;

	// Inform caller

	printf("\nLoading msgn %d in area: %s, type=%d\n", msgn, pszPath, type);

	// Open and close an area and check if ok

	if ((harea = FmaOpenArea(pszPath, type, OPENAREA_ALWAYS, fl, &netAddr)) == NULL) {
	  printf("FmaOpenArea(\"%s\") failed, code=%u\n", pszPath, FmaGetLastError());
	  return FALSE;
	}

	// Open the first message

	if (!msgn) msgn = FmaGetFirstMsg(harea);
	if ((hmsg = FmaOpenMsg(harea, OPENMSG_READ, msgn, 0)) == NULL) {
	  printf("FmaOpenMsg() failed, code=%d\n", FmaGetLastError());
	} else {

	  // Load message and check if ok

	  if (!FmaLoadMsg(hmsg, &pfmm, 0)) {
		printf("FmaLoadMsg() failed, code=%d\n", FmaGetLastError());
	  } else {
		FmaDumpMsg(pfmm);
		FmaDestroyMsg(pfmm);
	  }

	  // Close message

	  if (!FmaCloseMsg(hmsg)) {
		printf("FmaCloseMsg() failed, code=%d\n", FmaGetLastError());
	  }
	}

	// Close an area and check if ok

	if (!FmaCloseArea(harea)) {
	  printf("FmaCloseArea(\"%s\") failed, code=%d\n", pszPath, FmaGetLastError());
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message saving
//

  static BOOL FMASUB DoSaveMsg(PSZ pszPath, UINT type, MSGN msgn)
  {
	UMSG aReplies[10];
	ULONG crc, fl = 0;
	PSZ psz = NULL;
	HAREA harea;
	HMSG hmsg;
	PFMM pfmm;
	DTTM dttm;
	UINT n;

	// Inform caller

	printf("\nSaving msgn %d in area: %s, type=%d\n", msgn, pszPath, type);

	// Open and close an area and check if ok

	if ((harea = FmaOpenArea(pszPath, type, OPENAREA_ALWAYS, fl, &netAddr)) == NULL) {
	  printf("FmaOpenArea(\"%s\") failed, code=%u\n", pszPath, FmaGetLastError());
	  return FALSE;
	}

	// Create new message

	msgn = 0;
	if ((hmsg = FmaOpenMsg(harea, OPENMSG_CREATE, msgn, 0)) == NULL) {
	  printf("FmaOpenMsg() failed, code=%d\n", FmaGetLastError());
	} else {

	  // Create a message and check if ok

	  if ((pfmm = FmaCreateMsg(0)) == NULL) {
		printf("FmaCreateMsg() failed, code=%u\n", FmaGetLastError());
		return FALSE;
	  } 

	  // Add some message attrivutes

	  pfmm->attr|= MA_LOCAL | MA_PRIVATE | MA_KILLSENT | MA_FORCEPICKUP | MA_ENCRYPT;

	  // Add some string properties

	  FmaAddFmp(pfmm, FMP_SENDERNAME, "Pete Kvitek", -1);
	  FmaAddFmp(pfmm, FMP_ORIGADDR,   "2:5020/6.1@fidonet", -1);
	  FmaAddFmp(pfmm, FMP_RECEVRNAME, "All", -1);
	  FmaAddFmp(pfmm, FMP_DESTADDR,   "1:23/45.67@fidonet.org", -1);
	  FmaAddFmp(pfmm, FMP_SUBJECT,    "Testing FMA", -1);
	  FmaAddFmp(pfmm, FMP_GATEADDR,   "11:22/33.44@gate.org", -1);

	  // Add some binary properties

	  FmaGetLocalDttm(&dttm);
	  FmaAddFmp(pfmm, FMP_DTTMWRITTEN, &dttm, sizeof(dttm));

	  crc = JamGetCrc32("Hacking", -1, -1);
	  FmaAddFmp(pfmm, FMP_PASSWORDCRC, &crc, sizeof(crc));

	  for (n = 0; n < numbof(aReplies); n++) aReplies[n] = 1 + n;
	  FmaAddFmp(pfmm, FMP_MSGREPLIES,  aReplies, sizeof(aReplies));

	  // Add some text properties

	  FmaAllocString(&psz, "Here goes a message text\r\r/Pete\r", -1);
	  FmaAddFmp(pfmm, FMP_MESSAGETEXT, &psz, sizeof(psz));

	  // Add some kludge properties

//    FmaAddFmp(pfmm, FMP_MSGID,      "2:5020/6.1@fidonet f727a8c7", -1);
	  FmaAddFmp(pfmm, FMP_REPLYID,    "2:5020/122.1 7ed8ce50", -1);
	  FmaAddFmp(pfmm, FMP_FLAGS,      "XX YYY ZZZZ LOK", -1);
//    FmaAddFmp(pfmm, FMP_PID,        "FMA v??.??.???", -1);
	  FmaAddFmp(pfmm, FMP_REALNAME,   "Blues player", -1);
	  FmaAddFmp(pfmm, FMP_TZUTC,      "-0060", -1);
	  FmaAddFmp(pfmm, FMP_REPLYADDR,  "pete@kvitek.com", -1);
	  FmaAddFmp(pfmm, FMP_REPLYTO,    "2:5020/400 UUCP", -1);
	  FmaAddFmp(pfmm, FMP_FTSKLUDGE,  "RFC-Distribution: fido", -1);

	  // Add some eom kludge properties

	  FmaAddFmp(pfmm, FMP_SEENBY2D,   "5020/204 238 242 362 400 604 758", -1);
	  FmaAddFmp(pfmm, FMP_SEENBY2D,   "5020/1851 1978 5021/7 5022/5 12", -1);
	  FmaAddFmp(pfmm, FMP_PATH2D,     "5020/400 238 204 52 6", -1);
	  FmaAddFmp(pfmm, FMP_EOMKLUDGE,  "Via UniMail/386 0.45b 2:5020/268, 15 Apr 1999 00:15:20 UTC +0400", -1);
	  FmaAddFmp(pfmm, FMP_EOMKLUDGE,  "Via ifmail 2:5020/52@fidonet, Thu Apr 15 1999 at 06:05 (2.11)", -1);
	  FmaAddFmp(pfmm, FMP_EOMKLUDGE,  "Via 2:5020/6@fidonet @19990415.074252.UTC+4 T-Mail 2604.NT", -1);
	  FmaAddFmp(pfmm, FMP_EOMKLUDGE,  "Recd from 2:5020/6 @19990415.082013.UTC+3", -1);

	  // Dump message abstraction

	  FmaDumpMsg(pfmm);

	  // Load message and check if ok

	  if (!FmaSaveMsg(hmsg, pfmm, FSM_AUTOMSGID | FSM_AUTOPID, -1)) {
		printf("FmaSaveMsg() failed, code=%d\n", FmaGetLastError());
	  } 

	  // Destroy message abstraction

	  FmaDestroyMsg(pfmm);

	  // Close message

	  if (!FmaCloseMsg(hmsg)) {
		printf("FmaCloseMsg() failed, code=%d\n", FmaGetLastError());
	  }
	}

	// Close an area and check if ok

	if (!FmaCloseArea(harea)) {
	  printf("FmaCloseArea(\"%s\") failed, code=%d\n", pszPath, FmaGetLastError());
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Main...
//

  void main(int argc, char * argv[])
  {
	// Force debugging log

	if (!getenv("FMA-LOG")) putenv("FMA-LOG=*");

	// Initialize the app and check if ok

	if (!DoInitApp()) {
	  exit(EXIT_FAILURE);
	}

	// Basic functionality tests

//	DoOpenArea("..\\m\\test", MBTYPE_FTS);
//	DoOpenArea("..\\m\\test", MBTYPE_SQU);
//	DoOpenArea("..\\m\\test", MBTYPE_JAM);
//	DoOpenArea("..\\m\\test", MBTYPE_PKT);

	DoLoadMsg("..\\m\\test", MBTYPE_FTS, 1);
//	DoLoadMsg("..\\m\\test", MBTYPE_SQU, 10);
//	DoLoadMsg("..\\m\\test", MBTYPE_JAM, 1);
//	DoLoadMsg("..\\m\\test", MBTYPE_PKT, 1);

//	DoSaveMsg("..\\m\\test", MBTYPE_FTS, 0);
//	DoSaveMsg("..\\m\\test", MBTYPE_SQU, 0);
//	DoSaveMsg("..\\m\\test", MBTYPE_JAM, 0);
//	DoSaveMsg("..\\m\\test", MBTYPE_PKT, 0);

	// Miscellaneous tests

//	DoTestDttm();
//	DoTestFmm();
//	DoTestMsgIdStamp();

	// Cleanup and exit

	FmaTerminate();

	exit(EXIT_SUCCESS);
  }

/////////////////////////////////////////////////////////////////////////////
// End of Test.c
/////////////////////////////////////////////////////////////////////////////
