/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Utility routines header file
//
// FmaUti.h
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 17/Feb/1999
//
// History: 17/Jan/1999 kvitekp - Initially created for IMA project.
//
//          16/Feb/2000 kvitekp - Added FMATM_DTTMTIMEBASE
//
//          17/Feb/2000 kvitekp - Renamed FormatNetAddr to FmaFormatNetAddr
//
//          22/Feb/2000 kvitekp - Added namespace compatibility #ifdefs
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _FMA_UTI_INCLUDED
#define _FMA_UTI_INCLUDED

  // Declare C/C++ frame

#ifdef __cplusplus
extern "C" {
#endif

  // Include private headers

#include "FmaDef.h"

/////////////////////////////////////////////////////////////////////////////
// List management declarations
/////////////////////////////////////////////////////////////////////////////

#ifndef FMA_NO_LST_DEFS

  // List entry structure

  typedef struct _LST_ENTRY {   /* le */
	struct _LST_ENTRY * plePrev;        // Pointer to prev entry
	struct _LST_ENTRY * pleNext;        // Pointer to next entry
  } LST_ENTRY, FAR * PLST_ENTRY;

  // Function prototypes

  FMAAPI_(UINT) FmaListAdd(PLST_ENTRY * pple, PLST_ENTRY ple, UINT ile);
  FMAAPI_(PLST_ENTRY) FmaListDel(PLST_ENTRY * pple, UINT ile);
  FMAAPI_(UINT) FmaListCount(PLST_ENTRY ple);

  // Special list entry index values

#define LIST_END    ((UINT)-1)          // Used by FmaListAdd() 

  // Map functions to common names providing appropriate casting

#ifndef FMA_NO_FUNC_DEFS

#define ListAdd(pple, ple, ile)     \
	 FmaListAdd((PLST_ENTRY *)(pple), (PLST_ENTRY)(ple), (ile))

#define ListDel(pple, ile) (PVOID)  \
	 FmaListDel((PLST_ENTRY *)(pple), (ile))

#define ListCount(ple)              \
	 FmaListCount((PLST_ENTRY)(ple))

#endif

#endif // #ifndef FMA_NO_LST_DEFS

/////////////////////////////////////////////////////////////////////////////
// String manipulation declarations
/////////////////////////////////////////////////////////////////////////////

#ifndef FMA_NO_STR_DEFS

  // Private character manipulation macros (ascii only)

#define ISLOWER(ch) ((ch) >= 'a' && (ch) <= 'z')
#define ISUPPER(ch) ((ch) >= 'A' && (ch) <= 'Z')
#define ISDIGIT(ch) ((ch) >= '0' && (ch) <= '9')
#define ISALPHA(ch) (ISLOWER(ch) || ISUPPER(ch))
#define TOLOWER(ch) (ISUPPER(ch) ? (ch) + ('a' - 'A') : (ch))
#define TOUPPER(ch) (ISLOWER(ch) ? (ch) - ('a' - 'A') : (ch))

  // Private string formatting

#ifdef FMA_PRIVATE_FORMAT

  FMAAPIV_(INT) FmaFormat(PCH pch, UINT cch, PSZ pszFormat, ...);
  FMAAPI_(INT)  FmaFormatV(PCH pch, UINT cch, PSZ pszFormat, va_list parg);

#ifndef FMA_NO_FUNC_DEFS
#define Format  FmaFormat
#define FormatV FmaFormatV
#endif 

#else

#ifndef FMA_NO_FUNC_DEFS
#define Format  _snprintf
#define FormatV _vsnprintf
#endif 

#endif 

  // Function prototypes

  FMAAPI_(CHAR) FmaSkipSpaces(PCH * ppch);
  FMAAPI_(CHAR) FmaTrimSpaces(PCH * ppch);
  FMAAPI_(PSZ)  FmaAllocString(PSZ * ppsz, PCH pch, UINT cch);
  FMAAPI_(BOOL) FmaEqPsz(PSZ psz1, PSZ psz2);
  FMAAPI_(BOOL) FmaEqPch(PCH pch1, PCH pch2, UINT cch);
  FMAAPI_(UINT) FmaCpyStrMax(PCH pchTo, PCH pchFrom, UINT cchMax);
  FMAAPI_(ULONG) FmaGetHash(PSZ psz);
  FMAAPI_(ULONG) FmaGetCrc32(PBYTE pb, ULONG cb, ULONG crc, BOOL bLowerCase);

  // Map functions to common names

#ifndef FMA_NO_FUNC_DEFS
#define SkipSpaces  FmaSkipSpaces
#define TrimSpaces  FmaTrimSpaces
#define AllocString FmaAllocString
#define EqPsz       FmaEqPsz
#define EqPch       FmaEqPch
#define CpyStrMax   FmaCpyStrMax
#define GetHash     FmaGetHash
#define GetCrc32    FmaGetCrc32
#endif

#endif // #ifndef FMA_NO_STR_DEFS

/////////////////////////////////////////////////////////////////////////////
// Linked string list management declarations
/////////////////////////////////////////////////////////////////////////////

#ifndef FMA_NO_LSZ_DEFS

  // Linked string list entry structure

  typedef struct _LSZ {         /* lsz */
	struct _LSZ * plszPrev;             // Pointer to the prev entry in list
	struct _LSZ * plszNext;             // Pointer to the next entry in list
	ULONG   hash;                       // String hash value
//  PVOID   pv;                         // Custom pointer
//  UINT    id;                         // Custom id
	CHAR    ach[1];                     // String (asciiz)
  } LSZ, FAR * PLSZ;

  // Function prototypes

  FMAAPI_(PLSZ) FmaAddLsz(PLSZ * pplsz, PCH pch, UINT cch, UINT ilsz);
  FMAAPI_(PLSZ) FmaGetLsz(PLSZ plsz, PSZ psz);
  FMAAPI_(BOOL) FmaDelLsz(PLSZ * pplsz, UINT ilsz);
  FMAAPI_(VOID) FmaDelLszList(PLSZ * pplsz);

  // Map functions to common names

#ifndef FMA_NO_FUNC_DEFS
#define AddLsz       FmaAddLsz
#define GetLsz       FmaGetLsz
#define DelLsz       FmaDelLsz
#define DelLszList   FmaDelLszList
#endif

#endif // #ifndef FMA_NO_LSZ_DEFS

/////////////////////////////////////////////////////////////////////////////
// Dynamic string management declarations
/////////////////////////////////////////////////////////////////////////////

#ifndef FMA_NO_DSZ_DEFS

  // Dynamic string structure

  typedef struct _DSZ {         /* dsz */
	UINT    cch;                        // Max string length (not incl null)
	UINT    ich;                        // Current string length
	PSZ     psz;                        // String pointer (asciiz)
  } DSZ, FAR * PDSZ;

  // Dynamic string definitions

#define DSZ_INIT     {0,0,NULL}         // Dynamic string initializer
#define DSZ_GROW      127               // Allocation grow granularity
#define DSZ_GROWTWEAK 8                 // Allocation grow tweak factor

  // Function prototypes

  FMAAPI_(BOOL)  FmaAddDsz(PDSZ pdsz, PCH pch, UINT cch);
  FMAAPIV_(BOOL) FmaAddDszFormat(PDSZ pdsz, PSZ pszFormat, ...);
  FMAAPI_(BOOL)  FmaAddDszFormatV(PDSZ pdsz, PSZ pszFormat, va_list parg);
  FMAAPI_(BOOL)  FmaSetDsz(PDSZ pdsz, PCH pch, UINT cch);
  FMAAPIV_(BOOL) FmaSetDszFormat(PDSZ pdsz, PSZ pszFormat, ...);
  FMAAPI_(BOOL)  FmaSetDszFormatV(PDSZ pdsz, PSZ pszFormat, va_list parg);
  FMAAPI_(BOOL)  FmaFixDsz(PDSZ pdsz);
  FMAAPI_(VOID)  FmaDelDsz(PDSZ pdsz);

  // Map functions to common names

#ifndef FMA_NO_FUNC_DEFS
#define AddDsz         FmaAddDsz
#define AddDszFormat   FmaAddDszFormat
#define AddDszFormatV  FmaAddDszFormatV
#define SetDsz         FmaSetDsz
#define SetDszFormat   FmaSetDszFormat
#define SetDszFormatV  FmaSetDszFormatV
#define FixDsz         FmaFixDsz
#define DelDsz         FmaDelDsz
#endif

#endif // #ifndef FMA_NO_DSZ_DEFS

/////////////////////////////////////////////////////////////////////////////
// Network address routines declarations
/////////////////////////////////////////////////////////////////////////////

  // Function prototypes

  FMAAPI_(PCH)  FmaScanNetAddr(NETADDR * pnetAddr, PSZ psz);
  FMAAPI_(UINT) FmaFormatNetAddr(PCH pch, UINT cchMax, NETADDR * pnetAddr, 
							     PSZ pszDomain);

/////////////////////////////////////////////////////////////////////////////
// Date and time routines declarations
/////////////////////////////////////////////////////////////////////////////

  // Number of seconds since 00:00:00 01/Jan/1970 to 23:59:59 01/Jan/1980

#define FMATM_DTTMTIMEBASE	   315532800
  
  // FmaFormatDttm() flags

#define FDTTM_DATEONLY            0x0001        // Format date only, no time
#define FDTTM_TIMEONLY            0x0002        // Format time only, no date
#define FDTTM_TIMEFIRST           0x0004        // Format time before date
#define FDTTM_TIMESECONDS         0x0008        // Format time with seconds
#define FDTTM_MONTHNUMBER         0x0010        // Format month as number
#define FDTTM_4DIGITYEAR          0x0020        // Format full 4 digit year

  // Function prototypes

  FMAAPI_(BOOL) FmaCheckDttm(PDTTM pdttm);
  FMAAPI_(BOOL) FmaAsciiToDttm(PSZ psz, PDTTM pdttm);
  FMAAPI_(ULONG) FmaGetTimeFromDttm(PDTTM pdttm);
  FMAAPI_(BOOL) FmaGetDttmFromTime(PDTTM pdttm, ULONG time);

  FMAAPI_(UINT) FmaFormatDttm(PCH pchDttm, UINT cchMax, PDTTM pdttm, ULONG fl,
							  CHAR chDateSep, CHAR chTimeSep, 
							  PSZ pszDateTimeSep);

/////////////////////////////////////////////////////////////////////////////
// Header cleanup
/////////////////////////////////////////////////////////////////////////////

  // Terminate C/C++ frame

#ifdef __cplusplus
}
#endif

#endif  // #ifndef _FMA_UTI_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// End of FmaUti.h
/////////////////////////////////////////////////////////////////////////////
