/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Fts api declarations module
//
// FtsApi.h
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 03/Oct/1999
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////
//
// Note: These declaratons were derived from original
//       specifications published in FTS-0001.16 and 
//       some common real life practices.
//
/////////////////////////////////////////////////////////////////////////////
//
// NOTE: ALL STRUCTURES IN THIS MODULE ASSUME SINGLE BYTE PACKING !!!
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _FTS_API_INCLUDED
#define _FTS_API_INCLUDED

  // Declare C/C++ frame

#ifdef __cplusplus
extern "C" {
#endif

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

  // Set structure packing

#ifdef FMA_PUSHPOP_PACKING
#pragma pack(push, 1)
#endif

/////////////////////////////////////////////////////////////////////////////
// Fts base declarations
/////////////////////////////////////////////////////////////////////////////

  // Fts network address structure

#ifndef _NETADDR_DEFINED
#define _NETADDR_DEFINED

  typedef struct _NETADDR { /* net */
	USHORT  zone;               // Zone number
	USHORT  net;                // Network number
	USHORT  node;               // Node number
	USHORT  point;              // Point number
  } NETADDR, FAR * PNETADDR;

#endif 

  // Fts date time structure

#ifndef _DTTM_DEFINED
#define _DTTM_DEFINED

  typedef struct _DTTM { /* dttm */
	UINT    da:5;               // Day of the month (131)
	UINT    mo:4;               // Month (1 = January, 2 = February, and so on)
	UINT    yr:7;               // Year offset from 1980 (add 1980 to get actual year) 
	UINT    ss:5;               // Second divided by 2
	UINT    mm:6;               // Minute (059)
	UINT    hh:5;               // Hour (023 on a 24-hour clock)
  } DTTM, FAR * PDTTM;

#endif 

  // Fts base file extensions

#define FTS_MSG_FILEEXT  ".MSG" // Message file extension
#define FTS_LRD_FILEEXT  ".BBS" // Lastread file extension

  // Fts base file names

#define FTS_LRD_FILENAME "LASTREAD"  // Lastread file name

  // Fts message header structure mostly used

  typedef struct _FMH { /* fmh */
	CHAR    achBy[36];          // From name
	CHAR    achTo[36];          // To name
	CHAR    achRe[72];          // Message subject
	CHAR    achDate[20];        // ASCII date information
	USHORT  timesRead;          // Number of times read
	USHORT  destNode;           // Destination node number
	USHORT  origNode;           // Origination node number
	USHORT  cost;               // Unit cost charged to send the message
	USHORT  origNet;            // Origination net number
	USHORT  destNet;            // Destination net number
	DTTM    dttmWritten;        // Message written date/time stamp
	DTTM    dttmArrived;        // Message arrived date/time stamp
	USHORT  replyTo;            // Msg is a reply to this msg number
	USHORT  attr;               // Message attributes (FMA_)
	USHORT  nextReply;          // Next message in the reply thread
  } FMH, FAR * PFMH;

  // Fts message header structure according to FTS-001.16 with
  // zone/point info in place of dttm's, nice, but rarely used

  typedef struct _FMH2 { /* fmh2 */
	CHAR    achBy[36];          // From name
	CHAR    achTo[36];          // To name
	CHAR    achRe[72];          // Message subject
	CHAR    achDate[20];        // ASCII date information
	USHORT  timesRead;          // Number of times read
	USHORT  destNode;           // Destination node number
	USHORT  origNode;           // Origination node number
	USHORT  cost;               // Unit cost charged to send the message
	USHORT  origNet;            // Origination net number
	USHORT  destNet;            // Destination net number
	USHORT  destZone;           // Destination zone number
	USHORT  origZone;           // Origination zone number
	USHORT  destPoint;          // Destination point number
	USHORT  origPoint;          // Origination point number
	USHORT  replyTo;            // Msg is a reply to this msg number
	USHORT  attr;               // Message attributes (FMA_)
	USHORT  nextReply;          // Next message in the reply thread
  } FMH2, FAR * PFMH2;

  // Fts message attributes in fmh.attr

#define FMA_PRVT     0x0001u    // Private                 +
#define FMA_CRSH     0x0002u    // Crash                   | These
#define FMA_READ     0x0004u    // Read                    | flags are
#define FMA_SENT     0x0008u    // Sent                    | standard for
#define FMA_FILE     0x0010u    // File attach             | FidoNet(tm)
#define FMA_TRAN     0x0020u    // Transit                 | message attr
#define FMA_ORPH     0x0040u    // Orphan                  |
#define FMA_KILL     0x0080u    // Kill/sent               |
#define FMA_LOCL     0x0100u    // Local                   |
#define FMA_HOLD     0x0200u    // Hold                    |
#define FMA_RSV2     0x0400u    // Reserved                |
#define FMA_FREQ     0x0800u    // File request            |
#define FMA_RREQ     0x1000u    // Return receipt request  |
#define FMA_RRCP     0x2000u    // Confirmation receipt msg|
#define FMA_AREQ     0x4000u    // Audit request           |
#define FMA_UREQ     0x8000u    // File update request     +

  // Fts lastread file record structure

  typedef struct _FLR { /* flr */
	UMSG    umsg;               // Last read message uid
  } FLR, FAR * PFLR;

/////////////////////////////////////////////////////////////////////////////
// Fts api declarations
/////////////////////////////////////////////////////////////////////////////

  // Special values used by FtsReadMsg()

#define FRM_GETCTRLCOPY          0      // *pcchCtrl: Allocate a copy of msg control info
#define FRM_GETCTRLPTR  ((ULONG)-1)     // *pcchCtrl: Return a readonly pointer to msg control info
#define FRM_GETALLTEXT  ((ULONG)-1)     // *pcchText: read entire message text

  // Function prototypes

  FMAAPI_(PAPI) FtsGetApi(VOID);
  FMAAPI_(HAREA) FtsOpenArea(PSZ pszPath, UINT fOpenMode, ULONG fl, 
							 NETADDR * pnetAddr);

  FMAAPI_(BOOL) FtsCloseArea(HAREA harea);
  FMAAPI_(BOOL) FtsLockArea(HAREA harea, BOOL bLock);

  FMAAPI_(BOOL) FtsValidateMsg(HAREA harea, MSGN msgn, PULONG pimsg);
  FMAAPI_(BOOL) FtsCreateMsgMap(HAREA harea, UMSG * paumsg[], PULONG pcumsg);
  FMAAPI_(BOOL) FtsEnumMsgs(HAREA harea, PFNENUMMSGS pfnCallback, 
							PVOID pvCallbackData, ULONG fl);

  FMAAPI_(BOOL) FtsDeleteMsg(HAREA harea, MSGN msgn);
  FMAAPI_(BOOL) FtsSetCurMsg(HAREA harea, MSGN msgn);
  FMAAPI_(MSGN) FtsGetCurMsg(HAREA harea);
  FMAAPI_(MSGN) FtsGetFirstMsg(HAREA harea);
  FMAAPI_(MSGN) FtsGetHighMsg(HAREA harea);
  FMAAPI_(LONG) FtsGetMsgCount(HAREA harea);

  FMAAPI_(UMSG) FtsMsgnToUid(HAREA harea, MSGN msgn);
  FMAAPI_(MSGN) FtsUidToMsgn(HAREA harea, UMSG umsg, UINT fMode);

  FMAAPI_(HMSG) FtsOpenMsg(HAREA harea, UINT mode, MSGN msgn, ULONG fl);
  FMAAPI_(BOOL) FtsCloseMsg(HMSG hmsg);

  FMAAPI_(BOOL) FtsLoadMsg(HMSG hmsg, PFMM * ppfmm, ULONG fl);
  FMAAPI_(BOOL) FtsSaveMsg(HMSG hmsg, PFMM pfmm, ULONG fl, ULONG cchTextTotal);

  FMAAPI_(BOOL) FtsReadMsg(HMSG hmsg, PFMH pfmh, PCH * ppchCtrl, ULONG * pcchCtrl,
						   PCH * ppchText, ULONG * pcchText, ULONG * poffText);

  FMAAPI_(BOOL) FtsWriteMsg(HMSG hmsg, PFMH pfmh, PCH pchCtrl, ULONG cchCtrl,
							PCH pchText, ULONG cchText, ULONG * poffText, 
							ULONG cchTextTotal);

  FMAAPI_(BOOL) FtsGetMsgCtrlLen(HMSG hmsg, ULONG * pcchCtrl);
  FMAAPI_(BOOL) FtsGetMsgTextLen(HMSG hmsg, ULONG * pcchText);

  FMAAPI_(BOOL) FtsGetLRInfo(HAREA harea, ULONG uid, PLRINFO plri);
  FMAAPI_(BOOL) FtsSetLRInfo(HAREA harea, ULONG uid, PLRINFO plri, 
							 PSZ pszUserName);

/////////////////////////////////////////////////////////////////////////////
// Header cleanup
/////////////////////////////////////////////////////////////////////////////

  // Restore structure packing

#ifdef FMA_PUSHPOP_PACKING
#pragma pack(pop)
#endif

  // Terminate C/C++ frame

#ifdef __cplusplus
}
#endif

#endif  // #ifndef _FTS_API_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// End of FtsApi.h
/////////////////////////////////////////////////////////////////////////////
