/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Fts api private declarations module
//
// FtsApiP.h
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 23/Sep/1999
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _FTS_APIP_INCLUDED
#define _FTS_APIP_INCLUDED

  // Declare C/C++ frame

#ifdef __cplusplus
extern "C" {
#endif

  // Include types declaratons

#include "FmaDef.h"
#include "FtsApi.h"

/////////////////////////////////////////////////////////////////////////////
// Fts area declarations
/////////////////////////////////////////////////////////////////////////////

  // Forward declarations

  struct _FMSG;

  // Fts message area object structure

  typedef struct _FAREA {       /* area */
	ULONG   magic;                      // Magic number (FAREA_MAGIC)
	PAPI    papi;                       // Fts api pointer
	ULONG   cmsg;                       // Number of messages
	MSGN    msgnCur;                    // Current message number
	MSGN    msgnHigh;                   // Highest message number
	UINT *  aumsg;                      // Message numbers array
	UINT    cumsg;                      // Message numbers array size
	BOOL    bAreaLocked;                // Area lock count
	ULONG   fl;                         // Open area flags (FOA_)
	struct _FMSG * pmsgList;            // Open messges list
	NETADDR netAddr;                    // Default network address
	CHAR    achPath[1];                 // Area path (asciiz)
  } FAREA, FAR * PFAREA;

  // Fts message area object magic number

#define FAREA_MAGIC ((ULONG)0x45524146) // Long equivalent of "FARE"

  // Fts message object structure

  typedef struct _FMSG {        /* msg */
	struct _FMSG * pmsgPrev;            // Pointer to prev open msg in area
	struct _FMSG * pmsgNext;            // Pointer to next open msg in area
	ULONG   magic;                      // Magic number (FMSG_MAGIC)
	PFAREA  parea;                      // Message area handle
	MSGN    msgn;                       // Message number
	UINT    mode;                       // Message open mode
	ULONG   fl;                         // Message open flags (FOM_)
	PSZ     pszCtrl;                    // Message control info
	UINT    cchCtrl;                    // Message control info length
	FMH     fmh;                        // Message header
	UINT    fs;                         // Message status flags (FMF_)
	LONG    lmt;                        // Message text offset
	LONG    lmtEnd;                     // Message text end offset
	LONG    lmtMax;                     // Message text max offset
	HFILE   hfile;                      // Message file handle
  } FMSG, FAR * PFMSG;

  // Fts message object magic number

#define FMSG_MAGIC ((ULONG)0x47534D46)  // Long equivalent of "FMSG"

  // Fts message status flags in msg.fs

#define FMF_GOTMSGH       0x0001        // Message header has been associated
#define FMF_GOTCTRL       0x0002        // Control info has been associated
#define FMF_BOGUS         0x0004        // -- reserved
#define FMF_GOTFRAME      0x0008        // Writing frame selected for new message
#define FMF_MSGHWRITTEN   0x0100        // Message header has been written
#define FMF_CTRLWRITTEN   0x0200        // Message contrl info has been written
#define FMF_TEXTWRITTEN   0x0400        // Message text has been written
#define FMF_CTRLHASAREA   0x1000        // Message contrl info has AREA w/o ^a

  // Message has been written to mask

#define FMF_MASKWRITTEN   (FMF_MSGHWRITTEN | FMF_CTRLWRITTEN | FMF_TEXTWRITTEN)

/////////////////////////////////////////////////////////////////////////////
// Miscellaneous declarations
/////////////////////////////////////////////////////////////////////////////

  // Function prototypes

  FMAAPI_(BOOL) FtsCheckArea(PFAREA parea);
  FMAAPI_(BOOL) FtsCheckMsg(PFMSG pmsg);
  FMAAPI_(VOID) FtsSortMsgMap(PFAREA parea);
  FMAAPI_(BOOL) FtsRescanArea(PFAREA parea);
  FMAAPI_(BOOL) FtsAddMsg(PFAREA parea, MSGN msgn);
  FMAAPI_(BOOL) FtsCheckBase(PSZ pszPath);

  // Fts logging control defines

#ifdef  FMA_LOG
#define FMA_LOG_FTS_AREA
#define FMA_LOG_FTS_LAST
#define FMA_LOG_FTS_MSG
#define FMA_LOG_FTS_READ
#define FMA_LOG_FTS_WRITE
#endif

/////////////////////////////////////////////////////////////////////////////
// Header cleanup
/////////////////////////////////////////////////////////////////////////////

  // Terminate C/C++ frame

#ifdef __cplusplus
}
#endif

#endif  // #ifndef _FTS_APIP_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// End of FtsApiP.h
/////////////////////////////////////////////////////////////////////////////
