/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Fts base integrity check routines module
//
// FtsCheck.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 06/Oct/1999
// Updated: 06/Oct/1999
//
// History: 06/Oct/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "FtsApi.h"
#include "FtsApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine is a callback for FmaEnumFiles to check message files
//

  static BOOL FMACALLBACK DoCheckMsg(PSZ pszFile, ULONG cbFile, ULONG attr, 
									 ULONG * pcmsg)
  {
	BOOL bOk = TRUE;
	UINT msgn;

	ASSERT(pszFile != NULL);

	// Check the size if the message file and leave if too small

	if (cbFile < sizeof(FMH)) {
#ifdef FMA_LOG
Log("FtsCheckBase -----: invalid message file size=%ld", cbFile);
#endif
	  bOk = FALSE;
	}

	// Convert message number and check if ok

	if ((msgn = atoi(pszFile)) == 0) {
#ifdef FMA_LOG
Log("FtsCheckBase -----: invalid message file number=%ld", msgn);
#endif
	  bOk = FALSE;
	}

	// Increment caller's message count

	if (pcmsg) (*pcmsg)++;

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine checks the Fts message base integrity
//

  FMAAPI_(BOOL) FtsCheckBase(PSZ pszPath)
  {
	ULONG fl = FOA_LOCKAREA | FOA_DENYALL;
	BOOL bOk = TRUE;
	ULONG cmsg = 0;
	PFAREA parea;

	// Open area in exclusive mode and check if ok

	if ((parea = FtsOpenArea(pszPath, OPENAREA_EXISTING, fl, NULL)) == NULL) {
#ifdef FMA_LOG
Log("FtsCheckBase -----: Can't open %s", pszPath);
#endif
	  return TRUE;
	}

	// Enumerate all messsage files and check if ok

	if (parea->cmsg) {
	  if (!FmaEnumFiles(parea->achPath, "*"FTS_MSG_FILEEXT, 
					   (PFNENUMFILES)DoCheckMsg, &cmsg, 0)) {
		bOk = FALSE;
	  }
	}

	// Check the message count

	if (parea->cmsg != cmsg) {
#ifdef FMA_LOG
Log("FtsCheckBase -----: invalid cmsg=%ld, expected=%ld", cmsg, parea->cmsg);
#endif
	  bOk = FALSE;
	}

	// Close area 

	FtsCloseArea(parea);

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// End of FtsCheck.c
/////////////////////////////////////////////////////////////////////////////
