/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Fts lastread routines module
//
// FtsLast.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 03/Oct/1999
// Updated: 03/Oct/1999
//
// History: 03/Oct/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "FtsApi.h"
#include "FtsApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine opens fts lastread file
//

  static HFILE FMASUB DoOpenLRFile(PFAREA parea, ULONG uid, HFILE * phfile)
  {
	UINT fOpen = FOPENFILE_READWRITE | FOPENFILE_CREATE | FOPENFILE_DENYNONE;
	CHAR achLRFile[MAX_PATH];

	// Make up the lastread file name which is LASTREAD. for zero uid,
	// and LASTREAD.BBS for non zero uids

	strcpy(achLRFile, parea->achPath);
	strcat(achLRFile, FTS_LRD_FILENAME);
	if (uid) strcat(achLRFile, FTS_LRD_FILEEXT);

	// Open file and check if ok

	if (!FmaOpenFile(achLRFile, fOpen, phfile))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine gets fts lastread file record for the specified uid
//

  FMAAPI_(BOOL) FtsGetFlr(PFAREA parea, ULONG uid, PFLR pflr, UINT cflr)
  {
	ULONG cb, cbRead;
	HFILE hfile = -1;
	LONG offs;

	ASSERT(parea != NULL);

	// Open lastread file and check if ok

	if (!DoOpenLRFile(parea, uid, &hfile)) {
Fail: if (hfile != -1) FmaCloseFile(&hfile);
	  return FALSE;
	}

	// Find out offset of the lastread record and check if ok

	offs = uid * sizeof(FLR); cb = cflr * sizeof(FLR);
	if ((ULONG)offs + cb > (ULONG)FmaGetFileSize(hfile)) {
	  FMAERROR(NOT_FOUND);
	  goto Fail;
	}

	// Read lastread file records and check if ok

	if (!FmaReadFile(hfile, offs, pflr, cb, &cbRead)) {
	  goto Fail;
	} else 
	  if (cbRead != cb) {
		FMAERROR(INVALID_DATA);
		goto Fail;
	  }

	// Close lastread file

	FmaCloseFile(&hfile);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets fts lastread file record for the specified uid
//

  FMAAPI_(BOOL) FtsSetFlr(PFAREA parea, ULONG uid, PFLR pflr, UINT cflr)
  {
	ULONG cb, cbWritten;
	HFILE hfile = -1;
	LONG offs;

	ASSERT(parea != NULL);

	// Open lastread file and check if ok

	if (!DoOpenLRFile(parea, uid, &hfile)) {
Fail: if (hfile != -1) FmaCloseFile(&hfile);
	  return FALSE;
	}

	// Find out offset of the lastread record and check if ok

	offs = uid * sizeof(FLR); cb = cflr * sizeof(FLR);

	// Write lastread file records and check if ok

	if (!FmaWriteFile(hfile, offs, pflr, cb, &cbWritten))
	  goto Fail;

	ASSERT(cbWritten == cb);

	// Close lastread file

	FmaCloseFile(&hfile);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets lastread info
//

  FMAAPI_(BOOL) FtsGetLRInfo(HAREA harea, ULONG uid, PLRINFO plri)
  {
	PFAREA parea;
	FLR aflr[2];
	UINT cflr;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check if we need to get high read message and
	// set number of lastread records to set

	if (parea->fl & FOA_HIGHREAD) {
	  cflr = 2;

	  // Make sure user id is even

	  if (uid & 1) {
		FMAERROR(INVALID_USERID);
		return FALSE;
	  }
	} else 
	  cflr = 1;

	// Get lastread file record(s) and check if ok

	if (!FtsGetFlr(parea, uid, aflr, cflr)) 
	  return FALSE;

	// Pass info to the caller

	if (plri) {
	  if (cflr == 2) {
		plri->umsgLast = aflr[0].umsg;
		plri->umsgHigh = aflr[1].umsg;
	  } else {
		plri->umsgLast = aflr[0].umsg;
		plri->umsgHigh = aflr[0].umsg;
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets lastread info
//

  FMAAPI_(BOOL) FtsSetLRInfo(HAREA harea, ULONG uid, PLRINFO plri, PSZ pszUserName)
  {
	PFAREA parea;
	FLR aflr[2];
	UINT cflr;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea) || !plri) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check if we need to get high read message and
	// set number of lastread records to set

	if (parea->fl & FOA_HIGHREAD) {
	  cflr = 2;

	  // Make sure user id is even

	  if (uid & 1) {
		FMAERROR(INVALID_USERID);
		return FALSE;
	  }
	} else 
	  cflr = 1;

	// Make up the lastread file record(s)

	aflr[0].umsg = plri->umsgLast;
	aflr[1].umsg = plri->umsgHigh;

	// Set the lastread file record(s0 and check if ok

	if (!FtsSetFlr(parea, uid, aflr, cflr))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of FtsLast.c
/////////////////////////////////////////////////////////////////////////////
