/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Fts message routines module
//
// FtsMsg.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 09/Feb/2000
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
//          09/Feb/2000 kvitekp - Fixed msgnHigh/Cur/cmsg not being adjusted
//                                in DoFixStrayNewMsg()
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "FtsApi.h"
#include "FtsApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Message file name template

  static CHAR achMsgFmt[] = "%s%lu"FTS_MSG_FILEEXT;

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine creates message object
//

  static PFMSG FMASUB DoCreateMsg(PFAREA parea, MSGN msgn, UINT mode, ULONG fl)
  {
	PFMSG pmsg;

	ASSERT(parea != NULL);

	// Create message object and check if ok

	if ((pmsg = MemAlloc(sizeof(FMSG), MA_CLEAR)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return NULL;
	} 

	// Initialize message object

	pmsg->magic = FMSG_MAGIC;
	pmsg->parea = parea;
	pmsg->msgn = msgn;
	pmsg->mode = mode;
	pmsg->fl = fl;
	pmsg->hfile = -1;

	return pmsg;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine destroys message object
//

  static BOOL FMASUB DoDestroyMsg(PFMSG pmsg)
  {
	ASSERT(pmsg != NULL);

	// Close the message file and reset its handle

	FmaCloseFile(&pmsg->hfile);

	// Free associated message control info

	if (pmsg->pszCtrl) MemFree(pmsg->pszCtrl);

	// Reset message object magic number

	pmsg->magic = ~FMSG_MAGIC;

	// Free message object

	MemFree(pmsg);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine translates special message numbers
//

  static BOOL FMASUB DoXlateMsgn(PFAREA parea, MSGN * pmsgn)
  {
	MSGN msgn = *pmsgn;
	ULONG imsg;

	ASSERT(parea != NULL);

	// Select current message

	if (msgn == MSGN_CUR || msgn == MSGN_CURORPREV || msgn == MSGN_CURORNEXT) {
	  if (!FtsValidateMsg(parea, parea->msgnCur, &imsg)) {
		if (msgn == MSGN_CURORPREV) goto Prev;
		if (msgn == MSGN_CURORNEXT) goto Next;
NotF:   FMAERROR(NOT_FOUND);
		return FALSE;
	  } else {
		msgn = parea->msgnCur;
	  }
	} else

	if (msgn == MSGN_PREV) {

	  // Get current message index and if it's gone, scan the map
	  // looking for a lower message, otherwise just select previous 
	  // message in the map

Prev: if (!parea->cumsg) goto NotF;
	  if (!FtsValidateMsg(parea, msgn = parea->msgnCur, &imsg)) {
		for (imsg = parea->cumsg - 1; (MSGN)parea->aumsg[imsg] >= msgn; ) {
		  if (!imsg--) 
			goto NotF;
		}
		msgn = parea->aumsg[imsg]; 
	  } else {
		if (imsg-- > 0) {
		  msgn = parea->aumsg[imsg];
		} else
		  goto NotF;
	  }
	} else

	if (msgn == MSGN_NEXT) {

	  // Get current message index and if it's gone, scan the map
	  // looking for a higher message, otherwise just select next 
	  // message in the map

Next: if (!parea->cumsg) goto NotF;
	  if (!FtsValidateMsg(parea, msgn = parea->msgnCur, &imsg)) {
		for (imsg = 0; (MSGN)parea->aumsg[imsg] <= msgn; ) {
		  if (++imsg >= parea->cumsg) 
			goto NotF;
		}
		msgn = parea->aumsg[imsg];
	  } else {
		if (++imsg < parea->cumsg) {
		  msgn = parea->aumsg[imsg]; 
		} else
		  goto NotF;
	  }
	} 

	// Set message number for caller

	if (pmsgn) *pmsgn = msgn;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns index of the given message in the open messages list
//

  static BOOL FMASUB DoGetOpenMsgIndex(PFMSG pmsg, UINT * pimsg)
  {
	PFAREA parea = pmsg->parea;
	PFMSG pmsgThis;
	UINT imsg;

	// Scan through the open messages list looking for a match

	for (pmsgThis = parea->pmsgList, imsg = 0; pmsgThis != NULL; 
		 pmsgThis = pmsgThis->pmsgNext, imsg++) {
	  if (pmsgThis == pmsg) {
		if (pimsg) *pimsg = imsg;
		return TRUE;
	  }
	}

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine fixes new message which has not been written
//

  static BOOL FMASUB DoFixStrayNewMsg(PFMSG pmsg)
  {
	PFAREA parea = pmsg->parea;
	CHAR achMsg[MAX_PATH];

	ASSERT(pmsg->hfile != -1);

	// Check if message file length is valid and leave if so

	if ((ULONG)FmaGetFileSize(pmsg->hfile) >= sizeof(FMH))
	  return TRUE;

	// Close the message file and reset its handle

	FmaCloseFile(&pmsg->hfile);

	// Make up the message file name and delete the message file

	Format(achMsg, lengof(achMsg), achMsgFmt, parea->achPath, pmsg->msgn);

	FmaDeleteFile(achMsg);

	// Adjust area message count

	parea->cmsg--;

	// Check if we just deleted current message and reset it

	if (parea->msgnCur == pmsg->msgn) {
	  parea->msgnCur = 0;
	}

	// Check if we just deleted highest message and update it

	if (parea->msgnHigh == pmsg->msgn) {
	  parea->msgnHigh = parea->cmsg ? parea->aumsg[parea->cmsg - 1] : 0;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine opens fts area message
//

  FMAAPI_(HMSG) FtsOpenMsg(HAREA harea, UINT mode, MSGN msgn, ULONG fl)
  {
	UINT nErr, fOpen = FOPENFILE_DENYNONE;
	CHAR achMsg[MAX_PATH];
	BOOL bNewMsg = FALSE;
	PFMSG pmsg = NULL;
	HFILE hfile = -1;
	PFAREA parea;

	// Validate parameters

	if ((parea = (PFAREA)harea) == NULL || !FtsCheckArea(parea)) {
BadP: FMAERROR(INVALID_PARAM);
	  return NULL;
	}

#ifdef FMA_LOG_FTS_MSG
Log("\nFtsOpenMsg       >: [%09Fp] msgn=%ld in %s, mode=%d", pmsg, msgn, parea->achPath, mode);
#endif

	// Translate special message numbers and check if ok

	if (!DoXlateMsgn(parea, &msgn))
	  return NULL;

	// Check the message open mode and make up the open flags

	if (mode == OPENMSG_CREATE) {
	  fOpen|= FOPENFILE_CREATE | FOPENFILE_TRUNCATE | FOPENFILE_READWRITE;

	  // Check if creating a new message and generate new message number

	  if (!msgn) {

		// Check to see if the message number we're going to create already exists and if so,
		// rescan the message area. This is necessary since fts areas do not have locks, thus
		// some other process could create new messages since we rescanned the area.

		Format(achMsg, lengof(achMsg), achMsgFmt, parea->achPath, parea->msgnHigh + 1);
		if (FmaFileExists(achMsg)) {
		  if (!FtsRescanArea(parea))
			return NULL;
		} 

		// Generate new message number but postpone updating highest message untill later

		msgn = parea->msgnHigh + 1; bNewMsg = TRUE;
	  } else {

		// If creating specific message check to see if it's in the message map and if not,
		// force new message to be added to the message map

		if (!FtsValidateMsg(harea, msgn, NULL)) {
		  bNewMsg = TRUE;
		}
	  }
	} else 

	if (mode == OPENMSG_READ) {
	  fOpen|= FOPENFILE_READONLY;
	} else 

	if (mode == OPENMSG_WRITE) {
	  fOpen|= FOPENFILE_WRITEONLY;
	} else 

	if (mode == OPENMSG_READWRITE) {
	  fOpen|= FOPENFILE_READWRITE;
	} else 
	  goto BadP;

	// Create message object and check if ok

	if ((pmsg = DoCreateMsg(parea, msgn, mode, fl)) == NULL)
	  return NULL;

	// Make up the message file name and open it, then check if ok

	Format(achMsg, lengof(achMsg), achMsgFmt, parea->achPath, msgn);

	if (!FmaOpenFile(achMsg, fOpen, &hfile)) {
	  DoDestroyMsg(pmsg);
	  nErr = FmaGetLastError();
	  if (nErr == FMAERROR_CANNOT_OPEN && !FmaFileExists(achMsg)) FMAERROR(NOT_FOUND);
	  return NULL;
	} 

	// Check if we are creating a new message and add it to the message map

	if (bNewMsg) {

	  // Append an entry to the message map and check if ok. Note that we're not
	  // shrinking message map allocation since there might be more messages to add.

	  if (!FtsAddMsg(parea, msgn)) {
		nErr = FmaGetLastError();
		DoDestroyMsg(pmsg); 
		FmaCloseFile(&hfile); 
		FmaDeleteFile(achMsg);
		FmaSetLastError(nErr);
		return NULL;
	  }

	  // Check if this is the highest entry and update highest message,
	  // otherwise sort the message map

	  if (msgn > parea->msgnHigh) {
		parea->msgnHigh = msgn;
	  } else {
		FtsSortMsgMap(parea);
	  }
	}

	// Finalize message object initialization

	pmsg->hfile = hfile;

	// Set current message number and increment open message count. Note that we're
	// thread safe here since every thread is supposed to have its own open areas

	parea->msgnCur = msgn;

	// Add message to the top of the list of open messages for this area

	ListAdd(&parea->pmsgList, pmsg, 0);    

	ASSERT(FtsCheckMsg(pmsg));

#ifdef FMA_LOG_FTS_MSG
Log("FtsOpenMsg       <: [%09Fp] msgn=%ld in %s", pmsg, msgn, parea->achPath);
#endif

	return pmsg;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine closes fts area message
//

  FMAAPI_(BOOL) FtsCloseMsg(HMSG hmsg)
  {
	BOOL bOk = TRUE;
	PFAREA parea;
	PFMSG pmsg;
	UINT imsg;

	// Validate parameters

	if ((pmsg = (PFMSG)hmsg) == NULL || !FtsCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: 

#ifdef FMA_LOG_FTS_MSG
Log("FtsCloseMsg      <: [%09Fp] FAILED, error=%d", pmsg, FmaGetLastError());
#endif

	  return FALSE;
	} else
	  parea = pmsg->parea;

#ifdef FMA_LOG_FTS_MSG
Log("\nFtsCloseMsg       : [%09Fp] msgn=%ld in %s", pmsg, pmsg->msgn, parea->achPath);
#endif

	// Check if this is a new message which has not been written to and
	// update base to forget about it

	if (pmsg->mode == OPENMSG_CREATE && !(pmsg->fs & FMF_MASKWRITTEN)) {
	  if (!DoFixStrayNewMsg(pmsg))
		bOk = FALSE;
	}

	// Delete message from the list of open messages for this area

	if (DoGetOpenMsgIndex(pmsg, &imsg)) {
	  ListDel(&parea->pmsgList, imsg);
	} else {
	  ASSERT(DoGetOpenMsgIndex(pmsg, &imsg));
	}

	// Destroy message object

	DoDestroyMsg(pmsg);

	// Check if failed and jump to failure return

	if (!bOk) goto Fail;

#ifdef FMA_LOG_FTS_MSG
Log("FtsCloseMsg      <: [%09Fp] in %s", pmsg, parea->achPath);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine validates message object
//

  FMAAPI_(BOOL) FtsCheckMsg(PFMSG pmsg)
  {
	ASSERT(pmsg != NULL);

	// Check the message object magic number

	if (pmsg->magic != FMSG_MAGIC)
	  return FALSE;

	ASSERT(pmsg->hfile != -1);

	// Check the area object in the debugging build

	ASSERT(pmsg->parea != NULL);
	ASSERT(FtsCheckArea(pmsg->parea));

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of FtsMsg.c
/////////////////////////////////////////////////////////////////////////////
