/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Testing app -- Fts usage simulation
//
// FtsStress.c
//
// Copyright (c) 1999 by Pete Kvitek
//
// Created: 23/Sep/1999
// Updated: 23/Sep/1999
//
// History: 23/Sep/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include standard headers

#include <time.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

  // Include common and private headers

#include "FmaApi.h"
#include "FtsApi.h"
#include "FtsApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Get random number in the specified interval

#define RANDOM(num) (UINT)((((ULONG)(UINT)rand())*(num))/((ULONG)RAND_MAX+1))

  // Message info structure

  typedef struct _MSGI {
	FMH     fmh;                      // Message header
	PCH     pchCtrl;                  // Message header subfields
	ULONG   cchCtrl;                  // Message header subfields length
	PCH     pchText;                  // Message text
	ULONG   cchText;                  // Message text length
  } MSGI, FAR * PMSGI;

  // Check for operation macro

#define CHECKOP(ch) (strchr(achOps, TOLOWER(ch)) != NULL)

  // Command line info

  static CHAR  achBase[MAX_PATH];     // Message base path
  static ULONG nRunCount;             // Max run count
  static ULONG nRunTime;              // Max run time in minutes
  static ULONG nRunDelay;             // Delay between run in seconds
  static ULONG nRepeat = 1;           // Count of operations per run
  static ULONG xUpdate = 75;          // Percentage of base updates
  static ULONG xLock = 25;            // Percentage of base locks
  static ULONG xMemFail = 25;         // Percentage of memory allocation failures
  static ULONG xDskFail = 25;         // Percentage of disk space allocation failures
  static ULONG nCheck = 0;            // Integrity check frequency
  static BOOL  bVerbose;              // Verbose mode
  static CHAR  achOps[] = "rucd";     // Operations to perform

  // Statistics

  static ULONG nRun;                  // Current run number
  static ULONG nOpenArea;             // Number of area opens
  static ULONG nLockArea;             // Number of area locks
  static ULONG nReadMsg;              // Number of msg read
  static ULONG nUpdateMsg;            // Number of msg updated
  static ULONG nCreateMsg;            // Number of msg created
  static ULONG nDeleteMsg;            // Number of msg deleted
  static ULONG nMemFail;              // Number of memory allocation failures
  static ULONG nDskFail;              // Number of disk space allocation failures

  // Allocation failure controls

  static BOOL  bMemFail;              // TRUE if ok to simulate memory allocation failure
  static BOOL  bDskFail;              // TRUE if ok to simulate disk space allocation failure

  // External references

  extern ULONG cMemAlloc;             // Debugging alloc count in FmaSys.c

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine displays usage help
//

  static VOID DoShowHelp(VOID)
  {
	printf( "\n"
			"Fts message base stress test utility built "__DATE__", "__TIME__"\n"
			"Written by Pete Kvitek of JV Dialogue 1st BBS, 2:5020/6@fidonet\n"
			"\n"
			"Usage: FtsStress <area> [options]\n"
			"\n"
			"       /C[nn] - maximum run count\n"
			"       /T[nn] - maximum run time in minutes\n"
			"       /D[nn] - delay between runs in seconds, default is %ld\n"
			"       /R[nn] - Repeat operations per run, default is %ld\n"
			"       /U[nn] - percentage of base updates, default is %ld%%\n"
			"       /L[nn] - percentage of base locks, default is %ld%%\n"
			"       /M[nn] - percentage of memory allocation failures, default is %ld%%\n"
			"       /S[nn] - percentage of disk space allocation failures, default is %ld%%\n"
			"       /I[nn] - check base integrity every nn runs, default is %ld\n"
			"       /V     - verbose mode\n"
			"       /G[file|*|#] - log to file, * to debugger, # to console\n"
			"       /O[rucd]     - operations to perform, default is all\n"
			"\n",
			nRunDelay,
			nRepeat,
			xUpdate,
			xLock,
			xMemFail,
			xDskFail,
			nCheck
		  );
  }

/////////////////////////////////////////////////////////////////////////////
// This routine displays usage help
//

  static BOOL DoProcCmdLine(int argc, char * argv[])
  {
	int iarg;

	// Check if there are no command line options

	if (argc <= 1) {
	  DoShowHelp();
	  return FALSE;
	}

	// Scan the command line 

	for (iarg = 1; iarg < argc; iarg++) {
	  if (argv[iarg][0] == '/' || argv[iarg][0] == '-') {

		// Process the options

		switch (TOLOWER(argv[iarg][1])) {

		  case 'c': // Run count
			nRunCount = atol(&argv[iarg][2]);
			break;

		  case 't': // Run time in minutes
			nRunTime = atol(&argv[iarg][2]);
			break;

		  case 'd': // Run delay in seconds
			nRunDelay = atol(&argv[iarg][2]);
			break;

		  case 'r': // Repeat operations per run
			nRepeat = atol(&argv[iarg][2]);
			break;

		  case 'u': // Percentage of base updates
			xUpdate = atol(&argv[iarg][2]);
			if (xUpdate > 100) xUpdate = 100;
			break;

		  case 'l': // Percentage of base locks
			xLock = atol(&argv[iarg][2]);
			if (xLock > 100) xLock = 100;
			break;

		  case 'm': // Percentage of memory allocation failures
			xMemFail = atol(&argv[iarg][2]);
			if (xMemFail > 100) xMemFail = 100;
			break;

		  case 's': // Percentage of disk space allocation failures
			xDskFail = atol(&argv[iarg][2]);
			if (xDskFail > 100) xDskFail = 100;
			break;

		  case 'i': // Integrity check frequency
			nCheck = atol(&argv[iarg][2]);
			break;

		  case 'g': // Log options
			{ CHAR ach[MAX_PATH];
			  Format(ach, lengof(ach), "FMA-LOG=%s", &argv[iarg][2]);
			  putenv(ach);
			  Log("Logging to: %s", &argv[iarg][2]);
			}
			break;

		  case 'o': // Operations to perform
			strncpy(achOps, &argv[iarg][2], lengof(achOps));
			strlwr(achOps);
			break;

		  case 'v': // Verbose mode
			bVerbose = TRUE;
			break;

		  default:
			fprintf(stderr, "Invalid option: %s\n", argv[iarg]);
			return FALSE;

		}

	  } else {

		// Process message base specification

		if (!achBase[0]) {
		  strncpy(achBase, argv[iarg], lengof(achBase));
		} else {
		  fprintf(stderr, "Duplicate message base specification: %s\n", argv[iarg]);
		  return FALSE;
		}
	  }
	}

	// Check if there is no message base specified and assume default one

	if (!achBase[0]) strcpy(achBase, "_Test");

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns FMA error code string 
//

  static PSZ DoGetErrText(VOID)
  {
	static CHAR ach[128];
	UINT err = FmaGetLastError();

	#define FMAERRCASE(err) case FMAERROR_##err: return #err;

	// Return the known error string

	switch (err) {
	  FMAERRCASE(NONE);
	  FMAERRCASE(INTERNAL_ERROR);
	  FMAERRCASE(NOT_SUPPORTED);
	  FMAERRCASE(INVALID_PARAM);
	  FMAERRCASE(INVALID_DATA);
	  FMAERRCASE(INVALID_SIZE);
	  FMAERRCASE(NOT_ENOUGH_MEMORY);
	  FMAERRCASE(NOT_FOUND);
	  FMAERRCASE(ALREADY_EXISTS);
	  FMAERRCASE(STILL_OPEN);
	  FMAERRCASE(CANNOT_CREATE);
	  FMAERRCASE(CANNOT_OPEN);
	  FMAERRCASE(CANNOT_SEEK);
	  FMAERRCASE(CANNOT_READ);
	  FMAERRCASE(CANNOT_WRITE);
	  FMAERRCASE(CANNOT_DELETE);
	  FMAERRCASE(CANNOT_LOCK);
	  FMAERRCASE(CANNOT_UNLOCK);
	  FMAERRCASE(NO_DISK_SPACE);
	  FMAERRCASE(NOT_LOCKED);
	  FMAERRCASE(NOT_EXCLUSIVE);
	  FMAERRCASE(NOT_CACHING);
	  FMAERRCASE(FRAME_IN_USE);
	  FMAERRCASE(INVALID_FRAME);
	  FMAERRCASE(INVALID_INDEX);
	  FMAERRCASE(ACCESS_DENIED);
	  FMAERRCASE(MISSING_FILES);
	  FMAERRCASE(INVALID_ADDRESS);
	  FMAERRCASE(INVALID_KLUDGE);
	  FMAERRCASE(INVALID_USERID);
	  FMAERRCASE(NOT_REGISTERED);
	  FMAERRCASE(DATA_SIZE_LIMIT);
  }

	// Format the unknown error

	Format(ach, lengof(ach), "UNKNOWN ERROR %d", err);

	return ach;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine prints error unless it's result of simulated memory failure
//

  static VOID DoPrintErr(PSZ pszFormat, ...)
  {
	va_list parg;
	va_start(parg, pszFormat);

	// Due to FTS-001 design not found messages are unavoidable
	// when recovering from various failures, so shut these doen

	if (FmaGetLastError() == FMAERROR_NOT_FOUND) return;

	// Check if this is a simulated memory allocation failure

	if (bMemFail && (FmaGetLastError() == FMAERROR_NOT_ENOUGH_MEMORY)) {
	  nMemFail++;
	} else
	if (bDskFail && (FmaGetLastError() == FMAERROR_NO_DISK_SPACE)) {
	  nDskFail++;
	} else {
	  vprintf(pszFormat, parg);
	}

	va_end(parg);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates message info
//

  static PMSGI DoCreateMsgi(VOID)
  {
	BOOL bMemFailPrev = bMemFail;
	PMSGI pmsgi;

	// Disable memory allocation failure simulation

	bMemFail = FALSE;

	// Allocate message info and check if ok

	if ((pmsgi = MemAlloc(sizeof(MSGI), MA_CLEAR)) == NULL) {
	  printf("\nCan't allocate message info\n");
	  bMemFail = bMemFailPrev;
	  return NULL;
	} else
	  bMemFail = bMemFailPrev;

	return pmsgi;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine destroys message info
//

  static VOID DoDestroyMsgi(PMSGI pmsgi)
  {
	// Destroy message info

	if (pmsgi) {
	  if (pmsgi->pchCtrl) MemFree(pmsgi->pchCtrl);
	  if (pmsgi->pchText) MemFree(pmsgi->pchText);
	  MemFree(pmsgi);
	}
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns random message number
//

  static MSGN DoGetRandomMsgn(HAREA harea)
  {
	MSGN msgnFirst = FtsGetFirstMsg(harea);
	MSGN cmsg = FtsGetMsgCount(harea);
	MSGN msgn;

	// Generate random message number

	msgn = msgnFirst + RANDOM(cmsg);

	return msgn;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates random message control info
//

  static BOOL DoCreateRandomCtrl(PMSGI pmsgi)
  {
	BOOL bMemFailPrev = bMemFail;
	ULONG cch;
	PCH pch;

	// Once in a while create no message control info, 
	// normal size message text and a huge one

	if (RANDOM(100) > 15) {

	  if (RANDOM(100) > 10)
		pmsgi->cchCtrl = RANDOM(1024);
	  else
		pmsgi->cchCtrl = RANDOM(16) * 1024;

	} else 
	  return TRUE;

	// Check if we have any control info

	if (pmsgi->cchCtrl) {

	  // Disable memory allocation failure simulation

	  bMemFail = FALSE;

	  // Allocate message control info and check if ok

	  if ((pmsgi->pchCtrl = MemAlloc((int)pmsgi->cchCtrl, 0)) == NULL) {
		printf("\nCan't allocate message control info, cb=%ld\n", pmsgi->cchCtrl);
		bMemFail = bMemFailPrev;
		return FALSE;
	  } else
		bMemFail = bMemFailPrev;

	  // Generate message text

	  if (pmsgi->cchCtrl) {
		for (pch = pmsgi->pchCtrl, cch = 0; cch < pmsgi->cchCtrl; pch++, cch++) {
		  *pch = 'c';
		}

		pmsgi->pchCtrl[0] = '\r';
		pmsgi->pchCtrl[pmsgi->cchCtrl - 1] = '\r';
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates random message text
//

  static BOOL DoCreateRandomText(PMSGI pmsgi)
  {
	BOOL bMemFailPrev = bMemFail;
	ULONG cch;
	PCH pch;

	// Once in a while create no message text, 
	// normal size message text and a huge one

	if (RANDOM(100) > 15) {

	  if (RANDOM(100) > 10)
		pmsgi->cchText = RANDOM(4) * 1024;
	  else
		pmsgi->cchText = RANDOM(64) * 1024;

	} else 
	  return TRUE;

	// Check if we have any message text

	if (pmsgi->cchText) {

	  // Disable memory allocation failure simulation

	  bMemFail = FALSE;

	  // Allocate message text and check if ok

	  if ((pmsgi->pchText = MemAlloc((int)pmsgi->cchText, 0)) == NULL) {
		printf("\nCan't allocate message text, cb=%ld\n", pmsgi->cchText);
		bMemFail = bMemFailPrev;
		return FALSE;
	  } else
		bMemFail = bMemFailPrev;

	  // Generate message text

	  if (pmsgi->cchText) {
		for (pch = pmsgi->pchText, cch = 0; cch < pmsgi->cchText; pch++, cch++) {
		  *pch = 't';
		}

		pmsgi->pchText[0] = '\r';
		pmsgi->pchText[pmsgi->cchText - 1] = '\r';
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates random message
//

  static BOOL DoCreateRandomMsg(PMSGI * ppmsgi)
  {
	PMSGI pmsgi;

	// Create message info structure and check if ok

	if ((pmsgi = DoCreateMsgi()) == NULL) {
Fail: DoDestroyMsgi(pmsgi); *ppmsgi = NULL;
	  return FALSE;
	}

	// Create message header

	memset(&pmsgi->fmh, 0, sizeof(pmsgi->fmh));
	strcpy(pmsgi->fmh.achBy, "Fts stress simulator");
	strcpy(pmsgi->fmh.achTo, "All");

	// Create random message subfields and text

	if (!DoCreateRandomCtrl(pmsgi)) goto Fail;
	if (!DoCreateRandomText(pmsgi)) goto Fail;

	// Store message info pointer for caller

	*ppmsgi = pmsgi;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine updates random message control info
//

  static BOOL DoUpdateRandomCtrl(PMSGI pmsgi)
  {             
	BOOL bMemFailPrev = bMemFail;
	ULONG r = RANDOM(100);
	ULONG cch;
	PCH pch;

	// Once in a while don't touch message control info

	if (r < 20)
	  return TRUE;

	// Once in a while delete message control info

	if (r > 80) {
	  if (pmsgi->pchCtrl) MemFree(pmsgi->pchCtrl);
	  pmsgi->pchCtrl = NULL;
	  pmsgi->cchCtrl = 0;
	  return TRUE;
	}

	// Randomize message control info size

	pmsgi->cchCtrl = RANDOM(pmsgi->cchCtrl);

	// Check if we've got no message control info and free its buffer,
	// otherwise reallocate message control info

	if (!pmsgi->cchCtrl) {
	  if (pmsgi->pchCtrl) MemFree(pmsgi->pchCtrl);
	  pmsgi->pchCtrl = NULL;
	} else {

	  // Disable memory allocation failure simulation

	  bMemFail = FALSE;

	  // Reallocate memory to hold message control info

	  if ((pch = MemRealloc(pmsgi->pchCtrl, (int)pmsgi->cchCtrl)) == NULL) {
		printf("\nCan't reallocate message control info, cb=%ld\n", pmsgi->cchCtrl);
		bMemFail = bMemFailPrev;
		return FALSE;
	  } else {
		bMemFail = bMemFailPrev;
		pch = pmsgi->pchCtrl;
	  }
	}

	// Generate message control info

	if (pmsgi->pchCtrl) {
	  for (pch = pmsgi->pchCtrl, cch = 0; cch < pmsgi->cchCtrl; pch++, cch++) {
		*pch = 'C';
	  }

	  pmsgi->pchCtrl[0] = '\r';
	  pmsgi->pchCtrl[pmsgi->cchCtrl - 1] = '\r';
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine updates random message text
//

  static BOOL DoUpdateRandomText(PMSGI pmsgi)
  {             
	BOOL bMemFailPrev = bMemFail;
	ULONG r = RANDOM(100);
	ULONG cch;
	PCH pch;

	// Once in a while don't touch message text

	if (r < 20)
	  return TRUE;

	// Once in a while delete message text

	if (r > 80) {
	  if (pmsgi->pchText) MemFree(pmsgi->pchText);
	  pmsgi->pchText = NULL;
	  pmsgi->cchText = 0;
	  return TRUE;
	}

	// Randomize message text size

	pmsgi->cchText = RANDOM(pmsgi->cchText);

	// Check if we've got no message text and free its buffer,
	// otherwise reallocate message text

	if (!pmsgi->cchText) {
	  if (pmsgi->pchText) MemFree(pmsgi->pchText);
	  pmsgi->pchText = NULL;
	} else {

	  // Disable memory allocation failure simulation

	  bMemFail = FALSE;

	  // Reallocate memory to hold message text

	  if ((pch = MemRealloc(pmsgi->pchText, (int)pmsgi->cchText)) == NULL) {
		printf("\nCan't reallocate message text, cb=%ld\n", pmsgi->cchText);
		bMemFail = bMemFailPrev;
		return FALSE;
	  } else {
		bMemFail = bMemFailPrev;
		pch = pmsgi->pchText;
	  }
	}

	// Generate message text

	if (pmsgi->pchText) {
	  for (pch = pmsgi->pchText, cch = 0; cch < pmsgi->cchText; pch++, cch++) {
		*pch = 'T';
	  }

	  pmsgi->pchText[0] = '\r';
	  pmsgi->pchText[pmsgi->cchText - 1] = '\r';
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine updates random message
//

  static BOOL DoUpdateRandomMsg(PMSGI pmsgi)
  {
	// Update message header

	FmaGetLocalDttm(&pmsgi->fmh.dttmWritten);

	// Create random message subfields and text

	DoUpdateRandomCtrl(pmsgi);
	DoUpdateRandomText(pmsgi);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads message
//

  static BOOL DoLoadMsg(HMSG hmsg, MSGN msgn, PMSGI * ppmsgi)
  {
	BOOL bMemFailPrev = bMemFail;
	ULONG cchText, offText = 0;
	CHAR achText[512];
	PMSGI pmsgi;
	PCH pchText;

	// Create message info structure and check if ok

	if ((pmsgi = DoCreateMsgi()) == NULL) {
Fail: DoDestroyMsgi(pmsgi); *ppmsgi = NULL;
	  return FALSE;
	}

	// Randomize entire and by chunks message reading

	if (RANDOM(100) < 80) {

	  // Read entire message and check if ok

	  pmsgi->cchCtrl = FRM_GETCTRLCOPY;
	  pmsgi->cchText = FRM_GETALLTEXT;
	  if (!FtsReadMsg(hmsg, &pmsgi->fmh, &pmsgi->pchCtrl, &pmsgi->cchCtrl,
					  &pmsgi->pchText, &pmsgi->cchText, 0)) {
		DoPrintErr("\nFtsReadMsg(%ld) failed, error=%s\n", msgn, DoGetErrText());
		goto Fail;
	  }
	} else {

	  // Get message text length and check if ok

	  if (!FtsGetMsgTextLen(hmsg, &pmsgi->cchText)) {
		DoPrintErr("\nFtsGetMsgTextLen(%ld) failed, error=%s\n", msgn, DoGetErrText());
		goto Fail;
	  }

	  // Allocate memory to hold message text and check if ok

	  if (pmsgi->cchText) {

		// Disable memory allocation failure simulation

		bMemFail = FALSE;

		if ((pmsgi->pchText = MemAlloc((int)pmsgi->cchText, 0)) == NULL) {
		  printf("\nCan't allocate message text, cb=%ld\n", pmsgi->cchText);
		  bMemFail = bMemFailPrev;
		  goto Fail;
		} else
		  bMemFail = bMemFailPrev;
	  }

	  // Read message header and check if ok

	  pmsgi->cchCtrl = FRM_GETCTRLCOPY;
	  if (!FtsReadMsg(hmsg, &pmsgi->fmh, &pmsgi->pchCtrl, &pmsgi->cchCtrl, NULL, 0, 0)) {
		DoPrintErr("\nFtsReadMsg(%ld) failed, error=%s\n", msgn, DoGetErrText());
		goto Fail;
	  }

	  // Read message text in chunks

	  if (pmsgi->pchText) {
		loop {

		  // Read a chunk of text into stack buffer

		  pchText = achText;
		  cchText = sizeof(achText);
		  if (!FtsReadMsg(hmsg, 0, 0, 0, &pchText, &cchText, &offText)) {
			DoPrintErr("\nFtsReadMsg(%ld) failed, error=%s\n", msgn, DoGetErrText());
			goto Fail;
		  }

		  // Check if read something and copy a chunk of text into allocated buffer,
		  // othrewise leave loop

		  if (cchText) 
			memcpy(pmsgi->pchText + offText - cchText, achText, (int)cchText);
		  else
			break;
		}
	  }
	}

	// Store message info pointer for caller

	*ppmsgi = pmsgi;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message
//

  static BOOL DoSaveMsg(HMSG hmsg, MSGN msgn, PMSGI pmsgi)
  {
	ULONG cchText, offText = 0;
	PCH pchText, pchCtrl;

	// When updating existing message in a OPENMSG_READWRITE mode and
	// the new message has no ctrl or text, it's important to pass 
	// non null pointer so that ctrl or text gets rewritten

	pchCtrl = pmsgi->cchCtrl ? pmsgi->pchCtrl : (PCH) -1;
	pchText = pmsgi->cchText ? pmsgi->pchText : (PCH) -1;

	// Randomize entire and by chunks message writing

	if (RANDOM(100) < 50) {

	  // Write entire message and check if ok

	  if (!FtsWriteMsg(hmsg, &pmsgi->fmh, pchCtrl, pmsgi->cchCtrl,
					   pchText, pmsgi->cchText, 0, pmsgi->cchText)) {
		DoPrintErr("\nFtsWriteMsg(%ld) failed, error=%s\n", msgn, DoGetErrText());
		return FALSE;
	  }

	} else {

	  // Write message header and check if ok

	  if (!FtsWriteMsg(hmsg, &pmsgi->fmh, pchCtrl, pmsgi->cchCtrl,
					   NULL, 0, 0, pmsgi->cchText)) {
		DoPrintErr("\nFtsWriteMsg(%ld) failed, error=%s\n", msgn, DoGetErrText());
		return FALSE;
	  }

	  // Write message text in chunks

	  if (pmsgi->pchText) {

		loop {

		  // Get size of the chunk to write and check if ok

		  if ((cchText = min(256, pmsgi->cchText - offText)) == 0)
			break;

		  // Write chunk of text and check if ok

		  if (!FtsWriteMsg(hmsg, 0, 0, 0, pmsgi->pchText + offText, cchText, &offText, 0)) {
			DoPrintErr("\nFtsWriteMsg(%ld) failed, error=%s\n", msgn, DoGetErrText());
			return FALSE;
		  }
		}
	  } else {

		// Zero out the mesage text and check if ok

		if (!FtsWriteMsg(hmsg, 0, 0, 0, pchText, 0, &offText, 0)) {
		  DoPrintErr("\nFtsWriteMsg(%ld) failed, error=%s\n", msgn, DoGetErrText());
		  return FALSE;
		}
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine reads a message
//

  static BOOL DoReadMsg(HAREA harea)
  {
	UINT mode = OPENMSG_READ;
	PMSGI pmsgi;
	MSGN msgn;
	HMSG hmsg;
	BOOL bOk;

	// Check if we're doing reads

	if (!CHECKOP('r')) return TRUE;

	// Check if there are no messages and leave

	if (!FtsGetMsgCount(harea)) return TRUE;

	// Get random message number

	msgn = DoGetRandomMsgn(harea);

	// Inform user if in verbose mode

	if (bVerbose) printf("\nReading msgn=%ld\n", msgn);

	// Open message and check if ok

	if ((hmsg = FtsOpenMsg(harea, mode, msgn, 0)) == NULL) {
	  DoPrintErr("\nFtsOpenMsg(msgn=%ld mode=%d) failed, error=%s\n", msgn, mode, DoGetErrText());
	  return FALSE;
	} 

	// Load message info

	bOk = DoLoadMsg(hmsg, msgn, &pmsgi);

	// Destroy message info

	DoDestroyMsgi(pmsgi);

	// Close message and check if ok

	if (!FtsCloseMsg(hmsg)) {
	  DoPrintErr("\nFtsCloseMsg(msgn=%ld mode=%d) failed, error=%s\n", msgn, mode, DoGetErrText());
	  return FALSE;
	}

	// Increment read message count

	if (bOk) nReadMsg++;

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates a message
//

  static BOOL DoCreateMsg(HAREA harea)
  {
	UINT mode = OPENMSG_CREATE;
	BOOL bOk = TRUE;
	PMSGI pmsgi;
	MSGN msgn;
	HMSG hmsg;

	// Check if we're doing creates

	if (!CHECKOP('c')) return TRUE;

	// Randomize message number creating fresh new message or
	// new message on top of the old one

	if (RANDOM(100) > 50 || !FtsGetMsgCount(harea)) {
	  msgn = 0;
	} else {
	  msgn = DoGetRandomMsgn(harea);
	}

	// Inform user if in verbose mode

	if (bVerbose) printf("\nCreating msgn=%ld\n", msgn);

	// Open message and check if ok

	if ((hmsg = FtsOpenMsg(harea, mode, msgn, 0)) == NULL) {
	  DoPrintErr("\nFtsOpenMsg(msgn=%ld mode=%d) failed, error=%s\n", msgn, mode, DoGetErrText());
	  return FALSE;
	} 

	// Create random message

	if (DoCreateRandomMsg(&pmsgi)) {

	  // Save message

	  if (!DoSaveMsg(hmsg, msgn, pmsgi))
		bOk = FALSE;

	  // Destroy message info

	  DoDestroyMsgi(pmsgi);
	} else
	  bOk = FALSE;

	// Close message and check if ok

	if (!FtsCloseMsg(hmsg)) {
	  DoPrintErr("\nFtsCloseMsg(msgn=%ld mode=%d) failed, error=%s\n", msgn, mode, DoGetErrText());
	  return FALSE;
	}

	// Increment created message count

	if (bOk) nCreateMsg++;

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine updates a message
//

  static BOOL DoUpdateMsg(HAREA harea)
  {
	UINT mode = OPENMSG_READWRITE;
	BOOL bOk = TRUE;
	PMSGI pmsgi;
	MSGN msgn;
	HMSG hmsg;

	// Check if we're doing updates

	if (!CHECKOP('u')) return TRUE;

	// Chech if there are no messages and leave

	if (!FtsGetMsgCount(harea)) return TRUE;

	// Get random message number

	msgn = DoGetRandomMsgn(harea);

	// Inform user if in verbose mode

	if (bVerbose) printf("\nUpdating msgn=%ld\n", msgn);

	// Open message and check if ok

	if ((hmsg = FtsOpenMsg(harea, mode, msgn, 0)) == NULL) {
	  DoPrintErr("\nFtsOpenMsg(msgn=%ld mode=%d) failed, error=%s\n", msgn, mode, DoGetErrText());
	  return FALSE;
	} 

	// Load message info

	bOk = DoLoadMsg(hmsg, msgn, &pmsgi);

	// Update message info and check if ok

	if (bOk) {
	  if (!DoUpdateRandomMsg(pmsgi)) {
		bOk = FALSE;
	  } else 
	  if (!DoSaveMsg(hmsg, msgn, pmsgi)) {
		bOk = FALSE;
	  }
	}

	// Destroy message info

	DoDestroyMsgi(pmsgi);

	// Close message and check if ok

	if (!FtsCloseMsg(hmsg)) {
	  DoPrintErr("\nFtsCloseMsg(msgn=%ld mode=%d) failed, error=%s\n", msgn, mode, DoGetErrText());
	  return FALSE;
	}

	// Increment updated message count

	if (bOk) nUpdateMsg++;

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes random message
//

  static BOOL DoDeleteMsg(HAREA harea)
  {
	MSGN msgn;

	// Check if we're doing deletes

	if (!CHECKOP('d')) return TRUE;

	// Chech if there are no messages and leave

	if (!FtsGetMsgCount(harea)) return TRUE;

	// Get random message number

	msgn = DoGetRandomMsgn(harea);

	// Inform user if in verbose mode

	if (bVerbose) printf("\nDeleting msgn=%ld\n", msgn);

	// Delete message and check if ok

	if (!FtsDeleteMsg(harea, msgn)) {
	  DoPrintErr("\nFtsDeleteMsg(msgn=%ld) failed, error=%s\n", msgn, DoGetErrText());
	  return FALSE;
	} 

	// Increment deleted message count

	nDeleteMsg++;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine performs a simulation run
//

  static BOOL DoRun(VOID)
  {
	static ULONG cdel;
	static BOOL bInUse = FALSE;
	BOOL bLock, bUpdate, bReadBefore;
	ULONG cmsg, fl = 0;
	HAREA harea;
	UINT n;

	// Check if we will lock and/or update area on this run

	bLock = (RANDOM(100) < xLock);
	bUpdate = (RANDOM(100) < xUpdate);

	// Open area and if failed, don't terminate simulation if 
	// this is because other process updating the base

	if (bLock) fl|= FOA_LOCKAREA;
	if ((harea = FtsOpenArea(achBase, OPENAREA_ALWAYS, fl, NULL)) == NULL) {
	  if (FmaGetLastError() == FMAERROR_ACCESS_DENIED ||
		  FmaGetLastError() == FMAERROR_CANNOT_LOCK) {
		printf("Base %s is in use (%ld)\t\r", achBase, ++bInUse);
		return TRUE;
	  }
	  return FALSE;
	} else {
	  nOpenArea++;
	  if (bLock) nLockArea++;
	  bInUse = FALSE;
	}

	// Increment and display run number and current message count

	printf("Run %ld, cmsg=%ld, cdel=%ld, nMemFail=%ld, nDskFail=%ld\t\r", ++nRun, FtsGetMsgCount(harea), cdel, nMemFail, nDskFail);

	// Do random message reads on every run before or
	// after message base updates procedure

	bReadBefore = (RANDOM(100) > 50);
	if (bReadBefore)
	  for  (n = RANDOM(1 + nRepeat); n >= 1; n--) {
		DoReadMsg(harea);
	  }

	// Check if we're updating the area on this run

	if (bUpdate) {

	  // Do random message updates/creates

	  if (RANDOM(100) < 30) {
		for  (n = RANDOM(1 + nRepeat); n >= 1; n--) {
		  DoUpdateMsg(harea);
		}
	  } else {
		for  (n = RANDOM(1 + nRepeat); n >= 1; n--) {
		  DoCreateMsg(harea);
		}
	  }

	  // Do random message deletes in a sweeping manner

	  cmsg = FtsGetMsgCount(harea);
	  if (cmsg > 20) {
		cdel = 90;
	  } else
	  if (cmsg < 10) {
		cdel = 10;
	  }

	  // Delete random message

	  if (RANDOM(100) < cdel) {
		for  (n = RANDOM(1 + nRepeat); n >= 1; n--) {
		  DoDeleteMsg(harea);
		}
	  }
	}

	// Check if we did not do read run before updating base and
	// to it after 

	if (!bReadBefore)
	  for  (n = RANDOM(1 + nRepeat); n >= 1; n--) {
		DoReadMsg(harea);
	  }

	// Close area and check if ok

	if (!FtsCloseArea(harea)) {
	  DoPrintErr("\nFtsCloseArea(\"%s\") failed, error=%s\n", achBase, DoGetErrText());
	  return FALSE;
	} 

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// Memory management routines for allocation stress testing
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine allocates specified memory block 
//

  FMAAPI_(PVOID) StressMemAlloc(size_t cb, BOOL fZero)
  {
	PVOID pv;

	if (bMemFail && (RANDOM(100) < xMemFail)) return NULL;

	pv = FmaMemAlloc(cb, fZero);

#ifdef BREAK_MEMADDR
	if (pv == BREAK_MEMADDR) {
	  UINT i = 0;
	}
#endif

	return pv;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine reallocates specified memory block
//

  FMAAPI_(PVOID) StressMemRealloc(PVOID p, size_t cb)
  {
	PVOID pv;

	if (bMemFail && (RANDOM(100) < xMemFail)) return NULL;

	pv = FmaMemRealloc(p, cb);

#ifdef BREAK_MEMADDR
	if (pv == BREAK_MEMADDR) {
	  UINT i = 0;
	}
#endif

	return pv;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine frees specified memory block
//

  FMAAPI_(VOID) StressMemFree(PVOID p)
  {
#ifdef BREAK_MEMADDR
	if (p == BREAK_MEMADDR) {
	  UINT i = 0;
	}
#endif

	FmaMemFree(p);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine is a disk allocation failure callback
//

  FMAAPI_(BOOL) FmaQueryDiskAllocFailure(HFILE hfile, ULONG cb)
  {
	if (bDskFail && (RANDOM(100) < xDskFail)) {
#ifdef FMA_LOG
Log("FmaQueryDiskAllocFailure: FAILING hfile=%d cb=%ld", hfile, cb);
#endif
	  FMAERROR(NO_DISK_SPACE);
	  return TRUE;
	}

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Main...
//

  void main(int argc, char * argv[])
  {
	time_t tm, tmStart, tmEnd;
	int tm_sec, tm_min, tm_hour;
	ULONG iCheck = 0;
	FMAINIT fmaInit;

	// Process commad line

	if (!DoProcCmdLine(argc, argv)) {
	  exit(EXIT_FAILURE);
	}

	// Seed random number generator

	srand((int)time(NULL));

	// Make up initialization structure

	memset(&fmaInit, 0, sizeof(fmaInit));
	fmaInit.pfnMemAlloc   = StressMemAlloc;
	fmaInit.pfnMemRealloc = StressMemRealloc;
	fmaInit.pfnMemFree    = StressMemFree;

	// Initialize and check if ok

	if (!FmaInitialize(&fmaInit)) {
	  printf("FmaInitialize() failed, code=%s\n", DoGetErrText());
	  exit(EXIT_FAILURE);
	}

	// Display simulation parameters

	printf("Simulation base: %s\n", achBase);

	// Setup timing

	tmStart = time(NULL);
	tmEnd = nRunTime ? tmStart + nRunTime * 60L : 0;

	// Flush keyboard input

	while (kbhit()) getch();

	// Perform runs until key pressed

	while (!kbhit()) {

	  // Check message base integrity

	  if (!iCheck) {
		iCheck = nCheck;
		if (!FtsCheckBase(achBase)) {
		  printf("MESSAGE BASE INTEGRITY CHECK FAILED\a\n");
		  break;
		}
	  } else
		iCheck--;

	  // Perform a run and sleep if requested

	  bMemFail = bDskFail = TRUE;
	  DoRun(); if (nRunDelay) FmaSleep(nRunDelay * 1000);
	  bMemFail = bDskFail = FALSE;

	  // Check if reached run count limit if any

	  if (nRunCount && !(nRun < nRunCount))
		break;

	  // Check if reached run time limit if any

	  if (nRunTime && !(time(NULL) < tmEnd))
		break;
	}

	// Flush keyboard input again

	while (kbhit()) getch();

	// Calculate elapsed time

	tm = (time(NULL) - tmStart);
	tm_sec  = (int) (tm % 60); tm /= 60;
	tm_min  = (int) (tm % 60); tm /= 60;
	tm_hour = (int) (tm % 24); tm /= 24;

	// Display statistics

	printf("\n"
		   "Simulation runs: %ld\n"
		   "Simulation time: %02u:%02u:%02u\n"
		   "Memory failures: %ld\n"
		   "DskSpc failures: %ld\n"
		   "\n"
		   "Area opens     : %ld\n"
		   "Area locks     : %ld\n"
		   "Msgs read      : %ld\n"
		   "Msgs updated   : %ld\n"
		   "Msgs created   : %ld\n"
		   "Msgs deleted   : %ld\n"
		   ,
			nRun,
			tm_hour, tm_min, tm_sec,
			nMemFail, nDskFail,
			nOpenArea, nLockArea,
			nReadMsg, nUpdateMsg, nCreateMsg,
			nDeleteMsg
		  );

	// Cleanup and exit

	if (!FmaTerminate()) {
	  printf("FmaTerminate() failed, code=%s\n", DoGetErrText());
	  exit(EXIT_FAILURE);
	}

	// Check allocation count

	if (cMemAlloc) printf("cMemAlloc=%ld\a\n", cMemAlloc);

	exit(EXIT_SUCCESS);
  }

/////////////////////////////////////////////////////////////////////////////
// End of FtsStress.c
/////////////////////////////////////////////////////////////////////////////
