/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Testing app
//
// Test.c
//
// Copyright (c) 1999 by Pete Kvitek
//
// Created: 27/Aug/1999
// Updated: 27/Aug/1999
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include standard headers

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

  // Include common and private headers

#include "FmaApi.h"
#include "FtsApi.h"
#include "FtsApiP.h"
//#include "SquApi.h"
//#include "SquApiP.h"
//#include "JamApi.h"
//#include "JamApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine is a callback for FtsEnumMsgs()
//

  static BOOL FMACALLBACK DoEnumMsg(PENUMMSGINFO pemi, PVOID pvCallbackData)
  {
	ULONG fl = (ULONG) pvCallbackData;

	// Log the enum call

	Log("DoEnumMsg: msgn=%lu umsg=%lu imsg=%u offs=%08lx hash=%08lx pvCallbackData=%08lx",
		 pemi->msgn, pemi->umsg, pemi->imsg, pemi->offs, pemi->hash, 
		 pvCallbackData);

	//ASSERT(FtsValidateMsg(pemi->harea, pemi->msgn, NULL));

	//if (fl & FENUMMSGS_SAFEENUM) FtsDeleteMsg(pemi->harea, pemi->msgn);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message opening
//

  static BOOL FMASUB DoOpenMsg(HAREA harea)
  {
	HMSG hmsg;
	MSGN msgn;

	// Display what these return

	printf("FtsGetFirstMsg()=%d\n", FtsGetFirstMsg(harea));
	printf("FtsGetHighMsg()=%d\n", FtsGetHighMsg(harea));
	printf("FtsGetMsgCount()=%d\n", FtsGetMsgCount(harea));
	printf("FtsGetCurMsg()=%d\n", FtsGetCurMsg(harea));

	// Open the current message (this should fail after open)

	msgn = MSGN_CUR;
	if ((hmsg = FtsOpenMsg(harea, OPENMSG_READ, msgn, 0)) == NULL) {
	  printf("FtsOpenMsg() failed, code=%d\n", FmaGetLastError());
	} else
	  FtsCloseMsg(hmsg);

	// Loop reading next msgs

	msgn = MSGN_NEXT;
	while ((hmsg = FtsOpenMsg(harea, OPENMSG_READ, msgn, 0)) != NULL) {
	  printf("Opened msgn=%d\n", FtsGetCurMsg(harea));
	  FtsCloseMsg(hmsg);
	}

	// Loop reading prev msgs

	msgn = MSGN_PREV;
	while ((hmsg = FtsOpenMsg(harea, OPENMSG_READ, msgn, 0)) != NULL) {
	  printf("Opened msgn=%d\n", FtsGetCurMsg(harea));
	  FtsCloseMsg(hmsg);
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message reading
//

  static BOOL FMASUB DoReadMsg(HAREA harea, MSGN msgn)
  {
	HMSG hmsg;
	BOOL bOk;
	FMH fmh;

	// Open message and check if ok

	if ((hmsg = FtsOpenMsg(harea, OPENMSG_READ, msgn, 0)) == NULL) {
	  printf("FtsOpenMsg() failed, code=%d\n", FmaGetLastError());
	  return FALSE;
	}

	// Get message control info and text length
#if 1
	{ ULONG cchCtrl, cchText;

	  if (!FtsGetMsgCtrlLen(hmsg, &cchCtrl)) {
		printf("FtsGetMsgCtrlLen() failed, code=%d\n", FmaGetLastError());
	  }	else
		printf("FtsGetMsgCtrlLen()=%d\n", cchCtrl);

	  if (!FtsGetMsgTextLen(hmsg, &cchText)) {
		printf("FtsGetMsgTextLen() failed, code=%d\n", FmaGetLastError());
	  } else
		printf("FtsGetMsgTextLen()=%d\n", cchText);
	}
#endif

	// Read entire message into allocated buffer
#if 1
	{ PCH pchCtrl = NULL; ULONG cchCtrl = FRM_GETCTRLCOPY;
	  PCH pchText = NULL; ULONG cchText = FRM_GETALLTEXT;

	  bOk = FtsReadMsg(hmsg, &fmh, &pchCtrl, &cchCtrl, &pchText, &cchText, 0);

	  printf("FtsReadMsg() cchCtrl=%d cchText=%d\n", cchCtrl, cchText);

	  if (pchCtrl) MemFree(pchCtrl);
	  if (pchText) MemFree(pchText);
	}
#endif

	// Read message by chunks into allocated buffer
#if 0
	{ PCH pchCtrl = NULL; ULONG cchCtrl = FRM_GETCTRLPTR;
	  PCH pchText = NULL; ULONG cchText = 32; ULONG offText = 0;

	  bOk = FtsReadMsg(hmsg, &fmh, &pchCtrl, &cchCtrl, 0, 0, 0);

	  while (bOk) {
		bOk = FtsReadMsg(hmsg, 0, 0, 0, &pchText, &cchText, &offText);
		if (!cchText) break;
	  }

	  if (pchText) MemFree(pchText);
	}
#endif        

	// Read message by chunks into buffer on stack
#if 0
	{ PCH pchCtrl = NULL; ULONG cchCtrl = FRM_GETCTRLPTR; CHAR achText[32];
	  PCH pchText = achText; ULONG cchText = sizeof(achText); ULONG offText = 0;

	  bOk = FtsReadMsg(hmsg, &fmh, &pchCtrl, &cchCtrl, 0, 0, 0);

	  while (bOk) {
		bOk = FtsReadMsg(hmsg, 0, 0, 0, &pchText, &cchText, &offText);
		if (!cchText) break;
	  }
	}
#endif        

	// Close message 

	if (!FtsCloseMsg(hmsg)) {
	  printf("FtsCloseMsg() failed, code=%d\n", FmaGetLastError());
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message creation and writing
//

  static BOOL FMASUB DoCreateMsg(HAREA harea, MSGN msgn)
  {
	CHAR achCtrl[] = "\x01MSGID: 2:5020/6 12345678\r" "\x01PID: FMA\r";
//  CHAR achText[] = "\rThis is a message text\r";
	CHAR achText[] = "\rThis is a message text\r\r-Pete";
	HMSG hmsg;
	FMH fmh;

	// Open message and check if ok

	if ((hmsg = FtsOpenMsg(harea, OPENMSG_CREATE, msgn, 0)) == NULL) {
	  printf("FtsOpenMsg() failed, code=%d\n", FmaGetLastError());
	  return FALSE;
	} else
	  printf("FtsGetCurMsg()=%d\n", FtsGetCurMsg(harea));

	// Create message header

	memset(&fmh, 0, sizeof(fmh));
	strcpy(fmh.achBy, "From name");
	strcpy(fmh.achTo, "To name");
	Format(fmh.achRe, lengof(fmh.achRe), "Msg %d", FtsGetCurMsg(harea));
	fmh.attr = FMA_LOCL | FMA_READ;

	// Write message

	if (!FtsWriteMsg(hmsg, &fmh, achCtrl, lengof(achCtrl), 
					 achText, lengof(achText), 0, lengof(achText))) {
	  printf("FtsWriteMsg() failed, code=%d\n", FmaGetLastError());
	}

	// Close message 

	if (!FtsCloseMsg(hmsg)) {
	  printf("FtsCloseMsg() failed, code=%d\n", FmaGetLastError());
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message updating and writing
//

  static BOOL FMASUB DoUpdateMsg(HAREA harea, MSGN msgn)
  {
	PCH pchCtrl = NULL; ULONG cchCtrl = 0;
	CHAR achText[] = "\rThis is a message text\r";
	PCH pchText = NULL; ULONG cchText = FRM_GETALLTEXT;
	HMSG hmsg;
	FMH fmh;

	// Open message and check if ok

	if ((hmsg = FtsOpenMsg(harea, OPENMSG_READWRITE, msgn, 0)) == NULL) {
	  printf("FtsOpenMsg() failed, code=%d\n", FmaGetLastError());
	  return FALSE;
	} else
	  printf("FtsGetCurMsg()=%d\n", FtsGetCurMsg(harea));

	// Read message

	FtsReadMsg(hmsg, &fmh, &pchCtrl, &cchCtrl, &pchText, &cchText, 0);

	// Change message header

	// Change control info

	// Write message
#if 1
	if (!FtsWriteMsg(hmsg, &fmh, pchCtrl, cchCtrl, achText, lengof(achText), 0, lengof(achText))) {
	  printf("FtsWriteMsg() failed, code=%d\n", FmaGetLastError());
	}
#else
	if (!FtsWriteMsg(hmsg, &fmh, pchCtrl, cchCtrl, achText, 0, 0, 0)) {
	  printf("FtsWriteMsg() failed, code=%d\n", FmaGetLastError());
	}
#endif

	// Free allocated memory

	if (pchCtrl) MemFree(pchCtrl);
	if (pchText) MemFree(pchText);

	// Close message 

	if (!FtsCloseMsg(hmsg)) {
	  printf("FtsCloseMsg() failed, code=%d\n", FmaGetLastError());
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message creation and writing
//

  static BOOL FMASUB DoCheckUid(HAREA harea)
  {
	  // Test msgn to uid conversion
#if 1
	{ UMSG umsg;
	  umsg = FtsMsgnToUid(harea, 0);
	  umsg = FtsMsgnToUid(harea, 1);
	  umsg = FtsMsgnToUid(harea, 2);
	  umsg = FtsMsgnToUid(harea, 3);
	  umsg = FtsMsgnToUid(harea, 4);
	}
#endif

	// Test uid to msgn conversion
#if 1
	{ MSGN msgn; UMSG umsg;
	  msgn = FtsUidToMsgn(harea, 0, UMSG_EXACT);
	  msgn = FtsUidToMsgn(harea, 1, UMSG_EXACT);
	  msgn = FtsUidToMsgn(harea, 2, UMSG_EXACT);
	  msgn = FtsUidToMsgn(harea, 3, UMSG_EXACT);
	  msgn = FtsUidToMsgn(harea, 4, UMSG_EXACT);

	  umsg = 1;
	  msgn = FtsUidToMsgn(harea, umsg,  UMSG_EXACT);
	  msgn = FtsUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_PREV | UMSG_NEXT);
	  msgn = FtsUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_PREV);
	  msgn = FtsUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_NEXT);
	  msgn = FtsUidToMsgn(harea, umsg,  UMSG_PREV);
	  msgn = FtsUidToMsgn(harea, umsg,  UMSG_NEXT);

	  umsg = 2;
	  msgn = FtsUidToMsgn(harea, umsg,  UMSG_EXACT);
	  msgn = FtsUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_PREV | UMSG_NEXT);
	  msgn = FtsUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_PREV);
	  msgn = FtsUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_NEXT);
	  msgn = FtsUidToMsgn(harea, umsg,  UMSG_PREV);
	  msgn = FtsUidToMsgn(harea, umsg,  UMSG_NEXT);
	}
#endif

	 return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Main...
//

  void main(int argc, char * argv[])
  {
	HAREA harea = NULL;
	HMSG hmsg = NULL;
	PSZ pszArea;
	LONG cb = 0;

	// Force debugging log

	if (!getenv("FMA-LOG")) putenv("FMA-LOG=*");

	// Initialize and check if ok

	if (!FmaInitialize(NULL)) {
	  printf("FmaInitialize() failed, code=%u\n", FmaGetLastError());
	  exit(EXIT_FAILURE);
	}

	// Open area and check if ok

	pszArea = "..\\m\\test";
	if ((harea = FtsOpenArea(pszArea, OPENAREA_ALWAYS, 0, NULL)) == NULL) {
	  printf("FtsOpenArea() failed, code=%d\n", FmaGetLastError());
	} else {
	  BOOL bLock = 1;

	  // Lock area

	  if (bLock) bLock = FtsLockArea(harea, 1);

	  // Miscellaneous tests

//	  DoOpenMsg(harea);
	  DoReadMsg(harea, 9);
//	  DoCreateMsg(harea, 0);
//	  DoUpdateMsg(harea, FtsGetHighMsg(harea));
//	  DoCheckUid(harea);

	  // Enumerate area messages
#if 0
	  { ULONG fl = 0;
//      fl|= FENUMMSGS_REVERSE;
//      fl|= FENUMMSGS_SAFEENUM;
		FtsEnumMsgs(harea, DoEnumMsg, (PVOID)fl, fl);
	  }
#endif

#if 0
	  // Delete area messages

	  { BOOL bOk;
		bOk = FtsDeleteMsg(harea, 5);
		bOk = FtsDeleteMsg(harea, 6);
	  }
#endif

#if 0
	  // Lastread info test

	  { BOOL bOk;
		LRINFO lri;

		((PFAREA)harea)->fl|= FOA_HIGHREAD;

		lri.umsgLast = 1;
		lri.umsgHigh = 2;

		bOk = FtsSetLRInfo(harea, 2, &lri, NULL);
		bOk = FtsGetLRInfo(harea, 2, &lri);
	  }
#endif

	 // Unlock area

	  if (bLock) FtsLockArea(harea, 0);

	  // Close area

	  FtsCloseArea(harea);
	}

	// Cleanup and exit

	FmaTerminate();

	exit(EXIT_SUCCESS);
  }

/////////////////////////////////////////////////////////////////////////////
// End of Test.c
/////////////////////////////////////////////////////////////////////////////
