/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: JAM api routines module
//
// JamApi.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 25/Sep/1999
// Updated: 29/Jan/2000
//
// History: 25/Sep/1999 kvitekp - Initially created.
//
//          29/Jan/2000 kvitekp - Fixed space between ^aAREA: and area tag
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"
#include "FmaCtrl.h"

#include "JamApi.h"
#include "JamApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Jam message base api

  static API api = {
	sizeof(API),
	JamOpenArea,
	JamCloseArea,
	JamLockArea,
	JamValidateMsg,
	JamCreateMsgMap,
	JamEnumMsgs,
	JamDeleteMsg,
	JamSetCurMsg,
	JamGetCurMsg,
	JamGetFirstMsg,
	JamGetHighMsg,
	JamGetMsgCount,
	JamMsgnToUid,
	JamUidToMsgn,
	JamOpenMsg,
	JamCloseMsg,
	JamLoadMsg,
	JamSaveMsg,
XPV JamReadMsg,
XPV JamWriteMsg,
	JamGetMsgCtrlLen,
	JamGetMsgTextLen,
	JamGetLRInfo,
	JamSetLRInfo,
  };

  // JAM message attribute mapping table. Note: These are 
  // attributes and ^aFLAGS flags we generate when saving message.

  static ATTRMAP aAttr[] = {
	ATTR_(MA_LOCAL,             JMA_LOCAL,              NULL)
	ATTR_(MA_INTRANSIT,         JMA_INTRANSIT,          NULL)
	ATTR_(MA_PRIVATE,           JMA_PRIVATE,            NULL)
	ATTR_(MA_READ,              JMA_READ,               NULL)
	ATTR_(MA_SENT,              JMA_SENT,               NULL)
	ATTR_(MA_KILLSENT,          JMA_KILLSENT,           NULL)
	ATTR_(MA_ARCHIVESENT,       JMA_ARCHIVESENT,        NULL)
	ATTR_(MA_HOLD,              JMA_HOLD,               NULL)
	ATTR_(MA_CRASH,             JMA_CRASH,              NULL)
	ATTR_(MA_IMMEDIATE,         JMA_IMMEDIATE,          NULL)
	ATTR_(MA_DIRECT,            JMA_DIRECT,             NULL)
	ATTR_(MA_GATE,              JMA_GATE,               NULL)
	ATTR_(MA_FILEREQUEST,       JMA_FILEREQUEST,        NULL)
	ATTR_(MA_FILEATTACH,        JMA_FILEATTACH,         NULL)
	ATTR_(MA_TRUNCFILE,         JMA_TRUNCFILE,          NULL)
	ATTR_(MA_KILLFILE,          JMA_KILLFILE,           NULL)
	ATTR_(MA_RECEIPTREQ,        JMA_RECEIPTREQ,         NULL)
	ATTR_(MA_CONFIRMREQ,        JMA_CONFIRMREQ,         NULL)
	ATTR_(MA_ORPHAN,            JMA_ORPHAN,             NULL)
	ATTR_(MA_ENCRYPT,           JMA_ENCRYPT,            NULL)
	ATTR_(MA_COMPRESS,          JMA_COMPRESS,           NULL)
	ATTR_(MA_ESCAPED,           JMA_ESCAPED,            NULL)
	ATTR_(MA_FORCEPICKUP,       JMA_FPU,                NULL)
	ATTR_(MA_TYPELOCAL,         JMA_TYPELOCAL,          NULL)
	ATTR_(MA_TYPEECHO,          JMA_TYPEECHO,           NULL)
	ATTR_(MA_TYPENET,           JMA_TYPENET,            NULL)
//  ATTR_(MA_SCANNED,           0,                      NULL)
//  ATTR_(MA_FILEUPDATEREQ,     0,                      NULL)
//  ATTR_(MA_AUDITREQUEST,      0,                      NULL)
	ATTR_(MA_NODISPLAY,         JMA_NODISP,             NULL)
	ATTR_(MA_LOCKED,            JMA_LOCKED,             NULL)
	ATTR_(MA_DELETED,           JMA_DELETED,            NULL)
  };

  // Jam message header mapping table

  static MHDRMAP aMhdr[] = {
	MHDR_(JMH, cTimesRead,      MHDR_UL32,  FMP_TIMESREAD)
	MHDR_(JMH, msgnReplyTo,     MHDR_UL32,  FMP_MSGREPLYTO)
	MHDR_(JMH, msgnReply1st,    MHDR_UL32,  FMP_MSGREPLIES)
	MHDR_(JMH, msgnReplyNext,   MHDR_UL32,  FMP_MSGREPLYNEXT)
	MHDR_(JMH, timeWritten,     MHDR_TIME,  FMP_DTTMWRITTEN)
	MHDR_(JMH, timeReceived,    MHDR_TIME,  FMP_DTTMRECEIVED)
	MHDR_(JMH, timeProcessed,   MHDR_TIME,  FMP_DTTMPROCESSED)
	MHDR_(JMH, crcp,            MHDR_UL32,  FMP_PASSWORDCRC)
	MHDR_(JMH, cost,            MHDR_UL32,  FMP_MESSAGECOST)
  };

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine loads message attributes
//

  static BOOL FMASUB DoLoadAttr(PFMM pfmm, PJMH pjmh, PJSF pjsf, ULONG cbjsf)
  {
	UINT iAttr;

	// Scan the attribute mapping table translating attributes

	for (iAttr = 0; iAttr < numbof(aAttr); iAttr++) {
	  if (pjmh->attr & aAttr[iAttr].attrNative) {
		SETFLAG(pfmm->attr, aAttr[iAttr].attr);
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message attributes
//

  static BOOL FMASUB DoSaveAttr(PFMM pfmm, PJMH pjmh, PJSF * ppjsf, ULONG * pcbjsf)
  {
	UINT iAttr;

	// Scan the attribute mapping table translating attributes

	for (iAttr = 0; iAttr < numbof(aAttr); iAttr++) {
	  if (pfmm->attr & aAttr[iAttr].attr) {
		SETFLAG(pjmh->attr, aAttr[iAttr].attrNative);
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads message header
//

  static BOOL FMASUB DoLoadMhdr(PFMM pfmm, PJMH pjmh, PJSF pjsf, ULONG cbjsf)
  {
	// Load the message number as unique mesage id

	pfmm->umsg = pjmh->msgn;

	// Load the message header mapping table adding properties

	if (!FmaLoadMhdr(pfmm, pjmh, aMhdr, numbof(aMhdr)))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message header
//

  static BOOL FMASUB DoSaveMhdr(PFMM pfmm, PJMH pjmh, PJSF * ppjsf, ULONG * pcbjsf)
  {
	// Save the message header

	if (!FmaSaveMhdr(pfmm, pjmh, aMhdr, numbof(aMhdr)))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks if JSF_ENCLFREQ is actually a file udpate request
//

  static ULONG FMASUB DoCheckFreq(PJSF pjsf)
  {
	BOOL bFileUpdateReq;
	UINT cch;
	PCH pch;

	// Check if there is a password appended and 
	// get the file name length

	if ((pch = memchr(PJSFDATA(pjsf), 0, (int)pjsf->cb)) != NULL) {
	  cch = (UINT)(pch - (PCH)PJSFDATA(pjsf));
	} else
	  cch = (UINT)pjsf->cb;

	// Jam specs state the file request should be treaten as a file
	// update request if there is a complete path specified, so
	// check if there are any backslashes... what else?

	bFileUpdateReq = (memchr(PJSFDATA(pjsf), PATH_SEP, cch) != NULL);

	return bFileUpdateReq ? FMP_FILEUPDATEREQ : FMP_FILEREQUEST;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks if JSF_FTSKLUDGE contains leading or trailing kludges
//

  static ULONG FMASUB DoCheckKludge(PJSF pjsf)
  {
	PCH pch, pchKludge;
	UINT cch;

	// Scan past the kludge name looking for the space and if found,
	// check if this kludge is one of known eom kludges

	pchKludge = (PCH)PJSFDATA(pjsf); 
	for (pch = pchKludge, cch = (UINT)pjsf->cb; cch > 0; pch++, cch--) {
	  if (*pch == ' ') {
		if (FmaGetEomKludgeFromString(pchKludge, (int)pjsf->cb - cch + 1, TRUE))
		  return FMP_EOMKLUDGE;
		else
		  return FMP_FTSKLUDGE;
	  }
	}

	return FMP_FTSKLUDGE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads message header subfields
//

  static BOOL FMASUB DoLoadSubf(PFMM pfmm, PJMH pjmh, PJSF pjsf, ULONG cbjsf)
  {
	PFMP pfmp;
	ULONG id;

	// Scan the message header subfields

	while (pjsf && cbjsf) {

	  // Process the subfields

	  switch (pjsf->id) {
		case JSF_OADDRESS:    id = FMP_ORIGADDR;        goto Load;
		case JSF_DADDRESS:    id = FMP_DESTADDR;        goto Load;
		case JSF_SENDERNAME:  id = FMP_SENDERNAME;      goto Load;
		case JSF_RECVRNAME:   id = FMP_RECEVRNAME;      goto Load;
		case JSF_MSGID:       id = FMP_MSGID;           goto Load;
		case JSF_REPLYID:     id = FMP_REPLYID;         goto Load;
		case JSF_SUBJECT:     id = FMP_SUBJECT;         goto Load;
		case JSF_PID:         id = FMP_PID;             goto Load;

		case JSF_ENCLFILE:    
		case JSF_ENCLFWALIAS: 
		case JSF_ENCLFILEWC:  id = FMP_FILEATTACH;      goto Load;
		case JSF_ENCLFREQ:    id = DoCheckFreq(pjsf);   goto Load;

		case JSF_FTSKLUDGE:   id = DoCheckKludge(pjsf); goto Load;
		case JSF_TRACE:       id = FMP_EOMKLUDGE;       goto Load;

		case JSF_SEENBY2D:    id = FMP_SEENBY2D;        goto Load;
		case JSF_PATH2D:      id = FMP_PATH2D;          goto Load;
		case JSF_FLAGS:       id = FMP_FLAGS;           goto Load;
		case JSF_TZUTCINFO:   id = FMP_TZUTC;           goto Load;

		  // Load the subfield data and check if ok

Load:     if (!FmaAddFmp(pfmm, id, PJSFDATA(pjsf), pjsf->cb)) 
			return FALSE;

		  break;

		case JSF_ENCLINDFILE:
		case JSF_EMBINDAT:

		  // These are not supported

		  break;
	  }

	  // Advance to the next subfield if any

	  cbjsf-= sizeof(JSF) + pjsf->cb;
	  pjsf = (PJSF)(PJSFDATA(pjsf) + pjsf->cb);
	}

	// Load ^aFLAGS kludge info and check if ok. Note: ^aFLAGS kludge info
	// is already loaded but it's in raw form, so we need to make sure that
	// no flags we have binary representation form for show up in propery

	if ((pfmp = FmaGetFmp(pfmm, FMP_FLAGS, 0)) != NULL) {
	  if (!FmaLoadFLAGS(pfmm, PFMPDATA(pfmp), (int)pfmp->cb))
		return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message header subfields
//

  static BOOL FMASUB DoSaveSubf(PFMM pfmm, PJMH pjmh, PJSF * ppjsf, ULONG * pcbjsf)
  {
	PSZ psz, pszSpc;
	USHORT id;
	PFMP pfmp;

	// Loop through the property list adding appropriate subfields

	for (pfmp = pfmm->pfmp; pfmp != NULL; pfmp = pfmp->pfmpNext) {
	  pszSpc = " ";
	  switch (pfmp->id) {

		// These are saved as is

		case FMP_SENDERNAME:    id = JSF_SENDERNAME;       goto Save;
		case FMP_RECEVRNAME:    id = JSF_RECVRNAME;        goto Save;
		case FMP_ORIGADDR:      id = JSF_OADDRESS;         goto Save;
		case FMP_DESTADDR:      id = JSF_DADDRESS;         goto Save;
		case FMP_SUBJECT:       id = JSF_SUBJECT;          goto Save;
		case FMP_FILEATTACH:    id = JSF_ENCLFILE;         goto Save;
		case FMP_FILEREQUEST:   id = JSF_ENCLFREQ;         goto Save;
		case FMP_FILEUPDATEREQ: id = JSF_ENCLFREQ;         goto Save;
		case FMP_MSGID:         id = JSF_MSGID;            goto Save;
		case FMP_REPLYID:       id = JSF_REPLYID;          goto Save;
		case FMP_PID:           id = JSF_PID;              goto Save;
		case FMP_TZUTC:         id = JSF_TZUTCINFO;        goto Save;
		case FMP_FTSKLUDGE:     id = JSF_FTSKLUDGE;        goto Save;
		case FMP_SEENBY2D:      id = JSF_SEENBY2D;         goto Save;
		case FMP_PATH2D:        id = JSF_PATH2D;           goto Save;
		case FMP_FLAGS:         id = JSF_FLAGS;            goto Save;

		  // Add the subfield and check if ok

Save:     if (!JamAddJsf(ppjsf, pcbjsf, id, PFMPDATA(pfmp), pfmp->cb))
			return FALSE;

		  break;

		case FMP_AREA:          psz = "AREA:"; pszSpc= ""; goto SaveKludge;
		case FMP_TID:           psz = "TID:";              goto SaveKludge;
		case FMP_CHARSET:       psz = "CHRS:";             goto SaveKludge;
		case FMP_REALNAME:      psz = "REALNAME:";         goto SaveKludge;
		case FMP_REPLYADDR:     psz = "REPLYADDR:";        goto SaveKludge;
		case FMP_REPLYTO:       psz = "REPLYTO:";          goto SaveKludge;

		  // Format and save the kludge info and check if ok

SaveKludge: if (!JamAddJsfFormat(ppjsf, pcbjsf, JSF_FTSKLUDGE, 
							   "%s%s%s", psz, pszSpc, PFMPDATA(pfmp))) 
			  return FALSE;

		  break;

		// Here we could check if this is a ^aVia kludge and if so, add it as JSF_TRACE,
		// however, JSF_TRACE info is supposed to have a certain format (see jamdoc), but
		// in real life Via does not follow this format, so we add it as generic kludge.

		case FMP_EOMKLUDGE:     id = JSF_FTSKLUDGE;        goto Save;
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads message text
//

  static BOOL FMASUB DoLoadText(PFMM pfmm, HMSG hmsg, ULONG fl)
  {
	PCH pchText = NULL; ULONG cchText = JRM_GETALLTEXT;
	PCH pch;

	// Check if we need to load message text

	if (fl & FLM_NOTEXT) return TRUE;

	// Load text and check if ok

	if (!JamReadMsg(hmsg, NULL, NULL, NULL, &pchText, &cchText, 0))
	  return FALSE;

	// Check if no text loaded and leave

	if (!cchText) {
	  if (pchText) MemFree(pchText);
	  return TRUE;
	}

	// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
	if (!CHECK_DATA_SIZE_LIMIT(cchText)) {
	  FMAERROR(DATA_SIZE_LIMIT);
	  return FALSE;
	}
#endif

	// Check if text is not null terminated and add a trailing null

	if (pchText[cchText - 1]) {
	  if ((pch = MemRealloc(pchText, (int)cchText + 1)) == NULL) {
		FMAERROR(NOT_ENOUGH_MEMORY);
		MemFree(pchText);
		return FALSE;
	  } else {
		pchText = pch;
		pchText[cchText] = '\0';
	  }
	}

	// Add message text property and check if ok

	if (!FmaAddFmp(pfmm, FMP_MESSAGETEXT, &pchText, sizeof(pchText))) {
	  MemFree(pchText);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message text
//

  static BOOL FMASUB DoSaveText(PFMM pfmm, HMSG hmsg, ULONG fl)
  {
	PCH pchText; ULONG cchText;
	PSZ * ppsz;
	ULONG cpsz;
	PFMP pfmp;

	// Check if we need to save message text

	if (fl & FSM_NOTEXT) return TRUE;

	// Get the message text property and check if ok

	if ((pfmp = FmaGetFmp(pfmm, FMP_MESSAGETEXT, 0)) == NULL) 
	  return TRUE;

	// Loop through all the message text strings writing them out

	ppsz = PFMPDATA(pfmp);
	cpsz = pfmp->cb / sizeof(PSZ);

	while (cpsz--) {
	  if (ppsz && *ppsz) {
		pchText = *ppsz;
		cchText = strlen(pchText);
		if (!JamWriteMsg(hmsg, NULL, NULL, 0, pchText, cchText, 0, 0))
		  return FALSE;
	  }

	  ppsz++;
	}

	// Note: for jam eom kludges are part of subfields, not message text

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns total message text length
//

  static ULONG FMASUB DoGetTotalTextLen(PFMM pfmm)
  {
	ULONG cchTextTotal = 0;
	PFMP pfmp;

	// Get the message text property and add its length

	if ((pfmp = FmaGetFmp(pfmm, FMP_MESSAGETEXT, 0)) != NULL) 
	  cchTextTotal+= FmaGetFmpTextLen(pfmp);

	return cchTextTotal;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine returns api pointer
//

  FMAAPI_(PAPI) JamGetApi(VOID)
  {
	return &api;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads abstracted fidonet mail message
//

  FMAAPI_(BOOL) JamLoadMsg(HMSG hmsg, PFMM * ppfmm, ULONG fl)
  {
	ULONG cbjsf = JRM_GETJSFPTR;
	PJSF pjsf = NULL; 
	PFMM pfmm;
	JMH jmh;

	ASSERT(hmsg != NULL);
	ASSERT(ppfmm != NULL);

	// Create fmm object and check if ok

	if ((pfmm = FmaCreateMsg(0)) == NULL) {
Fail: FmaDestroyMsg(pfmm);      
	  return FALSE;
	} else
	  pfmm->type = MBTYPE_JAM;

	// Read the message header and subfields, then check if ok

	if (!JamReadMsg(hmsg, &jmh, &pjsf, &cbjsf, NULL, NULL, NULL))
	  goto Fail;

	// Load message attributes and check if ok

	if (!DoLoadAttr(pfmm, &jmh, pjsf, cbjsf))
	  goto Fail;

	// Load message header

	if (!DoLoadMhdr(pfmm, &jmh, pjsf, cbjsf))
	  goto Fail;

	// Load message header subfields and check if ok

	if (!DoLoadSubf(pfmm, &jmh, pjsf, cbjsf))
	  goto Fail;

	// Load message text and check if ok

	if (!DoLoadText(pfmm, hmsg, fl))
	  goto Fail;

	// Pass fmm object to caller

	*ppfmm = pfmm;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves abstracted fidonet mail message
//

  FMAAPI_(BOOL) JamSaveMsg(HMSG hmsg, PFMM pfmm, ULONG fl, ULONG cchTextTotal)
  {
	PJSF pjsf = NULL; ULONG cbjsf = 0;
	JMH jmh;

	ASSERT(hmsg != NULL);
	ASSERT(pfmm != NULL);

	// Clear target message header

	memset(&jmh, 0, sizeof(JMH));

	// Add automatic properties and check if ok

	if (!FmaAddAutoProp(pfmm, fl)) {
Fail: if (pjsf) MemFree(pjsf);
	  return FALSE;
	}

	// Save message attributes and check if ok

	if (!DoSaveAttr(pfmm, &jmh, &pjsf, &cbjsf))
	  goto Fail;

	// Save message header

	if (!DoSaveMhdr(pfmm, &jmh, &pjsf, &cbjsf))
	  goto Fail;

	// Save message header subfields and check if ok

	if (!DoSaveSubf(pfmm, &jmh, &pjsf, &cbjsf))
	  goto Fail;

	// Check if caller did not specify total message text length and get it now

	if (cchTextTotal == (ULONG)-1) cchTextTotal = DoGetTotalTextLen(pfmm);

	// Write the message header and subfields 

	if (!JamWriteMsg(hmsg, &jmh, pjsf, cbjsf, NULL, 0, 0, cchTextTotal))
	  goto Fail;

	// Save message text and check if ok

	if (!DoSaveText(pfmm, hmsg, fl))
	  goto Fail;

	// Free memory allocated for subfields info

	if (pjsf) MemFree(pjsf);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of JamApi.c
/////////////////////////////////////////////////////////////////////////////
