/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: JAM api declarations module
//
// JamApi.h
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 03/Oct/1999
//
// History: 07/Sep/1998 kvitekp - Initially created for MBS project.
//
/////////////////////////////////////////////////////////////////////////////
//
// JAM(mbp) - Copyright 1993 Joaquim Homrighausen,
//            Andrew Milner, Mats Birch, Mats Wallin.
//            ALL RIGHTS RESERVED.
//
// Note: These declaratons were derived from original
//       specifications published by JAM message base
//       authors in JAM(mbp) API documentation.
//
/////////////////////////////////////////////////////////////////////////////
//
// NOTE: ALL STRUCTURES IN THIS MODULE ASSUME SINGLE BYTE PACKING !!!
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _JAM_API_INCLUDED
#define _JAM_API_INCLUDED

  // Declare C/C++ frame

#ifdef __cplusplus
extern "C" {
#endif

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

  // Set structure packing

#ifdef FMA_PUSHPOP_PACKING
#pragma pack(push, 1)
#endif

/////////////////////////////////////////////////////////////////////////////
// JAM base declarations
/////////////////////////////////////////////////////////////////////////////

  // JAM base types

#define UMSG ULONG              // Unique message id
#define MSGN ULONG              // Message number

  // JAM base file extensions

#define JAM_JHR_FILEEXT  ".JHR" // Header file extension
#define JAM_JDT_FILEEXT  ".JDT" // Text file extension
#define JAM_JDX_FILEEXT  ".JDX" // Index file extension
#define JAM_JLR_FILEEXT  ".JLR" // Lastread file extension

  // JAM structures magic number and revision level

#define JAM_MAGIC  0x004D414Alu // Equals to string <J><A><M><nul>
#define JAM_REVLEVEL 1          // Current JAM revision level

  // JAM base header structure at the beginning of <area>.jhr

  typedef struct _JBH { /* jbh */
	ULONG   magic;              // Magic number JAM_MAGIC
	ULONG   time;               // Creation time
	ULONG   cmod;               // Modification counter
	ULONG   cmsgActive;         // Number of active (not deleted) msgs
	ULONG   crcp;               // CRC-32 of password to access
	MSGN    msgnBase;           // Lowest message number in index file
	BYTE    rsvd[1000];         // --- reserved
  } JBH, FAR * PJBH;

  // JAM message header structure in <area>.jhr

  typedef struct _JMH { /* jmh */
	ULONG   magic;              // Magic number (JAM_MAGIC)
	USHORT  revLevel;           // Current revision level (JAM_REVLEVEL)
	USHORT  rsvd;               // --- reserved
	ULONG   cbjsf;              // Length of subfields
	ULONG   cTimesRead;         // Number of times message was read
	ULONG   crcMsgId;           // CRC-32 of MSGID line
	ULONG   crcReply;           // CRC-32 of REPLY line
	MSGN    msgnReplyTo;        // This msg is a reply to
	MSGN    msgnReply1st;       // First reply to this msg
	MSGN    msgnReplyNext;      // Next msg in reply chain
	ULONG   timeWritten;        // When msg was written
	ULONG   timeReceived;       // When msg was received/read
	ULONG   timeProcessed;      // When msg was processed by packer
	MSGN    msgn;               // Message number (1-based)
	ULONG   attr;               // Msg attributes (JMA_)
	ULONG   attr2;              // --- reserved attributes
	ULONG   offText;            // Offset of text in <area>.JDT
	ULONG   cchText;            // Length of message text in <area>.JDT
	ULONG   crcp;               // CRC-32 of password to access msg
	ULONG   cost;               // Cost of message
  } JMH, FAR * PJMH;

  // JAM message attributes in jmh.attr

#define JMA_LOCAL       0x00000001L     // Msg created locally
#define JMA_INTRANSIT   0x00000002L     // Msg is in-transit
#define JMA_PRIVATE     0x00000004L     // Private message
#define JMA_READ        0x00000008L     // Read by addressee
#define JMA_SENT        0x00000010L     // Sent to remote
#define JMA_KILLSENT    0x00000020L     // Kill when sent
#define JMA_ARCHIVESENT 0x00000040L     // Archive when sent
#define JMA_HOLD        0x00000080L     // Hold for pick-up
#define JMA_CRASH       0x00000100L     // Crash
#define JMA_IMMEDIATE   0x00000200L     // Send Msg now, ignore restrictions
#define JMA_DIRECT      0x00000400L     // Send directly to destination
#define JMA_GATE        0x00000800L     // Send via gateway
#define JMA_FILEREQUEST 0x00001000L     // File request
#define JMA_FILEATTACH  0x00002000L     // File(s) attached to Msg
#define JMA_TRUNCFILE   0x00004000L     // Truncate file(s) when sent
#define JMA_KILLFILE    0x00008000L     // Delete file(s) when sent
#define JMA_RECEIPTREQ  0x00010000L     // Return receipt requested
#define JMA_CONFIRMREQ  0x00020000L     // Confirmation receipt requested
#define JMA_ORPHAN      0x00040000L     // Unknown destination
#define JMA_ENCRYPT     0x00080000L     // Msg text is encrypted
#define JMA_COMPRESS    0x00100000L     // Msg text is compressed
#define JMA_ESCAPED     0x00200000L     // Msg text is seven bit ASCII
#define JMA_FPU         0x00400000L     // Force pickup
#define JMA_TYPELOCAL   0x00800000L     // Msg is for local use only (not for export)
#define JMA_TYPEECHO    0x01000000L     // Msg is for conference distribution
#define JMA_TYPENET     0x02000000L     // Msg is direct network mail
#define JMA_0x04000000  0x04000000L     // -- reserved
#define JMA_0x08000000  0x08000000L     // -- reserved
#define JMA_0x10000000  0x10000000L     // -- reserved
#define JMA_NODISP      0x20000000L     // Msg may not be displayed to user
#define JMA_LOCKED      0x40000000L     // Msg is locked, no editing possible
#define JMA_DELETED     0x80000000L     // Msg is deleted

  // JAM message subfield structure

  typedef struct _JSF { /* jsf */
	USHORT  id;                 // Subfield ID, 0-65535
	USHORT  id2;                // Reserved for future use
	ULONG   cb;                 // Subfield data length
//  BYTE    ab[cb];             // Subfield data bytes
  } JSF, FAR * PJSF;

#define PJSFDATA(pjsf) ((PBYTE)(pjsf) + sizeof(JSF))    // Poiner to subfield data

  // JAM message subfield types

#define JSF_OADDRESS    0       // Originating address          100 bytes max
#define JSF_DADDRESS    1       // Destination address          100 bytes max
#define JSF_SENDERNAME  2       // Author's name                100 bytes max
#define JSF_RECVRNAME   3       // Recipient's name             100 bytes max
#define JSF_MSGID       4       // Message identification data  100 bytes max
#define JSF_REPLYID     5       // Message reply data           100 bytes max
#define JSF_SUBJECT     6       // Message subject              100 bytes max
#define JSF_PID         7       // FTN PID kludge line           40 bytes max
#define JSF_TRACE       8       // FTN ^aVia information
#define JSF_ENCLFILE    9       // File attachment
#define JSF_ENCLFWALIAS 10      // File attachment asciiz alias
#define JSF_ENCLFREQ    11      // File request
#define JSF_ENCLFILEWC  12      // One or more file attachment (wild cards)
#define JSF_ENCLINDFILE 13      // File attachment reference file
#define JSF_EMBINDAT    1000    // Reserved for future use
#define JSF_FTSKLUDGE   2000    // FTS-compliant "kludge" line
#define JSF_SEENBY2D    2001    // 2D (net/node) SEEN-BY information
#define JSF_PATH2D      2002    // 2D (net/node) PATH information
#define JSF_FLAGS       2003    // FTN FLAGS kludge information
#define JSF_TZUTCINFO   2004    // Time zone information
#define JSF_UNKNOWN     0xffff  // Unknown subfield type

  // JAM index file record structure

  typedef struct _JDX { /* jdx */
	ULONG   crc;                // CRC-32 of destination username
	ULONG   lmh;                // Offset of header in .JHR file
  } JDX, FAR * PJDX;

  // JAM index empty entry initialization

#define JDX_INIT_EMPTY  {(ULONG)-1,(ULONG)-1}

  // JAM lastread file record structure

  typedef struct _JLR { /* jlr */
	ULONG   crc;                // CRC-32 of user name (lowercase)
	ULONG   uid;                // Unique UserID
	MSGN    msgnLast;           // Last read message number
	MSGN    msgnHigh;           // Highest read message number
  } JLR, FAR * PJLR;

  // JAM message base locking of <area>.JHR

#define JBH_LOCK_OFFS   0       // Locking region offset
#define JBH_LOCK_LENG   1       // Locking region length
#define JBH_LOCK_RETRY  5       // Max number of locking attempts

/////////////////////////////////////////////////////////////////////////////
// JAM api declarations
/////////////////////////////////////////////////////////////////////////////

  // Special values used by JamReadMsg()

#define JRM_GETJSFCOPY   0            // *pccbjsf : Allocate a copy of subfields info
#define JRM_GETJSFPTR   ((ULONG)-1)   // *pccbjsf : Return a readonly pointer to subfields info
#define JRM_GETALLTEXT  ((ULONG)-1)   // *pcchText: read entire message text

  // Function prototypes

  FMAAPI_(PAPI) JamGetApi(VOID);
  FMAAPI_(HAREA) JamOpenArea(PSZ pszPath, UINT fOpenMode, ULONG fl, 
							 NETADDR * pnetAddr);

  FMAAPI_(BOOL) JamCloseArea(HAREA harea);
  FMAAPI_(BOOL) JamLockArea(HAREA harea, BOOL bLock);

  FMAAPI_(BOOL) JamValidateMsg(HAREA harea, MSGN msgn, PULONG pimsg);
  FMAAPI_(BOOL) JamCreateMsgMap(HAREA parea, UMSG * paumsg[], PULONG pcumsg);
  FMAAPI_(BOOL) JamEnumMsgs(HAREA harea, PFNENUMMSGS pfnCallback, 
							PVOID pvCallbackData, ULONG fl);

  FMAAPI_(BOOL) JamDeleteMsg(HAREA harea, MSGN msgn);
  FMAAPI_(BOOL) JamSetCurMsg(HAREA harea, MSGN msgn);
  FMAAPI_(MSGN) JamGetCurMsg(HAREA harea);
  FMAAPI_(MSGN) JamGetFirstMsg(HAREA harea);
  FMAAPI_(MSGN) JamGetHighMsg(HAREA harea);
  FMAAPI_(LONG) JamGetMsgCount(HAREA harea);

  FMAAPI_(UMSG) JamMsgnToUid(HAREA harea, MSGN msgn);
  FMAAPI_(MSGN) JamUidToMsgn(HAREA harea, UMSG umsg, UINT fMode);

  FMAAPI_(HMSG) JamOpenMsg(HAREA harea, UINT mode, MSGN msgn, ULONG fl);
  FMAAPI_(BOOL) JamCloseMsg(HMSG hmsg);

  FMAAPI_(BOOL) JamLoadMsg(HMSG hmsg, PFMM * ppfmm, ULONG fl);
  FMAAPI_(BOOL) JamSaveMsg(HMSG hmsg, PFMM pfmm, ULONG fl, ULONG cchTextTotal);

  FMAAPI_(BOOL) JamReadMsg(HMSG hmsg, PJMH pjmh, PJSF * ppjsf, ULONG * pcbjsf,
						   PCH * ppchText, ULONG * pcchText, ULONG * poffText);

  FMAAPI_(BOOL) JamWriteMsg(HMSG hmsg, PJMH pjmh, PJSF pjsf, ULONG cbjsf,
							PCH pchText, ULONG cchText, ULONG * poffText, 
							ULONG cchTextTotal);

  FMAAPI_(BOOL) JamGetMsgCtrlLen(HMSG hmsg, ULONG * pcbjsf);
  FMAAPI_(BOOL) JamGetMsgTextLen(HMSG hmsg, ULONG * pcchText);

  FMAAPI_(BOOL) JamGetLRInfo(HAREA harea, ULONG uid, PLRINFO plri);
  FMAAPI_(BOOL) JamSetLRInfo(HAREA harea, ULONG uid, PLRINFO plri, 
							 PSZ pszUserName);

/////////////////////////////////////////////////////////////////////////////
// Header cleanup
/////////////////////////////////////////////////////////////////////////////

  // Restore structure packing

#ifdef FMA_PUSHPOP_PACKING
#pragma pack(pop)
#endif

  // Terminate C/C++ frame

#ifdef __cplusplus
}
#endif

#endif  // #ifndef _JAM_API_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// End of JamApi.h
/////////////////////////////////////////////////////////////////////////////
