/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: JAM api private declarations
//
// JamApiP.h
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 18/Feb/2000
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
//          10/Feb/2000 kvitekp - Added JamAddJsfFormatV()
//
//          18/Feb/2000 kvitekp - Added JamSet/CheckPassword()
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _JAM_APIP_INCLUDED
#define _JAM_APIP_INCLUDED

  // Declare C/C++ frame

#ifdef __cplusplus
extern "C" {
#endif

  // Include types declaratons

#include "FmaDef.h"
#include "JamApi.h"

/////////////////////////////////////////////////////////////////////////////
// JAM area declarations
/////////////////////////////////////////////////////////////////////////////

  // Forward declarations

  struct _JMSG;

  // JAM index cache page structure

  typedef struct _JIC {         /* jic */
	struct _JIC * pjicPrev;             // Pointer to prev entry in list
	struct _JIC * pjicNext;             // Pointer to next entry in list
	UINT    cjdx;                       // Number of allocated index entries
	UINT    ijdx;                       // Index of the next unused index entry
	JDX     ajdx[1];                    // Index entries array
  } JIC, FAR * PJIC;

  // JAM message area object structure

  typedef struct _JAREA {       /* area */
	ULONG   magic;                      // Magic number (JAREA_MAGIC)
	PAPI    papi;                       // JAM api pointer
	LONG    lmhCur;                     // Current header offset
	MSGN    msgnCur;                    // Current message number
	MSGN    msgnBase;                   // Base message number
	MSGN    msgnHigh;                   // High message number
	ULONG   cmsgActive;                 // Number of active msgs
	JBH     jbh;                        // JAM base header
	HFILE   hfileJhr;                   // JAM header file handle
	HFILE   hfileJdt;                   // JAM text file handle
	HFILE   hfileJdx;                   // JAM index file handle
	HFILE   hfileJlr;                   // JAM lastread file handle
	BOOL    bAreaLocked;                // Area lock count
	BOOL    bBaseLocked;                // Base lock count
	BOOL    bExclusive;                 // Exclusive count
	BOOL    bJdxCached;                 // Index caching count
	PJIC    pjicList;                   // Index cache list
	MSGN    msgnLoDirty;                // Index cache low dirty (inclusive)
	MSGN    msgnHiDirty;                // Index cache high dirty (inclusive)
	ULONG   fl;                         // Open area flags (FOA_)
	struct _JMSG * pmsgList;            // Open messges list
	NETADDR netAddr;                    // Default network address
	CHAR    achPath[1];                 // Area path (asciiz)
  } JAREA, FAR * PJAREA;

  // JAM message area object magic number

#define JAREA_MAGIC ((ULONG)0x4552414A) // Long equivalent of "JARE"

  // JAM message object structure

  typedef struct _JMSG {        /* msg */
	struct _JMSG * pmsgPrev;            // Pointer to prev open msg in area
	struct _JMSG * pmsgNext;            // Pointer to next open msg in area
	ULONG   magic;                      // Magic number (JMSG_MAGIC)
	PJAREA  parea;                      // Message area handle
	MSGN    msgn;                       // Message number
	UINT    mode;                       // Message open mode
	ULONG   fl;                         // Message open flags (FOM_)
	LONG    lmh;                        // Message header offset
	JMH     jmh;                        // Message header
	PJSF    pjsf;                       // Message header subfields
	ULONG   cbjsf;                      // Message header subfields size
	ULONG   cbjsfMax;                   // Message header subfields size maximum
	LONG    lmt;                        // Message text offset
	LONG    lmtEnd;                     // Message text end offset
	LONG    lmtMax;                     // Message text max offset
	UINT    fs;                         // Message status flags (JMF_)
  } JMSG, FAR * PJMSG;

  // JAM message object magic number

#define JMSG_MAGIC ((ULONG)0x47534D4A)  // Long equivalent of "JMSG"

  // JAM message status flags in msg.fs

#define JMF_GOTMSGH       0x0001        // Message header has been associated
#define JMF_GOTSUBF       0x0002        // Subfield info has been associated
#define JMF_SAVEJMH       0x0004        // Save jam message header to disk
#define JMF_GOTFRAME      0x0008        // Writing frame selected for new message
#define JMF_EXCLUSIVE     0x0010        // Entered exclusive when opened
#define JMF_MSGHWRITTEN   0x0100        // Message header has been written
#define JMF_SUBFWRITTEN   0x0200        // Message contrl info has been written
#define JMF_TEXTWRITTEN   0x0400        // Message text has been written
#define JMF_READDELETED   0x8000        // Message had JMA_DELETED when last read

  // Message has been written to mask

#define JMF_MASKWRITTEN   (JMF_MSGHWRITTEN | JMF_SUBFWRITTEN | JMF_TEXTWRITTEN)

/////////////////////////////////////////////////////////////////////////////
// Miscellaneous declarations
/////////////////////////////////////////////////////////////////////////////

  // Function prototypes

  FMAAPI_(BOOL) JamCheckArea(PJAREA parea);
  FMAAPI_(BOOL) JamCheckMsg(PJMSG pmsg);
  FMAAPI_(BOOL) JamCheckMsgReadMode(PJMSG pmsg);
  FMAAPI_(BOOL) JamCheckMsgWriteMode(PJMSG pmsg);

  FMAAPI_(BOOL) JamLockBase(PJAREA parea, BOOL bLock);
  FMAAPI_(BOOL) JamEnterExclusive(PJAREA parea);
  FMAAPI_(BOOL) JamLeaveExclusive(PJAREA parea);

  FMAAPI_(BOOL) JamLoadJbh(PJAREA parea);
  FMAAPI_(BOOL) JamSaveJbh(PJAREA parea);
  FMAAPI_(BOOL) JamCheckJbh(PJAREA parea);

  FMAAPI_(BOOL) JamResetCurMsgCache(PJAREA parea);
  FMAAPI_(BOOL) JamSetCurMsgCache(PJMSG pmsg);

  FMAAPI_(PJMSG) JamGetOpenMsg(PJAREA parea, MSGN msgn);

  FMAAPI_(BOOL) JamCheckLmh(PJAREA parea, LONG lmh);
  FMAAPI_(BOOL) JamLoadJmh(PJAREA parea, PJMH pjmh, PJSF * ppjsf, LONG lmh);
  FMAAPI_(BOOL) JamSaveJmh(PJAREA parea, PJMH pjmh, PJSF pjsf, LONG lmh);

  FMAAPI_(PJSF) JamGetJsf(PJSF pjsf, ULONG cbjsf, USHORT id, LONG jsfIndex);
  FMAAPI_(BOOL) JamDelJsf(PJSF * ppjsf, ULONG * pcbjsf, USHORT id);
  FMAAPI_(PJSF) JamAddJsf(PJSF * ppjsf, ULONG * pcbjsf, USHORT id, PVOID pb, ULONG cb);
  FMAAPIV_(PJSF) JamAddJsfFormat(PJSF * ppjsf, ULONG * pcbjsf, USHORT id, PSZ pszFormat, ...);
  FMAAPI_(PJSF) JamAddJsfFormatV(PJSF * ppjsf, ULONG * pcbjsf, USHORT id, PSZ pszFormat, va_list parg);

  FMAAPI_(BOOL) JamBeginJdxCache(PJAREA parea);
  FMAAPI_(BOOL) JamEndJdxCache(PJAREA parea);

  FMAAPI_(ULONG) JamGetJdxCount(PJAREA parea);

  FMAAPI_(BOOL) JamGetJdx(PJAREA parea, PJDX pjdx, MSGN msgn);
  FMAAPI_(BOOL) JamSetJdx(PJAREA parea, PJDX pjdx, MSGN msgn);
  FMAAPI_(BOOL) JamDelJdx(PJAREA parea, PJDX pjdx, MSGN msgn);
  FMAAPI_(BOOL) JamAddJdx(PJAREA parea, PJDX pjdx, MSGN * pmsgn);

  FMAAPI_(ULONG) JamGetCrc32(PCH pch, ULONG cch, ULONG crc);

  FMAAPI_(BOOL) JamGetJlr(PJAREA parea, ULONG uid, PJLR pjlr, LONG * poffs);
  FMAAPI_(BOOL) JamSetJlr(PJAREA parea, PJLR pjlr);

  FMAAPI_(BOOL) JamCheckBase(PSZ pszPath);

  FMAAPI_(BOOL) JamSetPassword(PJAREA parea, PSZ pszPassword);
  FMAAPI_(BOOL) JamCheckPassword(PJAREA parea, PSZ pszPassword);

  // Logging control defines

#ifdef  FMA_LOG
#define FMA_LOG_JAM_AREA
//#define FMA_LOG_JAM_HELP
//#define FMA_LOG_JAM_INDEX
//#define FMA_LOG_JAM_LAST
//#define FMA_LOG_JAM_LOCK
#define FMA_LOG_JAM_MSG
#define FMA_LOG_JAM_READ
#define FMA_LOG_JAM_WRITE
#endif

/////////////////////////////////////////////////////////////////////////////
// Header cleanup
/////////////////////////////////////////////////////////////////////////////

  // Restore structure packing

#ifdef FMA_PACKING
#pragma pack(pop)
#endif

  // Terminate C/C++ frame

#ifdef __cplusplus
}
#endif

#endif  // #ifndef _JAM_APIP_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// End of JamApiP.h
/////////////////////////////////////////////////////////////////////////////
