/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: JAM area routines module
//
// JamArea.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 19/Sep/1999
// Updated: 27/Oct/1999
//
// History: 19/Sep/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "JamApi.h"
#include "JamApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Macro to check if message number is valid

#define GOODMSGN(msgn) ((msgn) >= parea->msgnBase && (msgn) <= parea->msgnHigh)

  // JAM message area files structure

  typedef struct _JFILES {      /* files */
	CHAR    achJhrFile[MAX_PATH];       // JAM header file name
	CHAR    achJdtFile[MAX_PATH];       // JAM text file name
	CHAR    achJdxFile[MAX_PATH];       // JAM index file name
	CHAR    achJlrFile[MAX_PATH];       // JAM lastread file name
  } JFILES, FAR * PJFILES;

  // Forward declarations

  static VOID FMASUB DoCloseAreaFiles(PJAREA parea);

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine creates area object
//

  static PJAREA FMASUB DoCreateArea(PSZ pszPath, ULONG fl, NETADDR * pnetAddr)
  {
	PJAREA parea;
	PCH pchExt;
	int cch;

	// Get area path length and check if ok

	if (!pszPath || (cch = strlen(pszPath)) == 0) {
BadP: FMAERROR(INVALID_PARAM);
	  return NULL;
	}

	// Check if there is an ".j??" extension specified and remove it

	if (cch > lengof(JAM_JHR_FILEEXT) &&
	   (pchExt = &pszPath[cch - lengof(JAM_JHR_FILEEXT)])[0] == '.' &&
		TOLOWER(pchExt[1]) == TOLOWER(JAM_JHR_FILEEXT[1])) {
	  cch-= lengof(JAM_JHR_FILEEXT);
	}

	// Get area path length and check if ok

	if (cch >= MAX_PATH - lengof(JAM_JHR_FILEEXT))
	  goto BadP;

	// Create message area object and check if ok

	if ((parea = MemAlloc(sizeof(JAREA) + cch, MA_CLEAR)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return NULL;
	} 

	// Initialize new area object

	parea->magic = JAREA_MAGIC;
	parea->papi = JamGetApi();
	parea->fl = fl;
	if (pnetAddr) parea->netAddr = *pnetAddr;
	memcpy(parea->achPath, pszPath, cch);
	parea->hfileJhr = -1;
	parea->hfileJdt = -1;
	parea->hfileJdx = -1;
	parea->hfileJlr = -1;

	return parea;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine destroys area object
//

  static BOOL FMASUB DoDestroyArea(PJAREA parea)
  {
	// Check if we have area object and leave if not

	if (!parea) return FALSE;

	// Close associated files

	DoCloseAreaFiles(parea);

	// Delete associated index list

	ASSERT(parea->pjicList == NULL);

	// Reset message area object magic number

	parea->magic = ~JAREA_MAGIC;

	// Free message area object

	MemFree(parea);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine makes up the area file names
//

  static BOOL FMASUB DoMakeAreaFileNames(PJAREA parea, PJFILES pfiles)
  {
	ASSERT(parea != NULL);
	ASSERT(pfiles != NULL);

	strcpy(pfiles->achJhrFile, parea->achPath); strcat(pfiles->achJhrFile, JAM_JHR_FILEEXT);
	strcpy(pfiles->achJdtFile, parea->achPath); strcat(pfiles->achJdtFile, JAM_JDT_FILEEXT);
	strcpy(pfiles->achJdxFile, parea->achPath); strcat(pfiles->achJdxFile, JAM_JDX_FILEEXT);
	strcpy(pfiles->achJlrFile, parea->achPath); strcat(pfiles->achJlrFile, JAM_JLR_FILEEXT);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine opens the area files
//

  static BOOL FMASUB DoOpenAreaFiles(PJAREA parea, PJFILES pfiles, ULONG fl, UINT fOpen)
  {
	ASSERT(parea != NULL);
	ASSERT(pfiles != NULL);

	// Make up the file open mode flags

	if (fl & FOA_READONLY)
	  fOpen|= FOPENFILE_READONLY;
	else
	  fOpen|= FOPENFILE_READWRITE;

	// Make up the file sharing flags

	if (fl & FOA_DENYALL)
	  fOpen|= FOPENFILE_DENYALL;
	else
	  fOpen|= FOPENFILE_DENYNONE;

	// Open the header file and check if ok

	if (!FmaOpenFile(pfiles->achJhrFile, fOpen, &parea->hfileJhr)) {
Fail: DoCloseAreaFiles(parea);
	  return FALSE;
	}

	// Open the text file and check if ok

	if (!FmaOpenFile(pfiles->achJdtFile, fOpen, &parea->hfileJdt)) {
	  goto Fail;
	}

	// Open the index file and check if ok

	if (!FmaOpenFile(pfiles->achJdxFile, fOpen, &parea->hfileJdx)) {
	  goto Fail;
	}

	// Open the lastread file and check if ok. Note that this file
	// is always created if it does not exist unless in readonly mode

	if (!(fl & FOA_READONLY)) fOpen|= FOPENFILE_CREATE;
	if (!FmaOpenFile(pfiles->achJlrFile, fOpen, &parea->hfileJlr)) {
	  goto Fail;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine closes the area files
//

  static VOID FMASUB DoCloseAreaFiles(PJAREA parea)
  {
	ASSERT(parea != NULL);

	// Close all open message base files

	FmaCloseFile(&parea->hfileJhr);
	FmaCloseFile(&parea->hfileJdt);
	FmaCloseFile(&parea->hfileJdx);
	FmaCloseFile(&parea->hfileJlr);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes area files
//

  static BOOL FMASUB DoDeleteAreaFiles(PJAREA parea, PJFILES pfiles)
  {
	BOOL bOk = TRUE;

	ASSERT(parea != NULL);
	ASSERT(pfiles != NULL);

	// Make sure the files area closed

	DoCloseAreaFiles(parea);

	// Delete area files and check if ok

	if (!FmaDeleteFile(pfiles->achJhrFile)) bOk = FALSE;
	if (!FmaDeleteFile(pfiles->achJdtFile)) bOk = FALSE;
	if (!FmaDeleteFile(pfiles->achJdxFile)) bOk = FALSE;
	if (!FmaDeleteFile(pfiles->achJlrFile)) bOk = FALSE;

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine initializes JAM base header
//

  static VOID FMASUB DoInitJbh(PJBH pjbh)
  {
	// Initialize JAM base header

	memset(pjbh, 0, sizeof(JBH));
	pjbh->magic = JAM_MAGIC;
	pjbh->time = FmaGetLocalTime();
	pjbh->msgnBase = 1;
	pjbh->crcp = (ULONG)-1;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates new area directory
//

  static BOOL FMASUB DoMakeAreaDir(PJAREA parea)
  {
	CHAR achPath[MAX_PATH];
	PCH pch;

	ASSERT(parea != NULL);

	// Copy path to the local storage and chop the base name

	strcpy(achPath, parea->achPath);
	if ((pch = strrchr(achPath, PATH_SEP)) != NULL) {
	  *pch = '\0';
	} else
	  return TRUE;

	// Check if at root or directory exists and leave if so

	if (!achPath[0] || FmaDirExists(achPath))
	  return TRUE;

	// Create directory and check if ok

	if (!FmaMakeDir(achPath)) {
	  FMAERROR(CANNOT_CREATE);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates new area
//

  static BOOL FMASUB DoCreateNewArea(PJAREA parea, ULONG fl, PJFILES pfiles)
  {
	ASSERT(parea != NULL);
	ASSERT(pfiles != NULL);

	ASSERT(parea->hfileJhr == -1);
	ASSERT(parea->hfileJdt == -1);
	ASSERT(parea->hfileJdx == -1);
	ASSERT(parea->hfileJlr == -1);

	// Create area directory if necessary

	DoMakeAreaDir(parea);

	// Create area files and check if ok

	if (!DoOpenAreaFiles(parea, pfiles, fl, FOPENFILE_CREATE | FOPENFILE_TRUNCATE))
	  return FALSE;

	// Initialize new area base header

	DoInitJbh(&parea->jbh);

	// Save empty base header and check if ok

	if (!JamSaveJbh(parea)) {
	  DoDeleteAreaFiles(parea, pfiles);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine clears existing area
//

  static BOOL FMASUB DoClearExistingArea(PJAREA parea, ULONG fl, PJFILES pfiles)
  {
	ASSERT(parea != NULL);
	ASSERT(pfiles != NULL);

	ASSERT(parea->hfileJhr == -1);
	ASSERT(parea->hfileJdt == -1);
	ASSERT(parea->hfileJdx == -1);
	ASSERT(parea->hfileJlr == -1);

	// Open and truncate area files and check if ok

	if (!DoOpenAreaFiles(parea, pfiles, fl, FOPENFILE_TRUNCATE))
	  return FALSE;

	// Initialize empty area base header

	DoInitJbh(&parea->jbh);

	// Save empty base header and check if ok

	if (!JamSaveJbh(parea)) {
	  DoCloseAreaFiles(parea);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine opens area
//

  FMAAPI_(HAREA) JamOpenArea(PSZ pszPath, UINT fOpenMode, ULONG fl, 
							 NETADDR * pnetAddr)
  {
	BOOL bAreaExists = FALSE;
	JFILES files;
	PJAREA parea;

	// Assert known structure sizes to ensure correct structure packing

	ASSERT(sizeof(JBH) == 1024);
	ASSERT(sizeof(JMH) == 76);
	ASSERT(sizeof(JDX) ==  8);
	ASSERT(sizeof(JLR) == 16);

#ifdef FMA_LOG_JAM_AREA
Log("\nJamOpenArea      >: %s", pszPath);
#endif

	// Create new area object and check if ok

	if ((parea = DoCreateArea(pszPath, fl, pnetAddr)) == NULL) {
Fail: DoDestroyArea(parea);

#ifdef FMA_LOG_JAM_AREA
Log("JamOpenArea      <: %s FAILED, error=%d", pszPath, FmaGetLastError());
#endif

	  return NULL;
	}

	// Make up the base file names

	DoMakeAreaFileNames(parea, &files);

	// Check if all area message base files exist

	if (FmaFileExists(files.achJhrFile)) bAreaExists++;
	if (FmaFileExists(files.achJdtFile)) bAreaExists++;
	if (FmaFileExists(files.achJdxFile)) bAreaExists++;
	if (FmaFileExists(files.achJlrFile)) bAreaExists++;

	// Check if some files are missing

	if (bAreaExists && bAreaExists != 4) {

	  // Check if it's a lastread file that is missing and
	  // don't fail if so, otherwise we're missing some 
	  // critical message base files

	  if (!(bAreaExists == 3 && !FmaFileExists(files.achJlrFile))) {
		FMAERROR(MISSING_FILES);
		goto Fail;
	  }
	}

	// Act upon an opening mode

	switch (fOpenMode) {

	  case OPENAREA_EXISTING:

		// Fail if does not exist, otherwise open

		if (!bAreaExists) {
		  FMAERROR(NOT_FOUND);
		  goto Fail;
		} else 
Open:   if (!DoOpenAreaFiles(parea, &files, fl, 0)) {
		  goto Fail;
		}

		break;

	  case OPENAREA_ALWAYS:

		// Create if does not exist, otherwise open

		if (!bAreaExists) {
		  if (!DoCreateNewArea(parea, fl, &files))
			goto Fail;
		} else
		  goto Open;

		break;

	  case OPENAREA_CREATE_NEW:

		// Fail if exists, otherwise create new

		if (bAreaExists) {
		  FMAERROR(ALREADY_EXISTS);
		  goto Fail;
		} else
		if (!DoCreateNewArea(parea, fl, &files)) {
		  goto Fail;
		}

		break;

	  case OPENAREA_CREATE_ALWAYS:

		// Clear if exists, otherwise create new

		if (bAreaExists) {
		  if (!DoClearExistingArea(parea, fl, &files)) {
			goto Fail;
		  }
		} else 
		if (!DoCreateNewArea(parea, fl, &files)) {
		  goto Fail;
		}

		break;
	}

	ASSERT(parea->hfileJhr != -1);
	ASSERT(parea->hfileJdt != -1);
	ASSERT(parea->hfileJdx != -1);
	ASSERT(parea->hfileJlr != -1);

	// Load base header and if failed, remove area files if
	// they were just created for the new area

	if (!JamLoadJbh(parea)) {
	  if (!bAreaExists) DoDeleteAreaFiles(parea, &files);
	  goto Fail;
	}

	// Calculate the highest message number. Note that if there are
	// no messages msgnHigh will be one less than msgnBase so inclusive
	// checks for message number validity will fail

	parea->msgnHigh = parea->msgnBase + JamGetJdxCount(parea) - 1;

	// Increment open area count in a thread safe manner

	FmaInterlockedIncrement(&fmaRtm.nOpenAreas);

	ASSERT(JamCheckArea(parea));

	// Check if caller wants to lock the area upon opening

	if (parea->fl & FOA_LOCKAREA) {
	  if (!JamLockArea(parea, TRUE)) {
		JamCloseArea(parea);
		parea = NULL;
		goto Fail;
	  }
	}

#ifdef FMA_LOG_JAM_AREA
Log("JamOpenArea      <: [%09Fp] msgnBase=%ld msgnHigh=%ld cmsgActive=%ld in %s", parea, parea->msgnBase, parea->msgnHigh, parea->cmsgActive, parea->achPath);
#endif

	return parea;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine closes area
//

  FMAAPI_(BOOL) JamCloseArea(HAREA harea)
  {
	PJAREA parea;

	// Validate parameters

	if ((parea = (PJAREA)harea) == NULL || !JamCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

#ifdef FMA_LOG_JAM_AREA
Log("\nJamCloseArea     >: [%09Fp] msgnBase=%ld msgnHigh=%ld cmsgActive=%ld in %s", parea, parea->msgnBase, parea->msgnHigh, parea->cmsgActive, parea->achPath);
#endif

	// Check if caller locked the area upon opening and
	// area is still locked and if so, unlock it

	if (parea->fl & FOA_LOCKAREA && parea->bAreaLocked) {
	  JamLockArea(parea, FALSE);
	}

	// Check if there are any open messages

	if (parea->pmsgList) {
	  FMAERROR(STILL_OPEN);
#ifdef FMA_LOG_JAM_AREA
Log("JamCloseArea      : [%09Fp] %d msg(s) still open", parea, ListCount(parea->pmsgList)); 
#endif

	  // Close all open messages

	  while (parea->pmsgList) {
		JamCloseMsg(parea->pmsgList);
	  }
	}

	// Check if we have exclusive access

	while (parea->bExclusive) {
#ifdef FMA_LOG_JAM_AREA
Log("JamCloseArea      : [%09Fp] bExclusive=%d", parea, parea->bExclusive); 
#endif
	  JamLeaveExclusive(parea);
	}

	// Check if we have area locked

	while (parea->bAreaLocked) {
#ifdef FMA_LOG_JAM_AREA
Log("JamCloseArea      : [%09Fp] bAreaLocked=%d", parea, parea->bAreaLocked); 
#endif
	  JamLockArea(parea, FALSE);
	}

	// Check if we have index caching

	while (parea->bJdxCached) {
#ifdef FMA_LOG_JAM_AREA
Log("JamCloseArea      : [%09Fp] bJdxCached=%d", parea, parea->bJdxCached); 
#endif
	  JamEndJdxCache(parea);
	}

	// Check if we have base locked

	while (parea->bBaseLocked) {
#ifdef FMA_LOG_JAM_AREA
Log("JamCloseArea      : [%09Fp] bBaseLocked=%d", parea, parea->bBaseLocked); 
#endif
	  JamLockBase(parea, FALSE);
	}

	// Destroy message area object

	DoDestroyArea(parea);

	// Decrement open area count in a thread safe manner

	FmaInterlockedDecrement(&fmaRtm.nOpenAreas);

#ifdef FMA_LOG_JAM_AREA
Log("JamCloseArea     <: [%09Fp] CLOSED", parea); 
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine validates message
//

  FMAAPI_(BOOL) JamValidateMsg(HAREA harea, MSGN msgn, PULONG pimsg)
  {
	PJAREA parea;
	JDX jdx;

	// Validate parameters

	if ((parea = (PJAREA)harea) == NULL || !JamCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check for the valid message number

	if (!GOODMSGN(msgn)) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Get the index entry for this message number and check if ok

	if (!JamGetJdx(parea, &jdx, msgn))
	  return FALSE;

	// Check if caller needs message index and calculate it

	if (pimsg) *pimsg = msgn - parea->msgnBase;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine creates message map which is array of unique message ids
//

  FMAAPI_(BOOL) JamCreateMsgMap(HAREA harea, UMSG * paumsg[], PULONG pcumsg)
  {
	ULONG cb, cumsg, iumsg = 0;
	PJAREA parea;
	UMSG * aumsg;
	MSGN msgn;
	JDX jdx;

	// Validate parameters

	if ((parea = (PJAREA)harea) == NULL || !JamCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	ASSERT(paumsg != NULL);
	ASSERT(pcumsg != NULL);

	// Check if no messages in this area and calculate message map size

	if ((cumsg = parea->msgnHigh - parea->msgnBase + 1) == 0) {
	  *paumsg = NULL; *pcumsg = 0;
	  return TRUE;
	} else
	  cb = cumsg * sizeof(UMSG);

	// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
	if (!CHECK_DATA_SIZE_LIMIT(cb)) {
	  FMAERROR(DATA_SIZE_LIMIT);
	  return FALSE;
	}
#endif

	// Allocate memory to hold message map and check if ok

	if ((aumsg = MemAlloc((int)cb, 0)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return FALSE;
	}

	// Loop through all the messages retrieving umsg ids

	for (msgn = parea->msgnBase; msgn <= parea->msgnHigh; msgn++) {
	  if (JamGetJdx(parea, &jdx, msgn)) {
		aumsg[iumsg++] = msgn;
	  }
	}

	ASSERT(iumsg == cumsg);

	// Set caller's pointers

	*paumsg = aumsg; *pcumsg = iumsg;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine enumerates messages
//

  FMAAPI_(BOOL) JamEnumMsgs(HAREA harea, PFNENUMMSGS pfnCallback, 
							PVOID pvCallbackData, ULONG fl)
  {
	ENUMMSGINFO emi;
	BOOL bOk = TRUE;
	ULONG cumsg, iumsg;
	UMSG * aumsg;
	PJAREA parea;
	MSGN msgn;
	JDX jdx;

	// Validate parameters

	if ((parea = (PJAREA)harea) == NULL || !JamCheckArea(parea) || !pfnCallback) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check if there are no messages in this area

	if ((parea->msgnHigh - parea->msgnBase + 1) == 0) return FALSE;

	// Prepare enum message info structure

	memset(&emi, 0, sizeof(emi));
	emi.harea = harea;

	// Check if safe enum which will work if messages area deleted

	if (fl & FENUMMSGS_SAFEENUM) {

	  // Create message map and check if ok

	  if (!JamCreateMsgMap(parea, &aumsg, &cumsg) || !cumsg)
		return FALSE;

	  // Check enumeration direction

	  if (fl & FENUMMSGS_REVERSE) {

		// Enumerate in reversed order

		for (iumsg = cumsg - 1; (LONG) iumsg >= 0; iumsg--) {
		  if (JamGetJdx(parea, &jdx, msgn = aumsg[iumsg])) {
			emi.msgn = emi.umsg = msgn; emi.imsg = msgn - parea->msgnBase;
			emi.offs = jdx.lmh; emi.hash = jdx.crc;
			bOk = pfnCallback(&emi, pvCallbackData);
			if (!bOk) break;
		  }
		}
	  } else {

		// Enumerate in normal order

		for (iumsg = 0; iumsg < cumsg; iumsg++) {
		  if (JamGetJdx(parea, &jdx, msgn = aumsg[iumsg])) {
			emi.msgn = emi.umsg = msgn; emi.imsg = msgn - parea->msgnBase;
			emi.offs = jdx.lmh; emi.hash = jdx.crc;
			bOk = pfnCallback(&emi, pvCallbackData);
			if (!bOk) break;
		  }
		}
	  }

	  // Free message map

	  MemFree(aumsg);

	} else {

	  // Check enumeration direction

	  if (fl & FENUMMSGS_REVERSE) {

		// Enumerate in reversed order

		for (msgn = parea->msgnHigh; msgn >= parea->msgnBase; msgn--) {
		  if (JamGetJdx(parea, &jdx, msgn)) {
			emi.msgn = emi.umsg = msgn; emi.imsg = msgn - parea->msgnBase;
			emi.offs = jdx.lmh; emi.hash = jdx.crc;
			bOk = pfnCallback(&emi, pvCallbackData);
			if (!bOk) break;
		  }
		}
	  } else {

		// Enumerate in normal order

		for (msgn = parea->msgnBase; msgn <= parea->msgnHigh; msgn++) {
		  if (JamGetJdx(parea, &jdx, msgn)) {
			emi.msgn = emi.umsg = msgn; emi.imsg = msgn - parea->msgnBase;
			emi.offs = jdx.lmh; emi.hash = jdx.crc;
			bOk = pfnCallback(&emi, pvCallbackData);
			if (!bOk) break;
		  }
		}
	  }
	}

	// Return whatever last callback bOkurned

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes message
//

  FMAAPI_(BOOL) JamDeleteMsg(HAREA harea, MSGN msgn)
  {
	BOOL bAlreadyDeleted = FALSE;
	BOOL bOk = FALSE;
	PJAREA parea;
	JDX jdx;
	JMH jmh;

	// Validate parameters

	if ((parea = (PJAREA)harea) == NULL || !JamCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

#ifdef FMA_LOG_JAM_AREA
Log("\nJamDeleteMsg     >: [%09Fp] msgn=%ld, msgnBase=%ld msgnHigh=%ld cmsgActive=%ld in %s", parea, msgn, parea->msgnBase, parea->msgnHigh, parea->cmsgActive, parea->achPath);
#endif

	// Check message number

	if (!GOODMSGN(msgn)) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Check if this message is open and leave if so

	if (JamGetOpenMsg(parea, msgn)) {
	  FMAERROR(STILL_OPEN);
	  return FALSE;
	}

	// Acquire exclusive access and check if ok

	if (!JamEnterExclusive(parea))
	  return FALSE;

	// Load the idnex entry and check if ok

	if (!JamGetJdx(parea, &jdx, msgn)) 
	  goto Out;

	ASSERT(jdx.lmh == (ULONG)-1 || jdx.lmh >= (ULONG)sizeof(JBH));
	ASSERT(jdx.lmh == (ULONG)-1 || jdx.lmh <= (ULONG)FmaGetFileSize(parea->hfileJhr) - sizeof(JMH));

	// Verify the message header 

	if (!JamCheckLmh(parea, jdx.lmh)) {
	  FMAERROR(INVALID_INDEX);
	  goto Out;
	}

	// Load the message header

	if (!JamLoadJmh(parea, &jmh, NULL, jdx.lmh)) 
	  goto Out;

	// Check if message is not already deleted and
	// update message header setting deleted flag

	if (jmh.attr & JMA_DELETED) {
	  bAlreadyDeleted = TRUE;
	} else {
	  SETFLAG(jmh.attr, JMA_DELETED);
	  if (!JamSaveJmh(parea, &jmh, NULL, jdx.lmh))
		goto Out;
	}

	// Update the message index entry

	jdx.crc = (ULONG)-1;
	if (!JamSetJdx(parea, &jdx, msgn))
	  goto Out;

	// Adjust active message count

	if (!bAlreadyDeleted) { 
	  ASSERT((LONG)parea->cmsgActive > 0);
	  parea->cmsgActive--;
	}

	// If we got here everything is fine

	bOk = TRUE;

	// Relinquish exclusive access and check if ok

Out:if (!JamLeaveExclusive(parea))
	  bOk = FALSE;

#ifdef FMA_LOG_JAM_AREA
Log("JamDeleteMsg     <: [%09Fp] msgn=%ld, msgnBase=%ld msgnHigh=%ld cmsgActive=%ld in %s", parea, msgn, parea->msgnBase, parea->msgnHigh, parea->cmsgActive, parea->achPath);
#endif

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets current message
//

  FMAAPI_(BOOL) JamSetCurMsg(HAREA harea, MSGN msgn)
  {
	PJAREA parea;
	JDX jdx;

	// Validate parameters

	if ((parea = (PJAREA)harea) == NULL || !JamCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check message number

	if (!GOODMSGN(msgn)) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Get the message index entry and check if ok

	if (!JamGetJdx(parea, &jdx, msgn))
	  return FALSE;

	// Set the current message cache

	parea->msgnCur = msgn;
	parea->lmhCur = jdx.lmh;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets current message
//

  FMAAPI_(MSGN) JamGetCurMsg(HAREA harea)
  {
	PJAREA parea;

	// Validate parameters

	if ((parea = (PJAREA)harea) == NULL || !JamCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Return current message

	return parea->msgnCur;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets first message number
//

  FMAAPI_(MSGN) JamGetFirstMsg(HAREA harea)
  {
	PJAREA parea;

	// Validate parameters

	if ((parea = (PJAREA)harea) == NULL || !JamCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Return first message number

	return (parea->msgnHigh >= parea->msgnBase) ? parea->msgnBase : 0;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets highest message number
//

  FMAAPI_(MSGN) JamGetHighMsg(HAREA harea)
  {
	PJAREA parea;

	// Validate parameters

	if ((parea = (PJAREA)harea) == NULL || !JamCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Return highest message number

	return parea->msgnHigh;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets message count
//

  FMAAPI_(LONG) JamGetMsgCount(HAREA harea)
  {
	PJAREA parea;

	// Validate parameters

	if ((parea = (PJAREA)harea) == NULL || !JamCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Return message count

	return parea->msgnHigh - parea->msgnBase + 1;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine converts message number to unique message id
//

  FMAAPI_(UMSG) JamMsgnToUid(HAREA harea, MSGN msgn)
  {
	PJAREA parea;
	JDX jdx;

	// Validate parameters

	if ((parea = (PJAREA)harea) == NULL || !JamCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check for the valid message number

	if (!GOODMSGN(msgn)) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Get the index entry for this message number and check if ok

	if (!JamGetJdx(parea, &jdx, msgn))
	  return FALSE;

	// Return unique message id which is the same as msgn for jam

	return msgn;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine converts unique message id to message number
//

  FMAAPI_(MSGN) JamUidToMsgn(HAREA harea, UMSG umsg, UINT fMode)
  {
	MSGN msgn = umsg;
	PJAREA parea;
	JDX jdx;

	// Validate parameters

	if ((parea = (PJAREA)harea) == NULL || !JamCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Get the index entry for this message number and if ok,
	// return message number if caller wants an exact one

	if (JamGetJdx(parea, &jdx, msgn)) {
	  if (fMode & UMSG_EXACT || !fMode)
		return msgn;
	}

	// Check if the caller wants next or previous message

	if (fMode & UMSG_NEXT) {
	  if (msgn < parea->msgnBase) 
		return parea->msgnBase;

	  while (++msgn <= parea->msgnHigh) {
		if (JamGetJdx(parea, &jdx, msgn))
		  return msgn;
	  }
	} else
	if (fMode & UMSG_PREV) {
	  if (msgn > parea->msgnHigh) 
		return parea->msgnHigh;

	  while (--msgn >= parea->msgnBase) {
		if (JamGetJdx(parea, &jdx, msgn))
		  return msgn;
	  }
	} 

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine validates message area object
//

  FMAAPI_(BOOL) JamCheckArea(PJAREA parea)
  {
	ASSERT(parea != NULL);

	// Check the message area object magic number

	if (parea->magic != JAREA_MAGIC)
	  return FALSE;

	ASSERT(JamCheckJbh(parea));

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of JamArea.c
/////////////////////////////////////////////////////////////////////////////
