/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: JAM base integrity check routines module
//
// JamCheck.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 05/Oct/1999
// Updated: 05/Oct/1999
//
// History: 05/Oct/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "JamApi.h"
#include "JamApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine checks jam message header subfields
//

  static BOOL FMASUB DoCheckJsf(PJAREA parea, LONG lmh)
  {
	PJSF pjsf, pjsfFirst = NULL;
	BOOL bOk = TRUE;
	ULONG cbjsf;
	JMH jmh;

	// Load the message header header with subfields and check if ok

	if (!JamLoadJmh(parea, &jmh, &pjsfFirst, lmh)) {
#ifdef FMA_LOG
Log("JamCheckBase -----: lmh=%08lx -- invalid jmh", lmh);
#endif
	  return FALSE;
	} else {
	  pjsf = pjsfFirst;
	  cbjsf = jmh.cbjsf;
	}

	// Loop through all the subfields

	while (pjsf && cbjsf) {

	  // Check subfield length

	  if (sizeof(JSF) + pjsf->cb > cbjsf) {
#ifdef FMA_LOG
Log("JamCheckBase -----: lmh=%08lx -- invalid subfield cb=%ld, id=%d", lmh, pjsf->cb, pjsf->id);
#endif
		bOk = FALSE;
		break;
	  }

	  // Advance to the next subfield if any

	  cbjsf-= sizeof(JSF) + pjsf->cb;
	  pjsf = (PJSF)(PJSFDATA(pjsf) + pjsf->cb);
	}

	// Free memory allocated for message header subfields

	if (pjsfFirst) MemFree(pjsfFirst);

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks jam header file
//

  static BOOL FMASUB DoCheckJhr(PJAREA parea)
  {
	ULONG cbJhrFile = FmaGetFileSize(parea->hfileJhr);
	ULONG cbJdtFile = FmaGetFileSize(parea->hfileJdt);
	ULONG cmsgActive = 0;
	BOOL bOk = TRUE;
	LONG lmh;
	JMH jmh;

	ASSERT(parea->hfileJhr != -1);

	// Check the header file size

	if (cbJhrFile < sizeof(JBH)) {
#ifdef FMA_LOG
Log("JamCheckBase -----: invalid jhr file size=%ld", cbJhrFile);
#endif
	  return FALSE;
	} else
	  lmh = sizeof(JBH);

	// Loop through all the message headers

	while (lmh < (LONG)cbJhrFile) {

	  // Load message header and check if ok

	  if (!JamLoadJmh(parea, &jmh, NULL, lmh)) {
#ifdef FMA_LOG
Log("JamCheckBase -----: lmh=%08lx -- invalid jmh", lmh);
#endif
		return FALSE;
	  }

	  // Check subfields length

	  if (jmh.cbjsf && (jmh.cbjsf < sizeof(JSF) || (ULONG)lmh + jmh.cbjsf > cbJhrFile)) {
#ifdef FMA_LOG
Log("JamCheckBase -----: lmh=%08lx -- invalid cbjsf=%ld", lmh, jmh.cbjsf);
#endif
		return FALSE;
	  }

	  // Check subfields info

	  if (!DoCheckJsf(parea, lmh)) bOk = FALSE;

	  // Check message number unless message is deleted

	  if (!(jmh.attr & JMA_DELETED) && (jmh.msgn < parea->msgnBase || jmh.msgn > parea->msgnHigh)) {
#ifdef FMA_LOG
Log("JamCheckBase -----: lmh=%08lx -- invalid msgn=%ld, msgnBase=%ld, msgnHigh=%ld", lmh, jmh.msgn, parea->msgnBase, parea->msgnHigh);
#endif
		bOk = FALSE;
	  }

	  // Check text offset and length

	  if (jmh.offText > cbJdtFile) {
#ifdef FMA_LOG
Log("JamCheckBase -----: lmh=%08lx -- invalid offText=%08lx, cbJdtFile=%08lx", lmh, jmh.offText, cbJdtFile);
#endif
		bOk = FALSE;
	  } else
	  if (jmh.offText + jmh.cchText > cbJdtFile) {
#ifdef FMA_LOG
Log("JamCheckBase -----: lmh=%08lx -- invalid cchText=%08lx, offText=%08lx, cbJdtFile=%08lx", lmh, jmh.cchText, jmh.offText, cbJdtFile);
#endif
		bOk = FALSE;
	  }

	  // Check if message is not deleted and increment active message count

	  if (!(jmh.attr & JMA_DELETED)) cmsgActive++;

	  // Advance to the next message header

	  lmh+= sizeof(JMH) + jmh.cbjsf;
	}

	// Check active message count

	if (cmsgActive != parea->cmsgActive) {
#ifdef FMA_LOG
Log("JamCheckBase -----: invalid jbh.cmsgActive=%ld, expected cmsgActive=%ld", parea->cmsgActive, cmsgActive);
#endif
		bOk = FALSE;
	}

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks jam index file
//

  static BOOL FMASUB DoCheckJdx(PJAREA parea)
  {
	BOOL bOk = TRUE;
	ULONG cjdx;
	MSGN msgn;
	JDX jdx;
	JMH jmh;

	// Loop through all the message numbers

	for (msgn = parea->msgnBase; msgn <= parea->msgnHigh; msgn++) {

	  // Load the index record and check if ok

	  if (!JamGetJdx(parea, &jdx, msgn)) {
#ifdef FMA_LOG
Log("JamCheckBase -----: invalid jdx, msgn=%ld", msgn);
#endif
		return FALSE;
	  }

	  // Skip entries with no associated message header

	  if (jdx.lmh == (ULONG)-1) continue;

	  // Check message header offset

	  if (!JamCheckLmh(parea, jdx.lmh)) {
#ifdef FMA_LOG
Log("JamCheckBase -----: invalid jdx.lmh=%08lx, msgn=%ld", jdx.lmh, msgn);
#endif
		bOk = FALSE;
		continue;
	  }

	  // Load message header and check if ok

	  if (!JamLoadJmh(parea, &jmh, NULL, jdx.lmh)) {
#ifdef FMA_LOG
Log("JamCheckBase -----: invalid jmh for msgn=%ld, jdx.lmh=%08lx", msgn, jdx.lmh);
#endif
		bOk = FALSE;
		continue;
	  }

	  // Check message number 

	  if (msgn != jmh.msgn) {
#ifdef FMA_LOG
Log("JamCheckBase -----: invalid jmh.msgn=%ld for msgn=%ld, jdx.lmh=%08lx", jmh.msgn, msgn, jdx.lmh);
#endif
		bOk = FALSE;
		continue;
	  }
	}

	// Check index file size

	cjdx = parea->msgnHigh - parea->msgnBase + 1;
	if (cjdx * sizeof(JDX) != (ULONG)FmaGetFileSize(parea->hfileJdx)) {
#ifdef FMA_LOG
Log("JamCheckBase -----: invalid jdx file size=%ld, expected=%ld", FmaGetFileSize(parea->hfileJdx), cjdx * sizeof(JDX));
#endif
	  return FALSE;
	}

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine checks the jam message base integrity
//

  FMAAPI_(BOOL) JamCheckBase(PSZ pszPath)
  {
	ULONG fl = FOA_LOCKAREA | FOA_DENYALL;
	BOOL bOk = TRUE;
	PJAREA parea;

	// Open area in exclusive mode and check if ok

	if ((parea = JamOpenArea(pszPath, OPENAREA_EXISTING, fl, NULL)) == NULL) {
#ifdef FMA_LOG
Log("JamCheckBase -----: Can't open %s, code=%d", pszPath, FmaGetLastError());
#endif
	  return TRUE;
	}

	// Acquire exclusive access and check if ok

	if (JamEnterExclusive(parea)) {

	  // Check message base integrity

	  if (!DoCheckJhr(parea)) bOk = FALSE;
	  if (!DoCheckJdx(parea)) bOk = FALSE;

	  // Relinquish exclusive access

	  if (!JamLeaveExclusive(parea))
		bOk = FALSE;
	}

	// Close area 

	JamCloseArea(parea);

#ifdef FMA_LOG
Log("JamCheckBase -----: %s integrity check %s", pszPath, bOk ? "passed" : "FAILED");
#endif

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// End of JamCheck.c
/////////////////////////////////////////////////////////////////////////////
