/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: JAM helper routines module
//
// JamHelp.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 19/Sep/1999
// Updated: 18/Feb/2000
//
// History: 19/Sep/1999 kvitekp - Initially created.
//
//          10/Feb/2000 kvitekp - Added JamAddJsfFormatV()
//
//          18/Feb/2000 kvitekp - Added JamSet/CheckPassword()
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "JamApi.h"
#include "JamApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine loads JAM message base header
//

  FMAAPI_(BOOL) JamLoadJbh(PJAREA parea)
  {
	ULONG cbRead;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileJhr != -1);

	// Seek to the beginning of the header file and load the base header

	if (!FmaReadFile(parea->hfileJhr, 0, &parea->jbh, sizeof(JBH), &cbRead)) {
	  return FALSE;
	} else 
	  if (cbRead != sizeof(JBH)) {
		FMAERROR(INVALID_DATA);
		return FALSE;
	  }

	// Check the JAM base header

	if (!JamCheckJbh(parea)) {
	  return FALSE;
	}

	// Update area info

	parea->cmsgActive = parea->jbh.cmsgActive;
	parea->msgnBase = parea->jbh.msgnBase ? parea->jbh.msgnBase : 1;

#ifdef FMA_LOG_JAM_HELP
Log("JamLoadJbh        : [%09Fp] bBaseLocked=%d bExclusive=%d cmsgActive=%ld msgnBase=%ld", parea, parea->bBaseLocked, parea->bExclusive, parea->jbh.cmsgActive, parea->jbh.msgnBase);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves JAM message base header
//

  FMAAPI_(BOOL) JamSaveJbh(PJAREA parea)
  {
	ULONG cbWritten;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileJhr != -1);
	ASSERT(parea->bExclusive || !parea->jbh.cmod);

	// Update area info and increment modifications counter

	parea->jbh.cmsgActive = parea->cmsgActive;
	parea->jbh.msgnBase = parea->msgnBase ? parea->msgnBase : 1;
	parea->jbh.cmod++;

	// Check the JAM base header

	if (!JamCheckJbh(parea)) 
	  return FALSE;

	// Seek to the beginning of the header file and save the base header

	if (!FmaWriteFile(parea->hfileJhr, 0, &parea->jbh, sizeof(JBH), &cbWritten))
	  return FALSE;

	ASSERT(cbWritten == sizeof(JBH));

#ifdef FMA_LOG_JAM_HELP
Log("JamSaveJbh        : [%09Fp] bBaseLocked=%d bExclusive=%d cmsgActive=%ld msgnBase=%ld", parea, parea->bBaseLocked, parea->bExclusive, parea->jbh.cmsgActive, parea->jbh.msgnBase);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks JAM message base header
//

  FMAAPI_(BOOL) JamCheckJbh(PJAREA parea)
  {
	ASSERT(parea != NULL);

	// Check JAM base header magic

	if (parea->jbh.magic != JAM_MAGIC) goto Fail;

	return TRUE;

	// Message base is corrupted

Fail: FMAERROR(INVALID_DATA);

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine resets the current message cache
//

  FMAAPI_(BOOL) JamResetCurMsgCache(PJAREA parea)
  {
	ASSERT(parea != NULL);

	// Reset current message cache

	parea->msgnCur = 0;
	parea->lmhCur = 0;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets the current message cache
//

  FMAAPI_(BOOL) JamSetCurMsgCache(PJMSG pmsg)
  {
	PJAREA parea;

	ASSERT(pmsg != NULL);

	parea = pmsg->parea;

	// Set current message cache

	parea->msgnCur = pmsg->msgn;
	parea->lmhCur = pmsg->lmh;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks message header offset
//

  FMAAPI_(BOOL) JamCheckLmh(PJAREA parea, LONG lmh)
  {
	ASSERT(parea != NULL);
	ASSERT(parea->hfileJhr != -1);

	ASSERT(lmh == -1 || lmh >= (LONG)sizeof(JBH));
	ASSERT(lmh == -1 || lmh < FmaGetFileSize(parea->hfileJhr));

	// Check if specified message header offset is valid

	if (lmh == -1) return FALSE;
	if (lmh < (LONG)sizeof(JBH)) return FALSE;
	if (lmh > FmaGetFileSize(parea->hfileJhr)) return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads message header
//

  FMAAPI_(BOOL) JamLoadJmh(PJAREA parea, PJMH pjmh, PJSF * ppjsf, LONG lmh)
  {
	ULONG cbjsf, cbRead;
	PJSF pjsf = NULL;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileJhr != -1);

	ASSERT(pjmh != NULL);

	ASSERT(JamCheckLmh(parea, lmh));

	// Read the message header and check if ok

	if (!FmaReadFile(parea->hfileJhr, lmh, pjmh, sizeof(JMH), &cbRead)) {
	  return FALSE;
	} else 
	  if (cbRead != sizeof(JMH)) {
		FMAERROR(INVALID_DATA);
		return FALSE;
	  }

	// Check if we've got a correct magic

	if (pjmh->magic != JAM_MAGIC) {
	  FMAERROR(INVALID_DATA);
	  return FALSE;
	}

	// Check if caller does needs subfields and leave now,
	// othrewise assume no subfields present

	if (!ppjsf) return TRUE;
	else *ppjsf = NULL;

	// Check if there are no subfields for this message and leave

	if ((cbjsf = pjmh->cbjsf) == 0) 
	  return TRUE;

	// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
	if (!CHECK_DATA_SIZE_LIMIT(cbjsf)) {
	  FMAERROR(DATA_SIZE_LIMIT);
	  return FALSE;
	}
#endif

	// Allocate memory to hold the subfields and check if ok

	if ((pjsf = MemAlloc((int)cbjsf, 0)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return FALSE;
	}

	// Read the subfields into the allocated memory

	if (!FmaReadFile(parea->hfileJhr, lmh + sizeof(JMH), pjsf, cbjsf, &cbRead)) {
Fail: MemFree(pjsf);
	  return FALSE;
	} else 
	  if (cbRead != cbjsf) {
		FMAERROR(INVALID_DATA);
		goto Fail;
	  }

	// Pass the pointer to caller

	*ppjsf = pjsf;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message header
//

  FMAAPI_(BOOL) JamSaveJmh(PJAREA parea, PJMH pjmh, PJSF pjsf, LONG lmh)
  {
	ULONG cbjsf, cbWritten;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileJhr != -1);

	ASSERT(pjmh != NULL);
	ASSERT(pjmh->magic == JAM_MAGIC);
	ASSERT(pjmh->msgn > 0);

	ASSERT(JamCheckLmh(parea, lmh));

	// Writne the message header and check if ok

	if (!FmaWriteFile(parea->hfileJhr, lmh, pjmh, sizeof(JMH), &cbWritten))
	  return FALSE;

	ASSERT(cbWritten == sizeof(JMH));

	// Write the subfields and check if ok

	if (pjsf && (cbjsf = pjmh->cbjsf) > 0) {

	  if (!FmaWriteFile(parea->hfileJhr, lmh + sizeof(JMH), pjsf, cbjsf, &cbWritten))
		return FALSE;

	  ASSERT(cbWritten == cbjsf);
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets the jam subfield with specified id and index
//

  FMAAPI_(PJSF) JamGetJsf(PJSF pjsf, ULONG cbjsf, USHORT id, LONG jsfIndex)
  {
	// Loop through all the subfields

	while (pjsf && cbjsf) {

#ifdef FMA_LOG_JAM_HELP
//Log("JamGetJsf         : [%09Fp] id=%d cb=%ld '%.*s'", pjsf, pjsf->id, pjsf->cb, (UINT)pjsf->cb, PJSFDATA(pjsf));
#endif

	  // Check for a subfield id match and skip specified
	  // amount of such ids before returning one

	  if (pjsf->id == id && !jsfIndex--)
		return pjsf;

	  // Advance to the next subfield if any

	  cbjsf-= sizeof(JSF) + pjsf->cb;
	  pjsf = (PJSF)(PJSFDATA(pjsf) + pjsf->cb);
	}

	return NULL;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds the jam message header subfield
//

  FMAAPI_(PJSF) JamAddJsf(PJSF * ppjsf, ULONG * pcbjsf, USHORT id, PVOID pb, ULONG cb)
  {
	LONG cbjsf;
	PJSF pjsf;

	ASSERT(ppjsf != NULL);
	ASSERT(pcbjsf != NULL);

	// Check if subfield is an asciiz string and get its length

	if (cb == (ULONG)-1) cb = pb ? strlen(pb) : 0;

	// Make sure old subfields size is null if no subfields present yet

	if (!*ppjsf) *pcbjsf = 0;

	// Calculate size of the memroty block to hold the subfield

	cbjsf = *pcbjsf + sizeof(JSF) + cb;

	// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
	if (!CHECK_DATA_SIZE_LIMIT(cbjsf)) {
	  FMAERROR(DATA_SIZE_LIMIT);
	  return FALSE;
	}
#endif

	// Reallocate subfield memory block and check if ok

	if ((pjsf = MemRealloc(*ppjsf, (int)cbjsf)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return NULL;
	} else
	  *ppjsf = pjsf;

	// Caluclate pointer to the new subfield

	pjsf = (PJSF)((PBYTE)pjsf + *pcbjsf);

	// Make up the new subfield header

	pjsf->id = id;
	pjsf->id2 = 0;
	pjsf->cb = cb;

	// Make up the new subfield data

	if (pb)
	  memcpy(PJSFDATA(pjsf), pb, (int)cb);
	else
	  memset(PJSFDATA(pjsf), 0, (int)cb);

#ifdef FMA_LOG_JAM_HELP
//Log("JamAddJsf         : [%09Fp] id=%d cb=%ld '%.*s'", pjsf, pjsf->id, pjsf->cb, (UINT)pjsf->cb, PJSFDATA(pjsf));
#endif

	// Set new subfields size for caller

	*pcbjsf = cbjsf;

	// Return pointer to newly added subfield 

	return pjsf;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds the jam formatted message header subfield given var arg
//

  FMAAPIV_(PJSF) JamAddJsfFormat(PJSF * ppjsf, ULONG * pcbjsf, USHORT id, 
								 PSZ pszFormat, ...)
  {
	va_list parg;
	va_start(parg, pszFormat);

	// Fall to arg ptr backend routine to do the job

	return JamAddJsfFormatV(ppjsf, pcbjsf, id, pszFormat, parg);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine adds the jam formatted message header subfield given arg ptr
//

  FMAAPI_(PJSF) JamAddJsfFormatV(PJSF * ppjsf, ULONG * pcbjsf, USHORT id, 
								 PSZ pszFormat, va_list parg)
  {
	DSZ dsz = DSZ_INIT;
	PJSF pjsf;

	// Format the subfield into dynamic string and check if ok

	if (!AddDszFormatV(&dsz, pszFormat, parg)) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return NULL;
	}

	// Add jam message header subfield

	pjsf = JamAddJsf(ppjsf, pcbjsf, id, dsz.psz, dsz.ich);

	// Free dynamic string

	DelDsz(&dsz);

	return pjsf;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine deletes jam message header subfield with specified id
//

  FMAAPI_(BOOL) JamDelJsf(PJSF * ppjsf, ULONG * pcbjsf, USHORT id)
  {
	LONG cbjsf, cbdel;
	BOOL bOk = FALSE;
	PJSF pjsf;

	ASSERT(ppjsf != NULL);
	ASSERT(pcbjsf != NULL);

	// Loop while we have subfields with specified id moving
	// the subsequent subfields over the deleted ones

	while ((pjsf = JamGetJsf(*ppjsf, *pcbjsf, id, 0)) != NULL) {

#ifdef FMA_LOG_JAM_HELP
//Log("JamDelJsf         : [%09Fp] id=%d cb=%ld '%.*s'", pjsf, pjsf->id, pjsf->cb, (UINT)pjsf->cb, PJSFDATA(pjsf));
#endif

	  cbdel = (sizeof(JSF) + pjsf->cb);
	  cbjsf = *pcbjsf - (UINT)((PBYTE)pjsf - (PBYTE)*ppjsf) - cbdel;
	  memmove(pjsf, PJSFDATA(pjsf) + pjsf->cb, (int)cbjsf);
	  *pcbjsf-= cbdel;
	  bOk = TRUE;
	}

	// Check if we deleted any subfields and shrink buffer

	if (bOk) {
	  if (!*pcbjsf) {
		MemFree(*ppjsf); *ppjsf = NULL;
	  } else {
		*ppjsf = MemRealloc(*ppjsf, (int)*pcbjsf);
		ASSERT(*ppjsf != NULL);
	  }
	}

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns index of the given message in the open messages list
//

  PJMSG FMAAPI JamGetOpenMsg(PJAREA parea, MSGN msgn)
  {
	PJMSG pmsg;

	ASSERT(parea  != NULL);

	// Scan through the open messages list looking for a match

	for (pmsg = parea->pmsgList; pmsg != NULL; pmsg = pmsg->pmsgNext) {
	  if (pmsg->msgn == msgn)
		return pmsg;
	}

	return NULL;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns updated jam crc value for the given string
//

  FMAAPI_(ULONG) JamGetCrc32(PCH pch, ULONG cch, ULONG crc)
  {
	return FmaGetCrc32((PBYTE)pch, cch, crc, TRUE);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets JAM area password
//

  FMAAPI_(BOOL) JamSetPassword(PJAREA parea, PSZ pszPassword)
  {
	ASSERT(parea != NULL);

	parea->jbh.crcp = JamGetCrc32(pszPassword, (ULONG)-1, (ULONG)-1);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks JAM area password
//

  FMAAPI_(BOOL) JamCheckPassword(PJAREA parea, PSZ pszPassword)
  {
	ASSERT(parea != NULL);

	return parea->jbh.crcp == JamGetCrc32(pszPassword, (ULONG)-1, (ULONG)-1);
  }

/////////////////////////////////////////////////////////////////////////////
// End of JamHelp.c
/////////////////////////////////////////////////////////////////////////////
