/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: JAM lastread routines module
//
// JamLast.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 03/Oct/1999
// Updated: 03/Oct/1999
//
// History: 03/Oct/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "JamApi.h"
#include "JamApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine gets jam lastread file record for the specified uid
//

  FMAAPI_(BOOL) JamGetJlr(PJAREA parea, ULONG uid, PJLR pjlr, LONG * poffs)
  {
	ULONG cjlr, cbRead;
	LONG offs = 0;
	JLR jlr;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileJlr != -1);

	// Get number of lastread records

	cjlr = FmaGetFileSize(parea->hfileJlr) / sizeof(JLR);

	// Loop reading lastread file records looking for uid match

	while (cjlr--) {

	  // Read lastread file record and check if ok

	  if (!FmaReadFile(parea->hfileJlr, offs, &jlr, sizeof(JLR), &cbRead)) {
		return FALSE;
	  } else 
		if (cbRead != sizeof(JLR)) {
		  FMAERROR(INVALID_DATA);
		  return FALSE;
		}

	  // Check for the uid match and leave if so setting the callers data

	  if (jlr.uid == uid) {
		if (poffs) *poffs = offs;
		if (pjlr) *pjlr = jlr;
		return TRUE;
	  }

	  // Advance to the next altread file record

	  offs+= sizeof(JLR);
	}

	// If we got here the record for this uid does not exist

	FMAERROR(NOT_FOUND);

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets jam lastread file record for the specified uid
//

  FMAAPI_(BOOL) JamSetJlr(PJAREA parea, PJLR pjlr)
  {
	ULONG cbWritten;
	BOOL bOk = TRUE;
	LONG offs;
	JLR jlr;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileJlr != -1);

	ASSERT(pjlr != NULL);

	// Acquire exclusive access and check if ok

	if (!JamEnterExclusive(parea))
	  return FALSE;

	// Check if there is record for this uid and if not,
	// create one at eof

	if (JamGetJlr(parea, pjlr->uid, &jlr, &offs)) {
	  ASSERT(jlr.uid == pjlr->uid);

	  // Check if existing record has the same user name crc

	  if (jlr.crc != pjlr->crc) {
		FMAERROR(ALREADY_EXISTS);
		bOk = FALSE;
		goto Out;
	  }

	} else
	  offs = FmaGetFileSize(parea->hfileJlr);

	// Write lastread file record and check if ok

	if (FmaWriteFile(parea->hfileJlr, offs, pjlr, sizeof(JLR), &cbWritten)) {
	  ASSERT(cbWritten == sizeof(JLR));
	} else
	  bOk = FALSE;

	// Relinquish exclusive access and check if ok

Out:if (!JamLeaveExclusive(parea))
	  bOk = FALSE;

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets lastread info
//

  FMAAPI_(BOOL) JamGetLRInfo(HAREA harea, ULONG uid, PLRINFO plri)
  {
	PJAREA parea;
	JLR jlr;

	// Validate parameters

	if ((parea = (PJAREA)harea) == NULL || !JamCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Get lastread file record and check if ok

	if (JamGetJlr(parea, uid, &jlr, NULL)) {
	  ASSERT(jlr.uid == uid);
	} else
	  return FALSE;

	// Pass info to the caller

	if (plri) {
	  plri->umsgLast = jlr.msgnLast;
	  plri->umsgHigh = jlr.msgnHigh;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets lastread info
//

  FMAAPI_(BOOL) JamSetLRInfo(HAREA harea, ULONG uid, PLRINFO plri, PSZ pszUserName)
  {
	PJAREA parea;
	JLR jlr;

	// Validate parameters

	if ((parea = (PJAREA)harea) == NULL || !JamCheckArea(parea) || !plri) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Make up the lastread file record

	jlr.uid = uid;
	jlr.crc = JamGetCrc32(pszUserName, (ULONG)-1, (ULONG)-1);
	jlr.msgnLast = plri->umsgLast;
	jlr.msgnHigh = plri->umsgHigh;

	// Set the lastread file record and check if ok

	if (!JamSetJlr(parea, &jlr))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of JamLast.c
/////////////////////////////////////////////////////////////////////////////
