/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: JAM message routines module
//
// JamMsg.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 19/Sep/1999
// Updated: 24/Sep/1999
//
// History: 19/Sep/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "JamApi.h"
#include "JamApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine creates message object
//

  static PJMSG FMASUB DoCreateMsg(PJAREA parea, MSGN msgn, UINT mode, ULONG fl)
  {
	PJMSG pmsg;

	ASSERT(parea != NULL);

	// Create message object and check if ok

	if ((pmsg = MemAlloc(sizeof(JMSG), MA_CLEAR)) == NULL) {
	  FMAERROR(NOT_ENOUGH_MEMORY);
	  return NULL;
	} 

	// Initialize message object

	pmsg->magic = JMSG_MAGIC;
	pmsg->parea = parea;
	pmsg->msgn = msgn;
	pmsg->mode = mode;
	pmsg->fl = fl;

	return pmsg;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine destroys message object
//

  static BOOL FMASUB DoDestroyMsg(PJMSG pmsg)
  {
	ASSERT(pmsg != NULL);

	// Destroy associated subfields info

	if (pmsg->pjsf) MemFree(pmsg->pjsf);

	// Reset message object magic number

	pmsg->magic = ~JMSG_MAGIC;

	// Free message object

	MemFree(pmsg);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine translates special message numbers
//

  static BOOL FMASUB DoXlateMsgn(PJAREA parea, MSGN * pmsgn)
  {
	MSGN msgn = *pmsgn;

	ASSERT(parea != NULL);

	// Select current message

	if (msgn == MSGN_CUR || msgn == MSGN_CURORPREV || msgn == MSGN_CURORNEXT) {

		if (parea->msgnCur < parea->msgnBase || 
			parea->msgnCur > parea->msgnHigh) {
		if (msgn == MSGN_CURORPREV) goto Prev;
		if (msgn == MSGN_CURORNEXT) goto Next;
NotF:   FMAERROR(NOT_FOUND);
		return FALSE;
	  } else {
		msgn = parea->msgnCur;
	  }
	} else

	// Select previous message

	if (msgn == MSGN_PREV) {

Prev: if (parea->msgnCur > parea->msgnHigh)
		msgn = parea->msgnHigh;
	  if (parea->msgnCur > parea->msgnBase) {
		msgn = parea->msgnCur - 1;
	  } else
		goto NotF;
	} else

	// Select next message

	if (msgn == MSGN_NEXT) {

Next: if (parea->msgnCur < parea->msgnBase) {
		msgn = parea->msgnBase;
	  } else 
	  if (parea->msgnCur < parea->msgnHigh) {
		msgn = parea->msgnCur + 1;
	  } else 
		goto NotF;
	}

	// Set message number for caller

	if (pmsgn) *pmsgn = msgn;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine fixes the new message at the end of the base
//

  static BOOL FMASUB DoFixNewMsg(PJMSG pmsg)
  {
	PJAREA parea = pmsg->parea;
	JDX jdx = JDX_INIT_EMPTY;
	MSGN msgn;

	ASSERT(parea->bExclusive);

	// Add empty index entry and check if ok

	if (!JamAddJdx(parea, &jdx, &msgn))
	  return FALSE;

	// Check if index is in sync and fail if not

	if (msgn != pmsg->msgn) {
#ifdef FMA_LOG_JAM_MSG
Log("DoFixNewMsg       : [%09Fp] msgn=%ld pmsg->msgn=%ld msgnHigh=%ld error=%d", pmsg, msgn, pmsg->msgn, parea->msgnHigh, FmaGetLastError());
#endif
	  JamDelJdx(parea, &jdx, msgn);
	  FMAERROR(INVALID_INDEX);
	  return FALSE;
	}

	// That's it... there is no actual message yet, it will be allocated 
	// later during the call to JamWriteMsg()

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine fixes the old message
//

  static BOOL FMASUB DoFixOldMsg(PJMSG pmsg)
  {
	PJAREA parea = pmsg->parea;
	JDX jdx;

	ASSERT(parea->bExclusive);

	// Get the message index entry and check if ok

	if (!JamGetJdx(parea, &jdx, pmsg->msgn))
	  return FALSE;

	ASSERT(jdx.lmh == (ULONG)-1 || jdx.lmh >= (ULONG)sizeof(JBH));
	ASSERT(jdx.lmh == (ULONG)-1 || jdx.lmh <= (ULONG)FmaGetFileSize(parea->hfileJhr) - sizeof(JMH));

	// Check if the message header offset is valid and 
	// save it in message handle, otherwise reset it so
	// that new space will be allocated

	if (!JamCheckLmh(parea, jdx.lmh)) {
	  pmsg->lmh = 0;
	} else
	  pmsg->lmh = jdx.lmh;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine opens new message
//

  static BOOL FMASUB DoOpenNewMsg(PJMSG pmsg)
  {
	PJAREA parea = pmsg->parea;
	BOOL bNew, bOk = TRUE;

	ASSERT(parea->bExclusive);

	// Set up the new message flag

	bNew = !pmsg->msgn || (pmsg->msgn > parea->msgnHigh);

	// Make up the new message number

	if (bNew) pmsg->msgn = parea->msgnHigh + 1;

	// Fix up the index and data files to indicate that 
	// message creation is in process

	if (bOk) bOk = bNew ? DoFixNewMsg(pmsg) : DoFixOldMsg(pmsg);

	// Increment number of messages and highest message number

	if (bOk && bNew) parea->msgnHigh++;

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine opens existing message
//

  static BOOL FMASUB DoOpenExistingMsg(PJMSG pmsg)
  {
	PJAREA parea = pmsg->parea;
	JDX jdx;

	ASSERT(pmsg->mode != OPENMSG_CREATE);

	// Check the message number and if invalid, reset current message cache 
	// if caller wants null message and leave

	if (pmsg->msgn < parea->msgnBase || 
		pmsg->msgn > parea->msgnHigh) {
	  if (!pmsg->msgn) JamResetCurMsgCache(parea);
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Get message index entry and check if ok

	if (!JamGetJdx(parea, &jdx, pmsg->msgn))
	  return FALSE;

	// Set message header offset into the message structure

	pmsg->lmh = jdx.lmh;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine fixes new message which has not been written
//

  static BOOL FMASUB DoFixStrayNewMsg(PJMSG pmsg)
  {
	PJAREA parea = pmsg->parea;
	JDX jdx = JDX_INIT_EMPTY;
	BOOL bOk = TRUE;
	LONG offEnd;
	JMH jmh;

	ASSERT(parea->bExclusive);
	ASSERT(pmsg->mode == OPENMSG_CREATE);

	// Get the associated index entry

	if (!JamGetJdx(parea, &jdx, pmsg->msgn)) {
	  jdx.crc = jdx.lmh = (ULONG)-1;
	}

	// Check if writing frame has been allocated but the message header has 
	// not been written to and thus contains invalid message header

	if ((pmsg->fs & JMF_GOTFRAME) && !(pmsg->fs & JMF_MSGHWRITTEN)) {

	  // Check if message header writing frame has been allocated 
	  // at the end of jhr file and if so, chop it

	  offEnd = pmsg->lmh + sizeof(JMH) + pmsg->cbjsfMax;
	  if (offEnd == FmaGetFileSize(parea->hfileJhr)) {
		if (FmaSetFileSize(parea->hfileJhr, pmsg->lmh)) {
#ifdef FMA_LOG_JAM_READ
Log("DoFixStrayNewMsg  : [%09Fp] msgn=%ld msgnHigh=%ld -- chopped jhr at %ld", pmsg, pmsg->msgn, parea->msgnHigh, pmsg->lmh);
#endif
		} else
		  bOk = FALSE;
	  } else 
	  if (offEnd < FmaGetFileSize(parea->hfileJhr)) {

		// Make up the deleted message header

		memset(&jmh, 0, sizeof(JMH));
		SETFLAG(jmh.attr, JMA_DELETED);
		jmh.cbjsf = pmsg->cbjsfMax; jmh.msgn = pmsg->msgn;
		jmh.magic = JAM_MAGIC; jmh.revLevel = JAM_REVLEVEL;

		// Write deleted message header with registered subfields size to
		// preserve header file integrity

		if (JamSaveJmh(parea, &jmh, NULL, pmsg->lmh)) {
		  jdx.lmh = pmsg->lmh;
#ifdef FMA_LOG_JAM_READ
Log("DoFixStrayNewMsg  : [%09Fp] msgn=%ld msgnHigh=%ld -- wrote deleted jmh", pmsg, pmsg->msgn, parea->msgnHigh);
#endif
		} else 
		  bOk = FALSE;
	  }

	  // Check if message text writing frame has been allocated 
	  // at the end of jdt file and chop it if so

	  offEnd = pmsg->lmt + pmsg->lmtMax;
	  if (offEnd == FmaGetFileSize(parea->hfileJdt)) {
		FmaSetFileSize(parea->hfileJdt, pmsg->lmt);
	  }

	  // Check if creating over existing message and adjust active message count

	  if (jdx.lmh != (ULONG)-1 && !(pmsg->fs & JMF_READDELETED)) {
		parea->cmsgActive--;
	  }
	}

	// Fix up the associated index entry

	if (jdx.lmh != (ULONG)-1) {
	  jdx.crc = jdx.lmh = (ULONG)-1;
	  if (!JamSetJdx(parea, &jdx, pmsg->msgn)) 
		bOk = FALSE;
	}

#ifdef FMA_LOG_JAM_READ
Log("DoFixStrayNewMsg  : [%09Fp] msgn=%ld msgnHigh=%ld cmsgActive=%ld", pmsg, pmsg->msgn, parea->msgnHigh, parea->cmsgActive);
#endif

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns index of the given message in the open messages list
//

  static BOOL FMASUB DoGetOpenMsgIndex(PJMSG pmsg, UINT * pimsg)
  {
	PJAREA parea = pmsg->parea;
	PJMSG pmsgThis;
	UINT imsg;

	// Scan through the open messages list looking for a match

	for (pmsgThis = parea->pmsgList, imsg = 0; pmsgThis != NULL; 
		 pmsgThis = pmsgThis->pmsgNext, imsg++) {
	  if (pmsgThis == pmsg) {
		if (pimsg) *pimsg = imsg;
		return TRUE;
	  }
	}

	return FALSE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine opens message
//

  FMAAPI_(HMSG) JamOpenMsg(HAREA harea, UINT mode, MSGN msgn, ULONG fl)
  {
	PJMSG pmsg = NULL;
	PJAREA parea;

	// Validate parameters

	if ((parea = (PJAREA)harea) == NULL || !JamCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
Fail: 

#ifdef FMA_LOG_JAM_MSG
Log("JamOpenMsg       <: [%09Fp] FAILED, error=%d", pmsg, FmaGetLastError());
#endif

	  // Clean up and leave

	  if (pmsg) {
		if (pmsg->fs & JMF_EXCLUSIVE) JamLeaveExclusive(parea);
		DoDestroyMsg(pmsg);
	  }

	  return NULL;
	}

#ifdef FMA_LOG_JAM_MSG
Log("\nJamOpenMsg       >: [%09Fp] msgn=%ld in %s, mode=%d", pmsg, msgn, parea->achPath, mode);
#endif

	// Translate special message numbers and check if ok

	if (!DoXlateMsgn(parea, &msgn))
	  goto Fail;

	// Create message object and check if ok

	if ((pmsg = DoCreateMsg(parea, msgn, mode, fl)) == NULL)
	  goto Fail;

	// Check if we need an exclusive access and acquire it

	if (mode == OPENMSG_CREATE || 
		mode == OPENMSG_WRITE ||
		mode == OPENMSG_READWRITE) {
	  if (JamEnterExclusive(parea)) 
		SETFLAG(pmsg->fs, JMF_EXCLUSIVE);
	  else 
		goto Fail;
	}

	// Create a new message, or open an existing message, as specified

	if (mode == OPENMSG_CREATE) {
	  if (!DoOpenNewMsg(pmsg)) 
		goto Fail;
	} else {
	  if (!DoOpenExistingMsg(pmsg)) 
		goto Fail;
	}

	// Set current message cache to this message

	JamSetCurMsgCache(pmsg);

	// Add message to the top of the list of open messages for this area

	ListAdd(&parea->pmsgList, pmsg, 0);    

	ASSERT(JamCheckMsg(pmsg));

#ifdef FMA_LOG_JAM_MSG
Log("JamOpenMsg       <: [%09Fp] msgn=%ld lmh=%08lx in %s", pmsg, pmsg->msgn, pmsg->lmh, parea->achPath);
#endif

	return pmsg;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine closes message
//

  FMAAPI_(BOOL) JamCloseMsg(HMSG hmsg)
  {
	BOOL bOk = TRUE;
	PJAREA parea;
	PJMSG pmsg;
	UINT imsg;

	// Validate parameters

	if ((pmsg = (PJMSG)hmsg) == NULL || !JamCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: 

#ifdef FMA_LOG_JAM_MSG
Log("JamCloseMsg      <: [%09Fp] FAILED, error=%d", pmsg, FmaGetLastError());
#endif

	  return FALSE;
	} else
	  parea = pmsg->parea;

#ifdef FMA_LOG_JAM_MSG
Log("\nJamCloseMsg      >: [%09Fp] msgn=%ld in %s", pmsg, pmsg->msgn, parea->achPath);
#endif

	// Check if this is a new message which has not been written to and
	// update base to forget about it

	if (pmsg->mode == OPENMSG_CREATE && !(pmsg->fs & JMF_MASKWRITTEN)) {
	  if (!DoFixStrayNewMsg(pmsg))
		bOk = FALSE;
	}

	// Check if we entered an exclusive access and relinquish it

	if (pmsg->fs & JMF_EXCLUSIVE) {
	  if (!JamLeaveExclusive(parea))
		bOk = FALSE;
	}

	// Delete message from the list of open messages for this area

	if (DoGetOpenMsgIndex(pmsg, &imsg)) {
	  ListDel(&parea->pmsgList, imsg);
	} else {
	  ASSERT(DoGetOpenMsgIndex(pmsg, &imsg));
	}

	// Destroy message object

	DoDestroyMsg(pmsg);

	// Check if failed and jump to failure return

	if (!bOk) goto Fail;

#ifdef FMA_LOG_JAM_MSG
Log("JamCloseMsg      <: [%09Fp] in %s", pmsg, parea->achPath);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine validates message object
//

  FMAAPI_(BOOL) JamCheckMsg(PJMSG pmsg)
  {
	ASSERT(pmsg != NULL);

	// Check the message object magic number

	if (pmsg->magic != JMSG_MAGIC)
	  return FALSE;

	// Check the area object in the debugging build

	ASSERT(pmsg->parea != NULL);
	ASSERT(JamCheckArea(pmsg->parea));

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks if given message is allowed to be read
//

  FMAAPI_(BOOL) JamCheckMsgReadMode(PJMSG pmsg)
  {
	ASSERT(pmsg != NULL);

	// Check if this message is allowed to be read

	if (pmsg->mode != OPENMSG_READ && pmsg->mode != OPENMSG_READWRITE) {
	  FMAERROR(CANNOT_READ);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks if given message is allowed to be written
//

  FMAAPI_(BOOL) JamCheckMsgWriteMode(PJMSG pmsg)
  {
	ASSERT(pmsg != NULL);

	// Check if this message is allowed to be written

	if (pmsg->mode != OPENMSG_CREATE && pmsg->mode != OPENMSG_WRITE &&
		pmsg->mode != OPENMSG_READWRITE) {
	  FMAERROR(CANNOT_WRITE);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of JamMsg.c
/////////////////////////////////////////////////////////////////////////////
