/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: JAM read routines module
//
// JamRead.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 19/Sep/1999
// Updated: 28/Sep/1999
//
// History: 19/Sep/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "JamApi.h"
#include "JamApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine reads message header
//

  static BOOL FMASUB DoReadJmh(PJMSG pmsg)
  {
	PJAREA parea = pmsg->parea;
	PJSF pjsf = NULL;

	ASSERT(parea->hfileJhr != -1);

	// Check if there is no message header

	if (!JamCheckLmh(parea, pmsg->lmh)) {
	  FMAERROR(INVALID_INDEX);
	  return FALSE;
	}

	// Load the message header and subfields

	if (!JamLoadJmh(parea, &pmsg->jmh, &pjsf, pmsg->lmh))
	  return FALSE;

	ASSERT(pmsg->jmh.msgn == pmsg->msgn);

	// Set message subfields info controls

	if (pmsg->pjsf) MemFree(pmsg->pjsf);
	pmsg->cbjsf = pmsg->cbjsfMax = pmsg->jmh.cbjsf;
	pmsg->pjsf = pjsf;

	// Set message text controls

	pmsg->lmt = pmsg->jmh.offText;
	pmsg->lmtEnd = pmsg->jmh.offText + pmsg->jmh.cchText;
	pmsg->lmtMax = pmsg->lmtEnd;

	// Preserve message deleted flag so that we can
	// adjust cmsgActive when writing

	SETFLAGTO(pmsg->fs, JMF_READDELETED, pmsg->jmh.attr & JMA_DELETED);

	// Set flags to prevent further header reading

	SETFLAG(pmsg->fs, JMF_GOTMSGH | JMF_GOTSUBF);

#ifdef FMA_LOG_JAM_READ
Log("DoReadJmh         : [%09Fp] msgn=%ld cbjsf=%ld cbText=%ld offText=%ld", pmsg, pmsg->jmh.msgn, pmsg->jmh.cbjsf, pmsg->jmh.cchText, pmsg->jmh.offText);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine reads message text
//

  static BOOL FMASUB DoReadText(PJMSG pmsg, PCH * ppchText, ULONG * pcchText, 
								ULONG * poffText)
  {
	PJAREA parea = pmsg->parea;
	ULONG cch, cbRead, cchAlloc;
	BOOL bAlloc = FALSE;
	LONG offs, offsEnd;

	ASSERT(parea->hfileJdt != -1);

	ASSERT(pmsg->jmh.offText <= (ULONG)FmaGetFileSize(parea->hfileJdt));

	// Calculate the message text to read and message end offsets

	offs = pmsg->jmh.offText; if (poffText) offs+= *poffText;
	offsEnd = pmsg->jmh.offText + pmsg->jmh.cchText;

	// Check if offset is past the end of the message text and fail,
	// otherwise calculate size of the message text we can read 

	if (offs > offsEnd) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	} else 
	  cch = offsEnd - offs;

	// Check if caller does not need all the message text and
	// adjust the read count if it's greater than requested 

	if (*pcchText != JRM_GETALLTEXT) {
	  if (cch > *pcchText) cch = *pcchText;
	}

	// Check if there is anything to read 

	if (cch > 0) {

	  // Check if we need to allocate text buffer for caller

	  if (*ppchText == NULL) {
		cchAlloc = (*pcchText != JRM_GETALLTEXT) ? *pcchText : cch;

		// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
		if (!CHECK_DATA_SIZE_LIMIT(cchAlloc)) {
		  FMAERROR(DATA_SIZE_LIMIT);
		  return FALSE;
		}
#endif
		// Allocate text buffer and check if ok 

		if ((*ppchText = MemAlloc((int)cchAlloc + 1, 0)) == NULL) {
		  FMAERROR(NOT_ENOUGH_MEMORY);
		  return FALSE;
		} else {
		  (*ppchText)[cchAlloc] = '\0';
		  bAlloc = TRUE;
		}
	  } 

	  // Read the message text and check if ok

	  if (!FmaReadFile(parea->hfileJdt, offs, *ppchText, cch, &cbRead)) {
Fail:   if (bAlloc) { MemFree(*ppchText); *ppchText = NULL; }
		return FALSE;
	  } else
		if (cbRead != cch) {
		  FMAERROR(INVALID_DATA);
		  goto Fail;
		}
	}

	// Update caller's pointers

	*pcchText = cch;
	if (poffText) *poffText+= cch;

#ifdef FMA_LOG_JAM_READ
Log("DoReadText        : [%09Fp] cchText=%ld", pmsg, cch);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine reads message
//

  FMAAPI_(BOOL) JamReadMsg(HMSG hmsg, PJMH pjmh, PJSF * ppjsf, ULONG * pcbjsf,
						   PCH * ppchText, ULONG * pcchText, ULONG * poffText)
  {
	PJAREA parea;
	PJMSG pmsg;
	PJSF pjsf;

	// Validate parameters

	if ((pmsg = (PJMSG)hmsg) == NULL || !JamCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: 

#ifdef FMA_LOG_JAM_READ
Log("JamReadMsg       <: [%09Fp] FAILED, error=%d", pmsg, FmaGetLastError());
#endif

	  return FALSE;
	} else
	  parea = pmsg->parea;

#ifdef FMA_LOG_JAM_READ
Log("\nJamReadMsg       >: [%09Fp] msgn=%ld lmh=%08lx in %s", pmsg, pmsg->msgn, pmsg->lmh, parea->achPath);
#endif

	// Check if message is allowed to be read

	if (!JamCheckMsgReadMode(pmsg))
	  goto Fail;

	// Read the message header unless it's cached and check if ok

	if ((pmsg->fs & (JMF_GOTMSGH | JMF_GOTSUBF)) != (JMF_GOTMSGH | JMF_GOTSUBF)) {
	  if (!DoReadJmh(pmsg))
		goto Fail;
	}

	// Check if caller needs message header and copy it

	if (pjmh) {
	  *pjmh = pmsg->jmh;
	}

	// Check if caller needs subfields info 

	if (ppjsf) {
	  if (!pmsg->pjsf || !pmsg->cbjsf) {
		*ppjsf = NULL; *pcbjsf = 0;
	  } else {

		// Check if caller needs a readonly pointer and set it,
		// otherwise allocate a copy of subfields info

		if (*pcbjsf == JRM_GETJSFPTR) {
		  *ppjsf = pmsg->pjsf; *pcbjsf = pmsg->cbjsf;
		} else 
		if ((pjsf = MemAlloc((int)pmsg->cbjsf, 0)) == NULL) {
		  FMAERROR(NOT_ENOUGH_MEMORY);
		  goto Fail;
		} else {
		  memcpy(pjsf, pmsg->pjsf, (int)pmsg->cbjsf);
		  *ppjsf = pjsf; *pcbjsf = pmsg->cbjsf;
		}
	  }
	}

	// Check if caller needs message text.

	if (ppchText) {
	  if (!DoReadText(pmsg, ppchText, pcchText, poffText))
		goto Fail;
	}

#ifdef FMA_LOG_JAM_READ
Log("JamReadMsg       <: [%09Fp] msgn=%ld lmh=%08lx in %s", pmsg, pmsg->msgn, pmsg->lmh, parea->achPath);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets message control info length
//

  FMAAPI_(BOOL) JamGetMsgCtrlLen(HMSG hmsg, ULONG * pcbjsf)
  {
	PJMSG pmsg;

	// Validate parameters

	if ((pmsg = (PJMSG)hmsg) == NULL || !JamCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: return FALSE;
	}

	// Check if message is allowed to be read

	if (!JamCheckMsgReadMode(pmsg))
	  goto Fail;

	// Read the message header unless it's cached and check if ok

	if ((pmsg->fs & (JMF_GOTMSGH | JMF_GOTSUBF)) != (JMF_GOTMSGH | JMF_GOTSUBF)) {
	  if (!DoReadJmh(pmsg))
		goto Fail;
	}

	// Set message subfields length for caller

	if (pcbjsf) *pcbjsf = pmsg->jmh.cbjsf;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets message text length
//

  FMAAPI_(BOOL) JamGetMsgTextLen(HMSG hmsg, ULONG * pcchText)
  {
	PJMSG pmsg;

	// Validate parameters

	if ((pmsg = (PJMSG)hmsg) == NULL || !JamCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: return FALSE;
	}

	// Check if message is allowed to be read

	if (!JamCheckMsgReadMode(pmsg))
	  goto Fail;

	// Read the message header unless it's cached and check if ok

	if ((pmsg->fs & (JMF_GOTMSGH | JMF_GOTSUBF)) != (JMF_GOTMSGH | JMF_GOTSUBF)) {
	  if (!DoReadJmh(pmsg))
		goto Fail;
	}

	// Set message text length for caller

	if (pcchText) *pcchText = pmsg->jmh.cchText;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of JamRead.c
/////////////////////////////////////////////////////////////////////////////
