/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Testing app
//
// Test.c
//
// Copyright (c) 1999 by Pete Kvitek
//
// Created: 27/Aug/1999
// Updated: 27/Aug/1999
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include standard headers

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

  // Include common and private headers

#include "FmaApi.h"
//#include "FtsApi.h"
//#include "FtsApiP.h"
//#include "JamApi.h"
//#include "JamApiP.h"
#include "JamApi.h"
#include "JamApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine is a callback for JamEnumMsgs()
//

  static BOOL FMACALLBACK DoEnumMsg(PENUMMSGINFO pemi, PVOID pvCallbackData)
  {
	ULONG fl = (ULONG) pvCallbackData;

	// Log the enum call

	Log("DoEnumMsg: msgn=%lu umsg=%lu imsg=%u offs=%08lx hash=%08lx pvCallbackData=%08lx",
		 pemi->msgn, pemi->umsg, pemi->imsg, pemi->offs, pemi->hash, 
		 pvCallbackData);

	//ASSERT(JamValidateMsg(pemi->harea, pemi->msgn, NULL));

	//if (fl & FENUMMSGS_SAFEENUM) JamDeleteMsg(pemi->harea, pemi->msgn);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message opening
//

  static BOOL FMASUB DoOpenMsg(HAREA harea)
  {
	HMSG hmsg;
	MSGN msgn;

	// Display what these return

	printf("JamGetFirstMsg()=%d\n", JamGetFirstMsg(harea));
	printf("JamGetHighMsg()=%d\n", JamGetHighMsg(harea));
	printf("JamGetMsgCount()=%d\n", JamGetMsgCount(harea));
	printf("JamGetCurMsg()=%d\n", JamGetCurMsg(harea));

	// Open the current message (this should fail after open)

	msgn = MSGN_CUR;
	if ((hmsg = JamOpenMsg(harea, OPENMSG_READ, msgn, 0)) == NULL) {
	  printf("JamOpenMsg() failed, code=%d\n", FmaGetLastError());
	} else
	  JamCloseMsg(hmsg);

	// Loop reading next msgs

	msgn = MSGN_NEXT;
	while ((hmsg = JamOpenMsg(harea, OPENMSG_READ, msgn, 0)) != NULL) {
	  printf("Opened msgn=%d\n", JamGetCurMsg(harea));
	  JamCloseMsg(hmsg);
	}

	// Loop reading prev msgs

	msgn = MSGN_PREV;
	while ((hmsg = JamOpenMsg(harea, OPENMSG_READ, msgn, 0)) != NULL) {
	  printf("Opened msgn=%d\n", JamGetCurMsg(harea));
	  JamCloseMsg(hmsg);
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message reading
//

  static BOOL FMASUB DoReadMsg(HAREA harea, MSGN msgn)
  {
	HMSG hmsg;
	BOOL bOk;
	JMH jmh;

	// Open message and check if ok

	if ((hmsg = JamOpenMsg(harea, OPENMSG_READ, msgn, 0)) == NULL) {
	  printf("JamOpenMsg() failed, code=%d\n", FmaGetLastError());
	  return FALSE;
	}

	// Get message subfields and text length
#if 0
	{ ULONG cbjsf, cchText;

	  if (!JamGetMsgCtrlLen(hmsg, &cbjsf)) {
		printf("JamGetMsgCtrlLen() failed, code=%d\n", FmaGetLastError());
	  }

	  if (!JamGetMsgTextLen(hmsg, &cchText)) {
		printf("JamGetMsgTextLen() failed, code=%d\n", FmaGetLastError());
	  }
	}
#endif

	// Read entire message into allocated buffer
#if 1
	{ PJSF pjsf = NULL; ULONG cbjsf = JRM_GETJSFCOPY;
	  PCH pchText = NULL; ULONG cchText = JRM_GETALLTEXT;

	  bOk = JamReadMsg(hmsg, &jmh, &pjsf, &cbjsf, &pchText, &cchText, 0);

	  if (pjsf) MemFree(pjsf);
	  if (pchText) MemFree(pchText);
	}
#endif

	// Read message by chunks into allocated buffer
#if 1
	{ PJSF pjsf = NULL; ULONG cbjsf = JRM_GETJSFPTR;
	  PCH pchText = NULL; ULONG cchText = 512; ULONG offText = 0;

	  bOk = JamReadMsg(hmsg, &jmh, &pjsf, &cbjsf, 0, 0, 0);

	  while (bOk) {
		bOk = JamReadMsg(hmsg, 0, 0, 0, &pchText, &cchText, &offText);
		if (!cchText) break;
	  }

	  if (pchText) MemFree(pchText);
	}
#endif        

	// Read message by chunks into buffer on stack
#if 1
	{ PJSF pjsf = NULL; ULONG cbjsf = JRM_GETJSFPTR; CHAR achText[512];
	  PCH pchText = achText; ULONG cchText = sizeof(achText); ULONG offText = 0;

	  bOk = JamReadMsg(hmsg, &jmh, &pjsf, &cbjsf, 0, 0, 0);

	  while (bOk) {
		bOk = JamReadMsg(hmsg, 0, 0, 0, &pchText, &cchText, &offText);
		if (!cchText) break;
	  }
	}
#endif        

	// Close message 

	if (!JamCloseMsg(hmsg)) {
	  printf("JamCloseMsg() failed, code=%d\n", FmaGetLastError());
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message creation and writing
//

  static BOOL FMASUB DoCreateMsg(HAREA harea, MSGN msgn)
  {
	PJSF pjsf = NULL; ULONG cbjsf = 0;
//  CHAR achText[] = "\rThis is a message text\r";
	CHAR achText[] = "\rThis is a message text\r\r-Pete";
	HMSG hmsg;
	JMH jmh;

	// Open message and check if ok

	if ((hmsg = JamOpenMsg(harea, OPENMSG_CREATE, msgn, 0)) == NULL) {
	  printf("JamOpenMsg() failed, code=%d\n", FmaGetLastError());
	  return FALSE;
	} else
	  printf("JamGetCurMsg()=%d\n", JamGetCurMsg(harea));

	// Create message header

	memset(&jmh, 0, sizeof(jmh));
	jmh.magic = JAM_MAGIC;
	jmh.timeWritten = FmaGetLocalTime();

	// Create subfields

	JamAddJsf(&pjsf, &cbjsf, JSF_RECVRNAME, "All", -1); 
	JamAddJsf(&pjsf, &cbjsf, JSF_SENDERNAME, "Pete Kvitek", -1); 
	JamAddJsf(&pjsf, &cbjsf, JSF_OADDRESS, "2:5020/6@fidonet", -1); 
	JamAddJsf(&pjsf, &cbjsf, JSF_DADDRESS, "2:5020/1234", -1); 
	JamAddJsf(&pjsf, &cbjsf, JSF_SUBJECT, "Hi there...", -1); 

	JamAddJsfFormat(&pjsf, &cbjsf, JSF_FTSKLUDGE, "COOL: My cool kludge #%d", 1234); 

	// Write message
#if 1    
	if (!JamWriteMsg(hmsg, &jmh, pjsf, cbjsf, achText, lengof(achText), 0, lengof(achText))) {
	  printf("JamWriteMsg() failed, code=%d\n", FmaGetLastError());
	}
#else
	if (!JamWriteMsg(hmsg, &jmh, pjsf, cbjsf, achText, 0, 0, 0)) {
	  printf("JamWriteMsg() failed, code=%d\n", FmaGetLastError());
	}
#endif

	// Destroy the subfields

	if (pjsf) MemFree(pjsf);

	// Close message 

	if (!JamCloseMsg(hmsg)) {
	  printf("JamCloseMsg() failed, code=%d\n", FmaGetLastError());
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message updating and writing
//

  static BOOL FMASUB DoUpdateMsg(HAREA harea, MSGN msgn)
  {
	PJSF pjsf = NULL; ULONG cbjsf = 0;
	CHAR achText[] = "\rThis is a message text\r";
	PCH pchText = NULL; ULONG cchText = JRM_GETALLTEXT;
	HMSG hmsg;
	JMH jmh;

	// Open message and check if ok

	if ((hmsg = JamOpenMsg(harea, OPENMSG_READWRITE, msgn, 0)) == NULL) {
	  printf("JamOpenMsg() failed, code=%d\n", FmaGetLastError());
	  return FALSE;
	} else
	  printf("JamGetCurMsg()=%d\n", JamGetCurMsg(harea));

	// Read message

	JamReadMsg(hmsg, &jmh, &pjsf, &cbjsf, &pchText, &cchText, 0);

	// Change message header

	jmh.timeWritten = FmaGetLocalTime();
	jmh.crcp = 0x12345678;

	// Change subfields

	JamDelJsf(&pjsf, &cbjsf, JSF_SENDERNAME);
	JamDelJsf(&pjsf, &cbjsf, JSF_RECVRNAME);
	JamDelJsf(&pjsf, &cbjsf, JSF_SUBJECT);
	JamGetJsf(pjsf, cbjsf, -1, 0);
	JamAddJsf(&pjsf, &cbjsf, 9999, "Weird subfield", -1); 
	JamAddJsf(&pjsf, &cbjsf, JSF_PID, "FMA v0.0", -1); 

	// Write message
#if 1
	if (!JamWriteMsg(hmsg, &jmh, pjsf, cbjsf, achText, lengof(achText), 0, lengof(achText))) {
	  printf("JamWriteMsg() failed, code=%d\n", FmaGetLastError());
	}
#else
	if (!JamWriteMsg(hmsg, &jmh, pjsf, cbjsf, achText, 0, 0, 0)) {
	  printf("JamWriteMsg() failed, code=%d\n", FmaGetLastError());
	}
#endif

	// Free allocated memory

	if (pjsf) MemFree(pjsf);
	if (pchText) MemFree(pchText);

	// Close message 

	if (!JamCloseMsg(hmsg)) {
	  printf("JamCloseMsg() failed, code=%d\n", FmaGetLastError());
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message creation and writing
//

  static BOOL FMASUB DoCheckUid(HAREA harea)
  {
	  // Test msgn to uid conversion
#if 1
	{ UMSG umsg;
	  umsg = JamMsgnToUid(harea, 0);
	  umsg = JamMsgnToUid(harea, 1);
	  umsg = JamMsgnToUid(harea, 2);
	  umsg = JamMsgnToUid(harea, 6);
	  umsg = JamMsgnToUid(harea, 7);
	}
#endif

	// Test uid to msgn conversion
#if 1
	{ MSGN msgn; UMSG umsg;
	  msgn = JamUidToMsgn(harea, 0,  UMSG_EXACT);
	  msgn = JamUidToMsgn(harea, 1,  UMSG_EXACT);
	  msgn = JamUidToMsgn(harea, 8,  UMSG_EXACT);
	  msgn = JamUidToMsgn(harea, 11, UMSG_EXACT);
	  msgn = JamUidToMsgn(harea, 14, UMSG_EXACT);

	  umsg = 2;
	  msgn = JamUidToMsgn(harea, umsg,  UMSG_EXACT);
	  msgn = JamUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_PREV | UMSG_NEXT);
	  msgn = JamUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_PREV);
	  msgn = JamUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_NEXT);
	  msgn = JamUidToMsgn(harea, umsg,  UMSG_PREV);
	  msgn = JamUidToMsgn(harea, umsg,  UMSG_NEXT);
	}
#endif

	 return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Main...
//

  void main(int argc, char * argv[])
  {
	HAREA harea = NULL;
	HMSG hmsg = NULL;
	PSZ pszArea;

	// Force debugging log

	if (!getenv("FMA-LOG")) putenv("FMA-LOG=*");

	// Initialize and check if ok

	if (!FmaInitialize(NULL)) {
	  printf("FmaInitialize() failed, code=%u\n", FmaGetLastError());
	  exit(EXIT_FAILURE);
	}

	// Open area and check if ok

	pszArea = "..\\jam\\test";
	if ((harea = JamOpenArea(pszArea, OPENAREA_ALWAYS, 0, NULL)) == NULL) {
	  printf("JamOpenArea() failed, code=%d\n", FmaGetLastError());
	} else {
	  BOOL bLock = 1;

	  // Lock area

	  if (bLock) bLock = JamLockArea(harea, 1);

	  // Miscellaneous tests

//    DoOpenMsg(harea);
//    DoReadMsg(harea, 1);
//	  DoCreateMsg(harea, 0);
//    DoUpdateMsg(harea, JamGetHighMsg(harea));
//    DoCheckUid(harea);

	  // Enumerate area messages
#if 0
	  { ULONG fl = 0;
//      fl|= FENUMMSGS_REVERSE;
		fl|= FENUMMSGS_SAFEENUM;
		JamEnumMsgs(harea, DoEnumMsg, (PVOID)fl, fl);
	  }
#endif

#if 0
	  // Delete area messages

	  { BOOL bOk;
		bOk = JamDeleteMsg(harea, 5);
		bOk = JamDeleteMsg(harea, 6);
	  }
#endif

#if 0
	  // Lastread info test

	  { BOOL bOk;
		LRINFO lri;

		lri.umsgLast = 1;
		lri.umsgHigh = 2;

		bOk = JamSetLRInfo(harea, 1, &lri, "Pete Kvitek");
		bOk = JamSetLRInfo(harea, 1, &lri, "Fidonet Mail Access");
	  }
#endif

#if 0
	  // Set/check password

	  { CHAR achPassword[] = "1234";
		JamSetPassword(harea, achPassword);
		JamCheckPassword(harea, achPassword);
	  }
#endif

	 // Unlock area

	  if (bLock) JamLockArea(harea, 0);

	  // Close area

	  JamCloseArea(harea);
	}

	// Cleanup and exit

	FmaTerminate();

	exit(EXIT_SUCCESS);
  }

/////////////////////////////////////////////////////////////////////////////
// End of Test.c
/////////////////////////////////////////////////////////////////////////////
