/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Pkt api routines module
//
// PktApi.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 08/Feb/2000
// Updated: 17/Feb/2000
//
// History: 08/Feb/2000 kvitekp - Initially created.
//
//          17/Feb/2000 kvitekp - Scan ASCII date if binary one is not valid
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"
#include "FmaCtrl.h"

#include "PktApi.h"
#include "PktApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Pkt message base api

  static API api = {
	sizeof(API),
	PktOpenArea,
	PktCloseArea,
	PktLockArea,
	PktValidateMsg,
	PktCreateMsgMap,
	PktEnumMsgs,
	PktDeleteMsg,
	PktSetCurMsg,
	PktGetCurMsg,
	PktGetFirstMsg,
	PktGetHighMsg,
	PktGetMsgCount,
	PktMsgnToUid,
	PktUidToMsgn,
	PktOpenMsg,
	PktCloseMsg,
	PktLoadMsg,
	PktSaveMsg,
XPV PktReadMsg,
XPV PktWriteMsg,
	PktGetMsgCtrlLen,
	PktGetMsgTextLen,
	PktGetLRInfo,
	PktSetLRInfo,
  };

  // Pkt message attribute mapping table. Note: These are 
  // attributes and ^aFLAGS flags we generate when saving message.

  static ATTRMAP aAttr[] = {
	ATTR_(MA_LOCAL,             FMA_LOCL,               NULL)
	ATTR_(MA_INTRANSIT,         FMA_TRAN,               NULL)
	ATTR_(MA_PRIVATE,           FMA_PRVT,               NULL)
	ATTR_(MA_READ,              FMA_READ,               NULL)
	ATTR_(MA_SENT,              FMA_SENT,               NULL)
	ATTR_(MA_KILLSENT,          FMA_KILL,               NULL)
	ATTR_(MA_ARCHIVESENT,       0,                     "A/S")
	ATTR_(MA_HOLD,              FMA_HOLD,               NULL)
	ATTR_(MA_CRASH,             FMA_CRSH,               NULL)
	ATTR_(MA_IMMEDIATE,         0,                     "IMM")
	ATTR_(MA_DIRECT,            0,                     "DIR")
	ATTR_(MA_GATE,              0,                     "ZON")
	ATTR_(MA_FILEREQUEST,       FMA_FREQ,               NULL)
	ATTR_(MA_FILEATTACH,        FMA_FILE,               NULL)
	ATTR_(MA_TRUNCFILE,         0,                     "TFS")
	ATTR_(MA_KILLFILE,          0,                     "KFS")
	ATTR_(MA_RECEIPTREQ,        FMA_RREQ,               NULL)
	ATTR_(MA_CONFIRMREQ,        FMA_RRCP,               NULL)
	ATTR_(MA_ORPHAN,            FMA_ORPH,               NULL)
	ATTR_(MA_ENCRYPT,           0,                     "ENC")
	ATTR_(MA_COMPRESS,          0,                     "CMP")
	ATTR_(MA_ESCAPED,           0,                     "ESC")
	ATTR_(MA_FORCEPICKUP,       0,                     "FPU")
//  ATTR_(MA_TYPELOCAL,         0,                      NULL)
//  ATTR_(MA_TYPEECHO,          0,                      NULL)
//  ATTR_(MA_TYPENET,           0,                      NULL)
//  ATTR_(MA_SCANNED,           0,                      NULL)
	ATTR_(MA_FILEUPDATEREQ,     FMA_UREQ,               NULL)
	ATTR_(MA_AUDITREQUEST,      FMA_AREQ,               NULL)
	ATTR_(MA_NODISPLAY,         0,                     "NOD")
	ATTR_(MA_LOCKED,            0,                     "LOK")
//  ATTR_(MA_DELETED,           0,                      NULL)
  };

  // Pkt message header mapping table

  static MHDRMAP aMhdr[] = {
	MHDR_(FMH, achBy,           MHDR_ASCZ,  FMP_SENDERNAME)
	MHDR_(FMH, achTo,           MHDR_ASCZ,  FMP_RECEVRNAME)
	MHDR_(FMH, achRe,           MHDR_ASCZ,  FMP_SUBJECT)
	MHDR_(FMH, timesRead,       MHDR_US16,  FMP_TIMESREAD)
	MHDR_(FMH, cost,            MHDR_US16,  FMP_MESSAGECOST)
	MHDR_(FMH, dttmWritten,     MHDR_DTTM,  FMP_DTTMWRITTEN)
	MHDR_(FMH, dttmArrived,     MHDR_DTTM,  FMP_DTTMRECEIVED)
	MHDR_(FMH, replyTo,         MHDR_US16,  FMP_MSGREPLYTO)
	MHDR_(FMH, nextReply,       MHDR_US16,  FMP_MSGREPLYNEXT)
  };

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine loads message attributes
//

  static BOOL FMASUB DoLoadAttr(PFMM pfmm, PFMH pfmh)
  {
	UINT iAttr;

	// Scan the attribute mapping table translating attributes

	for (iAttr = 0; iAttr < numbof(aAttr); iAttr++) {
	  if (pfmh->attr & aAttr[iAttr].attrNative) {
		SETFLAG(pfmm->attr, aAttr[iAttr].attr);
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message attributes
//

  static BOOL FMASUB DoSaveAttr(PFMM pfmm, PFMH pfmh)
  {
	UINT iAttr;

	// Scan the attribute mapping table translating attributes

	for (iAttr = 0; iAttr < numbof(aAttr); iAttr++) {
	  if (pfmm->attr & aAttr[iAttr].attr) {
		SETFLAG(pfmh->attr, (USHORT)aAttr[iAttr].attrNative);
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads message header
//

  static BOOL FMASUB DoLoadMhdr(PFMM pfmm, PFMH pfmh)
  {
	// Load the message header

	if (!FmaLoadMhdr(pfmm, pfmh, aMhdr, numbof(aMhdr)))
	  return FALSE;

	// Check if message written date is not valid and try
	// to use ascii date

	if (!FmaCheckDttm(&pfmh->dttmWritten)) {
		DTTM dttm;
		if (FmaAsciiToDttm(pfmh->achDate, &dttm)) {
			FmaAddFmp(pfmm, FMP_DTTMWRITTEN, &dttm, sizeof(dttm));
		}
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message header
//

  static BOOL FMASUB DoSaveMhdr(PFMM pfmm, PFMH pfmh)
  {
	UINT cch;

	// Save the message header

	if (!FmaSaveMhdr(pfmm, pfmh, aMhdr, numbof(aMhdr)))
	  return FALSE;

	// Format the asciiz date in message header

	cch = FmaFormatDttm(pfmh->achDate, sizeof(pfmh->achDate), &pfmh->dttmWritten, 
						FDTTM_TIMESECONDS, ' ', ':', "  ");

	ASSERT(cch == lengof(pfmh->achDate));

	// Note that origination and destination net/node info will be 
	// created later on when generating address kludges 

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks dttm
//

  static BOOL FMASUB DoCheckDttm(PDTTM pdttm)
  {
	// Empty dttm is valid

	if (*((PLONG)pdttm) == 0) return TRUE;

	// Check nonempty dttm

	return FmaCheckDttm(pdttm);
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads PKT-001.16 binary message header info
//

  static BOOL FMASUB DoLoadFmh2Maybe(PFMH pfmh, NETADDR * pnetOrig, NETADDR * pnetDest)
  {
	// Check if both dates are valid

	if (DoCheckDttm(&pfmh->dttmWritten) && DoCheckDttm(&pfmh->dttmArrived))
	  return FALSE;

	// Retrieve modern zone/point

	pnetOrig->zone  = ((PFMH2)pfmh)->origZone;
	pnetOrig->point = ((PFMH2)pfmh)->origPoint;

	pnetDest->zone  = ((PFMH2)pfmh)->destZone;
	pnetDest->point = ((PFMH2)pfmh)->destPoint;

	// Update the binary dates wannabes

	*((PULONG)&pfmh->dttmWritten) = 0;
	*((PULONG)&pfmh->dttmArrived) = 0;

	// Convert ascii date to binary if possible

	FmaAsciiToDttm(pfmh->achDate, &pfmh->dttmWritten);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads message control info
//

  static BOOL FMASUB DoLoadCtrl(PFMM pfmm, PFMH pfmh, PCH pchCtrl, ULONG cchCtrl,
								PPAREA parea)
  {
	NETADDR netOrig;
	NETADDR netDest;
	PCH pch, pchEnd;
	UINT cch;
	KLI kli;

	// Check if we have no control info

	if (!pchCtrl || !cchCtrl) return TRUE;

	// Prepare the kludge info structure

	memset(&kli, 0, sizeof(kli));

	// Scan message control info

	for (pch = pchCtrl; *pch == 0x01; pch = pchEnd) {

	  // Scan to the end of this kludge

	  for (pchEnd = ++pch; (UCHAR)*pchEnd > 0x01 && *pchEnd != '\r'; pchEnd++);

	  // Calculate the kludge length and if not zero,
	  // load the kludge and check if ok

	  if ((cch = (UINT)(pchEnd - pch)) > 0) {
		if (!FmaLoadKludge(pfmm, pch, cch, &kli))
		  return FALSE;
	  }

	  // Advance to the next kludge

	  if (*pchEnd == '\r') pchEnd++;
	}

	// Load ^aFLAGS kludge info and check if ok

	if (!FmaLoadFLAGS(pfmm, kli.pch[KLI_FLAGS], kli.cch[KLI_FLAGS]))
	  return FALSE;

	// Make up the binary origination network address

	netOrig.zone  = 0; 
	netOrig.net   = pfmh->origNet;
	netOrig.node  = pfmh->origNode;
	netOrig.point = 0;

	// Make up the binary destination network address

	netDest.zone  = 0; 
	netDest.net   = pfmh->destNet;
	netDest.node  = pfmh->destNode;
	netDest.point = 0;

	// Take care of SQDEV200 trick for 4D pointnets

	if (pfmh->timesRead && pfmh->timesRead == (USHORT)~pfmh->cost) {
	  netOrig.point = pfmh->timesRead;
	  pfmh->timesRead = 0;
	  pfmh->cost = 0;
	}

	// According to PKT-001.016 fmh.dttmWritten and fmh.dttmArrived 
	// may contain zone/point info so check dttm to see if it's weird

	DoLoadFmh2Maybe(pfmh, &netOrig, &netDest);

	// Load address kludges info and check if ok

	if (!FmaLoadAddrKludges(pfmm, &kli, &netOrig, &netDest, &parea->netAddr,
						   (parea->fl & FOA_ECHOMAIL) ? TRUE : FALSE))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves address info
//

  static BOOL FMASUB DoSaveAddr(PFMM pfmm, PFMH pfmh, PDSZ pdszCtrl,
								PPAREA parea, ULONG fl)
  {
	NETADDR oAddr, dAddr, gAddr;

	// Save address kludges and check if ok

	if (!FmaSaveAddrKludges(pfmm, pdszCtrl, &oAddr, &dAddr, &parea->netAddr,
							TRUE, TRUE, fl, TRUE))
	  return FALSE;

	// Update binary message header net/node numbers

	pfmh->destNet = dAddr.net;  pfmh->destNode = dAddr.node;
	pfmh->origNet = oAddr.net;  pfmh->origNode = oAddr.node;

	// Check if we have gate address

	if (FmaGetFmp(pfmm, FMP_GATEADDR, 0)) {
	  if (!FmaGetNetAddr(pfmm, FMP_GATEADDR, &gAddr, NULL)) {
		return FALSE;
	  } else {
		pfmh->destNet  = gAddr.net;  
		pfmh->destNode = gAddr.node;
	  }
	} 

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves ^aFLAGS kludge info
//

  static BOOL FMASUB DoSaveFLAGS(PFMM pfmm, PFMH pfmh, PDSZ pdszCtrl,
								 PPAREA parea, ULONG fl)
  {
	DSZ dsz = DSZ_INIT;
	PFMP pfmp;
	UINT iAttr;

	// Scan the attribute mapping table adding up flags

	for (iAttr = 0; iAttr < numbof(aAttr); iAttr++) {
	  if (pfmm->attr & aAttr[iAttr].attr && aAttr[iAttr].pszFlag) {
		if (!AddDszFormat(&dsz, " %s", aAttr[iAttr].pszFlag)) {
Fail:     FMAERROR(NOT_ENOUGH_MEMORY);
		  DelDsz(&dsz);
		  return FALSE;
		}
	  }
	}

	// Check if there is a flags property and add it

	if ((pfmp = FmaGetFmp(pfmm, FMP_FLAGS, 0)) != NULL) {
	  if (!AddDszFormat(&dsz, " %s", PFMPDATA(pfmp)))
		goto Fail;
	}

	// Check if we got anything to sore in ^aFLAGS kludge and
	// add it to the control info

	if (dsz.psz) {
	  if (!AddDszFormat(pdszCtrl, "\x01""FLAGS%s\r", dsz.psz))
		goto Fail;
	}

	// Delete flags info dynamic string

	DelDsz(&dsz);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves kludges info
//

  static BOOL FMASUB DoSaveKludges(PFMM pfmm, PFMH pfmh, PDSZ pdszCtrl,
								   PPAREA parea, ULONG fl)
  {
	PFMP pfmp;

	// Loop through all the properties

	for (pfmp = pfmm->pfmp; pfmp != NULL; pfmp = pfmp->pfmpNext) {

	  // Skip non kludge properties and eom kludges

	  if ((FMP_FLG(pfmp->id) & FMP_KLU) != FMP_KLU || (FMP_FLG(pfmp->id) & FMP_EOM))
		continue;

	  // Skip the flags property since ^aFLAGS kludge has already been added

	  if (pfmp->id == FMP_FLAGS)
		continue;

	  // Save kludge and check if ok

	  if (!FmaSaveKludge(pfmm, pfmp, pdszCtrl, TRUE))
		return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message control info
//

  static BOOL FMASUB DoSaveCtrl(PFMM pfmm, PFMH pfmh, PDSZ pdszCtrl,
								PPAREA parea, ULONG fl)
  {
	// Save address info and check if ok

	if (!DoSaveAddr(pfmm, pfmh, pdszCtrl, parea, fl))
	  return FALSE;

	// Save ^aFLAGS kludge and check if ok

	if (!DoSaveFLAGS(pfmm, pfmh, pdszCtrl,parea, fl))
	  return FALSE;

	// Save kludges and check if ok

	if (!DoSaveKludges(pfmm, pfmh, pdszCtrl,parea, fl))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads message text
//

  static BOOL FMASUB DoLoadText(PFMM pfmm, HMSG hmsg, ULONG fl)
  {
	PCH pchText = NULL; ULONG cchText = FRM_GETALLTEXT;
	PCH pch;

	// Check if we need to load message text

	if (fl & FLM_NOTEXT) return TRUE;

	// Load text and check if ok

	if (!PktReadMsg(hmsg, NULL, NULL, NULL, &pchText, &cchText, 0))
	  return FALSE;

	// Check if no text loaded and leave

	if (!cchText) {
	  if (pchText) MemFree(pchText);
	  return TRUE;
	}

	// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
	if (!CHECK_DATA_SIZE_LIMIT(cchText)) {
	  FMAERROR(DATA_SIZE_LIMIT);
	  return FALSE;
	}
#endif

	// Check if text is not null terminated and add a trailing null

	if (pchText[cchText - 1]) {
	  if ((pch = MemRealloc(pchText, (int)cchText + 1)) == NULL) {
		FMAERROR(NOT_ENOUGH_MEMORY);
		MemFree(pchText);
		return FALSE;
	  } else {
		pchText = pch;
		pchText[cchText] = '\0';
	  }
	}

	// Add message text property and check if ok

	if (!FmaAddFmp(pfmm, FMP_MESSAGETEXT, &pchText, sizeof(pchText))) {
	  MemFree(pchText);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message text
//

  static BOOL FMASUB DoSaveText(PFMM pfmm, HMSG hmsg, ULONG fl, PDSZ pdszEom)
  {
	PCH pchText; ULONG cchText, offText = 0;
	PSZ * ppsz;
	ULONG cpsz;
	PFMP pfmp;

	// Check if we need to save message text

	if (fl & FSM_NOTEXT) return TRUE;

	// Get the message text property and check if ok

	if ((pfmp = FmaGetFmp(pfmm, FMP_MESSAGETEXT, 0)) == NULL) 
	  return TRUE;

	// Loop through all the message text strings writing them out

	ppsz = PFMPDATA(pfmp);
	cpsz = pfmp->cb / sizeof(PSZ);

	while (cpsz--) {
	  if (ppsz && *ppsz) {
		pchText = *ppsz;
		cchText = strlen(pchText);
		if (!PktWriteMsg(hmsg, NULL, NULL, 0, pchText, cchText, &offText, 0))
		  return FALSE;
	  }

	  ppsz++;
	}

	// Check if we have eom kludges and write them as well

	if (pdszEom && pdszEom->ich) {
	  if (!PktWriteMsg(hmsg, NULL, NULL, 0, pdszEom->psz, pdszEom->ich, &offText, 0))
		return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns total message text length
//

  static ULONG FMASUB DoGetTotalTextLen(PFMM pfmm, PDSZ pdszEom)
  {
	ULONG cchTextTotal = 0;
	PFMP pfmp;

	// Get the message text property and add its length

	if ((pfmp = FmaGetFmp(pfmm, FMP_MESSAGETEXT, 0)) != NULL)
	  cchTextTotal+= FmaGetFmpTextLen(pfmp);

	// Add eom kludges length

	if (pdszEom) cchTextTotal+= pdszEom->ich;

	return cchTextTotal;
  }


/////////////////////////////////////////////////////////////////////////////
// This routine loads eom kludges
//

  static BOOL FMASUB DoLoadEomKludges(PFMM pfmm, HMSG hmsg, ULONG fl)
  {
	// Check if message text is loaded and fall to generic routine

	if (FmaGetFmp(pfmm, FMP_MESSAGETEXT, 0)) {
	  return FmaLoadEomKludges(pfmm, fl);
	}

	// No message text to load the eom kludges from

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine returns api pointer
//

  FMAAPI_(PAPI) PktGetApi(VOID)
  {
	return &api;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads abstracted fidonet mail message
//

  FMAAPI_(BOOL) PktLoadMsg(HMSG hmsg, PFMM * ppfmm, ULONG fl)
  {
	PPAREA parea = ((PPMSG)hmsg)->parea;
	ULONG cchCtrl = FRM_GETCTRLPTR;
	PCH pchCtrl = NULL; 
	PFMM pfmm;
	FMH fmh;

	ASSERT(hmsg != NULL);
	ASSERT(ppfmm != NULL);

	// Create fmm object and check if ok

	if ((pfmm = FmaCreateMsg(0)) == NULL) {
Fail: FmaDestroyMsg(pfmm);      
	  return FALSE;
	} else
	  pfmm->type = MBTYPE_PKT;

	// Read the message header and control info, then check if ok

	if (!PktReadMsg(hmsg, &fmh, &pchCtrl, &cchCtrl, NULL, NULL, NULL))
	  goto Fail;

	// Load message attributes and check if ok

	if (!DoLoadAttr(pfmm, &fmh))
	  goto Fail;

	// Load message header

	if (!DoLoadMhdr(pfmm, &fmh))
	  goto Fail;

	// Load message control info and check if ok

	if (!DoLoadCtrl(pfmm, &fmh, pchCtrl, cchCtrl, parea))
	  goto Fail;

	// Load message text and check if ok

	if (!DoLoadText(pfmm, hmsg, fl))
	  goto Fail;

	// Load eom kludges and check if ok

	if (!DoLoadEomKludges(pfmm, hmsg, fl))
	  goto Fail;

	// Pass fmm object to caller

	*ppfmm = pfmm;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves abstracted fidonet mail message
//

  FMAAPI_(BOOL) PktSaveMsg(HMSG hmsg, PFMM pfmm, ULONG fl, ULONG cchTextTotal)
  {
	PPAREA parea = ((PPMSG)hmsg)->parea;
	DSZ dszCtrl = DSZ_INIT;
	DSZ dszEom = DSZ_INIT;
	FMH fmh;

	ASSERT(hmsg != NULL);
	ASSERT(pfmm != NULL);

	// Clear target message header

	memset(&fmh, 0, sizeof(FMH));

	// Add automatic properties and check if ok

	if (!FmaAddAutoProp(pfmm, fl)) {
Fail: DelDsz(&dszCtrl); 
	  DelDsz(&dszEom);
	  return FALSE;
	}

	// Save message attributes and check if ok

	if (!DoSaveAttr(pfmm, &fmh)) 
	  goto Fail;

	// Save message header

	if (!DoSaveMhdr(pfmm, &fmh))
	  goto Fail;

	// Save message control info and check if ok

	if (!DoSaveCtrl(pfmm, &fmh, &dszCtrl, parea, fl))
	  goto Fail;

	// Save eom kludges early since their length has to be included
	// in the total message length which has to be known before the
	// first write to the new message

	if (!FmaSaveEomKludges(pfmm, &dszEom))
	  goto Fail;

	// Check if caller did not specify total message text length and get it now

	if (cchTextTotal == (ULONG)-1) cchTextTotal = DoGetTotalTextLen(pfmm, &dszEom);

	// Write the message header and control info

	if (!PktWriteMsg(hmsg, &fmh, dszCtrl.psz, dszCtrl.ich, NULL, 0, 0, cchTextTotal))
	  goto Fail;

	// Save message text and check if ok

	if (!DoSaveText(pfmm, hmsg, fl, &dszEom))
	  goto Fail;

	// Delete allocated dynamic strings

	DelDsz(&dszCtrl); 
	DelDsz(&dszEom);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of PktApi.c
/////////////////////////////////////////////////////////////////////////////
