/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Pkt api declarations module
//
// PktApi.h
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 08/Feb/2000
// Updated: 08/Feb/2000
//
// History: 08/Feb/2000 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////
//
// Note: These declaratons were derived from original specifications
//       published in FTS-0001.16, FSC-0039.004 and FSC-0048.002 and 
//       some common real life practices.
//
/////////////////////////////////////////////////////////////////////////////
//
// NOTE: ALL STRUCTURES IN THIS MODULE ASSUME SINGLE BYTE PACKING !!!
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _PKT_API_INCLUDED
#define _PKT_API_INCLUDED

  // Declare C/C++ frame

#ifdef __cplusplus
extern "C" {
#endif

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"
#include "FtsApi.h"

  // Set structure packing

#ifdef FMA_PUSHPOP_PACKING
#pragma pack(push, 1)
#endif

/////////////////////////////////////////////////////////////////////////////
// Pkt base declarations
/////////////////////////////////////////////////////////////////////////////

  // Packet file extensions

#define PKT_PKT_FILEEXT  ".PKT" // Packet file extension

  // Packet header structure (FSC-0048.002)

  typedef struct _PBH { /* pbh */
	USHORT  origNode;           // Origination node number
	USHORT  destNode;           // Destination node number
	USHORT  year;               // Year of packet creation (e.g. 1999)
	USHORT  month;              // Month of packet creation (0-11 for Jan-Dec)
	USHORT  day;                // Day of packet creation (1-31)
	USHORT  hour;               // Hour of packet creation (0-23)
	USHORT  minute;             // Minute of packet creation (0-59)
	USHORT  second;             // Second of packet creation (0-59)
	USHORT  baud;               // Max baud rate of orig and dest
	USHORT  type;               // Packet type (PKT_TYPE2)
	USHORT  origNet;            // Origination net number
	USHORT  destNet;            // Destination net number
	BYTE    prodCodeLow;        // Product code (low)
	BYTE    prodRevMajor;       // Product revision (major)
	CHAR    achPassword[8];     // Password (null padded)
	USHORT  origZone0;          // Origination zone number (ZMail,QMail)
	USHORT  destZone0;          // Destination zone number (ZMail,QMail)
	USHORT  auxNet;             // Auxiliary net number
	USHORT  capWordCopy;        // Byte swapped copy of capability word
	BYTE    prodCodeHigh;       // Product code (high)
	BYTE    prodRevMinor;       // Product revision (minor)
	USHORT  capWord;            // Capability word
	USHORT  origZone;           // Origination zone number (FrontDoor)
	USHORT  destZone;           // Destination zone number (FrontDoor)
	USHORT  origPoint;          // Origination point number
	USHORT  destPoint;          // Destination point number
	CHAR    achData[4];         // Product specific data
  } PBH, FAR * PPBH;

  // Packet type

#define PKT_TYPE2       0x0002  // Always version 2

  // Packet capability word 

#define PCAP_TYPE2PLUS  0x0001u // Type-2+ packet
#define PCAP_RFC822     0x8000u // Allows RFC-822

  // FMA product code... to be ascquired from FSC

#define FMA_PRODCODE    0x00FE  // None allocated

  // Packed message header structure mostly used

  typedef struct _PMH { /* pmh */
	USHORT  type;               // Packed message type  (PMH_TYPE2)
	USHORT  origNode;           // Origination node number
	USHORT  destNode;           // Destination node number
	USHORT  origNet;            // Origination net number
	USHORT  destNet;            // Destination net number
	USHORT  attr;               // Message attributes (FMA_)
	USHORT  cost;               // Unit cost charged to send the message
	CHAR    achDate[20];        // ASCII date information
//  CHAR    achTo[max 36];      // To name (asciiz)
//  CHAR    achBy[max 36];      // From name (asciiz)
//  CHAR    achRe[max 72];      // Message subject (asciiz)
  } PMH, FAR * PPMH;

  // Packed message stamp

#define PMH_TYPE2   0x0002      // Always version 2

  // Packet file locking

#define PBH_LOCK_OFFS   0       // Locking region offset
#define PBH_LOCK_LENG   1       // Locking region length
#define PBH_LOCK_RETRY  5       // Max number of locking attempts

/////////////////////////////////////////////////////////////////////////////
// Pkt api declarations
/////////////////////////////////////////////////////////////////////////////

  // Special values used by PktReadMsg()

#define PRM_GETCTRLCOPY          0      // *pcchCtrl: Allocate a copy of msg control info
#define PRM_GETCTRLPTR  ((ULONG)-1)     // *pcchCtrl: Return a readonly pointer to msg control info
#define PRM_GETALLTEXT  ((ULONG)-1)     // *pcchText: read entire message text

  // Function prototypes

  FMAAPI_(PAPI) PktGetApi(VOID);
  FMAAPI_(HAREA) PktOpenArea(PSZ pszPath, UINT fOpenMode, ULONG fl, 
							 NETADDR * pnetAddr);

  FMAAPI_(BOOL) PktCloseArea(HAREA harea);
  FMAAPI_(BOOL) PktLockArea(HAREA harea, BOOL bLock);

  FMAAPI_(BOOL) PktValidateMsg(HAREA harea, MSGN msgn, PULONG pimsg);
  FMAAPI_(BOOL) PktCreateMsgMap(HAREA harea, UMSG * paumsg[], PULONG pcumsg);
  FMAAPI_(BOOL) PktEnumMsgs(HAREA harea, PFNENUMMSGS pfnCallback, 
							PVOID pvCallbackData, ULONG fl);

  FMAAPI_(BOOL) PktDeleteMsg(HAREA harea, MSGN msgn);
  FMAAPI_(BOOL) PktSetCurMsg(HAREA harea, MSGN msgn);
  FMAAPI_(MSGN) PktGetCurMsg(HAREA harea);
  FMAAPI_(MSGN) PktGetFirstMsg(HAREA harea);
  FMAAPI_(MSGN) PktGetHighMsg(HAREA harea);
  FMAAPI_(LONG) PktGetMsgCount(HAREA harea);

  FMAAPI_(UMSG) PktMsgnToUid(HAREA harea, MSGN msgn);
  FMAAPI_(MSGN) PktUidToMsgn(HAREA harea, UMSG umsg, UINT fMode);

  FMAAPI_(HMSG) PktOpenMsg(HAREA harea, UINT mode, MSGN msgn, ULONG fl);
  FMAAPI_(BOOL) PktCloseMsg(HMSG hmsg);

  FMAAPI_(BOOL) PktLoadMsg(HMSG hmsg, PFMM * ppfmm, ULONG fl);
  FMAAPI_(BOOL) PktSaveMsg(HMSG hmsg, PFMM pfmm, ULONG fl, ULONG cchTextTotal);

  FMAAPI_(BOOL) PktReadMsg(HMSG hmsg, PFMH pfmh, PCH * ppchCtrl, ULONG * pcchCtrl,
						   PCH * ppchText, ULONG * pcchText, ULONG * poffText);

  FMAAPI_(BOOL) PktWriteMsg(HMSG hmsg, PFMH pfmh, PCH pchCtrl, ULONG cchCtrl,
							PCH pchText, ULONG cchText, ULONG * poffText, 
							ULONG cchTextTotal);

  FMAAPI_(BOOL) PktGetMsgCtrlLen(HMSG hmsg, ULONG * pcchCtrl);
  FMAAPI_(BOOL) PktGetMsgTextLen(HMSG hmsg, ULONG * pcchText);

  FMAAPI_(BOOL) PktGetLRInfo(HAREA harea, ULONG uid, PLRINFO plri);
  FMAAPI_(BOOL) PktSetLRInfo(HAREA harea, ULONG uid, PLRINFO plri, 
							 PSZ pszUserName);

/////////////////////////////////////////////////////////////////////////////
// Header cleanup
/////////////////////////////////////////////////////////////////////////////

  // Restore structure packing

#ifdef FMA_PUSHPOP_PACKING
#pragma pack(pop)
#endif

  // Terminate C/C++ frame

#ifdef __cplusplus
}
#endif

#endif  // #ifndef _PKT_API_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// End of PktApi.h
/////////////////////////////////////////////////////////////////////////////
