/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Pkt api private declarations module
//
// PktApiP.h
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 08/Feb/2000
// Updated: 18/Feb/2000
//
// History: 08/Feb/2000 kvitekp - Initially created.
//
//          18/Feb/2000 kvitekp - Added PktSet/CheckPassword()
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _PKT_APIP_INCLUDED
#define _PKT_APIP_INCLUDED

  // Declare C/C++ frame

#ifdef __cplusplus
extern "C" {
#endif

  // Include types declaratons

#include "FmaDef.h"
#include "PktApi.h"

/////////////////////////////////////////////////////////////////////////////
// Pkt area declarations
/////////////////////////////////////////////////////////////////////////////

  // Forward declarations

  struct _PMSG;

  // Packed message info structure

  typedef struct _PMI {         /* pmi */
	LONG    lmh;                        // Message header offset
	ULONG   cbHdr;                      // Message header length
	ULONG   cbMsg;                      // Message length (hdr + body)
	UMSG    umsg;                       // Message unique id
  } PMI, FAR * PPMI;

  // Pkt message area object structure

  typedef struct _PAREA {       /* area */
	ULONG   magic;                      // Magic number (PAREA_MAGIC)
	PAPI    papi;                       // Pkt api pointer
	PBH     pbh;                        // Packet header
	HFILE   hfilePkt;                   // Packet file handle
	BOOL    bAreaLocked;                // Area lock count
	BOOL    bBaseLocked;                // Base lock count
	BOOL    bExclusive;                 // Exclusive count
	ULONG   cmsg;                       // Number of messages
	MSGN    msgnCur;                    // Current message number
	MSGN    msgnBase;                   // Base message number
	MSGN    msgnHigh;                   // Highest message number
	PPMI    apmi;                       // Packed message info array
	UINT    cpmi;                       // Packed message info array size
	ULONG   fl;                         // Open area flags (FOA_)
	struct _PMSG * pmsgList;            // Open messges list
	NETADDR netAddr;                    // Default network address
	NETADDR netOrig;                    // Packet origination network address
	NETADDR netDest;                    // Packet destination network address
	DTTM    dttm;                       // Packet creation date and time
	CHAR    achPath[1];                 // Area path (asciiz)
  } PAREA, FAR * PPAREA;

  // Pkt message area object magic number

#define PAREA_MAGIC ((ULONG)0x45524150) // Long equivalent of "PARE"

  // Pkt message object structure

  typedef struct _PMSG {        /* msg */
	struct _PMSG * pmsgPrev;            // Pointer to prev open msg in area
	struct _PMSG * pmsgNext;            // Pointer to next open msg in area
	ULONG   magic;                      // Magic number (PMSG_MAGIC)
	PPAREA  parea;                      // Message area handle
	MSGN    msgn;                       // Message number
	UINT    mode;                       // Message open mode
	ULONG   fl;                         // Message open flags (FOM_)
	PSZ     pszCtrl;                    // Message control info
	UINT    cchCtrl;                    // Message control info length
	FMH     fmh;                        // Message header
	LONG    lmh;                        // Message header offset
	ULONG   cbHdr;                      // Message header length
	ULONG   cbMsg;                      // Message length (hdr + body)
	LONG    lmt;                        // Message text offset
	LONG    lmtEnd;                     // Message text end offset after writing
	UINT    fs;                         // Message status flags (PMF_)
  } /*PMSG,*/ FAR * PPMSG;              // Note: PMSG conflicts with Windows.h

  // Pkt message object magic number

#define PMSG_MAGIC ((ULONG)0x47534D50)  // Long equivalent of "PMSG"

  // Pkt message status flags in msg.fs

#define PMF_GOTMSGH       0x0001        // Message header has been associated
#define PMF_GOTCTRL       0x0002        // Control info has been associated
#define PMF_BOGUS         0x0004        // -- reserved
#define PMF_GOTFRAME      0x0008        // Writing frame selected for new message
#define PMF_EXCLUSIVE     0x0010        // Entered exclusive when opened
#define PMF_MSGHWRITTEN   0x0100        // Message header has been written
#define PMF_CTRLWRITTEN   0x0200        // Message contrl info has been written
#define PMF_TEXTWRITTEN   0x0400        // Message text has been written
#define PMF_CTRLHASAREA   0x1000        // Message contrl info has AREA w/o ^a

  // Message has been written to mask

#define PMF_MASKWRITTEN   (PMF_MSGHWRITTEN | PMF_CTRLWRITTEN | PMF_TEXTWRITTEN)

/////////////////////////////////////////////////////////////////////////////
// Miscellaneous declarations
/////////////////////////////////////////////////////////////////////////////

  // Function prototypes

  FMAAPI_(BOOL) PktCheckArea(PPAREA parea);
  FMAAPI_(BOOL) PktCheckMsg(PPMSG pmsg);
  FMAAPI_(BOOL) PktCheckMsgReadMode(PPMSG pmsg);
  FMAAPI_(BOOL) PktCheckMsgWriteMode(PPMSG pmsg);

  FMAAPI_(BOOL) PktLockBase(PPAREA parea, BOOL bLock);
  FMAAPI_(BOOL) PktEnterExclusive(PPAREA parea);
  FMAAPI_(BOOL) PktLeaveExclusive(PPAREA parea);
  FMAAPI_(BOOL) PktSetEnd(PPAREA parea, LONG offs);
  FMAAPI_(BOOL) PktRescanArea(PPAREA parea);

  FMAAPI_(BOOL) PktLoadPbh(PPAREA parea);
  FMAAPI_(BOOL) PktSavePbh(PPAREA parea);
  FMAAPI_(BOOL) PktCheckPbh(PPAREA parea);

  FMAAPI_(BOOL) PktResetCurMsgCache(PPAREA parea);
  FMAAPI_(BOOL) PktSetCurMsgCache(PPMSG pmsg);

  FMAAPI_(PPMSG) PktGetOpenMsg(PPAREA parea, MSGN msgn);

  FMAAPI_(BOOL) PktLoadPmh(PPAREA parea, PFMH pfmh, LONG lmh);
  FMAAPI_(BOOL) PktSavePmh(PPAREA parea, PFMH pfmh, LONG lmh);

  FMAAPI_(BOOL) PktAddMsg(PPAREA parea, LONG lmh, ULONG cbHdr, ULONG cbMsg);
  FMAAPI_(BOOL) PktDelMsg(PPAREA parea, UINT imsg);

  FMAAPI_(BOOL) PktShrinkPkt(PPAREA parea, LONG offs, ULONG cb);
  FMAAPI_(BOOL) PktExpandPkt(PPAREA parea, LONG offs, ULONG cb);

  FMAAPI_(BOOL) PktCheckBase(PSZ pszPath);

  FMAAPI_(BOOL) PktSetNetAddr(PPAREA parea, NETADDR * pnetOrig, NETADDR * pnetDest);
  FMAAPI_(BOOL) PktGetNetAddr(PPAREA parea, NETADDR * pnetOrig, NETADDR * pnetDest);

  FMAAPI_(BOOL) PktSetPassword(PPAREA parea, PSZ pszPassword);
  FMAAPI_(BOOL) PktCheckPassword(PPAREA parea, PSZ pszPassword);

  // Pkt logging control defines

#ifdef  FMA_LOG
#define FMA_LOG_PKT_AREA
#define FMA_LOG_PKT_HELP
//#define FMA_LOG_PKT_LAST
//#define FMA_LOG_PKT_LOCK
#define FMA_LOG_PKT_MSG
#define FMA_LOG_PKT_READ
#define FMA_LOG_PKT_WRITE
#endif

/////////////////////////////////////////////////////////////////////////////
// Header cleanup
/////////////////////////////////////////////////////////////////////////////

  // Terminate C/C++ frame

#ifdef __cplusplus
}
#endif

#endif  // #ifndef _PKT_APIP_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// End of PktApiP.h
/////////////////////////////////////////////////////////////////////////////
