/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Testing app
//
// Test.c
//
// Copyright (c) 1999 by Pete Kvitek
//
// Created: 08/Feb/2000
// Updated: 08/Feb/2000
//
// History: 08/Feb/2000 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include standard headers

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

  // Include common and private headers

#include "FmaApi.h"
#include "PktApi.h"
#include "PktApiP.h"
//#include "SquApi.h"
//#include "SquApiP.h"
//#include "JamApi.h"
//#include "JamApiP.h"
#include "PktApi.h"
#include "PktApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine is a callback for PktEnumMsgs()
//

  static BOOL FMACALLBACK DoEnumMsg(PENUMMSGINFO pemi, PVOID pvCallbackData)
  {
	ULONG fl = (ULONG) pvCallbackData;

	// Log the enum call

	Log("DoEnumMsg: msgn=%lu umsg=%lu imsg=%u offs=%08lx hash=%08lx pvCallbackData=%08lx",
		 pemi->msgn, pemi->umsg, pemi->imsg, pemi->offs, pemi->hash, 
		 pvCallbackData);

	//ASSERT(PktValidateMsg(pemi->harea, pemi->msgn, NULL));

	//if (fl & FENUMMSGS_SAFEENUM) PktDeleteMsg(pemi->harea, pemi->msgn);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message opening
//

  static BOOL FMASUB DoOpenMsg(HAREA harea)
  {
	HMSG hmsg;
	MSGN msgn;

	// Display what these return

	printf("PktGetFirstMsg()=%d\n", PktGetFirstMsg(harea));
	printf("PktGetHighMsg()=%d\n", PktGetHighMsg(harea));
	printf("PktGetMsgCount()=%d\n", PktGetMsgCount(harea));
	printf("PktGetCurMsg()=%d\n", PktGetCurMsg(harea));

	// Open the current message (this should fail after open)

	msgn = MSGN_CUR;
	if ((hmsg = PktOpenMsg(harea, OPENMSG_READ, msgn, 0)) == NULL) {
	  printf("PktOpenMsg() failed, code=%d\n", FmaGetLastError());
	} else
	  PktCloseMsg(hmsg);

	// Loop reading next msgs

	msgn = MSGN_NEXT;
	while ((hmsg = PktOpenMsg(harea, OPENMSG_READ, msgn, 0)) != NULL) {
	  printf("Opened msgn=%d\n", PktGetCurMsg(harea));
	  PktCloseMsg(hmsg);
	}

	// Loop reading prev msgs

	msgn = MSGN_PREV;
	while ((hmsg = PktOpenMsg(harea, OPENMSG_READ, msgn, 0)) != NULL) {
	  printf("Opened msgn=%d\n", PktGetCurMsg(harea));
	  PktCloseMsg(hmsg);
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message reading
//

  static BOOL FMASUB DoReadMsg(HAREA harea, MSGN msgn)
  {
	HMSG hmsg;
	BOOL bOk;
	FMH fmh;

	// Open message and check if ok

	if ((hmsg = PktOpenMsg(harea, OPENMSG_READ, msgn, 0)) == NULL) {
	  printf("PktOpenMsg() failed, code=%d\n", FmaGetLastError());
	  return FALSE;
	}

	// Get message control info and text length
#if 0
	{ ULONG cchCtrl, cchText;

	  if (!PktGetMsgCtrlLen(hmsg, &cchCtrl)) {
		printf("PktGetMsgCtrlLen() failed, code=%d\n", FmaGetLastError());
	  }	else
		printf("PktGetMsgCtrlLen()=%d\n", cchCtrl);

	  if (!PktGetMsgTextLen(hmsg, &cchText)) {
		printf("PktGetMsgTextLen() failed, code=%d\n", FmaGetLastError());
	  } else
		printf("PktGetMsgTextLen()=%d\n", cchText);
	}
#endif

	// Read entire message into allocated buffer
#if 1
	{ PCH pchCtrl = NULL; ULONG cchCtrl = FRM_GETCTRLCOPY;
	  PCH pchText = NULL; ULONG cchText = FRM_GETALLTEXT;

	  bOk = PktReadMsg(hmsg, &fmh, &pchCtrl, &cchCtrl, &pchText, &cchText, 0);

	  printf("PktReadMsg() cchCtrl=%d cchText=%d\n", cchCtrl, cchText);

	  if (pchCtrl) MemFree(pchCtrl);
	  if (pchText) MemFree(pchText);
	}
#endif

	// Read message by chunks into allocated buffer
#if 0
	{ PCH pchCtrl = NULL; ULONG cchCtrl = FRM_GETCTRLPTR;
	  PCH pchText = NULL; ULONG cchText = 32; ULONG offText = 0;

	  bOk = PktReadMsg(hmsg, &fmh, &pchCtrl, &cchCtrl, 0, 0, 0);

	  while (bOk) {
		bOk = PktReadMsg(hmsg, 0, 0, 0, &pchText, &cchText, &offText);
		if (!cchText) break;
	  }

	  if (pchText) MemFree(pchText);
	}
#endif        

	// Read message by chunks into buffer on stack
#if 0
	{ PCH pchCtrl = NULL; ULONG cchCtrl = FRM_GETCTRLPTR; CHAR achText[32];
	  PCH pchText = achText; ULONG cchText = sizeof(achText); ULONG offText = 0;

	  bOk = PktReadMsg(hmsg, &fmh, &pchCtrl, &cchCtrl, 0, 0, 0);

	  while (bOk) {
		bOk = PktReadMsg(hmsg, 0, 0, 0, &pchText, &cchText, &offText);
		if (!cchText) break;
	  }
	}
#endif        

	// Close message 

	if (!PktCloseMsg(hmsg)) {
	  printf("PktCloseMsg() failed, code=%d\n", FmaGetLastError());
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message creation and writing
//

  static BOOL FMASUB DoCreateMsg(HAREA harea, MSGN msgn)
  {
	CHAR achCtrl[] = //"\x01""AREA:PK.MADMED\r"
					 "\x01""MSGID: 2:5020/6 12345678\r" 
					 "\x01""PID: FMA\r";
//	CHAR achText[] = "1234";
	CHAR achText[] = "\rThis is a message text\r\r-Pete\r";
	HMSG hmsg;
	FMH fmh;

	// Open message and check if ok

	if ((hmsg = PktOpenMsg(harea, OPENMSG_CREATE, msgn, 0)) == NULL) {
	  printf("PktOpenMsg() failed, code=%d\n", FmaGetLastError());
	  return FALSE;
	} else
	  printf("PktGetCurMsg()=%d\n", PktGetCurMsg(harea));

	// Create message header

	memset(&fmh, 0, sizeof(fmh));
	strcpy(fmh.achBy, "From name");
	strcpy(fmh.achTo, "To name");
	Format(fmh.achRe, lengof(fmh.achRe), "Msg %d", PktGetCurMsg(harea));
	fmh.attr = FMA_LOCL | FMA_READ;

	fmh.origNet = 11; fmh.origNode = 111;
	fmh.destNet = 22; fmh.destNode = 222;

	// Write message header only

#if 0
	if (!PktWriteMsg(hmsg, &fmh, NULL, 0, NULL, 0, 0, 0)) {
	  printf("PktWriteMsg() failed, code=%d\n", FmaGetLastError());
	}
#endif

	// Write message header and control info

#if 0
	if (!PktWriteMsg(hmsg, &fmh, achCtrl, lengof(achCtrl), NULL, 0, 0, 0)) {
	  printf("PktWriteMsg() failed, code=%d\n", FmaGetLastError());
	}
#endif

	// Write message header, control info and text

#if 1
	if (!PktWriteMsg(hmsg, &fmh, achCtrl, lengof(achCtrl), 
					 achText, lengof(achText), 0, lengof(achText))) {
	  printf("PktWriteMsg() failed, code=%d\n", FmaGetLastError());
	}
#endif

	// Close message 

	if (!PktCloseMsg(hmsg)) {
	  printf("PktCloseMsg() failed, code=%d\n", FmaGetLastError());
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message updating and writing
//

  static BOOL FMASUB DoUpdateMsg(HAREA harea, MSGN msgn)
  {
	PCH pchCtrl = NULL; ULONG cchCtrl = 0;
	CHAR achText[] = "\rThis is a message text\r";
	PCH pchText = NULL; ULONG cchText = FRM_GETALLTEXT;
	HMSG hmsg;
	FMH fmh;

	// Open message and check if ok

	if ((hmsg = PktOpenMsg(harea, OPENMSG_READWRITE, msgn, 0)) == NULL) {
	  printf("PktOpenMsg() failed, code=%d\n", FmaGetLastError());
	  return FALSE;
	} else
	  printf("PktGetCurMsg()=%d\n", PktGetCurMsg(harea));

	// Read message

	PktReadMsg(hmsg, &fmh, &pchCtrl, &cchCtrl, &pchText, &cchText, 0);

	// Change message header

	// Change control info

	// Write message
#if 1
	if (!PktWriteMsg(hmsg, &fmh, pchCtrl, cchCtrl, achText, lengof(achText), 0, lengof(achText))) {
	  printf("PktWriteMsg() failed, code=%d\n", FmaGetLastError());
	}
#else
	if (!PktWriteMsg(hmsg, &fmh, pchCtrl, cchCtrl, achText, 0, 0, 0)) {
	  printf("PktWriteMsg() failed, code=%d\n", FmaGetLastError());
	}
#endif

	// Free allocated memory

	if (pchCtrl) MemFree(pchCtrl);
	if (pchText) MemFree(pchText);

	// Close message 

	if (!PktCloseMsg(hmsg)) {
	  printf("PktCloseMsg() failed, code=%d\n", FmaGetLastError());
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message creation and writing
//

  static BOOL FMASUB DoCheckUid(HAREA harea)
  {
	  // Test msgn to uid conversion
#if 0
	{ UMSG umsg;
	  umsg = PktMsgnToUid(harea, 0);
	  umsg = PktMsgnToUid(harea, 1);
	  umsg = PktMsgnToUid(harea, 2);
	  umsg = PktMsgnToUid(harea, 3);
	  umsg = PktMsgnToUid(harea, 4);
	}
#endif

	// Test uid to msgn conversion
#if 1
	{ MSGN msgn; UMSG umsg;
	  msgn = PktUidToMsgn(harea, 0, UMSG_EXACT);
	  msgn = PktUidToMsgn(harea, 1, UMSG_EXACT);
	  msgn = PktUidToMsgn(harea, 2, UMSG_EXACT);
	  msgn = PktUidToMsgn(harea, 6, UMSG_EXACT);
	  msgn = PktUidToMsgn(harea, 7, UMSG_EXACT);

	  umsg = 1;
	  msgn = PktUidToMsgn(harea, umsg,  UMSG_EXACT);
	  msgn = PktUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_PREV | UMSG_NEXT);
	  msgn = PktUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_PREV);
	  msgn = PktUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_NEXT);
	  msgn = PktUidToMsgn(harea, umsg,  UMSG_PREV);
	  msgn = PktUidToMsgn(harea, umsg,  UMSG_NEXT);

	  umsg = 2;
	  msgn = PktUidToMsgn(harea, umsg,  UMSG_EXACT);
	  msgn = PktUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_PREV | UMSG_NEXT);
	  msgn = PktUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_PREV);
	  msgn = PktUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_NEXT);
	  msgn = PktUidToMsgn(harea, umsg,  UMSG_PREV);
	  msgn = PktUidToMsgn(harea, umsg,  UMSG_NEXT);
	}
#endif

	 return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Main...
//

  void main(int argc, char * argv[])
  {
	NETADDR netAddr = {2,5020,6,0};
	HAREA harea = NULL;
	HMSG hmsg = NULL;
	PSZ pszArea;
	LONG cb = 0;

	// Force debugging log

	if (!getenv("FMA-LOG")) putenv("FMA-LOG=*");

	// Initialize and check if ok

	if (!FmaInitialize(NULL)) {
	  printf("FmaInitialize() failed, code=%u\n", FmaGetLastError());
	  exit(EXIT_FAILURE);
	}

	// Open area and check if ok

	pszArea = "pkt/!";
	if ((harea = PktOpenArea(pszArea, OPENAREA_ALWAYS, 0, &netAddr)) == NULL) {
	  printf("PktOpenArea() failed, code=%d\n", FmaGetLastError());
	} else {
	  BOOL bLock = 1;

	  // Lock area

	  if (bLock) bLock = PktLockArea(harea, 1);

	  // Miscellaneous tests

//	  DoOpenMsg(harea);
	  DoReadMsg(harea, 1);
//	  DoCreateMsg(harea, 0);
//	  DoUpdateMsg(harea, PktGetHighMsg(harea));
//	  DoCheckUid(harea);

#if 0
	  // Delete area messages

	  { BOOL bOk;
		bOk = PktDeleteMsg(harea, 2);
		bOk = PktDeleteMsg(harea, 3);
	  }
#endif

	  // Enumerate area messages
#if 0
	  { ULONG fl = 0;
//		fl|= FENUMMSGS_REVERSE;
//		fl|= FENUMMSGS_SAFEENUM;
		PktEnumMsgs(harea, DoEnumMsg, (PVOID)fl, fl);
	  }
#endif

	  // Set/check password
#if 0
	  { CHAR achPassword[] = "1234";
		PktSetPassword(harea, achPassword);
		PktCheckPassword(harea, achPassword);
	  }
#endif

	 // Unlock area

	  if (bLock) PktLockArea(harea, 0);

	  // Close area

	  PktCloseArea(harea);
	}

	// Cleanup and exit

	FmaTerminate();

	exit(EXIT_SUCCESS);
  }

/////////////////////////////////////////////////////////////////////////////
// End of Test.c
/////////////////////////////////////////////////////////////////////////////
