/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Pkt write routines module
//
// PktWrite.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 09/Feb/2000
// Updated: 15/Feb/2000
//
// History: 09/Feb/2000 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "PktApi.h"
#include "PktApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Module local data

  static CHAR achArea[] = "\x01""AREA:";

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine calculates message size
//

  static ULONG FMASUB DoGetMsgSize(PFMH pfmh, PCH pchCtrl, ULONG cchCtrl, ULONG cchTextTotal, 
								   ULONG * pcbHdr, ULONG * pcbText)
  {
	ULONG cbHdr = 0, cbText = 0;

	// Add message header size

	if (pfmh) {
	  cbHdr+= sizeof(PMH);
	  cbHdr+= strlen(pfmh->achBy) + 1;
	  cbHdr+= strlen(pfmh->achTo) + 1;
	  cbHdr+= strlen(pfmh->achRe) + 1;
	}

	// Add message control info size taking care of the leading area kludge
	// which is written without leading ^a

	if (pchCtrl) {
	  if (cchCtrl > lengof(achArea) && !memcmp(pchCtrl, achArea, lengof(achArea))) {
		cbText+= cchCtrl - 1;
	  } else
		cbText+= cchCtrl;
	}

	// Add total message text

	cbText+= cchTextTotal + 1;

	// Update caller's info

	if (pcbHdr) *pcbHdr = cbHdr;
	if (pcbText) *pcbText = cbText;

	// Return total message size

	return cbHdr + cbText;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets the frame to write the message to
//

  static BOOL FMASUB DoGetWriteFrame(PPMSG pmsg, PFMH pfmh, PCH pchCtrl, ULONG cchCtrl, 
									 ULONG cchTextTotal)
  {
	static CHAR achNull[] = "";
	PPAREA parea = pmsg->parea;
	ULONG cbMsg, cbHdr, cbText;
	PPMI ppmi;
	UINT imsg;
	LONG lmh; 

	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != -1);
	ASSERT(parea->bExclusive);

	// Check if creating a message 

	if (pmsg->mode == OPENMSG_CREATE) {

	  // Calculate size of the new message

	  cbMsg = DoGetMsgSize(pfmh, pchCtrl, cchCtrl, cchTextTotal, &cbHdr, &cbText);

	  // Determine message index

	  imsg = pmsg->msgn - parea->msgnBase;

	  ASSERT(parea->apmi != NULL);

	  // Check if this is a new message

	  if (!pmsg->lmh) {

		// Find out new message header offset. Note that message map entry
		// for this new message is already allocated. Also note that we don't 
		// want just subtract pkt null terminator from pkt file length since
		// this would prevent us for ability to work with unterminated files.

		if (!imsg) {
		  lmh = sizeof(PBH);
		} else {
		  ppmi = &parea->apmi[imsg - 1]; 
		  lmh = ppmi->lmh + ppmi->cbMsg;
		}

		// Now allocate space for the new message and check if ok

		if (!PktExpandPkt(parea, lmh, cbMsg))
		  return FALSE;

	  } else {

		// For the old message check if we need to shrink or
		// expand its frame and do it now. Note that when shrinking
		// the frame we write message trailing null, but when expanding
		// the frame the new space is zeroed out by default.

		ppmi = &parea->apmi[imsg];
		lmh = ppmi->lmh;
		if (cbMsg > ppmi->cbMsg) {
		  if (!PktExpandPkt(parea, lmh + ppmi->cbMsg, cbMsg - ppmi->cbMsg))
			return FALSE;
		} else
		if (cbMsg < ppmi->cbMsg) {
		  if (!PktShrinkPkt(parea, lmh + cbMsg, ppmi->cbMsg - cbMsg))
			return FALSE;
		  if (!FmaWriteFile(parea->hfilePkt, lmh + cbMsg - 1, achNull, 1, NULL))
			return FALSE;
		} 
	  }

	  // Update message object and associated message map entry

	  ppmi = &parea->apmi[imsg]; 
	  ppmi->lmh = pmsg->lmh = lmh;
	  ppmi->cbHdr = pmsg->cbHdr = cbHdr;
	  ppmi->cbMsg = pmsg->cbMsg = cbMsg;

	  // Set message text writing controls

	  pmsg->lmt = pmsg->lmh + pmsg->cbHdr + cchCtrl;

	} else 

	// Check if writing an existing message

	if (pmsg->mode == OPENMSG_WRITE || pmsg->mode == OPENMSG_READWRITE) {

	  ASSERT(pmsg->fs & PMF_GOTCTRL);

	  // Set message text writing controls

	  pmsg->lmt = pmsg->lmh + pmsg->cbHdr + cchCtrl;

	} else {

	  // Don't allow writing if message has been opened in read mode

	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Set flag that we've got the writing frame

	SETFLAG(pmsg->fs, PMF_GOTFRAME);

	// Update current message cache

	parea->msgnCur = pmsg->msgn;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes message header
//

  static BOOL FMASUB DoWritePmh(PPMSG pmsg, PFMH pfmh)
  {
	PPAREA parea = pmsg->parea;
	ULONG cbHdr;

	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != -1);
	ASSERT(parea->bExclusive);

	// Check if message header is not specified and 
	// assume the cached one

	if (!pfmh) pfmh = &pmsg->fmh;

	// Get size of the packed message header and
	// check if it goes past the writing frame.

	DoGetMsgSize(pfmh, NULL, 0, 0, &cbHdr, NULL);
	if (cbHdr >= pmsg->cbMsg) {
	  FMAERROR(INVALID_SIZE);
	  return FALSE;
	}

	// Write the message header and check if ok

	if (!PktSavePmh(parea, pfmh, pmsg->lmh))
	  return FALSE;

	ASSERT(pmsg->lmh + (LONG)cbHdr == FmaGetFilePos(parea->hfilePkt));

	// Update packed message header size

	pmsg->cbHdr = cbHdr;

	// Update message text offset so that text writing without
	// message control info will go right past the message header

	pmsg->lmt = pmsg->lmh + pmsg->cbHdr;

	// Check if wrote a new message header and 
	// update the cached one

	if (pfmh != &pmsg->fmh) {
	  SETFLAG(pmsg->fs, PMF_GOTMSGH);
	  pmsg->fmh = *pfmh;
	}

#ifdef FMA_LOG_PKT_WRITE
Log("DoWritePmh        : [%09Fp] %s to %s \"%s\"", pmsg, pmsg->fmh.achBy, pmsg->fmh.achTo, pmsg->fmh.achRe);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes message control info
//

  static BOOL FMASUB DoWriteCtrl(PPMSG pmsg, PCH pchCtrl, ULONG cchCtrl)
  {
	PPAREA parea = pmsg->parea;
	BOOL bArea = FALSE;
	BOOL bNew = FALSE;
	ULONG cbWritten;
	UINT cchWrite;
	PCH pchWrite;
	LONG offs;
	PCH pch;

	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != -1);
	ASSERT(parea->bExclusive);

	// Check if message contol info is not specified and 
	// assume the cached one, otherwise allocate a copy of
	// the new message control info unless it's empty

	 if (!pchCtrl) {
	  pchCtrl = pmsg->pszCtrl;
	  cchCtrl = pmsg->cchCtrl;
	} else {
	  bNew = TRUE;
	  if (!cchCtrl) {
		pchCtrl = NULL;
	  } else {

		// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
		if (!CHECK_DATA_SIZE_LIMIT(cchCtrl)) {
		  FMAERROR(DATA_SIZE_LIMIT);
		  return FALSE;
		}
#endif
		// Allocate memory to hold message control info

		if ((pch = MemAlloc((int)cchCtrl + 1, 0)) == NULL) {
		  FMAERROR(NOT_ENOUGH_MEMORY);
		  return FALSE;
		} else {
		  memcpy(pch, pchCtrl, (int)cchCtrl); pchCtrl = pch; 
		  pchCtrl[cchCtrl] = '\0';
		}
	  }
	}

	// Check if we have message control info

	if (pchCtrl && cchCtrl) {

	  // Check for the leading AREA kludge which should be written 
	  // without ^a prefix

	  if (cchCtrl > lengof(achArea) && !memcmp(pchCtrl, achArea, lengof(achArea))) {
		bArea = TRUE;
		pchWrite = pchCtrl + 1;
		cchWrite = cchCtrl - 1;
	  } else {
		pchWrite = pchCtrl;
		cchWrite = cchCtrl;
	  }

	  // Check if control info will go past the writing frame

	  offs = pmsg->lmh + pmsg->cbHdr;
	  if (offs + cchWrite >= pmsg->lmh + pmsg->cbMsg) {
		if (bNew) MemFree(pchCtrl);
		FMAERROR(INVALID_SIZE);
		return FALSE;
	  }

	  // Write the message control info and check if ok

	  if (!FmaWriteFile(parea->hfilePkt, offs, pchWrite, cchWrite, &cbWritten)) {
		if (bNew) MemFree(pchCtrl);
		return FALSE;
	  }

	  ASSERT(cbWritten == cchWrite);

	  // Update message text offset so that text writing will go 
	  // right past the message control info

	  pmsg->lmt = offs + cchWrite;
	}

	// Check if wrote a new message control info and update the cached one

	if (bNew) {
	  if (pmsg->pszCtrl) MemFree(pmsg->pszCtrl);
	  SETFLAG(pmsg->fs, PMF_GOTCTRL);
	  pmsg->pszCtrl = pchCtrl;
	  pmsg->cchCtrl = (UINT)cchCtrl;
	  SETFLAGTO(pmsg->fs, PMF_CTRLHASAREA, bArea);
	}

#ifdef FMA_LOG_PKT_WRITE
Log("DoWriteCtrl       : [%09Fp] cchCtrl=%d \"%s\"", pmsg, pmsg->cchCtrl, pmsg->pszCtrl);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine writes message text
//

  static BOOL FMASUB DoWriteText(PPMSG pmsg, PCH pchText, ULONG cchText, 
								 ULONG * poffText)
  {
	PPAREA parea = pmsg->parea;
	ULONG cbWritten;
	LONG offs;

	ASSERT(parea != NULL);
	ASSERT(parea->hfilePkt != -1);
	ASSERT(parea->bExclusive);

	// Calculate the message text offset

	offs = pmsg->lmt; if (poffText) offs+= *poffText;

	// Check if message text we're going to write will 
	// run past the end of the write frame

	if (offs + (LONG)cchText >= pmsg->lmh + (LONG)pmsg->cbMsg) {
	  FMAERROR(INVALID_SIZE);
	  return FALSE;
	}

	// Write the message text and check if ok

	if (pchText && cchText) {

	  if (!FmaWriteFile(parea->hfilePkt, offs, pchText, cchText, &cbWritten))
		return FALSE;

	  ASSERT(cbWritten == cchText);
	}

	// Update the message text end and update flags

	pmsg->lmtEnd = offs  + cchText;
	SETFLAG(pmsg->fs, PMF_TEXTWRITTEN);

	// Update caller's text offset

	if (poffText) *poffText+= cchText;

#ifdef FMA_LOG_PKT_WRITE
Log("DoWriteText       : [%09Fp] cchText=%d", pmsg, cchText);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine writes message
//

  FMAAPI_(BOOL) PktWriteMsg(HMSG hmsg, PFMH pfmh, PCH pchCtrl, ULONG cchCtrl,
							PCH pchText, ULONG cchText, ULONG * poffText, 
							ULONG cchTextTotal)
  {
	PPAREA parea;
	PPMSG pmsg;

	// Validate parameters

	if ((pmsg = (PPMSG)hmsg) == NULL || !PktCheckMsg(pmsg)) {
	  FMAERROR(INVALID_PARAM);
Fail: 

#ifdef FMA_LOG_PKT_WRITE
Log("PktWriteMsg      <: [%09Fp] FAILED, error=%d", pmsg, FmaGetLastError());
#endif

	  return FALSE;
	} else
	  parea = pmsg->parea;

#ifdef FMA_LOG_PKT_WRITE
Log("\nPktWriteMsg      >: [%09Fp] msgn=%ld in %s", pmsg, pmsg->msgn, parea->achPath);
#endif

	ASSERT(parea->hfilePkt != -1);

	// Check if message is allowed to be written

	if (!PktCheckMsgWriteMode(pmsg))
	  goto Fail;

	// Check if there is no frame for this messsage

	if (!(pmsg->fs & PMF_GOTFRAME)) {

	  // We need a message header and control info 
	  // on the first write to any message

	  if (!pfmh || (cchCtrl && !pchCtrl)) {
		FMAERROR(INVALID_PARAM);
		goto Fail;
	  }

	  // Get the frame to write and save result

	  if (!DoGetWriteFrame(pmsg, pfmh, pchCtrl, cchCtrl, cchTextTotal))
		goto Fail;
	}

	// Check if caller wants to write message header.

	if (pfmh) {
	  if (DoWritePmh(pmsg, pfmh)) {
		SETFLAG(pmsg->fs, PMF_MSGHWRITTEN);
	  } else 
		goto Fail;
	}

	// Check if caller wants to write message control info

	if (pchCtrl) {
	  if (DoWriteCtrl(pmsg, pchCtrl, cchCtrl)) {
		SETFLAG(pmsg->fs, PMF_CTRLWRITTEN);
	  } else 
		goto Fail;
	}

	// Check if caller wants to write message text

	if (pchText) {
	  if (DoWriteText(pmsg, pchText, cchText, poffText)) {
		SETFLAG(pmsg->fs, PMF_TEXTWRITTEN);
	  } else 
		goto Fail;
	}

#ifdef FMA_LOG_PKT_WRITE
Log("PktWriteMsg      <: [%09Fp] msgn=%ld in %s", pmsg, pmsg->msgn, parea->achPath);
#endif

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of PktWrite.c
/////////////////////////////////////////////////////////////////////////////
