/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Squish api routines module
//
// SquApi.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 25/Sep/1999
// Updated: 14/Oct/1999
//
// History: 25/Sep/1999 kvitekp - Initially created.
//
//          30/Jan/2000 kvitekp - Removed NOT_SUPPORTED in DoLoadEomKludges
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"
#include "FmaCtrl.h"

#include "SquApi.h"
#include "SquApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

  // Squish message base api

  static API api = {
	sizeof(API),
	SquOpenArea,
	SquCloseArea,
	SquLockArea,
	SquValidateMsg,
	SquCreateMsgMap,
	SquEnumMsgs,
	SquDeleteMsg,
	SquSetCurMsg,
	SquGetCurMsg,
	SquGetFirstMsg,
	SquGetHighMsg,
	SquGetMsgCount,
	SquMsgnToUid,
	SquUidToMsgn,
	SquOpenMsg,
	SquCloseMsg,
	SquLoadMsg,
	SquSaveMsg,
XPV SquReadMsg,
XPV SquWriteMsg,
	SquGetMsgCtrlLen,
	SquGetMsgTextLen,
	SquGetLRInfo,
	SquSetLRInfo,
  };

  // Squish message attribute mapping table. Note: These are 
  // attributes and ^aFLAGS flags we generate when saving message.

  static ATTRMAP aAttr[] = {
	ATTR_(MA_LOCAL,             SMA_LOCL,               NULL)
	ATTR_(MA_INTRANSIT,         SMA_TRAN,               NULL)
	ATTR_(MA_PRIVATE,           SMA_PRVT,               NULL)
	ATTR_(MA_READ,              SMA_READ,               NULL)
	ATTR_(MA_SENT,              SMA_SENT,               NULL)
	ATTR_(MA_KILLSENT,          SMA_KILL,               NULL)
	ATTR_(MA_ARCHIVESENT,       0,                     "A/S")
	ATTR_(MA_HOLD,              SMA_HOLD,               NULL)
	ATTR_(MA_CRASH,             SMA_CRSH,               NULL)
	ATTR_(MA_IMMEDIATE,         0,                     "IMM")
	ATTR_(MA_DIRECT,            0,                     "DIR")
	ATTR_(MA_GATE,              0,                     "ZON")
	ATTR_(MA_FILEREQUEST,       SMA_FREQ,               NULL)
	ATTR_(MA_FILEATTACH,        SMA_FILE,               NULL)
	ATTR_(MA_TRUNCFILE,         0,                     "TFS")
	ATTR_(MA_KILLFILE,          0,                     "KFS")
	ATTR_(MA_RECEIPTREQ,        SMA_RREQ,               NULL)
	ATTR_(MA_CONFIRMREQ,        SMA_RRCP,               NULL)
	ATTR_(MA_ORPHAN,            SMA_ORPH,               NULL)
	ATTR_(MA_ENCRYPT,           0,                     "ENC")
	ATTR_(MA_COMPRESS,          0,                     "CMP")
	ATTR_(MA_ESCAPED,           0,                     "ESC")
	ATTR_(MA_FORCEPICKUP,       0,                     "FPU")
//  ATTR_(MA_TYPELOCAL,         0,                      NULL)
//  ATTR_(MA_TYPEECHO,          0,                      NULL)
//  ATTR_(MA_TYPENET,           0,                      NULL)
	ATTR_(MA_SCANNED,           SMA_SCAN,               NULL)
	ATTR_(MA_FILEUPDATEREQ,     SMA_UREQ,               NULL)
	ATTR_(MA_AUDITREQUEST,      SMA_AREQ,               NULL)
	ATTR_(MA_NODISPLAY,         0,                     "NOD")
	ATTR_(MA_LOCKED,            0,                     "LOK")
//  ATTR_(MA_DELETED,           0,                      NULL)
  };

  // Squish message header mapping table

  static MHDRMAP aMhdr[] = {
	MHDR_(SMH, achBy,           MHDR_ASCZ,  FMP_SENDERNAME)
	MHDR_(SMH, achTo,           MHDR_ASCZ,  FMP_RECEVRNAME)
	MHDR_(SMH, achRe,           MHDR_ASCZ,  FMP_SUBJECT)
	MHDR_(SMH, netOrig,         MHDR_ADDR,  FMP_ORIGADDR)
	MHDR_(SMH, netDest,         MHDR_ADDR,  FMP_DESTADDR)
	MHDR_(SMH, dttmWritten,     MHDR_DTTM,  FMP_DTTMWRITTEN)
	MHDR_(SMH, dttmArrived,     MHDR_DTTM,  FMP_DTTMRECEIVED)
	MHDR_(SMH, umsgReplyTo,     MHDR_UL32,  FMP_MSGREPLYTO)
  };

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine loads message attributes
//

  static BOOL FMASUB DoLoadAttr(PFMM pfmm, PSMH psmh)
  {
	UINT iAttr;

	// Scan the attribute mapping table translating attributes

	for (iAttr = 0; iAttr < numbof(aAttr); iAttr++) {
	  if (psmh->attr & aAttr[iAttr].attrNative) {
		SETFLAG(pfmm->attr, aAttr[iAttr].attr);
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message attributes
//

  static BOOL FMASUB DoSaveAttr(PFMM pfmm, PSMH psmh)
  {
	UINT iAttr;

	// Scan the attribute mapping table translating attributes

	for (iAttr = 0; iAttr < numbof(aAttr); iAttr++) {
	  if (pfmm->attr & aAttr[iAttr].attr) {
		SETFLAG(psmh->attr, aAttr[iAttr].attrNative);
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads message header
//

  static BOOL FMASUB DoLoadMhdr(PFMM pfmm, PSMH psmh)
  {
	UINT cumsg, iumsg, cb;
	PFMP pfmp;

	// Load the message number as unique mesage id. Note: area has to be 
	// open with FOA_KEEPUMSG flag in order to umsg be non zero

	pfmm->umsg = psmh->umsg;

	// Load the message header mapping table adding properties

	if (!FmaLoadMhdr(pfmm, psmh, aMhdr, numbof(aMhdr)))
	  return FALSE;

	// Check if authors utc offset is specified and 
	// create appropriate property

	if (psmh->utcOffset) {
	  if (!FmaAddFmpFormat(pfmm, FMP_TZUTC, "%+05d", psmh->utcOffset))
		return FALSE;
	}

	// Count all the replies to this message if any

	for (iumsg = 0, cumsg = 0; iumsg < numbof(psmh->aumsgReplies); iumsg++) {
	  if (psmh->aumsgReplies[iumsg]) 
		cumsg++;
	  else
		break;
	}

	// Check if there are any replies to this message and add the property

	if (cumsg) {
	  cb = sizeof(UMSG) * cumsg;
	  if ((pfmp = FmaAddFmp(pfmm, FMP_MSGREPLIES, psmh->aumsgReplies, cb)) == NULL)
		return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message header
//

  static BOOL FMASUB DoSaveMhdr(PFMM pfmm, PSMH psmh)
  {
	UINT cumsg, iumsg;
	UMSG * pumsg;
	PFMP pfmp;
	UINT cch;

	// Save the message header

	if (!FmaSaveMhdr(pfmm, psmh, aMhdr, numbof(aMhdr)))
	  return FALSE;

	// Format the asciiz date in message header

	cch = FmaFormatDttm(psmh->achDate, sizeof(psmh->achDate), &psmh->dttmWritten, 
						FDTTM_TIMESECONDS, ' ', ':', "  ");

	ASSERT(cch == lengof(psmh->achDate));

	// Check if we have timezone property and convert it

	if ((pfmp = FmaGetFmp(pfmm, FMP_TZUTC, 0)) != NULL) {
	  psmh->utcOffset = (SHORT)atoi(PFMPDATA(pfmp));
	}

	// Check if there are any replies and save as many as we can

	if ((pfmp = FmaGetFmp(pfmm, FMP_MSGREPLIES, 0)) != NULL) {
	  pumsg = PFMPDATA(pfmp);
	  cumsg = min(pfmp->cb / sizeof(UMSG), numbof(psmh->aumsgReplies));
	  for (iumsg = 0; iumsg < cumsg; iumsg++, pumsg++) {
		psmh->aumsgReplies[iumsg] = *pumsg;
	  }
	}

	// Note that origination and destination address info will be 
	// created later on when generating address kludges 

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads message control info
//

  static BOOL FMASUB DoLoadCtrl(PFMM pfmm, PSMH psmh, PCH pchCtrl, ULONG cchCtrl,
								PSAREA parea)
  {
	PCH pch, pchEnd;
	UINT cch;
	KLI kli;

	// Check if we have no control info

	if (!pchCtrl || !cchCtrl) return TRUE;

	// Prepare the kludge info structure

	memset(&kli, 0, sizeof(kli));

	// Scan message control info

	for (pch = pchCtrl; *pch == 0x01; pch = pchEnd) {

	  // Scan to the end of this kludge

	  for (pchEnd = ++pch; (UCHAR)*pchEnd > 0x01; pchEnd++);

	  // Calculate the kludge length and if not zero,
	  // load the kludge and check if ok

	  if ((cch = (UINT)(pchEnd - pch)) > 0) {
		if (!FmaLoadKludge(pfmm, pch, cch, &kli))
		  return FALSE;
	  }
	}

	// Load ^aFLAGS kludge info and check if ok

	if (!FmaLoadFLAGS(pfmm, kli.pch[KLI_FLAGS], kli.cch[KLI_FLAGS]))
	  return FALSE;

	// Load address kludges info and check if ok

	if (!FmaLoadAddrKludges(pfmm, &kli, &psmh->netOrig, &psmh->netDest, &parea->netAddr,
						   (parea->fl & FOA_ECHOMAIL ? TRUE : FALSE)))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves address info
//

  static BOOL FMASUB DoSaveAddr(PFMM pfmm, PSMH psmh, PDSZ pdszCtrl,
								PSAREA parea, ULONG fl)
  {
	NETADDR gAddr;

	// Save address kludges and check if ok

	if (!FmaSaveAddrKludges(pfmm, pdszCtrl, &psmh->netOrig, &psmh->netDest, 
						   &parea->netAddr, FALSE, FALSE, fl, FALSE))
	  return FALSE;

	// Check if we have gate address

	if (FmaGetFmp(pfmm, FMP_GATEADDR, 0)) {
	  if (!FmaGetNetAddr(pfmm, FMP_GATEADDR, &gAddr, NULL)) {
		return FALSE;
	  } else {
		psmh->netDest.net  = gAddr.net;  
		psmh->netDest.node = gAddr.node;
	  }
	} 

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves ^aFLAGS kludge info
//

  static BOOL FMASUB DoSaveFLAGS(PFMM pfmm, PSMH psmh, PDSZ pdszCtrl,
								 PSAREA parea, ULONG fl)
  {
	DSZ dsz = DSZ_INIT;
	PFMP pfmp;
	UINT iAttr;

	// Scan the attribute mapping table adding up flags

	for (iAttr = 0; iAttr < numbof(aAttr); iAttr++) {
	  if (pfmm->attr & aAttr[iAttr].attr && aAttr[iAttr].pszFlag) {
		if (!AddDszFormat(&dsz, " %s", aAttr[iAttr].pszFlag)) {
Fail:     FMAERROR(NOT_ENOUGH_MEMORY);
		  DelDsz(&dsz);
		  return FALSE;
		}
	  }
	}

	// Check if there is a flags property and add it

	if ((pfmp = FmaGetFmp(pfmm, FMP_FLAGS, 0)) != NULL) {
	  if (!AddDszFormat(&dsz, " %s", PFMPDATA(pfmp)))
		goto Fail;
	}

	// Check if we got anything to sore in ^aFLAGS kludge and
	// add it to the control info

	if (dsz.psz) {
	  if (!AddDszFormat(pdszCtrl, "\x01""FLAGS%s", dsz.psz))
		goto Fail;
	}

	// Delete flags info dynamic string

	DelDsz(&dsz);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves kludges info
//

  static BOOL FMASUB DoSaveKludges(PFMM pfmm, PSMH psmh, PDSZ pdszCtrl,
								   PSAREA parea, ULONG fl)
  {
	PFMP pfmp;

	// Loop through all the properties

	for (pfmp = pfmm->pfmp; pfmp != NULL; pfmp = pfmp->pfmpNext) {

	  // Skip non kludge properties and eom kludges

	  if ((FMP_FLG(pfmp->id) & FMP_KLU) != FMP_KLU || (FMP_FLG(pfmp->id) & FMP_EOM))
		continue;

	  // Skip the flags property since ^aFLAGS kludge has already been added

	  if (pfmp->id == FMP_FLAGS)
		continue;

	  // Save kludge and check if ok

	  if (!FmaSaveKludge(pfmm, pfmp, pdszCtrl, FALSE))
		return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message control info
//

  static BOOL FMASUB DoSaveCtrl(PFMM pfmm, PSMH psmh, PDSZ pdszCtrl,
								PSAREA parea, ULONG fl)
  {
	// Save address info and check if ok

	if (!DoSaveAddr(pfmm, psmh, pdszCtrl, parea, fl))
	  return FALSE;

	// Save ^aFLAGS kludge and check if ok

	if (!DoSaveFLAGS(pfmm, psmh, pdszCtrl, parea, fl))
	  return FALSE;

	// Save kludges and check if ok

	if (!DoSaveKludges(pfmm, psmh, pdszCtrl,parea, fl))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads message text
//

  static BOOL FMASUB DoLoadText(PFMM pfmm, HMSG hmsg, ULONG fl)
  {
	PCH pchText = NULL; ULONG cchText = SRM_GETALLTEXT;
	PCH pch;

	// Check if we need to load message text

	if (fl & FLM_NOTEXT) return TRUE;

	// Load text and check if ok

	if (!SquReadMsg(hmsg, NULL, NULL, NULL, &pchText, &cchText, 0))
	  return FALSE;

	// Check if no text loaded and leave

	if (!cchText) {
	  if (pchText) MemFree(pchText);
	  return TRUE;
	}

	// Check data size limit if appropriate

#ifdef CHECK_DATA_SIZE_LIMIT
	if (!CHECK_DATA_SIZE_LIMIT(cchText)) {
	  FMAERROR(DATA_SIZE_LIMIT);
	  return FALSE;
	}
#endif

	// Check if text is not null terminated and add a trailing null

	if (pchText[cchText - 1]) {
	  if ((pch = MemRealloc(pchText, (int)cchText + 1)) == NULL) {
		FMAERROR(NOT_ENOUGH_MEMORY);
		MemFree(pchText);
		return FALSE;
	  } else {
		pchText = pch;
		pchText[cchText] = '\0';
	  }
	}

	// Add message text property and check if ok

	if (!FmaAddFmp(pfmm, FMP_MESSAGETEXT, &pchText, sizeof(pchText))) {
	  MemFree(pchText);
	  return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves message text
//

  static BOOL FMASUB DoSaveText(PFMM pfmm, HMSG hmsg, ULONG fl, PDSZ pdszEom)
  {
	PCH pchText; ULONG cchText, offText = 0;
	PSZ * ppsz;
	ULONG cpsz;
	PFMP pfmp;

	// Check if we need to save message text

	if (fl & FSM_NOTEXT) return TRUE;

	// Get the message text property and check if ok

	if ((pfmp = FmaGetFmp(pfmm, FMP_MESSAGETEXT, 0)) == NULL) 
	  return TRUE;

	// Loop through all the message text strings writing them out

	ppsz = PFMPDATA(pfmp);
	cpsz = pfmp->cb / sizeof(PSZ);

	while (cpsz--) {
	  if (ppsz && *ppsz) {
		pchText = *ppsz;
		cchText = strlen(pchText);
		if (!SquWriteMsg(hmsg, NULL, NULL, 0, pchText, cchText, &offText, 0))
		  return FALSE;
	  }

	  ppsz++;
	}

	// Check if we have eom kludges and write them as well

	if (pdszEom && pdszEom->ich) {
	  if (!SquWriteMsg(hmsg, NULL, NULL, 0, pdszEom->psz, pdszEom->ich, &offText, 0))
		return FALSE;
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine returns total message text length
//

  static ULONG FMASUB DoGetTotalTextLen(PFMM pfmm, PDSZ pdszEom)
  {
	ULONG cchTextTotal = 0;
	PFMP pfmp;

	// Get the message text property and add its length

	if ((pfmp = FmaGetFmp(pfmm, FMP_MESSAGETEXT, 0)) != NULL) 
	  cchTextTotal+= FmaGetFmpTextLen(pfmp);

	// Add eom kludges length

	if (pdszEom) cchTextTotal+= pdszEom->ich;

	return cchTextTotal;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads eom kludges
//

  static BOOL FMASUB DoLoadEomKludges(PFMM pfmm, HMSG hmsg, ULONG fl)
  {
	// Check if message text is loaded and fall to generic routine

	if (FmaGetFmp(pfmm, FMP_MESSAGETEXT, 0)) {
	  return FmaLoadEomKludges(pfmm, fl);
	}

	// No message text to load the eom kludges from

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine returns api pointer
//

  FMAAPI_(PAPI) SquGetApi(VOID)
  {
	return &api;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine loads abstracted fidonet mail message
//

  FMAAPI_(BOOL) SquLoadMsg(HMSG hmsg, PFMM * ppfmm, ULONG fl)
  {
	PSAREA parea = ((PSMSG)hmsg)->parea;
	ULONG cchCtrl = SRM_GETCTRLPTR;
	PCH pchCtrl = NULL; 
	PFMM pfmm;
	SMH smh;

	ASSERT(hmsg != NULL);
	ASSERT(ppfmm != NULL);

	// Create fmm object and check if ok

	if ((pfmm = FmaCreateMsg(0)) == NULL) {
Fail: FmaDestroyMsg(pfmm);      
	  return FALSE;
	} else
	  pfmm->type = MBTYPE_SQU;

	// Read the message header and control info, then check if ok

	if (!SquReadMsg(hmsg, &smh, &pchCtrl, &cchCtrl, NULL, NULL, NULL))
	  goto Fail;

	// Load message attributes and check if ok

	if (!DoLoadAttr(pfmm, &smh))
	  goto Fail;

	// Load message header

	if (!DoLoadMhdr(pfmm, &smh))
	  goto Fail;

	// Load message control info and check if ok

	if (!DoLoadCtrl(pfmm, &smh, pchCtrl, cchCtrl, parea))
	  goto Fail;

	// Load message text and check if ok

	if (!DoLoadText(pfmm, hmsg, fl))
	  goto Fail;

	// Load eom kludges and check if ok

	if (!DoLoadEomKludges(pfmm, hmsg, fl))
	  goto Fail;

	// Pass fmm object to caller

	*ppfmm = pfmm;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine saves abstracted fidonet mail message
//

  FMAAPI_(BOOL) SquSaveMsg(HMSG hmsg, PFMM pfmm, ULONG fl, ULONG cchTextTotal)
  {
	PSAREA parea = ((PSMSG)hmsg)->parea;
	DSZ dszCtrl = DSZ_INIT;
	DSZ dszEom = DSZ_INIT;
	SMH smh;

	ASSERT(hmsg != NULL);
	ASSERT(pfmm != NULL);

	// Clear target message header

	memset(&smh, 0, sizeof(SMH));

	// Add automatic properties and check if ok

	if (!FmaAddAutoProp(pfmm, fl)) {
Fail: DelDsz(&dszCtrl); 
	  DelDsz(&dszEom);
	  return FALSE;
	}

	// Save message attributes and check if ok

	if (!DoSaveAttr(pfmm, &smh))
	  goto Fail;

	// Save message header

	if (!DoSaveMhdr(pfmm, &smh))
	  goto Fail;

	// Save message control info and check if ok

	if (!DoSaveCtrl(pfmm, &smh, &dszCtrl, parea, fl))
	  goto Fail;

	// Save eom kludges early since their length has to be included
	// in the total message length which has to be known before the
	// first write to the new message

	if (!FmaSaveEomKludges(pfmm, &dszEom))
	  goto Fail;

	// Check if caller did not specify total message text length and get it now

	if (cchTextTotal == (ULONG)-1) cchTextTotal = DoGetTotalTextLen(pfmm, &dszEom);

	// Write the message header and control info

	if (!SquWriteMsg(hmsg, &smh, dszCtrl.psz, dszCtrl.ich, NULL, 0, 0, cchTextTotal))
	  goto Fail;

	// Save message text and check if ok

	if (!DoSaveText(pfmm, hmsg, fl, &dszEom))
	  goto Fail;

	// Delete allocated dynamic strings

	DelDsz(&dszCtrl); 
	DelDsz(&dszEom);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of SquApi.c
/////////////////////////////////////////////////////////////////////////////
