/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Squish api declarations module
//
// SquApi.h
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 03/Oct/1999
//
// History: 07/Sep/1998 kvitekp - Initially created for MBS project.
//
/////////////////////////////////////////////////////////////////////////////
//
// Squish base - Copyright 1989, 1995 by Lanius Corporation.
//               All rights reserved.
//
// Note: These declaratons were derived from original
//       specifications published by Squish message base
//       author Scott Dudley in Squish Developers Kit v2.00
//
/////////////////////////////////////////////////////////////////////////////
//
// NOTE: ALL STRUCTURES IN THIS MODULE ASSUME SINGLE BYTE PACKING !!!
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _SQU_API_INCLUDED
#define _SQU_API_INCLUDED

  // Declare C/C++ frame

#ifdef __cplusplus
extern "C" {
#endif

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

  // Set structure packing

#ifdef FMA_PUSHPOP_PACKING
#pragma pack(push, 1)
#endif

/////////////////////////////////////////////////////////////////////////////
// Squish base declarations
/////////////////////////////////////////////////////////////////////////////

  // Squish base types

#define UMSG ULONG              // Unique message id
#define MSGN ULONG              // Message number

  // Squish network address structure

#ifndef _NETADDR_DEFINED
#define _NETADDR_DEFINED

  typedef struct _NETADDR { /* net */
	USHORT  zone;               // Zone number
	USHORT  net;                // Network number
	USHORT  node;               // Node number
	USHORT  point;              // Point number
  } NETADDR, FAR * PNETADDR;

#endif 

  // Squish date time structure

#ifndef _DTTM_DEFINED
#define _DTTM_DEFINED

  typedef struct _DTTM { /* dttm */
	UINT    da:5;               // Day of the month (131)
	UINT    mo:4;               // Month (1 = January, 2 = February, and so on)
	UINT    yr:7;               // Year offset from 1980 (add 1980 to get actual year) 
	UINT    ss:5;               // Second divided by 2
	UINT    mm:6;               // Minute (059)
	UINT    hh:5;               // Hour (023 on a 24-hour clock)
  } DTTM, FAR * PDTTM;

#endif 

  // Squish base file extensions

#define SQU_SQD_FILEEXT  ".SQD" // Data file extension
#define SQU_SQI_FILEEXT  ".SQI" // Index file extension
#define SQU_SQL_FILEEXT  ".SQL" // Lastread file extension

  // Squish base header structure at the beginning of <area>.sqd

  typedef struct _SBH { /* sbh */
	USHORT  cb;                 // Size of the message base header
	USHORT  rsvd1;              // --- reserved
	ULONG   cmsg;               // Number of messages in area
	ULONG   msgnHigh;           // Highest msg in area. Same as cmsg.
	ULONG   cmsgSkip;           // Number of first messages to keep
	UMSG    umsgHwtr;           // Hight water mark message unique id
	UMSG    umsgNext;           // Next unique message id to use
	CHAR    achPath[80];        // Message base path
	LONG    lfmFirst;           // First frame offset
	LONG    lfmLast;            // Last frame offset
	LONG    lfmFree;            // First free frame offset
	LONG    lfmLastFree;        // Last free frame offset
	LONG    lfmFrameEnd;        // Frame end offset (eof)
	ULONG   cmsgMax;            // Max number of messages to keep
	USHORT  keepDays;           // Number of days to keep message
	USHORT  cbSfh;              // Size of the squish frame header
	CHAR    rsvd2[124];         // --- reserved for future use
  } SBH, FAR * PSBH;

  // Squish frame header which goes before each and every message

  typedef struct _SFH { /* sfh */
	ULONG   magic;              // Magic number SFH_MAGIC
	LONG    lfmNext;            // Next frame offset
	LONG    lfmPrev;            // Previous frame offset
	ULONG   cbFrame;            // Frame size not including this header
	ULONG   cbMsg;              // Msg size (msg header, ctrl info and text)
	ULONG   cbCtrl;             // Message control info size
	USHORT  type;               // Frame type (FTYPE_)
	USHORT  rsvd;               // --- reserved for future use
  } SFH, FAR * PSFH;

  // Squish frame header magic number in sfh.magic

#define SFH_MAGIC       0xAFAE4453lu

  // Squish frame types in sqfh.type

#define FRAME_USED      0x0000  // Used frame
#define FRAME_FREE      0x0001  // Free frame
#define FRAME_LZSS      0x0002  // N/A
#define FRAME_UPDATE    0x0003  // Frame is being updated

  // Squish message header structure

  typedef struct _SMH {         /* smh */
	ULONG   attr;               // Message attributes (SMA_)
	CHAR    achBy[36];          // From name string
	CHAR    achTo[36];          // To name string
	CHAR    achRe[72];          // Subject string
	NETADDR netOrig;            // Origination address
	NETADDR netDest;            // Destination address
	DTTM    dttmWritten;        // Written date/time (UTC)
	DTTM    dttmArrived;        // Arrived date/time (UTC)
	SHORT   utcOffset;          // Author's UTC offset in minutes
	UMSG    umsgReplyTo;        // Original message uid
	UMSG    aumsgReplies[9];    // Array  of the reply msgs
	UMSG    umsg;               // Msg uid if (attr & SMA_UMSG)
	CHAR    achDate[20];        // Obsolete date information
  } SMH, FAR * PSMH;

  // Squish message attributes in smh.attr

#define SMA_PRVT     0x0001u    // Private                 +
#define SMA_CRSH     0x0002u    // Crash                   | These
#define SMA_READ     0x0004u    // Read                    | flags are
#define SMA_SENT     0x0008u    // Sent                    | standard for
#define SMA_FILE     0x0010u    // File attach             | FidoNet(tm)
#define SMA_TRAN     0x0020u    // Transit                 | message attr
#define SMA_ORPH     0x0040u    // Orphan                  |
#define SMA_KILL     0x0080u    // Kill/sent               |
#define SMA_LOCL     0x0100u    // Local                   |
#define SMA_HOLD     0x0200u    // Hold                    |
#define SMA_RSV2     0x0400u    // Reserved                |
#define SMA_FREQ     0x0800u    // File request            |
#define SMA_RREQ     0x1000u    // Return receipt request  |
#define SMA_RRCP     0x2000u    // Confirmation receipt msg|
#define SMA_AREQ     0x4000u    // Audit request           |
#define SMA_UREQ     0x8000u    // File update request     +
#define SMA_SCAN 0x00010000lu   // Squish scanned          - *.SQ? only
#define SMA_UMSG 0x00020000lu   // Valid xmsg.umsg         - *.SQ? only

#define SMA_XSQU 0x0003fffflu   // Known squish message attributes

  // Squish message index structure

  typedef struct _SQI {         /* sqi */
	LONG    lfm;                // Frame offset in <area>.sqd
	UMSG    umsg;               // Message unique id
	ULONG   hash;               // To name hash value
  } SQI, FAR * PSQI;

  // Squish index read message hash flag

#define SQI_HASH_READ   0x80000000lu

  // Squish index empty entry initialization

#define SQI_INIT_EMPTY  {0,(ULONG)-1,(ULONG)-1}

  // Squish lastread file record structure

  typedef struct _SLR { /* slr */
	UMSG    umsg;               // Last read message uid
  } SLR, FAR * PSLR;

  // Squish message base locking of <area>.sqd

#define SBH_LOCK_OFFS   0       // Locking region offset
#define SBH_LOCK_LENG   1       // Locking region length
#define SBH_LOCK_RETRY  5       // Max number of locking attempts

/////////////////////////////////////////////////////////////////////////////
// Squish api declarations
/////////////////////////////////////////////////////////////////////////////

  // Special values used by SquReadMsg()

#define SRM_GETCTRLCOPY  0          // *pcchCtrl: Allocate a copy of msg control info
#define SRM_GETCTRLPTR  ((ULONG)-1) // *pcchCtrl: Return a readonly pointer to msg control info
#define SRM_GETALLTEXT  ((ULONG)-1) // *pcchText: read entire message text

  // Function prototypes

  FMAAPI_(PAPI) SquGetApi(VOID);
  FMAAPI_(HAREA) SquOpenArea(PSZ pszPath, UINT fOpenMode, ULONG fl, 
							 NETADDR * pnetAddr);

  FMAAPI_(BOOL) SquCloseArea(HAREA harea);
  FMAAPI_(BOOL) SquLockArea(HAREA harea, BOOL bLock);

  FMAAPI_(BOOL) SquValidateMsg(HAREA harea, MSGN msgn, PULONG pimsg);
  FMAAPI_(BOOL) SquCreateMsgMap(HAREA harea, UMSG * paumsg[], PULONG pcumsg);
  FMAAPI_(BOOL) SquEnumMsgs(HAREA harea, PFNENUMMSGS pfnCallback, 
							PVOID pvCallbackData, ULONG fl);

  FMAAPI_(BOOL) SquDeleteMsg(HAREA harea, MSGN msgn);
  FMAAPI_(BOOL) SquSetCurMsg(HAREA harea, MSGN msgn);
  FMAAPI_(MSGN) SquGetCurMsg(HAREA harea);
  FMAAPI_(MSGN) SquGetFirstMsg(HAREA harea);
  FMAAPI_(MSGN) SquGetHighMsg(HAREA harea);
  FMAAPI_(LONG) SquGetMsgCount(HAREA harea);

  FMAAPI_(UMSG) SquMsgnToUid(HAREA harea, MSGN msgn);
  FMAAPI_(MSGN) SquUidToMsgn(HAREA harea, UMSG umsg, UINT fMode);

  FMAAPI_(HMSG) SquOpenMsg(HAREA harea, UINT mode, MSGN msgn, ULONG fl);
  FMAAPI_(BOOL) SquCloseMsg(HMSG hmsg);

  FMAAPI_(BOOL) SquLoadMsg(HMSG hmsg, PFMM * ppfmm, ULONG fl);
  FMAAPI_(BOOL) SquSaveMsg(HMSG hmsg, PFMM pfmm, ULONG fl, ULONG cchTextTotal);

  FMAAPI_(BOOL) SquReadMsg(HMSG hmsg, PSMH psmh, PCH * ppchCtrl, ULONG * pcchCtrl,
						   PCH * ppchText, ULONG * pcchText, ULONG * poffText);

  FMAAPI_(BOOL) SquWriteMsg(HMSG hmsg, PSMH psmh, PCH pchCtrl, ULONG cchCtrl,
							PCH pchText, ULONG cchText, ULONG * poffText, 
							ULONG cchTextTotal);

  FMAAPI_(BOOL) SquGetMsgCtrlLen(HMSG hmsg, ULONG * pcchCtrl);
  FMAAPI_(BOOL) SquGetMsgTextLen(HMSG hmsg, ULONG * pcchText);

  FMAAPI_(BOOL) SquGetLRInfo(HAREA harea, ULONG uid, PLRINFO plri);
  FMAAPI_(BOOL) SquSetLRInfo(HAREA harea, ULONG uid, PLRINFO plri, 
							 PSZ pszUserName);

/////////////////////////////////////////////////////////////////////////////
// Header cleanup
/////////////////////////////////////////////////////////////////////////////

  // Restore structure packing

#ifdef FMA_PUSHPOP_PACKING
#pragma pack(pop)
#endif

  // Terminate C/C++ frame

#ifdef __cplusplus
}
#endif

#endif  // #ifndef _SQU_API_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// End of SquApi.h
/////////////////////////////////////////////////////////////////////////////
