/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Squish api private declarations
//
// SquApiP.h
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 27/Aug/1999
// Updated: 28/Sep/1999
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _SQU_APIP_INCLUDED
#define _SQU_APIP_INCLUDED

  // Declare C/C++ frame

#ifdef __cplusplus
extern "C" {
#endif

  // Include types declaratons

#include "FmaDef.h"
#include "SquApi.h"

/////////////////////////////////////////////////////////////////////////////
// Squish area declarations
/////////////////////////////////////////////////////////////////////////////

  // Forward declarations

  struct _SMSG;

  // Squish index cache page structure

  typedef struct _SIC {         /* sic */
	struct _SIC * psicPrev;             // Pointer to prev entry in list
	struct _SIC * psicNext;             // Pointer to next entry in list
	UINT    csqi;                       // Number of allocated index entries
	UINT    isqi;                       // Index of the next unused index entry
	SQI     asqi[1];                    // Index entries array
  } SIC, * PSIC;

  // Squish message area object structure

  typedef struct _SAREA {       /* area */
	ULONG   magic;                      // Magic number (SAREA_MAGIC)
	PAPI    papi;                       // Squish api pointer
	SBH     sbh;                        // Squish base header
	HFILE   hfileSqd;                   // Squish data file handle
	HFILE   hfileSqi;                   // Squish index file handle
	HFILE   hfileSql;                   // Squish lastread file handle
	BOOL    bAreaLocked;                // Area lock count
	BOOL    bBaseLocked;                // Base lock count
	BOOL    bExclusive;                 // Exclusive count
	BOOL    bSqiCached;                 // Index caching count
	PSIC    psicList;                   // Index cache list
	MSGN    msgnLoDirty;                // Index cache low dirty (inclusive)
	MSGN    msgnHiDirty;                // Index cache high dirty (inclusive)
	MSGN    msgnCur;                    // Current message number
	LONG    lfmCur;                     // Current message frame offset
	LONG    lfmPrev;                    // Prev message frame offset
	LONG    lfmNext;                    // Next message frame offset
	ULONG   fl;                         // Open area flags (FOA_)
	struct _SMSG * pmsgList;            // Open messges list
	NETADDR netAddr;                    // Default network address
	CHAR    achPath[1];                 // Area path (asciiz)
  } SAREA, FAR * PSAREA;

  // Squish message area object magic number

#define SAREA_MAGIC ((ULONG)0x45524153) // Long equivalent of "SARE"

  // Squish message object structure

  typedef struct _SMSG {        /* msg */
	struct _SMSG * pmsgPrev;            // Pointer to prev open msg in area
	struct _SMSG * pmsgNext;            // Pointer to next open msg in area
	ULONG   magic;                      // Magic number (SMSG_MAGIC)
	PSAREA  parea;                      // Message area handle
	MSGN    msgn;                       // Message number
	UINT    mode;                       // Message open mode
	ULONG   fl;                         // Message open flags (FOM_)
	PSZ     pszCtrl;                    // Message control info
	UINT    cchCtrl;                    // Message control info length
	SMH     smh;                        // Message header
	SFH     sfh;                        // Message frame header
	UINT    fs;                         // Message status flags (SMF_)
	LONG    lfm;                        // Message frame offset
	UMSG    umsg;                       // Message unique id
  } SMSG, FAR * PSMSG;

  // Squish message object magic number

#define SMSG_MAGIC ((ULONG)0x47534D53)  // Long equivalent of "SMSG"

  // Squish message status flags in msg.fs

#define SMF_GOTMSGH       0x0001        // Message header has been associated
#define SMF_GOTCTRL       0x0002        // Control info has been associated
#define SMF_SAVESFH       0x0004        // Save squish frame header to disk
#define SMF_GOTFRAME      0x0008        // Writing frame selected for new message
#define SMF_EXCLUSIVE     0x0010        // Entered exclusive when opened
#define SMF_MSGHWRITTEN   0x0100        // Message header has been written
#define SMF_CTRLWRITTEN   0x0200        // Message contrl info has been written
#define SMF_TEXTWRITTEN   0x0400        // Message text has been written

  // Message has been written to mask

#define SMF_MASKWRITTEN   (SMF_MSGHWRITTEN | SMF_CTRLWRITTEN | SMF_TEXTWRITTEN)

/////////////////////////////////////////////////////////////////////////////
// Miscellaneous declarations
/////////////////////////////////////////////////////////////////////////////

  // Function prototypes

  FMAAPI_(BOOL) SquCheckArea(PSAREA parea);
  FMAAPI_(BOOL) SquCheckMsg(PSMSG pmsg);
  FMAAPI_(BOOL) SquCheckMsgReadMode(PSMSG pmsg);
  FMAAPI_(BOOL) SquCheckMsgWriteMode(PSMSG pmsg);

  FMAAPI_(BOOL) SquLockBase(PSAREA parea, BOOL bLock);
  FMAAPI_(BOOL) SquEnterExclusive(PSAREA parea);
  FMAAPI_(BOOL) SquLeaveExclusive(PSAREA parea);

  FMAAPI_(BOOL) SquLoadSbh(PSAREA parea);
  FMAAPI_(BOOL) SquSaveSbh(PSAREA parea);
  FMAAPI_(BOOL) SquCheckSbh(PSAREA parea);

  FMAAPI_(BOOL) SquResetCurMsgCache(PSAREA parea);
  FMAAPI_(BOOL) SquSetCurMsgCache(PSMSG pmsg);

  FMAAPI_(BOOL) SquLoadSfh(PSAREA parea, PSFH psfh, LONG lfm);
  FMAAPI_(BOOL) SquSaveSfh(PSAREA parea, PSFH psfh, LONG lfm);

  FMAAPI_(BOOL) SquAddFrame(PSAREA parea, PLONG plfm, ULONG cbFrame);
  FMAAPI_(BOOL) SquGetFrame(PSAREA parea, PLONG plfm, MSGN msgn);
  FMAAPI_(BOOL) SquSetNextFrame(PSAREA parea, LONG lfm, LONG lfmNext);
  FMAAPI_(BOOL) SquSetPrevFrame(PSAREA parea, LONG lfm, LONG lfmPrev);

  FMAAPI_(BOOL) SquCheckUsed(PSAREA parea, LONG lfm);
  FMAAPI_(BOOL) SquAddUsed(PSAREA parea, PSFH psfh, LONG lfm);
  FMAAPI_(BOOL) SquDelUsed(PSAREA parea, PSFH psfh, LONG lfm);

  FMAAPI_(BOOL) SquCheckFree(PSAREA parea, LONG lfm);
  FMAAPI_(BOOL) SquAddFree(PSAREA parea, PSFH psfh, LONG lfm);
  FMAAPI_(BOOL) SquDelFree(PSAREA parea, PSFH psfh, LONG lfm);
  FMAAPI_(BOOL) SquFindFree(PSAREA parea, PSFH psfh, PLONG plfm, ULONG cbFrame);

  FMAAPI_(PSMSG) SquGetOpenMsg(PSAREA parea, MSGN msgn);
  FMAAPI_(BOOL) SquDelMsg(PSAREA parea, MSGN msgn, PSFH psfh, LONG lfm);
  FMAAPI_(BOOL) SquTrimBase(PSAREA parea, BOOL bAdding);

  FMAAPI_(BOOL) SquLoadSmh(PSAREA parea, PSMH psmh, LONG lfm);
  FMAAPI_(BOOL) SquSaveSmh(PSAREA parea, PSMH psmh, LONG lfm);
  FMAAPI_(BOOL) SquGetSafeMsgCtrlLen(PSMSG pmsg, PULONG pcbMsg, PULONG pcbCtrl);

  FMAAPI_(BOOL) SquBeginSqiCache(PSAREA parea);
  FMAAPI_(BOOL) SquEndSqiCache(PSAREA parea);

  FMAAPI_(ULONG) SquGetSqiCount(PSAREA parea);

  FMAAPI_(BOOL) SquGetSqi(PSAREA parea, PSQI psqi, MSGN msgn);
  FMAAPI_(BOOL) SquSetSqi(PSAREA parea, PSQI psqi, MSGN msgn);
  FMAAPI_(BOOL) SquDelSqi(PSAREA parea, PSQI psqi, MSGN msgn);
  FMAAPI_(BOOL) SquAddSqi(PSAREA parea, PSQI psqi, MSGN * pmsgn);
  FMAAPI_(BOOL) SquAppSqi(PSAREA parea, PSQI psqi, MSGN * pmsgn);

  FMAAPI_(BOOL) SquSetBaseTrimming(PSAREA parea, ULONG cmsgMax, ULONG cmsgSkip, 
										 USHORT keepDays);

  FMAAPI_(ULONG) SquGetHash(PCH pchText, UINT cchText);

  FMAAPI_(BOOL) SquCheckBase(PSZ pszPath);

  // Squish logging control defines

#ifdef  FMA_LOG
#define FMA_LOG_SQU_AREA
//#define FMA_LOG_SQU_HELP
//#define FMA_LOG_SQU_INDEX
//#define FMA_LOG_SQU_LAST
//#define FMA_LOG_SQU_LOCK
#define FMA_LOG_SQU_MSG
#define FMA_LOG_SQU_READ
#define FMA_LOG_SQU_WRITE
#endif

/////////////////////////////////////////////////////////////////////////////
// Header cleanup
/////////////////////////////////////////////////////////////////////////////

  // Terminate C/C++ frame

#ifdef __cplusplus
}
#endif

#endif  // #ifndef _SQU_APIP_INCLUDED

/////////////////////////////////////////////////////////////////////////////
// End of SquApiP.h
/////////////////////////////////////////////////////////////////////////////
