/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Squish base integrity check routines module
//
// SquCheck.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 06/Oct/1999
// Updated: 05/Nov/1999
//
// History: 06/Oct/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "SquApi.h"
#include "SquApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine checks used frame chain
//

  static ULONG FMASUB DoCheckUsedChain(PSAREA parea)
  {
	LONG lfm = parea->sbh.lfmFirst;
	LONG lfmPrev = 0;
	BOOL bOk = TRUE;
	SFH sfh;

	// Loop trough the used frame chain

	while (lfm) {

	  // Load frame and check if ok

	  if (!SquLoadSfh(parea, &sfh, lfm)) {
#ifdef FMA_LOG
Log("SquCheckBase -----: lfm=%08lx -- invalid sfh", lfm);
#endif
		return FALSE;
	  }

	  // Check previous frame link 

	  if (lfmPrev != sfh.lfmPrev) {
#ifdef FMA_LOG
Log("SquCheckBase -----: lfm=%08lx -- invalid sfh.lfmPrev=%08lx, expected=%08lx", lfm, sfh.lfmPrev, lfmPrev);
#endif
		return FALSE;
	  }

	  // Check if frame is marked as used

	  if (sfh.type != FRAME_USED) {
#ifdef FMA_LOG
Log("SquCheckBase -----: lfm=%08lx -- invalid sfh.type=%d in used chain", lfm, sfh.type);
#endif
		bOk = FALSE;
	  }

	  // Check the last frame link in sbh

	  if (!sfh.lfmNext && lfm != parea->sbh.lfmLast) {
#ifdef FMA_LOG
Log("SquCheckBase -----: lfm=%08lx -- invalid sbh.lfmLast=%08lx, expected=%08lx", lfm, parea->sbh.lfmLast, lfm);
#endif
	  }

	  // Advance to the next frame

	  lfmPrev = lfm;
	  lfm = sfh.lfmNext;
	}

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks free frame chain
//

  static ULONG FMASUB DoCheckFreeChain(PSAREA parea)
  {
	LONG lfm = parea->sbh.lfmFree;
	LONG lfmPrev = 0;
	BOOL bOk = TRUE;
	SFH sfh;

	// Loop trough the free frame chain

	while (lfm) {

	  // Load frame and check if ok

	  if (!SquLoadSfh(parea, &sfh, lfm)) {
#ifdef FMA_LOG
Log("SquCheckBase -----: lfm=%08lx -- invalid sfh", lfm);
#endif
		return FALSE;
	  }

	  // Check previous frame link 

	  if (lfmPrev != sfh.lfmPrev) {
#ifdef FMA_LOG
Log("SquCheckBase -----: lfm=%08lx -- invalid sfh.lfmPrev=%08lx, expected=%08lx", lfm, sfh.lfmPrev, lfmPrev);
#endif
		return FALSE;
	  }

	  // Check if frame is marked as free

	  if (sfh.type != FRAME_FREE) {
#ifdef FMA_LOG
Log("SquCheckBase -----: lfm=%08lx -- invalid sfh.type=%d in free chain", lfm, sfh.type);
#endif
		bOk = FALSE;
	  }

	  // Check the last frame link in sbh

	  if (!sfh.lfmNext && lfm != parea->sbh.lfmLastFree) {
#ifdef FMA_LOG
Log("SquCheckBase -----: lfm=%08lx -- invalid sbh.lfmLastFree=%08lx, expected=%08lx", lfm, parea->sbh.lfmLastFree, lfm);
#endif
	  }

	  // Advance to the next frame

	  lfmPrev = lfm;
	  lfm = sfh.lfmNext;
	}

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks squish header file
//

  static BOOL FMASUB DoCheckSqd(PSAREA parea)
  {
	LONG cbSqdFile = FmaGetFileSize(parea->hfileSqd);
	BOOL bOk = TRUE;
	LONG lfm;
	SFH sfh;

	ASSERT(parea->hfileSqd != -1);

	// Get first physical frame offset and check it

	lfm = parea->sbh.cb;
	if (lfm > cbSqdFile || (parea->sbh.lfmFrameEnd != cbSqdFile)) {
#ifdef FMA_LOG
Log("SquCheckBase -----: invalid sqd file size=%ld", cbSqdFile);
#endif
	  return FALSE;
	}

	// Loop through all physical frames

	while (lfm < cbSqdFile) {

	  // Load frame and check if ok

	  if (!SquLoadSfh(parea, &sfh, lfm)) {
#ifdef FMA_LOG
Log("SquCheckBase -----: lfm=%08lx -- invalid sfh", lfm);
#endif
		return FALSE;
	  }

	  // Check the frame type

	  if (sfh.type == FRAME_USED) {
		if (!SquCheckUsed(parea, lfm)) {
#ifdef FMA_LOG
Log("SquCheckBase -----: lfm=%08lx -- sfh.type=FRAME_USED not in used frames chain", lfm);
#endif
		  bOk = FALSE;
		}
	  } else 

	  if (sfh.type == FRAME_FREE) {
		if (!SquCheckFree(parea, lfm)) {
#ifdef FMA_LOG
Log("SquCheckBase -----: lfm=%08lx -- sfh.type=FRAME_FREE not in free frames chain", lfm);
#endif
		  bOk = FALSE;
		}
	  } else 

	  if (sfh.type == FRAME_UPDATE) {
#ifdef FMA_LOG
Log("SquCheckBase -----: lfm=%08lx -- unexpected sfh.type=FRAME_UPDATE", lfm);
#endif
		bOk = FALSE;
	  } else {
#ifdef FMA_LOG
Log("SquCheckBase -----: lfm=%08lx -- invalid sfh.type=%d", lfm, sfh.type);
#endif
		return FALSE;
	  }

	  // Advance to the next frame

	  lfm+= parea->sbh.cbSfh + sfh.cbFrame;
	}

	// Check the used and free frame chains

	if (!DoCheckUsedChain(parea) || !DoCheckFreeChain(parea))
	  return FALSE;

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine checks squish index file
//

  static BOOL FMASUB DoCheckSqi(PSAREA parea)
  {
	UMSG umsgPrev = 0;
	BOOL bOk = TRUE;
	MSGN msgn;
	SQI sqi;
	SMH smh;

	// Loop through all the message numbers

	for (msgn = 1; msgn <= parea->sbh.msgnHigh; msgn++) {

	  // Load the index record and check if ok

	  if (!SquGetSqi(parea, &sqi, msgn)) {
#ifdef FMA_LOG
Log("SquCheckBase -----: invalid sqi, msgn=%ld", msgn);
#endif
		return FALSE;
	  }

	  // Check message frame offset

	  if (sqi.lfm && (sqi.lfm < (LONG)parea->sbh.cb || sqi.lfm > parea->sbh.lfmFrameEnd)) {
#ifdef FMA_LOG
Log("SquCheckBase -----: invalid sqi.lfm=%08lx, msgn=%ld", sqi.lfm, msgn);
#endif
		bOk = FALSE;
		continue;
	  }

	  // Load message header and check if ok

	  if (!SquLoadSmh(parea, &smh, sqi.lfm)) {
#ifdef FMA_LOG
Log("SquCheckBase -----: invalid smh for msgn=%ld, sqi.lfm=%08lx", msgn, sqi.lfm);
#endif
		bOk = FALSE;
		continue;
	  }

	  // Check unique message id

	  if ((smh.attr & SMA_UMSG) && (smh.umsg != sqi.umsg)) {
#ifdef FMA_LOG
Log("SquCheckBase -----: invalid sqi.umsg=%ld for msgn=%ld, smh.umsg=%ld", sqi.umsg, msgn, smh.umsg);
#endif
		bOk = FALSE;
		continue;
	  }

	  // Check unique message id ascendance

	  if (!(sqi.umsg > umsgPrev)) {
#ifdef FMA_LOG
Log("SquCheckBase -----: not ascending sqi.umsg=%ld, expected > %ld", sqi.umsg, umsgPrev);
#endif
		bOk = FALSE;
		continue;
	  } else
		umsgPrev = sqi.umsg;
	}

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine checks the squish message base integrity
//

  FMAAPI_(BOOL) SquCheckBase(PSZ pszPath)
  {
	ULONG fl = FOA_LOCKAREA | FOA_DENYALL;
	BOOL bOk = TRUE;
	PSAREA parea;

	// Open area in exclusive mode and check if ok

	if ((parea = SquOpenArea(pszPath, OPENAREA_EXISTING, fl, NULL)) == NULL) {
#ifdef FMA_LOG
Log("SquCheckBase -----: Can't open %s, code=%d", pszPath, FmaGetLastError());
#endif
	  return TRUE;
	}

	// Acquire exclusive access and check if ok

	if (SquEnterExclusive(parea)) {

	  // Check message base integrity

	  if (!DoCheckSqd(parea)) bOk = FALSE;
	  if (!DoCheckSqi(parea)) bOk = FALSE;

	  // Relinquish exclusive access

	  if (!SquLeaveExclusive(parea))
		bOk = FALSE;
	}

	// Close area 

	SquCloseArea(parea);

#ifdef FMA_LOG
Log("SquCheckBase -----: %s integrity check %s", pszPath, bOk ? "passed" : "FAILED");
#endif

	return bOk;
  }

/////////////////////////////////////////////////////////////////////////////
// End of SquCheck.c
/////////////////////////////////////////////////////////////////////////////
