/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Squish lastread routines module
//
// SquLast.c
//
// Copyright (c) 1999-2000 by Pete Kvitek
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 03/Oct/1999
// Updated: 03/Oct/1999
//
// History: 03/Oct/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include common and private headers

#include "FmaApi.h"
#include "FmaApiP.h"

#include "SquApi.h"
#include "SquApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine gets squish lastread file record for the specified uid
//

  FMAAPI_(BOOL) SquGetSlr(PSAREA parea, ULONG uid, PSLR pslr, UINT cslr)
  {
	ULONG cb, cbRead;
	LONG offs;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileSql != -1);

	// Find out offset of the lastread record and check if ok

	offs = uid * sizeof(SLR); cb = cslr * sizeof(SLR);
	if ((ULONG)offs + cb > (ULONG)FmaGetFileSize(parea->hfileSql)) {
	  FMAERROR(NOT_FOUND);
	  return FALSE;
	}

	// Read lastread file records and check if ok

	if (!FmaReadFile(parea->hfileSql, offs, pslr, cb, &cbRead)) {
	  return FALSE;
	} else 
	  if (cbRead != cb) {
		FMAERROR(INVALID_DATA);
		return FALSE;
	  }

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets squish lastread file record for the specified uid
//

  FMAAPI_(BOOL) SquSetSlr(PSAREA parea, ULONG uid, PSLR pslr, UINT cslr)
  {
	ULONG cb, cbWritten;
	LONG offs;

	ASSERT(parea != NULL);
	ASSERT(parea->hfileSql != -1);

	// Find out offset of the lastread record and check if ok

	offs = uid * sizeof(SLR); cb = cslr * sizeof(SLR);

	// Write lastread file records and check if ok

	if (!FmaWriteFile(parea->hfileSql, offs, pslr, cb, &cbWritten))
	  return FALSE;

	ASSERT(cbWritten == cb);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine gets lastread info
//

  FMAAPI_(BOOL) SquGetLRInfo(HAREA harea, ULONG uid, PLRINFO plri)
  {
	PSAREA parea;
	SLR aslr[2];
	UINT cslr;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea)) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check if we need to get high read message and
	// set number of lastread records to set

	if (parea->fl & FOA_HIGHREAD) {
	  cslr = 2;

	  // Make sure user id is even

	  if (uid & 1) {
		FMAERROR(INVALID_USERID);
		return FALSE;
	  }
	} else 
	  cslr = 1;

	// Get lastread file record(s) and check if ok

	if (!SquGetSlr(parea, uid, aslr, cslr)) 
	  return FALSE;

	// Pass info to the caller

	if (plri) {
	  if (cslr == 2) {
		plri->umsgLast = aslr[0].umsg;
		plri->umsgHigh = aslr[1].umsg;
	  } else {
		plri->umsgLast = aslr[0].umsg;
		plri->umsgHigh = aslr[0].umsg;
	  }
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine sets lastread info
//

  FMAAPI_(BOOL) SquSetLRInfo(HAREA harea, ULONG uid, PLRINFO plri, PSZ pszUserName)
  {
	PSAREA parea;
	SLR aslr[2];
	UINT cslr;

	// Validate parameters

	if ((parea = (PSAREA)harea) == NULL || !SquCheckArea(parea) || !plri) {
	  FMAERROR(INVALID_PARAM);
	  return FALSE;
	}

	// Check if we need to get high read message and
	// set number of lastread records to set

	if (parea->fl & FOA_HIGHREAD) {
	  cslr = 2;

	  // Make sure user id is even

	  if (uid & 1) {
		FMAERROR(INVALID_USERID);
		return FALSE;
	  }
	} else 
	  cslr = 1;

	// Make up the lastread file record(s)

	aslr[0].umsg = plri->umsgLast;
	aslr[1].umsg = plri->umsgHigh;

	// Set the lastread file record(s0 and check if ok

	if (!SquSetSlr(parea, uid, aslr, cslr))
	  return FALSE;

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// End of SquLast.c
/////////////////////////////////////////////////////////////////////////////
