/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Access: Testing app
//
// Test.c
//
// Copyright (c) 1999 by Pete Kvitek
//
// Created: 27/Aug/1999
// Updated: 27/Aug/1999
//
// History: 27/Aug/1999 kvitekp - Initially created.
//
/////////////////////////////////////////////////////////////////////////////

  // Include standard headers

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

  // Include common and private headers

#include "FmaApi.h"
//#include "FtsApi.h"
//#include "FtsApiP.h"
#include "SquApi.h"
#include "SquApiP.h"
//#include "JamApi.h"
//#include "JamApiP.h"

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   s p e c i f i c   d e c l a r a t i o n s                 //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// M o d u l e   l o c a l   r o u t i n e s                               //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This routine is a callback for SquEnumMsgs()
//

  static BOOL FMACALLBACK DoEnumMsg(PENUMMSGINFO pemi, PVOID pvCallbackData)
  {
	ULONG fl = (ULONG) pvCallbackData;

	// Log the enum call

	Log("DoEnumMsg: msgn=%lu umsg=%lu imsg=%u offs=%08lx hash=%08lx pvCallbackData=%08lx",
		 pemi->msgn, pemi->umsg, pemi->imsg, pemi->offs, pemi->hash, 
		 pvCallbackData);

	//ASSERT(SquValidateMsg(pemi->harea, pemi->msgn, NULL));

	//if (fl & FENUMMSGS_SAFEENUM) SquDeleteMsg(pemi->harea, pemi->msgn);

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message opening
//

  static BOOL FMASUB DoOpenMsg(HAREA harea)
  {
	HMSG hmsg;
	MSGN msgn;

	// Display what these return

	printf("SquGetFirstMsg()=%d\n", SquGetFirstMsg(harea));
	printf("SquGetHighMsg()=%d\n", SquGetHighMsg(harea));
	printf("SquGetMsgCount()=%d\n", SquGetMsgCount(harea));
	printf("SquGetCurMsg()=%d\n", SquGetCurMsg(harea));

	// Open the current message (this should fail after open)

	msgn = MSGN_CUR;
	if ((hmsg = SquOpenMsg(harea, OPENMSG_READ, msgn, 0)) == NULL) {
	  printf("SquOpenMsg() failed, code=%d\n", FmaGetLastError());
	} else
	  SquCloseMsg(hmsg);

	// Loop reading next msgs

	msgn = MSGN_NEXT;
	while ((hmsg = SquOpenMsg(harea, OPENMSG_READ, msgn, 0)) != NULL) {
	  printf("Opened msgn=%d\n", SquGetCurMsg(harea));
	  SquCloseMsg(hmsg);
	}

	// Loop reading prev msgs

	msgn = MSGN_PREV;
	while ((hmsg = SquOpenMsg(harea, OPENMSG_READ, msgn, 0)) != NULL) {
	  printf("Opened msgn=%d\n", SquGetCurMsg(harea));
	  SquCloseMsg(hmsg);
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message reading
//

  static BOOL FMASUB DoReadMsg(HAREA harea, MSGN msgn)
  {
	HMSG hmsg;
	BOOL bOk;
	SMH smh;

	// Open message and check if ok

	if ((hmsg = SquOpenMsg(harea, OPENMSG_READ, msgn, 0)) == NULL) {
	  printf("SquOpenMsg() failed, code=%d\n", FmaGetLastError());
	  return FALSE;
	}

	// Get message control info and text length
#if 0
	{ ULONG cchCtrl, cchText;

	  if (!SquGetMsgCtrlLen(hmsg, &cchCtrl)) {
		printf("SquGetMsgCtrlLen() failed, code=%d\n", FmaGetLastError());
	  }

	  if (!SquGetMsgTextLen(hmsg, &cchText)) {
		printf("SquGetMsgTextLen() failed, code=%d\n", FmaGetLastError());
	  }
	}
#endif

	// Read entire message into allocated buffer
#if 1
	{ PCH pchCtrl = NULL; ULONG cchCtrl = SRM_GETCTRLCOPY;
	  PCH pchText = NULL; ULONG cchText = SRM_GETALLTEXT;

	  bOk = SquReadMsg(hmsg, &smh, &pchCtrl, &cchCtrl, &pchText, &cchText, 0);

	  if (pchCtrl) MemFree(pchCtrl);
	  if (pchText) MemFree(pchText);
	}
#endif

	// Read message by chunks into allocated buffer
#if 0
	{ PCH pchCtrl = NULL; ULONG cchCtrl = SRM_GETCTRLPTR;
	  PCH pchText = NULL; ULONG cchText = 32; ULONG offText = 0;

	  bOk = SquReadMsg(hmsg, &smh, &pchCtrl, &cchCtrl, 0, 0, 0);

	  while (bOk) {
		bOk = SquReadMsg(hmsg, 0, 0, 0, &pchText, &cchText, &offText);
		if (!cchText) break;
	  }

	  if (pchText) MemFree(pchText);
	}
#endif        

	// Read message by chunks into buffer on stack
#if 0
	{ PCH pchCtrl = NULL; ULONG cchCtrl = SRM_GETCTRLPTR; CHAR achText[32];
	  PCH pchText = achText; ULONG cchText = sizeof(achText); ULONG offText = 0;

	  bOk = SquReadMsg(hmsg, &smh, &pchCtrl, &cchCtrl, 0, 0, 0);

	  while (bOk) {
		bOk = SquReadMsg(hmsg, 0, 0, 0, &pchText, &cchText, &offText);
		if (!cchText) break;
	  }
	}
#endif        

	// Close message 

	if (!SquCloseMsg(hmsg)) {
	  printf("SquCloseMsg() failed, code=%d\n", FmaGetLastError());
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message creation and writing
//

  static BOOL FMASUB DoCreateMsg(HAREA harea, MSGN msgn)
  {
	CHAR achCtrl[] = "\x01MSGID: 2:5020/6 12345678" "\x01PID: FMA";
//  CHAR achText[] = "\rThis is a message text\r";
	CHAR achText[] = "\rThis is a message text\r\r-Pete";
	HMSG hmsg;
	SMH smh;

	// Open message and check if ok

	if ((hmsg = SquOpenMsg(harea, OPENMSG_CREATE, msgn, 0)) == NULL) {
	  printf("SquOpenMsg() failed, code=%d\n", FmaGetLastError());
	  return FALSE;
	} else
	  printf("SquGetCurMsg()=%d\n", SquGetCurMsg(harea));

	// Create message header

	memset(&smh, 0, sizeof(smh));
	strcpy(smh.achBy, "From name");
	strcpy(smh.achTo, "To name");
	Format(smh.achRe, lengof(smh.achRe), "Msg %d", SquGetCurMsg(harea));
	smh.attr = SMA_LOCL | SMA_READ;

	// Write message

#if 1    
	if (!SquWriteMsg(hmsg, &smh, achCtrl, lengof(achCtrl), 
					 achText, lengof(achText), 0, lengof(achText))) {
	  printf("SquWriteMsg() failed, code=%d\n", FmaGetLastError());
	}
#else
	if (!SquWriteMsg(hmsg, &smh, achCtrl, lengof(achCtrl), 
					 achText, lengof(achText), 0, 0)) {
	  printf("SquWriteMsg() failed, code=%d\n", FmaGetLastError());
	}
#endif

	// Close message 

	if (!SquCloseMsg(hmsg)) {
	  printf("SquCloseMsg() failed, code=%d\n", FmaGetLastError());
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message updating and writing
//

  static BOOL FMASUB DoUpdateMsg(HAREA harea, MSGN msgn)
  {
	PCH pchCtrl = NULL; ULONG cchCtrl = 0;
	CHAR achText[] = "\rThis is a message text\r";
	PCH pchText = NULL; ULONG cchText = SRM_GETALLTEXT;
	HMSG hmsg;
	SMH smh;

	// Open message and check if ok

	if ((hmsg = SquOpenMsg(harea, OPENMSG_READWRITE, msgn, 0)) == NULL) {
	  printf("SquOpenMsg() failed, code=%d\n", FmaGetLastError());
	  return FALSE;
	} else
	  printf("SquGetCurMsg()=%d\n", SquGetCurMsg(harea));

	// Read message

	SquReadMsg(hmsg, &smh, &pchCtrl, &cchCtrl, &pchText, &cchText, 0);

	// Change message header

	// Change control info

	// Write message
#if 1
	if (!SquWriteMsg(hmsg, &smh, pchCtrl, cchCtrl, achText, lengof(achText), 0, lengof(achText))) {
	  printf("SquWriteMsg() failed, code=%d\n", FmaGetLastError());
	}
#else
	if (!SquWriteMsg(hmsg, &smh, pchCtrl, cchCtrl, achText, 0, 0, 0)) {
	  printf("SquWriteMsg() failed, code=%d\n", FmaGetLastError());
	}
#endif

	// Free allocated memory

	if (pchCtrl) MemFree(pchCtrl);
	if (pchText) MemFree(pchText);

	// Close message 

	if (!SquCloseMsg(hmsg)) {
	  printf("SquCloseMsg() failed, code=%d\n", FmaGetLastError());
	}

	return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// This routine tests message creation and writing
//

  static BOOL FMASUB DoCheckUid(HAREA harea)
  {
	  // Test msgn to uid conversion
#if 1
	{ UMSG umsg;
	  umsg = SquMsgnToUid(harea, 0);
	  umsg = SquMsgnToUid(harea, 1);
	  umsg = SquMsgnToUid(harea, 2);
	  umsg = SquMsgnToUid(harea, 3);
	  umsg = SquMsgnToUid(harea, 4);
	  umsg = 0;
	}
#endif

	// Test uid to msgn conversion
#if 1
	{ MSGN msgn; UMSG umsg;
	  msgn = SquUidToMsgn(harea, 0, UMSG_EXACT);
	  msgn = SquUidToMsgn(harea, 1, UMSG_EXACT);
	  msgn = SquUidToMsgn(harea, 2, UMSG_EXACT);
	  msgn = SquUidToMsgn(harea, 3, UMSG_EXACT);
	  msgn = SquUidToMsgn(harea, 4, UMSG_EXACT);

	  umsg = 1;
	  msgn = SquUidToMsgn(harea, umsg,  UMSG_EXACT);
	  msgn = SquUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_PREV | UMSG_NEXT);
	  msgn = SquUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_PREV);
	  msgn = SquUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_NEXT);
	  msgn = SquUidToMsgn(harea, umsg,  UMSG_PREV);
	  msgn = SquUidToMsgn(harea, umsg,  UMSG_NEXT);

	  umsg = 2;
	  msgn = SquUidToMsgn(harea, umsg,  UMSG_EXACT);
	  msgn = SquUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_PREV | UMSG_NEXT);
	  msgn = SquUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_PREV);
	  msgn = SquUidToMsgn(harea, umsg,  UMSG_EXACT | UMSG_NEXT);
	  msgn = SquUidToMsgn(harea, umsg,  UMSG_PREV);
	  msgn = SquUidToMsgn(harea, umsg,  UMSG_NEXT);
	  msgn = 0;
	}
#endif

	 return TRUE;
  }

/////////////////////////////////////////////////////////////////////////////
// P u b l i c   r o u t i n e s                                           //
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Main...
//

  void main(int argc, char * argv[])
  {
	HAREA harea = NULL;
	HMSG hmsg = NULL;
	PSZ pszArea;
	LONG cb = 0;

	// Force debugging log

	if (!getenv("FMA-LOG")) putenv("FMA-LOG=*");

	// Initialize and check if ok

	if (!FmaInitialize(NULL)) {
	  printf("FmaInitialize() failed, code=%u\n", FmaGetLastError());
	  exit(EXIT_FAILURE);
	}

	// Open area and check if ok

	pszArea = "..\\m\\test";
	if ((harea = SquOpenArea(pszArea, OPENAREA_ALWAYS, 0, NULL)) == NULL) {
	  printf("SquOpenArea() failed, code=%d\n", FmaGetLastError());
	} else {
	  BOOL bLock = 0;

	  // Lock area

	  if (bLock) bLock = SquLockArea(harea, 1);

	  // Miscellaneous tests

//	  DoOpenMsg(harea);
	  DoReadMsg(harea, 1);
//	  DoCreateMsg(harea, 0);
//	  DoUpdateMsg(harea, 1);
//	  DoCheckUid(harea);

	  // Enumerate area messages
#if 0
	  { ULONG fl = 0;
//      fl|= FENUMMSGS_REVERSE;
		fl|= FENUMMSGS_SAFEENUM;
		SquEnumMsgs(harea, DoEnumMsg, (PVOID)fl, fl);
	  }
#endif

#if 0
	  // Delete area messages

	  { BOOL bOk;
		bOk = SquDeleteMsg(harea, 5);
		bOk = SquDeleteMsg(harea, 6);
	  }
#endif

#if 0
	  // Lastread info test

	  { BOOL bOk;
		LRINFO lri;

		((PSAREA)harea)->fl|= FOA_HIGHREAD;

		lri.umsgLast = 1;
		lri.umsgHigh = 2;

		bOk = SquSetLRInfo(harea, 0, &lri, NULL);
		bOk = SquGetLRInfo(harea, 0, &lri);
	  }
#endif

	  // Unlock area

	  if (bLock) SquLockArea(harea, 0);

	  // Close area

	  SquCloseArea(harea);
	}

	// Cleanup and exit

	FmaTerminate();

	exit(EXIT_SUCCESS);
  }

/////////////////////////////////////////////////////////////////////////////
// End of Test.c
/////////////////////////////////////////////////////////////////////////////
